/* Generated by Cython 0.29.36 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/npy_math.h"
        ],
        "include_dirs": [
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include"
        ],
        "name": "treams.special._integrals",
        "sources": [
            "src/treams/special/_integrals.pyx"
        ]
    },
    "module_name": "treams.special._integrals"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_36"
#define CYTHON_HEX_VERSION 0x001D24F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS ((PY_VERSION_HEX >= 0x030600B1) && (PY_VERSION_HEX < 0x030C00A5))
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__treams__special___integrals
#define __PYX_HAVE_API__treams__special___integrals
/* Early includes */
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "numpy/npy_math.h"
#include <complex.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/treams/special/_integrals.pyx",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_jn;
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_jn;
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_yn;
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_yn;
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_in;
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_in;
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_kn;
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_kn;

/* "scipy/special/cython_special.pxd":8
 *     double
 * 
 * cpdef number_t spherical_jn(long n, number_t z, bint derivative=*) noexcept nogil             # <<<<<<<<<<<<<<
 * cpdef number_t spherical_yn(long n, number_t z, bint derivative=*) noexcept nogil
 * cpdef number_t spherical_in(long n, number_t z, bint derivative=*) noexcept nogil
 */
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_jn {
  int __pyx_n;
  int derivative;
};
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_jn {
  int __pyx_n;
  int derivative;
};

/* "scipy/special/cython_special.pxd":9
 * 
 * cpdef number_t spherical_jn(long n, number_t z, bint derivative=*) noexcept nogil
 * cpdef number_t spherical_yn(long n, number_t z, bint derivative=*) noexcept nogil             # <<<<<<<<<<<<<<
 * cpdef number_t spherical_in(long n, number_t z, bint derivative=*) noexcept nogil
 * cpdef number_t spherical_kn(long n, number_t z, bint derivative=*) noexcept nogil
 */
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_yn {
  int __pyx_n;
  int derivative;
};
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_yn {
  int __pyx_n;
  int derivative;
};

/* "scipy/special/cython_special.pxd":10
 * cpdef number_t spherical_jn(long n, number_t z, bint derivative=*) noexcept nogil
 * cpdef number_t spherical_yn(long n, number_t z, bint derivative=*) noexcept nogil
 * cpdef number_t spherical_in(long n, number_t z, bint derivative=*) noexcept nogil             # <<<<<<<<<<<<<<
 * cpdef number_t spherical_kn(long n, number_t z, bint derivative=*) noexcept nogil
 * 
 */
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_in {
  int __pyx_n;
  int derivative;
};
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_in {
  int __pyx_n;
  int derivative;
};

/* "scipy/special/cython_special.pxd":11
 * cpdef number_t spherical_yn(long n, number_t z, bint derivative=*) noexcept nogil
 * cpdef number_t spherical_in(long n, number_t z, bint derivative=*) noexcept nogil
 * cpdef number_t spherical_kn(long n, number_t z, bint derivative=*) noexcept nogil             # <<<<<<<<<<<<<<
 * 
 * ctypedef fused Dd_number_t:
 */
struct __pyx_fuse_0__pyx_opt_args_5scipy_7special_14cython_special_spherical_kn {
  int __pyx_n;
  int derivative;
};
struct __pyx_fuse_1__pyx_opt_args_5scipy_7special_14cython_special_spherical_kn {
  int __pyx_n;
  int derivative;
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_str(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_str(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_0_29_36(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_0_29_36(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'scipy.special.cython_special' */
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc)(__pyx_t_double_complex, int __pyx_skip_dispatch); /*proto*/
static double (*__pyx_fuse_1__pyx_f_5scipy_7special_14cython_special_erfc)(double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_exp1)(__pyx_t_double_complex, int __pyx_skip_dispatch); /*proto*/
static double (*__pyx_fuse_1__pyx_f_5scipy_7special_14cython_special_exp1)(double, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'numpy.math' */

/* Module declarations from 'cython' */

/* Module declarations from 'treams.special._misc' */
static double *__pyx_vp_6treams_7special_5_misc_SQPI = 0;
#define __pyx_v_6treams_7special_5_misc_SQPI (*__pyx_vp_6treams_7special_5_misc_SQPI)
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp)(__pyx_t_double_complex); /*proto*/
static double (*__pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp)(double); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_5_misc_pow)(__pyx_t_double_complex, __pyx_t_double_complex); /*proto*/
static double (*__pyx_fuse_1__pyx_f_6treams_7special_5_misc_pow)(double, double); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_5_misc_sqrt)(__pyx_t_double_complex); /*proto*/
static double (*__pyx_fuse_1__pyx_f_6treams_7special_5_misc_sqrt)(double); /*proto*/

/* Module declarations from 'treams.special._integrals' */
static __pyx_t_double_complex __pyx_f_6treams_7special_10_integrals__cintkambe_m2(__pyx_t_double_complex, __pyx_t_double_complex); /*proto*/
static double __pyx_f_6treams_7special_10_integrals__dintkambe_m2(double, double); /*proto*/
static __pyx_t_double_complex __pyx_f_6treams_7special_10_integrals__cintkambe_0(__pyx_t_double_complex, __pyx_t_double_complex); /*proto*/
static double __pyx_f_6treams_7special_10_integrals__dintkambe_0(double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals_incgamma(double, __pyx_t_double_complex); /*proto*/
static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals_incgamma(double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals_intkambe(long, __pyx_t_double_complex, __pyx_t_double_complex); /*proto*/
static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals_intkambe(long, double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m3(__pyx_t_double_complex, __pyx_t_double_complex); /*proto*/
static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m3(double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m1(__pyx_t_double_complex, __pyx_t_double_complex); /*proto*/
static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m1(double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m2(__pyx_t_double_complex, __pyx_t_double_complex); /*proto*/
static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m2(double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_0(__pyx_t_double_complex, __pyx_t_double_complex); /*proto*/
static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_0(double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe(long, __pyx_t_double_complex, __pyx_t_double_complex, __pyx_t_double_complex *); /*proto*/
static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe(long, double, double, double *); /*proto*/
#define __Pyx_MODULE_NAME "treams.special._integrals"
extern int __pyx_module_is_main_treams__special___integrals;
int __pyx_module_is_main_treams__special___integrals = 0;

/* Implementation of 'treams.special._integrals' */
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_is_neither_integer_nor_half_int[] = " is neither integer nor half-integer";
static const char __pyx_k_Numerical_evaluation_of_integral[] = "Numerical evaluation of integrals";
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_kp_u_is_neither_integer_nor_half_int;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_test;
/* Late includes */

/* "treams/special/_integrals.pyx":18
 * 
 * 
 * cdef number_t incgamma(double n, number_t z) nogil:             # <<<<<<<<<<<<<<
 *     r"""
 *     incgamma(n, z)
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals_incgamma(double __pyx_v_n, __pyx_t_double_complex __pyx_v_z) {
  long __pyx_v_twicen;
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __pyx_t_double_complex __pyx_t_4;
  __pyx_t_double_complex __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_0incgamma", 1);

  /* "treams/special/_integrals.pyx":19
 * 
 * cdef number_t incgamma(double n, number_t z) nogil:
 *     r"""             # <<<<<<<<<<<<<<
 *     incgamma(n, z)
 * 
 */
  /*try:*/ {

    /* "treams/special/_integrals.pyx":36
 *         - `DLMF: 8.2 <https://dlmf.nist.gov/8.2>`
 *     """
 *     cdef long twicen = <long>(2 * n)             # <<<<<<<<<<<<<<
 *     if twicen != 2 * n:
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 */
    __pyx_v_twicen = ((long)(2.0 * __pyx_v_n));

    /* "treams/special/_integrals.pyx":37
 *     """
 *     cdef long twicen = <long>(2 * n)
 *     if twicen != 2 * n:             # <<<<<<<<<<<<<<
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:
 */
    __pyx_t_1 = ((__pyx_v_twicen != (2.0 * __pyx_v_n)) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":38
 *     cdef long twicen = <long>(2 * n)
 *     if twicen != 2 * n:
 *         raise ValueError(f"{n} is neither integer nor half-integer")             # <<<<<<<<<<<<<<
 *     if z == 0:
 *         if twicen <= 0:
 */
      {
          #ifdef WITH_THREAD
          PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
          #endif
          /*try:*/ {
            __pyx_t_2 = PyFloat_FromDouble(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_t_3, __pyx_kp_u_is_neither_integer_nor_half_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_Raise(__pyx_t_3, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __PYX_ERR(0, 38, __pyx_L8_error)
          }
          /*finally:*/ {
            __pyx_L8_error: {
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              goto __pyx_L4_error;
            }
          }
      }

      /* "treams/special/_integrals.pyx":37
 *     """
 *     cdef long twicen = <long>(2 * n)
 *     if twicen != 2 * n:             # <<<<<<<<<<<<<<
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:
 */
    }

    /* "treams/special/_integrals.pyx":39
 *     if twicen != 2 * n:
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:             # <<<<<<<<<<<<<<
 *         if twicen <= 0:
 *             return INFINITY
 */
    __pyx_t_1 = ((__Pyx_c_eq_double(__pyx_v_z, __pyx_t_double_complex_from_parts(0, 0))) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":40
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:
 *         if twicen <= 0:             # <<<<<<<<<<<<<<
 *             return INFINITY
 *         else:
 */
      __pyx_t_1 = ((__pyx_v_twicen <= 0) != 0);
      if (__pyx_t_1) {

        /* "treams/special/_integrals.pyx":41
 *     if z == 0:
 *         if twicen <= 0:
 *             return INFINITY             # <<<<<<<<<<<<<<
 *         else:
 *             return expd(lgamma(n))
 */
        __pyx_r = __pyx_t_double_complex_from_parts(NPY_INFINITY, 0);
        goto __pyx_L3_return;

        /* "treams/special/_integrals.pyx":40
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:
 *         if twicen <= 0:             # <<<<<<<<<<<<<<
 *             return INFINITY
 *         else:
 */
      }

      /* "treams/special/_integrals.pyx":43
 *             return INFINITY
 *         else:
 *             return expd(lgamma(n))             # <<<<<<<<<<<<<<
 *     if twicen == 0:
 *         return cs.exp1(z)
 */
      /*else*/ {
        __pyx_r = __pyx_t_double_complex_from_parts(exp(lgamma(__pyx_v_n)), 0);
        goto __pyx_L3_return;
      }

      /* "treams/special/_integrals.pyx":39
 *     if twicen != 2 * n:
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:             # <<<<<<<<<<<<<<
 *         if twicen <= 0:
 *             return INFINITY
 */
    }

    /* "treams/special/_integrals.pyx":44
 *         else:
 *             return expd(lgamma(n))
 *     if twicen == 0:             # <<<<<<<<<<<<<<
 *         return cs.exp1(z)
 *     if twicen == 2:
 */
    __pyx_t_1 = ((__pyx_v_twicen == 0) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":45
 *             return expd(lgamma(n))
 *     if twicen == 0:
 *         return cs.exp1(z)             # <<<<<<<<<<<<<<
 *     if twicen == 2:
 *         return exp(-z)
 */
      __pyx_r = __pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_exp1(__pyx_v_z, 0);
      goto __pyx_L3_return;

      /* "treams/special/_integrals.pyx":44
 *         else:
 *             return expd(lgamma(n))
 *     if twicen == 0:             # <<<<<<<<<<<<<<
 *         return cs.exp1(z)
 *     if twicen == 2:
 */
    }

    /* "treams/special/_integrals.pyx":46
 *     if twicen == 0:
 *         return cs.exp1(z)
 *     if twicen == 2:             # <<<<<<<<<<<<<<
 *         return exp(-z)
 *     if twicen == 1:
 */
    __pyx_t_1 = ((__pyx_v_twicen == 2) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":47
 *         return cs.exp1(z)
 *     if twicen == 2:
 *         return exp(-z)             # <<<<<<<<<<<<<<
 *     if twicen == 1:
 *         return SQPI * cs.erfc(sqrt(z))
 */
      __pyx_r = __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_neg_double(__pyx_v_z));
      goto __pyx_L3_return;

      /* "treams/special/_integrals.pyx":46
 *     if twicen == 0:
 *         return cs.exp1(z)
 *     if twicen == 2:             # <<<<<<<<<<<<<<
 *         return exp(-z)
 *     if twicen == 1:
 */
    }

    /* "treams/special/_integrals.pyx":48
 *     if twicen == 2:
 *         return exp(-z)
 *     if twicen == 1:             # <<<<<<<<<<<<<<
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:
 */
    __pyx_t_1 = ((__pyx_v_twicen == 1) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":49
 *         return exp(-z)
 *     if twicen == 1:
 *         return SQPI * cs.erfc(sqrt(z))             # <<<<<<<<<<<<<<
 *     if twicen > 2:
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)
 */
      __pyx_r = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(__pyx_v_6treams_7special_5_misc_SQPI, 0), __pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc(__pyx_fuse_0__pyx_f_6treams_7special_5_misc_sqrt(__pyx_v_z), 0));
      goto __pyx_L3_return;

      /* "treams/special/_integrals.pyx":48
 *     if twicen == 2:
 *         return exp(-z)
 *     if twicen == 1:             # <<<<<<<<<<<<<<
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:
 */
    }

    /* "treams/special/_integrals.pyx":50
 *     if twicen == 1:
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:             # <<<<<<<<<<<<<<
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)
 *     return (incgamma(n + 1, z) - pow(z, n) * exp(-z)) / n
 */
    __pyx_t_1 = ((__pyx_v_twicen > 2) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":51
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)             # <<<<<<<<<<<<<<
 *     return (incgamma(n + 1, z) - pow(z, n) * exp(-z)) / n
 * 
 */
      __pyx_r = __Pyx_c_sum_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts((__pyx_v_n - 1.0), 0), __pyx_fuse_0__pyx_f_6treams_7special_10_integrals_incgamma((__pyx_v_n - 1.0), __pyx_v_z)), __Pyx_c_prod_double(__pyx_fuse_0__pyx_f_6treams_7special_5_misc_pow(__pyx_v_z, __pyx_t_double_complex_from_parts((__pyx_v_n - 1.0), 0)), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_neg_double(__pyx_v_z))));
      goto __pyx_L3_return;

      /* "treams/special/_integrals.pyx":50
 *     if twicen == 1:
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:             # <<<<<<<<<<<<<<
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)
 *     return (incgamma(n + 1, z) - pow(z, n) * exp(-z)) / n
 */
    }

    /* "treams/special/_integrals.pyx":52
 *     if twicen > 2:
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)
 *     return (incgamma(n + 1, z) - pow(z, n) * exp(-z)) / n             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_4 = __Pyx_c_diff_double(__pyx_fuse_0__pyx_f_6treams_7special_10_integrals_incgamma((__pyx_v_n + 1.0), __pyx_v_z), __Pyx_c_prod_double(__pyx_fuse_0__pyx_f_6treams_7special_5_misc_pow(__pyx_v_z, __pyx_t_double_complex_from_parts(__pyx_v_n, 0)), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_neg_double(__pyx_v_z))));
    __pyx_t_5 = __pyx_t_double_complex_from_parts(__pyx_v_n, 0);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_5))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 52, __pyx_L4_error)
    }
    __pyx_r = __Pyx_c_quot_double(__pyx_t_4, __pyx_t_5);
    goto __pyx_L3_return;
  }

  /* "treams/special/_integrals.pyx":19
 * 
 * cdef number_t incgamma(double n, number_t z) nogil:
 *     r"""             # <<<<<<<<<<<<<<
 *     incgamma(n, z)
 * 
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/special/_integrals.pyx":18
 * 
 * 
 * cdef number_t incgamma(double n, number_t z) nogil:             # <<<<<<<<<<<<<<
 *     r"""
 *     incgamma(n, z)
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.special._integrals.incgamma", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals_incgamma(double __pyx_v_n, double __pyx_v_z) {
  long __pyx_v_twicen;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_1incgamma", 1);

  /* "treams/special/_integrals.pyx":19
 * 
 * cdef number_t incgamma(double n, number_t z) nogil:
 *     r"""             # <<<<<<<<<<<<<<
 *     incgamma(n, z)
 * 
 */
  /*try:*/ {

    /* "treams/special/_integrals.pyx":36
 *         - `DLMF: 8.2 <https://dlmf.nist.gov/8.2>`
 *     """
 *     cdef long twicen = <long>(2 * n)             # <<<<<<<<<<<<<<
 *     if twicen != 2 * n:
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 */
    __pyx_v_twicen = ((long)(2.0 * __pyx_v_n));

    /* "treams/special/_integrals.pyx":37
 *     """
 *     cdef long twicen = <long>(2 * n)
 *     if twicen != 2 * n:             # <<<<<<<<<<<<<<
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:
 */
    __pyx_t_1 = ((__pyx_v_twicen != (2.0 * __pyx_v_n)) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":38
 *     cdef long twicen = <long>(2 * n)
 *     if twicen != 2 * n:
 *         raise ValueError(f"{n} is neither integer nor half-integer")             # <<<<<<<<<<<<<<
 *     if z == 0:
 *         if twicen <= 0:
 */
      {
          #ifdef WITH_THREAD
          PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
          #endif
          /*try:*/ {
            __pyx_t_2 = PyFloat_FromDouble(__pyx_v_n); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __pyx_t_2 = __Pyx_PyUnicode_Concat(__pyx_t_3, __pyx_kp_u_is_neither_integer_nor_half_int); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 38, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 38, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_Raise(__pyx_t_3, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __PYX_ERR(0, 38, __pyx_L8_error)
          }
          /*finally:*/ {
            __pyx_L8_error: {
              #ifdef WITH_THREAD
              __Pyx_PyGILState_Release(__pyx_gilstate_save);
              #endif
              goto __pyx_L4_error;
            }
          }
      }

      /* "treams/special/_integrals.pyx":37
 *     """
 *     cdef long twicen = <long>(2 * n)
 *     if twicen != 2 * n:             # <<<<<<<<<<<<<<
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:
 */
    }

    /* "treams/special/_integrals.pyx":39
 *     if twicen != 2 * n:
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:             # <<<<<<<<<<<<<<
 *         if twicen <= 0:
 *             return INFINITY
 */
    __pyx_t_1 = ((__pyx_v_z == 0.0) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":40
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:
 *         if twicen <= 0:             # <<<<<<<<<<<<<<
 *             return INFINITY
 *         else:
 */
      __pyx_t_1 = ((__pyx_v_twicen <= 0) != 0);
      if (__pyx_t_1) {

        /* "treams/special/_integrals.pyx":41
 *     if z == 0:
 *         if twicen <= 0:
 *             return INFINITY             # <<<<<<<<<<<<<<
 *         else:
 *             return expd(lgamma(n))
 */
        __pyx_r = NPY_INFINITY;
        goto __pyx_L3_return;

        /* "treams/special/_integrals.pyx":40
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:
 *         if twicen <= 0:             # <<<<<<<<<<<<<<
 *             return INFINITY
 *         else:
 */
      }

      /* "treams/special/_integrals.pyx":43
 *             return INFINITY
 *         else:
 *             return expd(lgamma(n))             # <<<<<<<<<<<<<<
 *     if twicen == 0:
 *         return cs.exp1(z)
 */
      /*else*/ {
        __pyx_r = exp(lgamma(__pyx_v_n));
        goto __pyx_L3_return;
      }

      /* "treams/special/_integrals.pyx":39
 *     if twicen != 2 * n:
 *         raise ValueError(f"{n} is neither integer nor half-integer")
 *     if z == 0:             # <<<<<<<<<<<<<<
 *         if twicen <= 0:
 *             return INFINITY
 */
    }

    /* "treams/special/_integrals.pyx":44
 *         else:
 *             return expd(lgamma(n))
 *     if twicen == 0:             # <<<<<<<<<<<<<<
 *         return cs.exp1(z)
 *     if twicen == 2:
 */
    __pyx_t_1 = ((__pyx_v_twicen == 0) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":45
 *             return expd(lgamma(n))
 *     if twicen == 0:
 *         return cs.exp1(z)             # <<<<<<<<<<<<<<
 *     if twicen == 2:
 *         return exp(-z)
 */
      __pyx_r = __pyx_fuse_1__pyx_f_5scipy_7special_14cython_special_exp1(__pyx_v_z, 0);
      goto __pyx_L3_return;

      /* "treams/special/_integrals.pyx":44
 *         else:
 *             return expd(lgamma(n))
 *     if twicen == 0:             # <<<<<<<<<<<<<<
 *         return cs.exp1(z)
 *     if twicen == 2:
 */
    }

    /* "treams/special/_integrals.pyx":46
 *     if twicen == 0:
 *         return cs.exp1(z)
 *     if twicen == 2:             # <<<<<<<<<<<<<<
 *         return exp(-z)
 *     if twicen == 1:
 */
    __pyx_t_1 = ((__pyx_v_twicen == 2) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":47
 *         return cs.exp1(z)
 *     if twicen == 2:
 *         return exp(-z)             # <<<<<<<<<<<<<<
 *     if twicen == 1:
 *         return SQPI * cs.erfc(sqrt(z))
 */
      __pyx_r = __pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp((-__pyx_v_z));
      goto __pyx_L3_return;

      /* "treams/special/_integrals.pyx":46
 *     if twicen == 0:
 *         return cs.exp1(z)
 *     if twicen == 2:             # <<<<<<<<<<<<<<
 *         return exp(-z)
 *     if twicen == 1:
 */
    }

    /* "treams/special/_integrals.pyx":48
 *     if twicen == 2:
 *         return exp(-z)
 *     if twicen == 1:             # <<<<<<<<<<<<<<
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:
 */
    __pyx_t_1 = ((__pyx_v_twicen == 1) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":49
 *         return exp(-z)
 *     if twicen == 1:
 *         return SQPI * cs.erfc(sqrt(z))             # <<<<<<<<<<<<<<
 *     if twicen > 2:
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)
 */
      __pyx_r = (__pyx_v_6treams_7special_5_misc_SQPI * __pyx_fuse_1__pyx_f_5scipy_7special_14cython_special_erfc(__pyx_fuse_1__pyx_f_6treams_7special_5_misc_sqrt(__pyx_v_z), 0));
      goto __pyx_L3_return;

      /* "treams/special/_integrals.pyx":48
 *     if twicen == 2:
 *         return exp(-z)
 *     if twicen == 1:             # <<<<<<<<<<<<<<
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:
 */
    }

    /* "treams/special/_integrals.pyx":50
 *     if twicen == 1:
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:             # <<<<<<<<<<<<<<
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)
 *     return (incgamma(n + 1, z) - pow(z, n) * exp(-z)) / n
 */
    __pyx_t_1 = ((__pyx_v_twicen > 2) != 0);
    if (__pyx_t_1) {

      /* "treams/special/_integrals.pyx":51
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)             # <<<<<<<<<<<<<<
 *     return (incgamma(n + 1, z) - pow(z, n) * exp(-z)) / n
 * 
 */
      __pyx_r = (((__pyx_v_n - 1.0) * __pyx_fuse_1__pyx_f_6treams_7special_10_integrals_incgamma((__pyx_v_n - 1.0), __pyx_v_z)) + (__pyx_fuse_1__pyx_f_6treams_7special_5_misc_pow(__pyx_v_z, (__pyx_v_n - 1.0)) * __pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp((-__pyx_v_z))));
      goto __pyx_L3_return;

      /* "treams/special/_integrals.pyx":50
 *     if twicen == 1:
 *         return SQPI * cs.erfc(sqrt(z))
 *     if twicen > 2:             # <<<<<<<<<<<<<<
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)
 *     return (incgamma(n + 1, z) - pow(z, n) * exp(-z)) / n
 */
    }

    /* "treams/special/_integrals.pyx":52
 *     if twicen > 2:
 *         return (n - 1) * incgamma(n - 1, z) + pow(z, n - 1) * exp(-z)
 *     return (incgamma(n + 1, z) - pow(z, n) * exp(-z)) / n             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_4 = (__pyx_fuse_1__pyx_f_6treams_7special_10_integrals_incgamma((__pyx_v_n + 1.0), __pyx_v_z) - (__pyx_fuse_1__pyx_f_6treams_7special_5_misc_pow(__pyx_v_z, __pyx_v_n) * __pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp((-__pyx_v_z))));
    if (unlikely(__pyx_v_n == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 52, __pyx_L4_error)
    }
    __pyx_r = (__pyx_t_4 / __pyx_v_n);
    goto __pyx_L3_return;
  }

  /* "treams/special/_integrals.pyx":19
 * 
 * cdef number_t incgamma(double n, number_t z) nogil:
 *     r"""             # <<<<<<<<<<<<<<
 *     incgamma(n, z)
 * 
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/special/_integrals.pyx":18
 * 
 * 
 * cdef number_t incgamma(double n, number_t z) nogil:             # <<<<<<<<<<<<<<
 *     r"""
 *     incgamma(n, z)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.special._integrals.incgamma", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":55
 * 
 * 
 * cdef number_t _intkambe_m3(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     r"""Kambe integral of order -3
 * 
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m3(__pyx_t_double_complex __pyx_v_z, __pyx_t_double_complex __pyx_v_eta) {
  __pyx_t_double_complex __pyx_v_acc;
  __pyx_t_double_complex __pyx_v_mult;
  __pyx_t_double_complex __pyx_v_macc;
  long __pyx_v_mu;
  __pyx_t_double_complex __pyx_r;
  long __pyx_t_1;
  __pyx_t_double_complex __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":66
 *        \Gamma(-n - 1, \frac{z^2 \eta^2}{2})
 *     """
 *     cdef number_t acc = 0             # <<<<<<<<<<<<<<
 *     cdef number_t mult = z * z * 0.25
 *     cdef number_t macc = mult
 */
  __pyx_v_acc = __pyx_t_double_complex_from_parts(0, 0);

  /* "treams/special/_integrals.pyx":67
 *     """
 *     cdef number_t acc = 0
 *     cdef number_t mult = z * z * 0.25             # <<<<<<<<<<<<<<
 *     cdef number_t macc = mult
 *     cdef long mu
 */
  __pyx_v_mult = __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_v_z, __pyx_v_z), __pyx_t_double_complex_from_parts(0.25, 0));

  /* "treams/special/_integrals.pyx":68
 *     cdef number_t acc = 0
 *     cdef number_t mult = z * z * 0.25
 *     cdef number_t macc = mult             # <<<<<<<<<<<<<<
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop
 */
  __pyx_v_macc = __pyx_v_mult;

  /* "treams/special/_integrals.pyx":70
 *     cdef number_t macc = mult
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop             # <<<<<<<<<<<<<<
 *         acc += macc * incgamma(-1 - mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 21; __pyx_t_1+=1) {
    __pyx_v_mu = __pyx_t_1;

    /* "treams/special/_integrals.pyx":71
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop
 *         acc += macc * incgamma(-1 - mu, z * z * eta * eta * 0.5)             # <<<<<<<<<<<<<<
 *         macc *= mult / (mu + 1)
 *     return acc
 */
    __pyx_v_acc = __Pyx_c_sum_double(__pyx_v_acc, __Pyx_c_prod_double(__pyx_v_macc, __pyx_fuse_0__pyx_f_6treams_7special_10_integrals_incgamma((-1L - __pyx_v_mu), __Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_v_z, __pyx_v_z), __pyx_v_eta), __pyx_v_eta), __pyx_t_double_complex_from_parts(0.5, 0)))));

    /* "treams/special/_integrals.pyx":72
 *     for mu in range(21):  # TODO: optimize loop
 *         acc += macc * incgamma(-1 - mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)             # <<<<<<<<<<<<<<
 *     return acc
 * 
 */
    __pyx_t_2 = __pyx_t_double_complex_from_parts((__pyx_v_mu + 1), 0);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_2))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 72, __pyx_L1_error)
    }
    __pyx_v_macc = __Pyx_c_prod_double(__pyx_v_macc, __Pyx_c_quot_double(__pyx_v_mult, __pyx_t_2));
  }

  /* "treams/special/_integrals.pyx":73
 *         acc += macc * incgamma(-1 - mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)
 *     return acc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_acc;
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":55
 * 
 * 
 * cdef number_t _intkambe_m3(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     r"""Kambe integral of order -3
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._intkambe_m3", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m3(double __pyx_v_z, double __pyx_v_eta) {
  double __pyx_v_acc;
  double __pyx_v_mult;
  double __pyx_v_macc;
  long __pyx_v_mu;
  double __pyx_r;
  long __pyx_t_1;
  long __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":66
 *        \Gamma(-n - 1, \frac{z^2 \eta^2}{2})
 *     """
 *     cdef number_t acc = 0             # <<<<<<<<<<<<<<
 *     cdef number_t mult = z * z * 0.25
 *     cdef number_t macc = mult
 */
  __pyx_v_acc = 0.0;

  /* "treams/special/_integrals.pyx":67
 *     """
 *     cdef number_t acc = 0
 *     cdef number_t mult = z * z * 0.25             # <<<<<<<<<<<<<<
 *     cdef number_t macc = mult
 *     cdef long mu
 */
  __pyx_v_mult = ((__pyx_v_z * __pyx_v_z) * 0.25);

  /* "treams/special/_integrals.pyx":68
 *     cdef number_t acc = 0
 *     cdef number_t mult = z * z * 0.25
 *     cdef number_t macc = mult             # <<<<<<<<<<<<<<
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop
 */
  __pyx_v_macc = __pyx_v_mult;

  /* "treams/special/_integrals.pyx":70
 *     cdef number_t macc = mult
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop             # <<<<<<<<<<<<<<
 *         acc += macc * incgamma(-1 - mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 21; __pyx_t_1+=1) {
    __pyx_v_mu = __pyx_t_1;

    /* "treams/special/_integrals.pyx":71
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop
 *         acc += macc * incgamma(-1 - mu, z * z * eta * eta * 0.5)             # <<<<<<<<<<<<<<
 *         macc *= mult / (mu + 1)
 *     return acc
 */
    __pyx_v_acc = (__pyx_v_acc + (__pyx_v_macc * __pyx_fuse_1__pyx_f_6treams_7special_10_integrals_incgamma((-1L - __pyx_v_mu), ((((__pyx_v_z * __pyx_v_z) * __pyx_v_eta) * __pyx_v_eta) * 0.5))));

    /* "treams/special/_integrals.pyx":72
 *     for mu in range(21):  # TODO: optimize loop
 *         acc += macc * incgamma(-1 - mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)             # <<<<<<<<<<<<<<
 *     return acc
 * 
 */
    __pyx_t_2 = (__pyx_v_mu + 1);
    if (unlikely(__pyx_t_2 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 72, __pyx_L1_error)
    }
    __pyx_v_macc = (__pyx_v_macc * (__pyx_v_mult / ((double)__pyx_t_2)));
  }

  /* "treams/special/_integrals.pyx":73
 *         acc += macc * incgamma(-1 - mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)
 *     return acc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_acc;
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":55
 * 
 * 
 * cdef number_t _intkambe_m3(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     r"""Kambe integral of order -3
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._intkambe_m3", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":76
 * 
 * 
 * cdef number_t _intkambe_m1(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     r"""Kambe integral of order -1
 * 
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m1(__pyx_t_double_complex __pyx_v_z, __pyx_t_double_complex __pyx_v_eta) {
  __pyx_t_double_complex __pyx_v_acc;
  __pyx_t_double_complex __pyx_v_mult;
  __pyx_t_double_complex __pyx_v_macc;
  long __pyx_v_mu;
  __pyx_t_double_complex __pyx_r;
  long __pyx_t_1;
  __pyx_t_double_complex __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":87
 *        \Gamma(-n, \frac{z^2 \eta^2}{2})
 *     """
 *     cdef number_t acc = 0             # <<<<<<<<<<<<<<
 *     cdef number_t mult = z * z * 0.25
 *     cdef number_t macc = 0.5
 */
  __pyx_v_acc = __pyx_t_double_complex_from_parts(0, 0);

  /* "treams/special/_integrals.pyx":88
 *     """
 *     cdef number_t acc = 0
 *     cdef number_t mult = z * z * 0.25             # <<<<<<<<<<<<<<
 *     cdef number_t macc = 0.5
 *     cdef long mu
 */
  __pyx_v_mult = __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_v_z, __pyx_v_z), __pyx_t_double_complex_from_parts(0.25, 0));

  /* "treams/special/_integrals.pyx":89
 *     cdef number_t acc = 0
 *     cdef number_t mult = z * z * 0.25
 *     cdef number_t macc = 0.5             # <<<<<<<<<<<<<<
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop
 */
  __pyx_v_macc = __pyx_t_double_complex_from_parts(0.5, 0);

  /* "treams/special/_integrals.pyx":91
 *     cdef number_t macc = 0.5
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop             # <<<<<<<<<<<<<<
 *         acc += macc * incgamma(-mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 21; __pyx_t_1+=1) {
    __pyx_v_mu = __pyx_t_1;

    /* "treams/special/_integrals.pyx":92
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop
 *         acc += macc * incgamma(-mu, z * z * eta * eta * 0.5)             # <<<<<<<<<<<<<<
 *         macc *= mult / (mu + 1)
 *     return acc
 */
    __pyx_v_acc = __Pyx_c_sum_double(__pyx_v_acc, __Pyx_c_prod_double(__pyx_v_macc, __pyx_fuse_0__pyx_f_6treams_7special_10_integrals_incgamma((-__pyx_v_mu), __Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_v_z, __pyx_v_z), __pyx_v_eta), __pyx_v_eta), __pyx_t_double_complex_from_parts(0.5, 0)))));

    /* "treams/special/_integrals.pyx":93
 *     for mu in range(21):  # TODO: optimize loop
 *         acc += macc * incgamma(-mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)             # <<<<<<<<<<<<<<
 *     return acc
 * 
 */
    __pyx_t_2 = __pyx_t_double_complex_from_parts((__pyx_v_mu + 1), 0);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_2))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 93, __pyx_L1_error)
    }
    __pyx_v_macc = __Pyx_c_prod_double(__pyx_v_macc, __Pyx_c_quot_double(__pyx_v_mult, __pyx_t_2));
  }

  /* "treams/special/_integrals.pyx":94
 *         acc += macc * incgamma(-mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)
 *     return acc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_acc;
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":76
 * 
 * 
 * cdef number_t _intkambe_m1(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     r"""Kambe integral of order -1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._intkambe_m1", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m1(double __pyx_v_z, double __pyx_v_eta) {
  double __pyx_v_acc;
  double __pyx_v_mult;
  double __pyx_v_macc;
  long __pyx_v_mu;
  double __pyx_r;
  long __pyx_t_1;
  long __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":87
 *        \Gamma(-n, \frac{z^2 \eta^2}{2})
 *     """
 *     cdef number_t acc = 0             # <<<<<<<<<<<<<<
 *     cdef number_t mult = z * z * 0.25
 *     cdef number_t macc = 0.5
 */
  __pyx_v_acc = 0.0;

  /* "treams/special/_integrals.pyx":88
 *     """
 *     cdef number_t acc = 0
 *     cdef number_t mult = z * z * 0.25             # <<<<<<<<<<<<<<
 *     cdef number_t macc = 0.5
 *     cdef long mu
 */
  __pyx_v_mult = ((__pyx_v_z * __pyx_v_z) * 0.25);

  /* "treams/special/_integrals.pyx":89
 *     cdef number_t acc = 0
 *     cdef number_t mult = z * z * 0.25
 *     cdef number_t macc = 0.5             # <<<<<<<<<<<<<<
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop
 */
  __pyx_v_macc = 0.5;

  /* "treams/special/_integrals.pyx":91
 *     cdef number_t macc = 0.5
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop             # <<<<<<<<<<<<<<
 *         acc += macc * incgamma(-mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 21; __pyx_t_1+=1) {
    __pyx_v_mu = __pyx_t_1;

    /* "treams/special/_integrals.pyx":92
 *     cdef long mu
 *     for mu in range(21):  # TODO: optimize loop
 *         acc += macc * incgamma(-mu, z * z * eta * eta * 0.5)             # <<<<<<<<<<<<<<
 *         macc *= mult / (mu + 1)
 *     return acc
 */
    __pyx_v_acc = (__pyx_v_acc + (__pyx_v_macc * __pyx_fuse_1__pyx_f_6treams_7special_10_integrals_incgamma((-__pyx_v_mu), ((((__pyx_v_z * __pyx_v_z) * __pyx_v_eta) * __pyx_v_eta) * 0.5))));

    /* "treams/special/_integrals.pyx":93
 *     for mu in range(21):  # TODO: optimize loop
 *         acc += macc * incgamma(-mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)             # <<<<<<<<<<<<<<
 *     return acc
 * 
 */
    __pyx_t_2 = (__pyx_v_mu + 1);
    if (unlikely(__pyx_t_2 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 93, __pyx_L1_error)
    }
    __pyx_v_macc = (__pyx_v_macc * (__pyx_v_mult / ((double)__pyx_t_2)));
  }

  /* "treams/special/_integrals.pyx":94
 *         acc += macc * incgamma(-mu, z * z * eta * eta * 0.5)
 *         macc *= mult / (mu + 1)
 *     return acc             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_acc;
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":76
 * 
 * 
 * cdef number_t _intkambe_m1(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     r"""Kambe integral of order -1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._intkambe_m1", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":97
 * 
 * 
 * cdef double complex _cintkambe_m2(double_complex z, double complex eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order -2 for complex double"""
 *     if creal(z) < 0:
 */

static __pyx_t_double_complex __pyx_f_6treams_7special_10_integrals__cintkambe_m2(__pyx_t_double_complex __pyx_v_z, __pyx_t_double_complex __pyx_v_eta) {
  __pyx_t_double_complex __pyx_v_faddp;
  __pyx_t_double_complex __pyx_v_faddm;
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;
  __pyx_t_double_complex __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":99
 * cdef double complex _cintkambe_m2(double_complex z, double complex eta) nogil:
 *     """Kambe integral of order -2 for complex double"""
 *     if creal(z) < 0:             # <<<<<<<<<<<<<<
 *         z = -z
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 */
  __pyx_t_1 = ((creal(__pyx_v_z) < 0.0) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":100
 *     """Kambe integral of order -2 for complex double"""
 *     if creal(z) < 0:
 *         z = -z             # <<<<<<<<<<<<<<
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 *     cdef double complex faddm = cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z)
 */
    __pyx_v_z = __Pyx_c_neg_double(__pyx_v_z);

    /* "treams/special/_integrals.pyx":99
 * cdef double complex _cintkambe_m2(double_complex z, double complex eta) nogil:
 *     """Kambe integral of order -2 for complex double"""
 *     if creal(z) < 0:             # <<<<<<<<<<<<<<
 *         z = -z
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 */
  }

  /* "treams/special/_integrals.pyx":101
 *     if creal(z) < 0:
 *         z = -z
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)             # <<<<<<<<<<<<<<
 *     cdef double complex faddm = cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z)
 *     return -0.5j * (faddp - faddm) * SQPI * M_SQRT1_2
 */
  __pyx_t_2 = __pyx_t_double_complex_from_parts(0, 1.0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_v_eta))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 101, __pyx_L1_error)
  }
  __pyx_v_faddp = __Pyx_c_prod_double(__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc(__Pyx_c_prod_double(__Pyx_c_diff_double(__Pyx_c_prod_double(__pyx_v_z, __pyx_v_eta), __Pyx_c_quot_double(__pyx_t_2, __pyx_v_eta)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0)), 0), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_v_z)));

  /* "treams/special/_integrals.pyx":102
 *         z = -z
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 *     cdef double complex faddm = cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z)             # <<<<<<<<<<<<<<
 *     return -0.5j * (faddp - faddm) * SQPI * M_SQRT1_2
 * 
 */
  __pyx_t_2 = __pyx_t_double_complex_from_parts(0, 1.0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_v_eta))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 102, __pyx_L1_error)
  }
  __pyx_v_faddm = __Pyx_c_prod_double(__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc(__Pyx_c_prod_double(__Pyx_c_sum_double(__Pyx_c_prod_double(__pyx_v_z, __pyx_v_eta), __Pyx_c_quot_double(__pyx_t_2, __pyx_v_eta)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0)), 0), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_v_z)));

  /* "treams/special/_integrals.pyx":103
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 *     cdef double complex faddm = cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z)
 *     return -0.5j * (faddp - faddm) * SQPI * M_SQRT1_2             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 0.5)), __Pyx_c_diff_double(__pyx_v_faddp, __pyx_v_faddm)), __pyx_t_double_complex_from_parts(__pyx_v_6treams_7special_5_misc_SQPI, 0)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0));
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":97
 * 
 * 
 * cdef double complex _cintkambe_m2(double_complex z, double complex eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order -2 for complex double"""
 *     if creal(z) < 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._cintkambe_m2", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":106
 * 
 * 
 * cdef double _dintkambe_m2(double z, double eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order -2 for double"""
 *     if z < 0:
 */

static double __pyx_f_6treams_7special_10_integrals__dintkambe_m2(double __pyx_v_z, double __pyx_v_eta) {
  double __pyx_v_faddp;
  double __pyx_v_faddm;
  double __pyx_r;
  int __pyx_t_1;
  __pyx_t_double_complex __pyx_t_2;
  __pyx_t_double_complex __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":108
 * cdef double _dintkambe_m2(double z, double eta) nogil:
 *     """Kambe integral of order -2 for double"""
 *     if z < 0:             # <<<<<<<<<<<<<<
 *         z = -z
 *     cdef double faddp = cimag(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 */
  __pyx_t_1 = ((__pyx_v_z < 0.0) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":109
 *     """Kambe integral of order -2 for double"""
 *     if z < 0:
 *         z = -z             # <<<<<<<<<<<<<<
 *     cdef double faddp = cimag(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 *     cdef double faddm = cimag(cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z))
 */
    __pyx_v_z = (-__pyx_v_z);

    /* "treams/special/_integrals.pyx":108
 * cdef double _dintkambe_m2(double z, double eta) nogil:
 *     """Kambe integral of order -2 for double"""
 *     if z < 0:             # <<<<<<<<<<<<<<
 *         z = -z
 *     cdef double faddp = cimag(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 */
  }

  /* "treams/special/_integrals.pyx":110
 *     if z < 0:
 *         z = -z
 *     cdef double faddp = cimag(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))             # <<<<<<<<<<<<<<
 *     cdef double faddm = cimag(cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z))
 *     return 0.5 * (faddp - faddm) * SQPI * M_SQRT1_2
 */
  __pyx_t_2 = __pyx_t_double_complex_from_parts(0, 1.0);
  __pyx_t_3 = __pyx_t_double_complex_from_parts(__pyx_v_eta, 0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_3))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 110, __pyx_L1_error)
  }
  __pyx_v_faddp = cimag(__Pyx_c_prod_double(__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc(__Pyx_c_prod_double(__Pyx_c_diff_double(__pyx_t_double_complex_from_parts((__pyx_v_z * __pyx_v_eta), 0), __Pyx_c_quot_double(__pyx_t_2, __pyx_t_3)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0)), 0), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_t_double_complex_from_parts(__pyx_v_z, 0)))));

  /* "treams/special/_integrals.pyx":111
 *         z = -z
 *     cdef double faddp = cimag(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 *     cdef double faddm = cimag(cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z))             # <<<<<<<<<<<<<<
 *     return 0.5 * (faddp - faddm) * SQPI * M_SQRT1_2
 * 
 */
  __pyx_t_3 = __pyx_t_double_complex_from_parts(0, 1.0);
  __pyx_t_2 = __pyx_t_double_complex_from_parts(__pyx_v_eta, 0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_2))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 111, __pyx_L1_error)
  }
  __pyx_v_faddm = cimag(__Pyx_c_prod_double(__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc(__Pyx_c_prod_double(__Pyx_c_sum_double(__pyx_t_double_complex_from_parts((__pyx_v_z * __pyx_v_eta), 0), __Pyx_c_quot_double(__pyx_t_3, __pyx_t_2)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0)), 0), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts(__pyx_v_z, 0)))));

  /* "treams/special/_integrals.pyx":112
 *     cdef double faddp = cimag(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 *     cdef double faddm = cimag(cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z))
 *     return 0.5 * (faddp - faddm) * SQPI * M_SQRT1_2             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (((0.5 * (__pyx_v_faddp - __pyx_v_faddm)) * __pyx_v_6treams_7special_5_misc_SQPI) * M_SQRT1_2);
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":106
 * 
 * 
 * cdef double _dintkambe_m2(double z, double eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order -2 for double"""
 *     if z < 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._dintkambe_m2", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":115
 * 
 * 
 * cdef number_t _intkambe_m2(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for fused type"""
 *     if number_t is double:
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m2(__pyx_t_double_complex __pyx_v_z, __pyx_t_double_complex __pyx_v_eta) {
  __pyx_t_double_complex __pyx_r;

  /* "treams/special/_integrals.pyx":120
 *         return _dintkambe_m2(z, eta)
 *     elif number_t is double_complex:
 *         return _cintkambe_m2(z, eta)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_f_6treams_7special_10_integrals__cintkambe_m2(__pyx_v_z, __pyx_v_eta);
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":115
 * 
 * 
 * cdef number_t _intkambe_m2(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for fused type"""
 *     if number_t is double:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m2(double __pyx_v_z, double __pyx_v_eta) {
  double __pyx_r;

  /* "treams/special/_integrals.pyx":118
 *     """Kambe integral of order 0 for fused type"""
 *     if number_t is double:
 *         return _dintkambe_m2(z, eta)             # <<<<<<<<<<<<<<
 *     elif number_t is double_complex:
 *         return _cintkambe_m2(z, eta)
 */
  __pyx_r = __pyx_f_6treams_7special_10_integrals__dintkambe_m2(__pyx_v_z, __pyx_v_eta);
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":115
 * 
 * 
 * cdef number_t _intkambe_m2(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for fused type"""
 *     if number_t is double:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":123
 * 
 * 
 * cdef double complex _cintkambe_0(double complex z, double complex eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for complex double"""
 *     if creal(z) < 0:
 */

static __pyx_t_double_complex __pyx_f_6treams_7special_10_integrals__cintkambe_0(__pyx_t_double_complex __pyx_v_z, __pyx_t_double_complex __pyx_v_eta) {
  __pyx_t_double_complex __pyx_v_faddp;
  __pyx_t_double_complex __pyx_v_faddm;
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;
  __pyx_t_double_complex __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":125
 * cdef double complex _cintkambe_0(double complex z, double complex eta) nogil:
 *     """Kambe integral of order 0 for complex double"""
 *     if creal(z) < 0:             # <<<<<<<<<<<<<<
 *         z = -z
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 */
  __pyx_t_1 = ((creal(__pyx_v_z) < 0.0) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":126
 *     """Kambe integral of order 0 for complex double"""
 *     if creal(z) < 0:
 *         z = -z             # <<<<<<<<<<<<<<
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 *     cdef double complex faddm = cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z)
 */
    __pyx_v_z = __Pyx_c_neg_double(__pyx_v_z);

    /* "treams/special/_integrals.pyx":125
 * cdef double complex _cintkambe_0(double complex z, double complex eta) nogil:
 *     """Kambe integral of order 0 for complex double"""
 *     if creal(z) < 0:             # <<<<<<<<<<<<<<
 *         z = -z
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 */
  }

  /* "treams/special/_integrals.pyx":127
 *     if creal(z) < 0:
 *         z = -z
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)             # <<<<<<<<<<<<<<
 *     cdef double complex faddm = cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z)
 *     return (faddp + faddm) * SQPI * 0.5 * M_SQRT1_2 / z
 */
  __pyx_t_2 = __pyx_t_double_complex_from_parts(0, 1.0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_v_eta))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 127, __pyx_L1_error)
  }
  __pyx_v_faddp = __Pyx_c_prod_double(__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc(__Pyx_c_prod_double(__Pyx_c_diff_double(__Pyx_c_prod_double(__pyx_v_z, __pyx_v_eta), __Pyx_c_quot_double(__pyx_t_2, __pyx_v_eta)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0)), 0), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_v_z)));

  /* "treams/special/_integrals.pyx":128
 *         z = -z
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 *     cdef double complex faddm = cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z)             # <<<<<<<<<<<<<<
 *     return (faddp + faddm) * SQPI * 0.5 * M_SQRT1_2 / z
 * 
 */
  __pyx_t_2 = __pyx_t_double_complex_from_parts(0, 1.0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_v_eta))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 128, __pyx_L1_error)
  }
  __pyx_v_faddm = __Pyx_c_prod_double(__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc(__Pyx_c_prod_double(__Pyx_c_sum_double(__Pyx_c_prod_double(__pyx_v_z, __pyx_v_eta), __Pyx_c_quot_double(__pyx_t_2, __pyx_v_eta)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0)), 0), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_v_z)));

  /* "treams/special/_integrals.pyx":129
 *     cdef double complex faddp = cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z)
 *     cdef double complex faddm = cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z)
 *     return (faddp + faddm) * SQPI * 0.5 * M_SQRT1_2 / z             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_sum_double(__pyx_v_faddp, __pyx_v_faddm), __pyx_t_double_complex_from_parts(__pyx_v_6treams_7special_5_misc_SQPI, 0)), __pyx_t_double_complex_from_parts(0.5, 0)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0));
  if (unlikely(__Pyx_c_is_zero_double(__pyx_v_z))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 129, __pyx_L1_error)
  }
  __pyx_r = __Pyx_c_quot_double(__pyx_t_2, __pyx_v_z);
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":123
 * 
 * 
 * cdef double complex _cintkambe_0(double complex z, double complex eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for complex double"""
 *     if creal(z) < 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._cintkambe_0", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":132
 * 
 * 
 * cdef double _dintkambe_0(double z, double eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for double"""
 *     if z < 0:  # TODO: necessary?
 */

static double __pyx_f_6treams_7special_10_integrals__dintkambe_0(double __pyx_v_z, double __pyx_v_eta) {
  double __pyx_v_faddp;
  double __pyx_v_faddm;
  double __pyx_r;
  int __pyx_t_1;
  __pyx_t_double_complex __pyx_t_2;
  __pyx_t_double_complex __pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":134
 * cdef double _dintkambe_0(double z, double eta) nogil:
 *     """Kambe integral of order 0 for double"""
 *     if z < 0:  # TODO: necessary?             # <<<<<<<<<<<<<<
 *         z = -z
 *     cdef double faddp = creal(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 */
  __pyx_t_1 = ((__pyx_v_z < 0.0) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":135
 *     """Kambe integral of order 0 for double"""
 *     if z < 0:  # TODO: necessary?
 *         z = -z             # <<<<<<<<<<<<<<
 *     cdef double faddp = creal(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 *     cdef double faddm = creal(cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z))
 */
    __pyx_v_z = (-__pyx_v_z);

    /* "treams/special/_integrals.pyx":134
 * cdef double _dintkambe_0(double z, double eta) nogil:
 *     """Kambe integral of order 0 for double"""
 *     if z < 0:  # TODO: necessary?             # <<<<<<<<<<<<<<
 *         z = -z
 *     cdef double faddp = creal(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 */
  }

  /* "treams/special/_integrals.pyx":136
 *     if z < 0:  # TODO: necessary?
 *         z = -z
 *     cdef double faddp = creal(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))             # <<<<<<<<<<<<<<
 *     cdef double faddm = creal(cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z))
 *     return (faddp + faddm) * SQPI * 0.5 * M_SQRT1_2 / z
 */
  __pyx_t_2 = __pyx_t_double_complex_from_parts(0, 1.0);
  __pyx_t_3 = __pyx_t_double_complex_from_parts(__pyx_v_eta, 0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_3))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 136, __pyx_L1_error)
  }
  __pyx_v_faddp = creal(__Pyx_c_prod_double(__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc(__Pyx_c_prod_double(__Pyx_c_diff_double(__pyx_t_double_complex_from_parts((__pyx_v_z * __pyx_v_eta), 0), __Pyx_c_quot_double(__pyx_t_2, __pyx_t_3)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0)), 0), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_t_double_complex_from_parts(__pyx_v_z, 0)))));

  /* "treams/special/_integrals.pyx":137
 *         z = -z
 *     cdef double faddp = creal(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 *     cdef double faddm = creal(cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z))             # <<<<<<<<<<<<<<
 *     return (faddp + faddm) * SQPI * 0.5 * M_SQRT1_2 / z
 * 
 */
  __pyx_t_3 = __pyx_t_double_complex_from_parts(0, 1.0);
  __pyx_t_2 = __pyx_t_double_complex_from_parts(__pyx_v_eta, 0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_2))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 137, __pyx_L1_error)
  }
  __pyx_v_faddm = creal(__Pyx_c_prod_double(__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc(__Pyx_c_prod_double(__Pyx_c_sum_double(__pyx_t_double_complex_from_parts((__pyx_v_z * __pyx_v_eta), 0), __Pyx_c_quot_double(__pyx_t_3, __pyx_t_2)), __pyx_t_double_complex_from_parts(M_SQRT1_2, 0)), 0), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts(__pyx_v_z, 0)))));

  /* "treams/special/_integrals.pyx":138
 *     cdef double faddp = creal(cs.erfc((z * eta - 1j / eta) * M_SQRT1_2) * exp(-1j * z))
 *     cdef double faddm = creal(cs.erfc((z * eta + 1j / eta) * M_SQRT1_2) * exp(1j * z))
 *     return (faddp + faddm) * SQPI * 0.5 * M_SQRT1_2 / z             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = ((((__pyx_v_faddp + __pyx_v_faddm) * __pyx_v_6treams_7special_5_misc_SQPI) * 0.5) * M_SQRT1_2);
  if (unlikely(__pyx_v_z == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 138, __pyx_L1_error)
  }
  __pyx_r = (__pyx_t_4 / __pyx_v_z);
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":132
 * 
 * 
 * cdef double _dintkambe_0(double z, double eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for double"""
 *     if z < 0:  # TODO: necessary?
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._dintkambe_0", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":141
 * 
 * 
 * cdef number_t _intkambe_0(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for fused type"""
 *     if number_t is double:
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_0(__pyx_t_double_complex __pyx_v_z, __pyx_t_double_complex __pyx_v_eta) {
  __pyx_t_double_complex __pyx_r;

  /* "treams/special/_integrals.pyx":146
 *         return _dintkambe_0(z, eta)
 *     elif number_t is double_complex:
 *         return _cintkambe_0(z, eta)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_f_6treams_7special_10_integrals__cintkambe_0(__pyx_v_z, __pyx_v_eta);
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":141
 * 
 * 
 * cdef number_t _intkambe_0(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for fused type"""
 *     if number_t is double:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_0(double __pyx_v_z, double __pyx_v_eta) {
  double __pyx_r;

  /* "treams/special/_integrals.pyx":144
 *     """Kambe integral of order 0 for fused type"""
 *     if number_t is double:
 *         return _dintkambe_0(z, eta)             # <<<<<<<<<<<<<<
 *     elif number_t is double_complex:
 *         return _cintkambe_0(z, eta)
 */
  __pyx_r = __pyx_f_6treams_7special_10_integrals__dintkambe_0(__pyx_v_z, __pyx_v_eta);
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":141
 * 
 * 
 * cdef number_t _intkambe_0(number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     """Kambe integral of order 0 for fused type"""
 *     if number_t is double:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":149
 * 
 * 
 * cdef number_t _intkambe(long n, number_t z, number_t eta, number_t *cache) nogil:             # <<<<<<<<<<<<<<
 *     r"""
 *     Cached recursion for the Kambe integral
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe(long __pyx_v_n, __pyx_t_double_complex __pyx_v_z, __pyx_t_double_complex __pyx_v_eta, __pyx_t_double_complex *__pyx_v_cache) {
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_double_complex __pyx_t_3;
  __pyx_t_double_complex __pyx_t_4;
  __pyx_t_double_complex __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":165
 *         .. [#] `K. Kambe, Zeitschrift Fuer Naturforschung A 23, 9 (1968). <https://doi.org/10.1515/zna-1968-0908>`_
 *     """
 *     if not isnan(creal(cache[labs(n)])):             # <<<<<<<<<<<<<<
 *         return cache[labs(n)]
 *     if eta == 0:
 */
  __pyx_t_1 = ((!(isnan(creal((__pyx_v_cache[labs(__pyx_v_n)]))) != 0)) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":166
 *     """
 *     if not isnan(creal(cache[labs(n)])):
 *         return cache[labs(n)]             # <<<<<<<<<<<<<<
 *     if eta == 0:
 *         cache[labs(n)] = INFINITY
 */
    __pyx_r = (__pyx_v_cache[labs(__pyx_v_n)]);
    goto __pyx_L0;

    /* "treams/special/_integrals.pyx":165
 *         .. [#] `K. Kambe, Zeitschrift Fuer Naturforschung A 23, 9 (1968). <https://doi.org/10.1515/zna-1968-0908>`_
 *     """
 *     if not isnan(creal(cache[labs(n)])):             # <<<<<<<<<<<<<<
 *         return cache[labs(n)]
 *     if eta == 0:
 */
  }

  /* "treams/special/_integrals.pyx":167
 *     if not isnan(creal(cache[labs(n)])):
 *         return cache[labs(n)]
 *     if eta == 0:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = INFINITY
 *     # n < -4 is handled regularly
 */
  __pyx_t_1 = ((__Pyx_c_eq_double(__pyx_v_eta, __pyx_t_double_complex_from_parts(0, 0))) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":168
 *         return cache[labs(n)]
 *     if eta == 0:
 *         cache[labs(n)] = INFINITY             # <<<<<<<<<<<<<<
 *     # n < -4 is handled regularly
 *     elif z == 0 and n == -4:
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_t_double_complex_from_parts(NPY_INFINITY, 0);

    /* "treams/special/_integrals.pyx":167
 *     if not isnan(creal(cache[labs(n)])):
 *         return cache[labs(n)]
 *     if eta == 0:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = INFINITY
 *     # n < -4 is handled regularly
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":170
 *         cache[labs(n)] = INFINITY
 *     # n < -4 is handled regularly
 *     elif z == 0 and n == -4:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:
 */
  __pyx_t_2 = ((__Pyx_c_eq_double(__pyx_v_z, __pyx_t_double_complex_from_parts(0, 0))) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_n == -4L) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":171
 *     # n < -4 is handled regularly
 *     elif z == 0 and n == -4:
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta             # <<<<<<<<<<<<<<
 *     elif z == 0 and n == -3:
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 */
    __pyx_t_3 = __pyx_t_double_complex_from_parts(0.5, 0);
    __pyx_t_4 = __Pyx_c_prod_double(__pyx_v_eta, __pyx_v_eta);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_4))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 171, __pyx_L1_error)
    }
    __pyx_t_5 = __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_quot_double(__pyx_t_3, __pyx_t_4));
    if (unlikely(__Pyx_c_is_zero_double(__pyx_v_eta))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 171, __pyx_L1_error)
    }
    (__pyx_v_cache[labs(__pyx_v_n)]) = __Pyx_c_sum_double(__Pyx_c_neg_double(__pyx_fuse_0__pyx_f_6treams_7special_10_integrals_intkambe(-2L, __pyx_t_double_complex_from_parts(0, 0), __pyx_v_eta)), __Pyx_c_quot_double(__pyx_t_5, __pyx_v_eta));

    /* "treams/special/_integrals.pyx":170
 *         cache[labs(n)] = INFINITY
 *     # n < -4 is handled regularly
 *     elif z == 0 and n == -4:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":172
 *     elif z == 0 and n == -4:
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 *     # n == -2 is handled regularly
 */
  __pyx_t_2 = ((__Pyx_c_eq_double(__pyx_v_z, __pyx_t_double_complex_from_parts(0, 0))) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_n == -3L) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":173
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1             # <<<<<<<<<<<<<<
 *     # n == -2 is handled regularly
 *     elif z == 0 and n > -2:
 */
    __pyx_t_5 = __pyx_t_double_complex_from_parts(1, 0);
    __pyx_t_4 = __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(2, 0), __pyx_v_eta), __pyx_v_eta);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_4))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 173, __pyx_L1_error)
    }
    (__pyx_v_cache[labs(__pyx_v_n)]) = __Pyx_c_diff_double(__pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_quot_double(__pyx_t_5, __pyx_t_4)), __pyx_t_double_complex_from_parts(1, 0));

    /* "treams/special/_integrals.pyx":172
 *     elif z == 0 and n == -4:
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 *     # n == -2 is handled regularly
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":175
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 *     # n == -2 is handled regularly
 *     elif z == 0 and n > -2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:
 */
  __pyx_t_2 = ((__Pyx_c_eq_double(__pyx_v_z, __pyx_t_double_complex_from_parts(0, 0))) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_n > -2L) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":176
 *     # n == -2 is handled regularly
 *     elif z == 0 and n > -2:
 *         cache[labs(n)] = INFINITY             # <<<<<<<<<<<<<<
 *     elif n == -3:
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_t_double_complex_from_parts(NPY_INFINITY, 0);

    /* "treams/special/_integrals.pyx":175
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 *     # n == -2 is handled regularly
 *     elif z == 0 and n > -2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":177
 *     elif z == 0 and n > -2:
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:
 */
  __pyx_t_1 = ((__pyx_v_n == -3L) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":178
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:
 *         cache[labs(n)] = _intkambe_m3(z, eta)             # <<<<<<<<<<<<<<
 *     elif n == -2:
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m3(__pyx_v_z, __pyx_v_eta);

    /* "treams/special/_integrals.pyx":177
 *     elif z == 0 and n > -2:
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":179
 *     elif n == -3:
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:
 */
  __pyx_t_1 = ((__pyx_v_n == -2L) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":180
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:
 *         cache[labs(n)] = _intkambe_m2(z, eta)             # <<<<<<<<<<<<<<
 *     elif n == -1:
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m2(__pyx_v_z, __pyx_v_eta);

    /* "treams/special/_integrals.pyx":179
 *     elif n == -3:
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":181
 *     elif n == -2:
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:
 */
  __pyx_t_1 = ((__pyx_v_n == -1L) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":182
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:
 *         cache[labs(n)] = _intkambe_m1(z, eta)             # <<<<<<<<<<<<<<
 *     elif n == 0:
 *         cache[labs(n)] = _intkambe_0(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m1(__pyx_v_z, __pyx_v_eta);

    /* "treams/special/_integrals.pyx":181
 *     elif n == -2:
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":183
 *     elif n == -1:
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_0(z, eta)
 *     # Simplified recursion for n == 1
 */
  __pyx_t_1 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":184
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:
 *         cache[labs(n)] = _intkambe_0(z, eta)             # <<<<<<<<<<<<<<
 *     # Simplified recursion for n == 1
 *     elif n == 1:
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_0(__pyx_v_z, __pyx_v_eta);

    /* "treams/special/_integrals.pyx":183
 *     elif n == -1:
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_0(z, eta)
 *     # Simplified recursion for n == 1
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":186
 *         cache[labs(n)] = _intkambe_0(z, eta)
 *     # Simplified recursion for n == 1
 *     elif n == 1:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 */
  __pyx_t_1 = ((__pyx_v_n == 1) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":188
 *     elif n == 1:
 *         cache[labs(n)] = (
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *             - _intkambe_m3(z, eta)
 *         ) / (z * z)
 */
    __pyx_t_4 = __pyx_t_double_complex_from_parts(1, 0);
    __pyx_t_5 = __Pyx_c_prod_double(__pyx_v_eta, __pyx_v_eta);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_5))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 188, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":189
 *         cache[labs(n)] = (
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *             - _intkambe_m3(z, eta)             # <<<<<<<<<<<<<<
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 */
    __pyx_t_3 = __Pyx_c_diff_double(__pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__Pyx_c_sum_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_v_z), __pyx_v_z), __pyx_v_eta), __pyx_v_eta), __Pyx_c_quot_double(__pyx_t_4, __pyx_t_5)), __pyx_t_double_complex_from_parts(0.5, 0))), __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m3(__pyx_v_z, __pyx_v_eta));

    /* "treams/special/_integrals.pyx":190
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *             - _intkambe_m3(z, eta)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     # Needed to avoid false cache hits
 *     elif n == 2:
 */
    __pyx_t_5 = __Pyx_c_prod_double(__pyx_v_z, __pyx_v_z);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_5))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 190, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":187
 *     # Simplified recursion for n == 1
 *     elif n == 1:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *             - _intkambe_m3(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __Pyx_c_quot_double(__pyx_t_3, __pyx_t_5);

    /* "treams/special/_integrals.pyx":186
 *         cache[labs(n)] = _intkambe_0(z, eta)
 *     # Simplified recursion for n == 1
 *     elif n == 1:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":192
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 *     elif n == 2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             _intkambe(0, z, eta, cache)
 */
  __pyx_t_1 = ((__pyx_v_n == 2) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":196
 *             _intkambe(0, z, eta, cache)
 *             - _intkambe_m2(z, eta)
 *             + eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 */
    __pyx_t_5 = __pyx_t_double_complex_from_parts(1, 0);
    __pyx_t_3 = __Pyx_c_prod_double(__pyx_v_eta, __pyx_v_eta);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_3))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 196, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_c_sum_double(__Pyx_c_diff_double(__pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe(0, __pyx_v_z, __pyx_v_eta, __pyx_v_cache), __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m2(__pyx_v_z, __pyx_v_eta)), __Pyx_c_prod_double(__pyx_v_eta, __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__Pyx_c_sum_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_v_z), __pyx_v_z), __pyx_v_eta), __pyx_v_eta), __Pyx_c_quot_double(__pyx_t_5, __pyx_t_3)), __pyx_t_double_complex_from_parts(0.5, 0)))));

    /* "treams/special/_integrals.pyx":197
 *             - _intkambe_m2(z, eta)
 *             + eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     # Needed to avoid false cache hits
 *     elif n == 3:
 */
    __pyx_t_3 = __Pyx_c_prod_double(__pyx_v_z, __pyx_v_z);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_3))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 197, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":193
 *     # Needed to avoid false cache hits
 *     elif n == 2:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             _intkambe(0, z, eta, cache)
 *             - _intkambe_m2(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __Pyx_c_quot_double(__pyx_t_4, __pyx_t_3);

    /* "treams/special/_integrals.pyx":192
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 *     elif n == 2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             _intkambe(0, z, eta, cache)
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":199
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 *     elif n == 3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             2 * _intkambe(1, z, eta, cache)
 */
  __pyx_t_1 = ((__pyx_v_n == 3) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":203
 *             2 * _intkambe(1, z, eta, cache)
 *             - _intkambe_m1(z, eta)
 *             + eta * eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *         ) / (z * z)
 *     elif n < -3:
 */
    __pyx_t_3 = __pyx_t_double_complex_from_parts(1, 0);
    __pyx_t_4 = __Pyx_c_prod_double(__pyx_v_eta, __pyx_v_eta);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_4))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 203, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_c_sum_double(__Pyx_c_diff_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(2, 0), __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe(1, __pyx_v_z, __pyx_v_eta, __pyx_v_cache)), __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe_m1(__pyx_v_z, __pyx_v_eta)), __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_v_eta, __pyx_v_eta), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__Pyx_c_sum_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_v_z), __pyx_v_z), __pyx_v_eta), __pyx_v_eta), __Pyx_c_quot_double(__pyx_t_3, __pyx_t_4)), __pyx_t_double_complex_from_parts(0.5, 0)))));

    /* "treams/special/_integrals.pyx":204
 *             - _intkambe_m1(z, eta)
 *             + eta * eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     elif n < -3:
 *         cache[labs(n)] = (
 */
    __pyx_t_4 = __Pyx_c_prod_double(__pyx_v_z, __pyx_v_z);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_4))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 204, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":200
 *     # Needed to avoid false cache hits
 *     elif n == 3:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             2 * _intkambe(1, z, eta, cache)
 *             - _intkambe_m1(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __Pyx_c_quot_double(__pyx_t_5, __pyx_t_4);

    /* "treams/special/_integrals.pyx":199
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 *     elif n == 3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             2 * _intkambe(1, z, eta, cache)
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":205
 *             + eta * eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)
 *     elif n < -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             (n + 3) * _intkambe(n + 2, z, eta, cache)
 */
  __pyx_t_1 = ((__pyx_v_n < -3L) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":209
 *             (n + 3) * _intkambe(n + 2, z, eta, cache)
 *             - z * z * _intkambe(n + 4, z, eta, cache)
 *             + pow(eta, n + 3) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *         )
 *     else:
 */
    __pyx_t_4 = __pyx_t_double_complex_from_parts(1, 0);
    __pyx_t_5 = __Pyx_c_prod_double(__pyx_v_eta, __pyx_v_eta);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_5))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 209, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":206
 *         ) / (z * z)
 *     elif n < -3:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             (n + 3) * _intkambe(n + 2, z, eta, cache)
 *             - z * z * _intkambe(n + 4, z, eta, cache)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __Pyx_c_sum_double(__Pyx_c_diff_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts((__pyx_v_n + 3), 0), __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe((__pyx_v_n + 2), __pyx_v_z, __pyx_v_eta, __pyx_v_cache)), __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_v_z, __pyx_v_z), __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe((__pyx_v_n + 4), __pyx_v_z, __pyx_v_eta, __pyx_v_cache))), __Pyx_c_prod_double(__pyx_fuse_0__pyx_f_6treams_7special_5_misc_pow(__pyx_v_eta, __pyx_t_double_complex_from_parts((__pyx_v_n + 3), 0)), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__Pyx_c_sum_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_v_z), __pyx_v_z), __pyx_v_eta), __pyx_v_eta), __Pyx_c_quot_double(__pyx_t_4, __pyx_t_5)), __pyx_t_double_complex_from_parts(0.5, 0)))));

    /* "treams/special/_integrals.pyx":205
 *             + eta * eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)
 *     elif n < -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             (n + 3) * _intkambe(n + 2, z, eta, cache)
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":216
 *             - _intkambe(n - 4, z, eta, cache)
 *             + pow(eta, n - 1) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     return cache[labs(n)]
 * 
 */
  /*else*/ {

    /* "treams/special/_integrals.pyx":215
 *             (n - 1) * _intkambe(n - 2, z, eta, cache)
 *             - _intkambe(n - 4, z, eta, cache)
 *             + pow(eta, n - 1) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *         ) / (z * z)
 *     return cache[labs(n)]
 */
    __pyx_t_5 = __pyx_t_double_complex_from_parts(1, 0);
    __pyx_t_4 = __Pyx_c_prod_double(__pyx_v_eta, __pyx_v_eta);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_4))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 215, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_c_sum_double(__Pyx_c_diff_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts((__pyx_v_n - 1), 0), __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe((__pyx_v_n - 2), __pyx_v_z, __pyx_v_eta, __pyx_v_cache)), __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe((__pyx_v_n - 4), __pyx_v_z, __pyx_v_eta, __pyx_v_cache)), __Pyx_c_prod_double(__pyx_fuse_0__pyx_f_6treams_7special_5_misc_pow(__pyx_v_eta, __pyx_t_double_complex_from_parts((__pyx_v_n - 1), 0)), __pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp(__Pyx_c_prod_double(__Pyx_c_sum_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_neg_double(__pyx_v_z), __pyx_v_z), __pyx_v_eta), __pyx_v_eta), __Pyx_c_quot_double(__pyx_t_5, __pyx_t_4)), __pyx_t_double_complex_from_parts(0.5, 0)))));

    /* "treams/special/_integrals.pyx":216
 *             - _intkambe(n - 4, z, eta, cache)
 *             + pow(eta, n - 1) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     return cache[labs(n)]
 * 
 */
    __pyx_t_4 = __Pyx_c_prod_double(__pyx_v_z, __pyx_v_z);
    if (unlikely(__Pyx_c_is_zero_double(__pyx_t_4))) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 216, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":212
 *         )
 *     else:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             (n - 1) * _intkambe(n - 2, z, eta, cache)
 *             - _intkambe(n - 4, z, eta, cache)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __Pyx_c_quot_double(__pyx_t_3, __pyx_t_4);
  }
  __pyx_L4:;

  /* "treams/special/_integrals.pyx":217
 *             + pow(eta, n - 1) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)
 *     return cache[labs(n)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_cache[labs(__pyx_v_n)]);
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":149
 * 
 * 
 * cdef number_t _intkambe(long n, number_t z, number_t eta, number_t *cache) nogil:             # <<<<<<<<<<<<<<
 *     r"""
 *     Cached recursion for the Kambe integral
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._intkambe", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe(long __pyx_v_n, double __pyx_v_z, double __pyx_v_eta, double *__pyx_v_cache) {
  double __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_integrals.pyx":165
 *         .. [#] `K. Kambe, Zeitschrift Fuer Naturforschung A 23, 9 (1968). <https://doi.org/10.1515/zna-1968-0908>`_
 *     """
 *     if not isnan(creal(cache[labs(n)])):             # <<<<<<<<<<<<<<
 *         return cache[labs(n)]
 *     if eta == 0:
 */
  __pyx_t_1 = ((!(isnan(creal(__pyx_t_double_complex_from_parts((__pyx_v_cache[labs(__pyx_v_n)]), 0))) != 0)) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":166
 *     """
 *     if not isnan(creal(cache[labs(n)])):
 *         return cache[labs(n)]             # <<<<<<<<<<<<<<
 *     if eta == 0:
 *         cache[labs(n)] = INFINITY
 */
    __pyx_r = (__pyx_v_cache[labs(__pyx_v_n)]);
    goto __pyx_L0;

    /* "treams/special/_integrals.pyx":165
 *         .. [#] `K. Kambe, Zeitschrift Fuer Naturforschung A 23, 9 (1968). <https://doi.org/10.1515/zna-1968-0908>`_
 *     """
 *     if not isnan(creal(cache[labs(n)])):             # <<<<<<<<<<<<<<
 *         return cache[labs(n)]
 *     if eta == 0:
 */
  }

  /* "treams/special/_integrals.pyx":167
 *     if not isnan(creal(cache[labs(n)])):
 *         return cache[labs(n)]
 *     if eta == 0:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = INFINITY
 *     # n < -4 is handled regularly
 */
  __pyx_t_1 = ((__pyx_v_eta == 0.0) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":168
 *         return cache[labs(n)]
 *     if eta == 0:
 *         cache[labs(n)] = INFINITY             # <<<<<<<<<<<<<<
 *     # n < -4 is handled regularly
 *     elif z == 0 and n == -4:
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = NPY_INFINITY;

    /* "treams/special/_integrals.pyx":167
 *     if not isnan(creal(cache[labs(n)])):
 *         return cache[labs(n)]
 *     if eta == 0:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = INFINITY
 *     # n < -4 is handled regularly
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":170
 *         cache[labs(n)] = INFINITY
 *     # n < -4 is handled regularly
 *     elif z == 0 and n == -4:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:
 */
  __pyx_t_2 = ((__pyx_v_z == 0.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_n == -4L) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":171
 *     # n < -4 is handled regularly
 *     elif z == 0 and n == -4:
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta             # <<<<<<<<<<<<<<
 *     elif z == 0 and n == -3:
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 */
    __pyx_t_3 = (__pyx_v_eta * __pyx_v_eta);
    if (unlikely(__pyx_t_3 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 171, __pyx_L1_error)
    }
    __pyx_t_4 = __pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp((0.5 / __pyx_t_3));
    if (unlikely(__pyx_v_eta == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 171, __pyx_L1_error)
    }
    (__pyx_v_cache[labs(__pyx_v_n)]) = ((-__pyx_fuse_1__pyx_f_6treams_7special_10_integrals_intkambe(-2L, 0.0, __pyx_v_eta)) + (__pyx_t_4 / __pyx_v_eta));

    /* "treams/special/_integrals.pyx":170
 *         cache[labs(n)] = INFINITY
 *     # n < -4 is handled regularly
 *     elif z == 0 and n == -4:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":172
 *     elif z == 0 and n == -4:
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 *     # n == -2 is handled regularly
 */
  __pyx_t_2 = ((__pyx_v_z == 0.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_n == -3L) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":173
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1             # <<<<<<<<<<<<<<
 *     # n == -2 is handled regularly
 *     elif z == 0 and n > -2:
 */
    __pyx_t_4 = ((2.0 * __pyx_v_eta) * __pyx_v_eta);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 173, __pyx_L1_error)
    }
    (__pyx_v_cache[labs(__pyx_v_n)]) = (__pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp((1.0 / __pyx_t_4)) - 1.0);

    /* "treams/special/_integrals.pyx":172
 *     elif z == 0 and n == -4:
 *         cache[labs(n)] = -intkambe(-2, 0, eta) + exp(0.5 / (eta * eta)) / eta
 *     elif z == 0 and n == -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 *     # n == -2 is handled regularly
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":175
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 *     # n == -2 is handled regularly
 *     elif z == 0 and n > -2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:
 */
  __pyx_t_2 = ((__pyx_v_z == 0.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_n > -2L) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":176
 *     # n == -2 is handled regularly
 *     elif z == 0 and n > -2:
 *         cache[labs(n)] = INFINITY             # <<<<<<<<<<<<<<
 *     elif n == -3:
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = NPY_INFINITY;

    /* "treams/special/_integrals.pyx":175
 *         cache[labs(n)] = exp(1 / (2 * eta * eta)) - 1
 *     # n == -2 is handled regularly
 *     elif z == 0 and n > -2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":177
 *     elif z == 0 and n > -2:
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:
 */
  __pyx_t_1 = ((__pyx_v_n == -3L) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":178
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:
 *         cache[labs(n)] = _intkambe_m3(z, eta)             # <<<<<<<<<<<<<<
 *     elif n == -2:
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m3(__pyx_v_z, __pyx_v_eta);

    /* "treams/special/_integrals.pyx":177
 *     elif z == 0 and n > -2:
 *         cache[labs(n)] = INFINITY
 *     elif n == -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":179
 *     elif n == -3:
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:
 */
  __pyx_t_1 = ((__pyx_v_n == -2L) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":180
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:
 *         cache[labs(n)] = _intkambe_m2(z, eta)             # <<<<<<<<<<<<<<
 *     elif n == -1:
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m2(__pyx_v_z, __pyx_v_eta);

    /* "treams/special/_integrals.pyx":179
 *     elif n == -3:
 *         cache[labs(n)] = _intkambe_m3(z, eta)
 *     elif n == -2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":181
 *     elif n == -2:
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:
 */
  __pyx_t_1 = ((__pyx_v_n == -1L) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":182
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:
 *         cache[labs(n)] = _intkambe_m1(z, eta)             # <<<<<<<<<<<<<<
 *     elif n == 0:
 *         cache[labs(n)] = _intkambe_0(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m1(__pyx_v_z, __pyx_v_eta);

    /* "treams/special/_integrals.pyx":181
 *     elif n == -2:
 *         cache[labs(n)] = _intkambe_m2(z, eta)
 *     elif n == -1:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":183
 *     elif n == -1:
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_0(z, eta)
 *     # Simplified recursion for n == 1
 */
  __pyx_t_1 = ((__pyx_v_n == 0) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":184
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:
 *         cache[labs(n)] = _intkambe_0(z, eta)             # <<<<<<<<<<<<<<
 *     # Simplified recursion for n == 1
 *     elif n == 1:
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_0(__pyx_v_z, __pyx_v_eta);

    /* "treams/special/_integrals.pyx":183
 *     elif n == -1:
 *         cache[labs(n)] = _intkambe_m1(z, eta)
 *     elif n == 0:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = _intkambe_0(z, eta)
 *     # Simplified recursion for n == 1
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":186
 *         cache[labs(n)] = _intkambe_0(z, eta)
 *     # Simplified recursion for n == 1
 *     elif n == 1:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 */
  __pyx_t_1 = ((__pyx_v_n == 1) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":188
 *     elif n == 1:
 *         cache[labs(n)] = (
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *             - _intkambe_m3(z, eta)
 *         ) / (z * z)
 */
    __pyx_t_4 = (__pyx_v_eta * __pyx_v_eta);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 188, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":189
 *         cache[labs(n)] = (
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *             - _intkambe_m3(z, eta)             # <<<<<<<<<<<<<<
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 */
    __pyx_t_3 = (__pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp(((((((-__pyx_v_z) * __pyx_v_z) * __pyx_v_eta) * __pyx_v_eta) + (1.0 / __pyx_t_4)) * 0.5)) - __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m3(__pyx_v_z, __pyx_v_eta));

    /* "treams/special/_integrals.pyx":190
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *             - _intkambe_m3(z, eta)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     # Needed to avoid false cache hits
 *     elif n == 2:
 */
    __pyx_t_4 = (__pyx_v_z * __pyx_v_z);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 190, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":187
 *     # Simplified recursion for n == 1
 *     elif n == 1:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *             - _intkambe_m3(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = (__pyx_t_3 / __pyx_t_4);

    /* "treams/special/_integrals.pyx":186
 *         cache[labs(n)] = _intkambe_0(z, eta)
 *     # Simplified recursion for n == 1
 *     elif n == 1:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":192
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 *     elif n == 2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             _intkambe(0, z, eta, cache)
 */
  __pyx_t_1 = ((__pyx_v_n == 2) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":196
 *             _intkambe(0, z, eta, cache)
 *             - _intkambe_m2(z, eta)
 *             + eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 */
    __pyx_t_4 = (__pyx_v_eta * __pyx_v_eta);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 196, __pyx_L1_error)
    }
    __pyx_t_3 = ((__pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe(0, __pyx_v_z, __pyx_v_eta, __pyx_v_cache) - __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m2(__pyx_v_z, __pyx_v_eta)) + (__pyx_v_eta * __pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp(((((((-__pyx_v_z) * __pyx_v_z) * __pyx_v_eta) * __pyx_v_eta) + (1.0 / __pyx_t_4)) * 0.5))));

    /* "treams/special/_integrals.pyx":197
 *             - _intkambe_m2(z, eta)
 *             + eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     # Needed to avoid false cache hits
 *     elif n == 3:
 */
    __pyx_t_4 = (__pyx_v_z * __pyx_v_z);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 197, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":193
 *     # Needed to avoid false cache hits
 *     elif n == 2:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             _intkambe(0, z, eta, cache)
 *             - _intkambe_m2(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = (__pyx_t_3 / __pyx_t_4);

    /* "treams/special/_integrals.pyx":192
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 *     elif n == 2:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             _intkambe(0, z, eta, cache)
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":199
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 *     elif n == 3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             2 * _intkambe(1, z, eta, cache)
 */
  __pyx_t_1 = ((__pyx_v_n == 3) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":203
 *             2 * _intkambe(1, z, eta, cache)
 *             - _intkambe_m1(z, eta)
 *             + eta * eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *         ) / (z * z)
 *     elif n < -3:
 */
    __pyx_t_4 = (__pyx_v_eta * __pyx_v_eta);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 203, __pyx_L1_error)
    }
    __pyx_t_3 = (((2.0 * __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe(1, __pyx_v_z, __pyx_v_eta, __pyx_v_cache)) - __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe_m1(__pyx_v_z, __pyx_v_eta)) + ((__pyx_v_eta * __pyx_v_eta) * __pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp(((((((-__pyx_v_z) * __pyx_v_z) * __pyx_v_eta) * __pyx_v_eta) + (1.0 / __pyx_t_4)) * 0.5))));

    /* "treams/special/_integrals.pyx":204
 *             - _intkambe_m1(z, eta)
 *             + eta * eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     elif n < -3:
 *         cache[labs(n)] = (
 */
    __pyx_t_4 = (__pyx_v_z * __pyx_v_z);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 204, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":200
 *     # Needed to avoid false cache hits
 *     elif n == 3:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             2 * _intkambe(1, z, eta, cache)
 *             - _intkambe_m1(z, eta)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = (__pyx_t_3 / __pyx_t_4);

    /* "treams/special/_integrals.pyx":199
 *         ) / (z * z)
 *     # Needed to avoid false cache hits
 *     elif n == 3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             2 * _intkambe(1, z, eta, cache)
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":205
 *             + eta * eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)
 *     elif n < -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             (n + 3) * _intkambe(n + 2, z, eta, cache)
 */
  __pyx_t_1 = ((__pyx_v_n < -3L) != 0);
  if (__pyx_t_1) {

    /* "treams/special/_integrals.pyx":209
 *             (n + 3) * _intkambe(n + 2, z, eta, cache)
 *             - z * z * _intkambe(n + 4, z, eta, cache)
 *             + pow(eta, n + 3) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *         )
 *     else:
 */
    __pyx_t_4 = (__pyx_v_eta * __pyx_v_eta);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 209, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":206
 *         ) / (z * z)
 *     elif n < -3:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             (n + 3) * _intkambe(n + 2, z, eta, cache)
 *             - z * z * _intkambe(n + 4, z, eta, cache)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = ((((__pyx_v_n + 3) * __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe((__pyx_v_n + 2), __pyx_v_z, __pyx_v_eta, __pyx_v_cache)) - ((__pyx_v_z * __pyx_v_z) * __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe((__pyx_v_n + 4), __pyx_v_z, __pyx_v_eta, __pyx_v_cache))) + (__pyx_fuse_1__pyx_f_6treams_7special_5_misc_pow(__pyx_v_eta, (__pyx_v_n + 3)) * __pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp(((((((-__pyx_v_z) * __pyx_v_z) * __pyx_v_eta) * __pyx_v_eta) + (1.0 / __pyx_t_4)) * 0.5))));

    /* "treams/special/_integrals.pyx":205
 *             + eta * eta * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)
 *     elif n < -3:             # <<<<<<<<<<<<<<
 *         cache[labs(n)] = (
 *             (n + 3) * _intkambe(n + 2, z, eta, cache)
 */
    goto __pyx_L4;
  }

  /* "treams/special/_integrals.pyx":216
 *             - _intkambe(n - 4, z, eta, cache)
 *             + pow(eta, n - 1) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     return cache[labs(n)]
 * 
 */
  /*else*/ {

    /* "treams/special/_integrals.pyx":215
 *             (n - 1) * _intkambe(n - 2, z, eta, cache)
 *             - _intkambe(n - 4, z, eta, cache)
 *             + pow(eta, n - 1) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)             # <<<<<<<<<<<<<<
 *         ) / (z * z)
 *     return cache[labs(n)]
 */
    __pyx_t_4 = (__pyx_v_eta * __pyx_v_eta);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 215, __pyx_L1_error)
    }
    __pyx_t_3 = ((((__pyx_v_n - 1) * __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe((__pyx_v_n - 2), __pyx_v_z, __pyx_v_eta, __pyx_v_cache)) - __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe((__pyx_v_n - 4), __pyx_v_z, __pyx_v_eta, __pyx_v_cache)) + (__pyx_fuse_1__pyx_f_6treams_7special_5_misc_pow(__pyx_v_eta, (__pyx_v_n - 1)) * __pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp(((((((-__pyx_v_z) * __pyx_v_z) * __pyx_v_eta) * __pyx_v_eta) + (1.0 / __pyx_t_4)) * 0.5))));

    /* "treams/special/_integrals.pyx":216
 *             - _intkambe(n - 4, z, eta, cache)
 *             + pow(eta, n - 1) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)             # <<<<<<<<<<<<<<
 *     return cache[labs(n)]
 * 
 */
    __pyx_t_4 = (__pyx_v_z * __pyx_v_z);
    if (unlikely(__pyx_t_4 == 0)) {
      #ifdef WITH_THREAD
      PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      #ifdef WITH_THREAD
      __Pyx_PyGILState_Release(__pyx_gilstate_save);
      #endif
      __PYX_ERR(0, 216, __pyx_L1_error)
    }

    /* "treams/special/_integrals.pyx":212
 *         )
 *     else:
 *         cache[labs(n)] = (             # <<<<<<<<<<<<<<
 *             (n - 1) * _intkambe(n - 2, z, eta, cache)
 *             - _intkambe(n - 4, z, eta, cache)
 */
    (__pyx_v_cache[labs(__pyx_v_n)]) = (__pyx_t_3 / __pyx_t_4);
  }
  __pyx_L4:;

  /* "treams/special/_integrals.pyx":217
 *             + pow(eta, n - 1) * exp((-z * z * eta * eta + 1 / (eta * eta)) * 0.5)
 *         ) / (z * z)
 *     return cache[labs(n)]             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = (__pyx_v_cache[labs(__pyx_v_n)]);
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":149
 * 
 * 
 * cdef number_t _intkambe(long n, number_t z, number_t eta, number_t *cache) nogil:             # <<<<<<<<<<<<<<
 *     r"""
 *     Cached recursion for the Kambe integral
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._integrals._intkambe", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_integrals.pyx":220
 * 
 * 
 * cdef number_t intkambe(long n, number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     r"""
 *     Integral appearing in the accelerated lattice summations.
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_7special_10_integrals_intkambe(long __pyx_v_n, __pyx_t_double_complex __pyx_v_z, __pyx_t_double_complex __pyx_v_eta) {
  __pyx_t_double_complex *__pyx_v_cache;
  long __pyx_v_i;
  __pyx_t_double_complex __pyx_v_res;
  __pyx_t_double_complex __pyx_r;
  long __pyx_t_1;
  long __pyx_t_2;
  long __pyx_t_3;

  /* "treams/special/_integrals.pyx":248
 *         .. [#] `K. Kambe, Zeitschrift Fuer Naturforschung A 23, 9 (1968). <https://doi.org/10.1515/zna-1968-0908>`_
 *     """
 *     cdef number_t *cache = <number_t*>malloc((labs(n) + 1) * sizeof(number_t))             # <<<<<<<<<<<<<<
 *     cdef long i
 *     for i in range(labs(n) + 1):
 */
  __pyx_v_cache = ((__pyx_t_double_complex *)malloc(((labs(__pyx_v_n) + 1) * (sizeof(__pyx_t_double_complex)))));

  /* "treams/special/_integrals.pyx":250
 *     cdef number_t *cache = <number_t*>malloc((labs(n) + 1) * sizeof(number_t))
 *     cdef long i
 *     for i in range(labs(n) + 1):             # <<<<<<<<<<<<<<
 *         cache[i] = NAN
 *     cdef number_t res = _intkambe(n, z, eta, cache)
 */
  __pyx_t_1 = (labs(__pyx_v_n) + 1);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "treams/special/_integrals.pyx":251
 *     cdef long i
 *     for i in range(labs(n) + 1):
 *         cache[i] = NAN             # <<<<<<<<<<<<<<
 *     cdef number_t res = _intkambe(n, z, eta, cache)
 *     free(cache)
 */
    (__pyx_v_cache[__pyx_v_i]) = __pyx_t_double_complex_from_parts(NPY_NAN, 0);
  }

  /* "treams/special/_integrals.pyx":252
 *     for i in range(labs(n) + 1):
 *         cache[i] = NAN
 *     cdef number_t res = _intkambe(n, z, eta, cache)             # <<<<<<<<<<<<<<
 *     free(cache)
 *     return res
 */
  __pyx_v_res = __pyx_fuse_0__pyx_f_6treams_7special_10_integrals__intkambe(__pyx_v_n, __pyx_v_z, __pyx_v_eta, __pyx_v_cache);

  /* "treams/special/_integrals.pyx":253
 *         cache[i] = NAN
 *     cdef number_t res = _intkambe(n, z, eta, cache)
 *     free(cache)             # <<<<<<<<<<<<<<
 *     return res
 */
  free(__pyx_v_cache);

  /* "treams/special/_integrals.pyx":254
 *     cdef number_t res = _intkambe(n, z, eta, cache)
 *     free(cache)
 *     return res             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":220
 * 
 * 
 * cdef number_t intkambe(long n, number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     r"""
 *     Integral appearing in the accelerated lattice summations.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static double __pyx_fuse_1__pyx_f_6treams_7special_10_integrals_intkambe(long __pyx_v_n, double __pyx_v_z, double __pyx_v_eta) {
  double *__pyx_v_cache;
  long __pyx_v_i;
  double __pyx_v_res;
  double __pyx_r;
  long __pyx_t_1;
  long __pyx_t_2;
  long __pyx_t_3;

  /* "treams/special/_integrals.pyx":248
 *         .. [#] `K. Kambe, Zeitschrift Fuer Naturforschung A 23, 9 (1968). <https://doi.org/10.1515/zna-1968-0908>`_
 *     """
 *     cdef number_t *cache = <number_t*>malloc((labs(n) + 1) * sizeof(number_t))             # <<<<<<<<<<<<<<
 *     cdef long i
 *     for i in range(labs(n) + 1):
 */
  __pyx_v_cache = ((double *)malloc(((labs(__pyx_v_n) + 1) * (sizeof(double)))));

  /* "treams/special/_integrals.pyx":250
 *     cdef number_t *cache = <number_t*>malloc((labs(n) + 1) * sizeof(number_t))
 *     cdef long i
 *     for i in range(labs(n) + 1):             # <<<<<<<<<<<<<<
 *         cache[i] = NAN
 *     cdef number_t res = _intkambe(n, z, eta, cache)
 */
  __pyx_t_1 = (labs(__pyx_v_n) + 1);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "treams/special/_integrals.pyx":251
 *     cdef long i
 *     for i in range(labs(n) + 1):
 *         cache[i] = NAN             # <<<<<<<<<<<<<<
 *     cdef number_t res = _intkambe(n, z, eta, cache)
 *     free(cache)
 */
    (__pyx_v_cache[__pyx_v_i]) = NPY_NAN;
  }

  /* "treams/special/_integrals.pyx":252
 *     for i in range(labs(n) + 1):
 *         cache[i] = NAN
 *     cdef number_t res = _intkambe(n, z, eta, cache)             # <<<<<<<<<<<<<<
 *     free(cache)
 *     return res
 */
  __pyx_v_res = __pyx_fuse_1__pyx_f_6treams_7special_10_integrals__intkambe(__pyx_v_n, __pyx_v_z, __pyx_v_eta, __pyx_v_cache);

  /* "treams/special/_integrals.pyx":253
 *         cache[i] = NAN
 *     cdef number_t res = _intkambe(n, z, eta, cache)
 *     free(cache)             # <<<<<<<<<<<<<<
 *     return res
 */
  free(__pyx_v_cache);

  /* "treams/special/_integrals.pyx":254
 *     cdef number_t res = _intkambe(n, z, eta, cache)
 *     free(cache)
 *     return res             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_res;
  goto __pyx_L0;

  /* "treams/special/_integrals.pyx":220
 * 
 * 
 * cdef number_t intkambe(long n, number_t z, number_t eta) nogil:             # <<<<<<<<<<<<<<
 *     r"""
 *     Integral appearing in the accelerated lattice summations.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_misc.pxd":29
 * cdef long ipow(long base, long exponent) nogil
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a > b else b)
 * 
 */

static CYTHON_INLINE double __pyx_fuse_0__pyx_f_6treams_7special_5_misc_max(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  double __pyx_t_1;

  /* "treams/special/_misc.pxd":30
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:
 *     return (a if a > b else b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 */
  if (((__pyx_v_a > __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":29
 * cdef long ipow(long base, long exponent) nogil
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a > b else b)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static CYTHON_INLINE long __pyx_fuse_1__pyx_f_6treams_7special_5_misc_max(long __pyx_v_a, long __pyx_v_b) {
  long __pyx_r;
  long __pyx_t_1;

  /* "treams/special/_misc.pxd":30
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:
 *     return (a if a > b else b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 */
  if (((__pyx_v_a > __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":29
 * cdef long ipow(long base, long exponent) nogil
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a > b else b)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

static CYTHON_INLINE double __pyx_fuse_0__pyx_f_6treams_7special_5_misc_min(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  double __pyx_t_1;

  /* "treams/special/_misc.pxd":33
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 *     return (a if a < b else b)             # <<<<<<<<<<<<<<
 */
  if (((__pyx_v_a < __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static CYTHON_INLINE long __pyx_fuse_1__pyx_f_6treams_7special_5_misc_min(long __pyx_v_a, long __pyx_v_b) {
  long __pyx_r;
  long __pyx_t_1;

  /* "treams/special/_misc.pxd":33
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 *     return (a if a < b else b)             # <<<<<<<<<<<<<<
 */
  if (((__pyx_v_a < __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__integrals(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__integrals},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_integrals",
    __pyx_k_Numerical_evaluation_of_integral, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_kp_u_is_neither_integer_nor_half_int, __pyx_k_is_neither_integer_nor_half_int, sizeof(__pyx_k_is_neither_integer_nor_half_int), 0, 1, 0, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 38, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 70, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("__pyx_fuse_0incgamma", (void (*)(void))__pyx_fuse_0__pyx_f_6treams_7special_10_integrals_incgamma, "__pyx_t_double_complex (double, __pyx_t_double_complex)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1incgamma", (void (*)(void))__pyx_fuse_1__pyx_f_6treams_7special_10_integrals_incgamma, "double (double, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_0intkambe", (void (*)(void))__pyx_fuse_0__pyx_f_6treams_7special_10_integrals_intkambe, "__pyx_t_double_complex (long, __pyx_t_double_complex, __pyx_t_double_complex)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1intkambe", (void (*)(void))__pyx_fuse_1__pyx_f_6treams_7special_10_integrals_intkambe, "double (long, double, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("treams.special._misc"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_0_29_36(__pyx_t_1, "SQPI", (void **)&__pyx_vp_6treams_7special_5_misc_SQPI, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("scipy.special.cython_special"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0erfc", (void (**)(void))&__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_erfc, "__pyx_t_double_complex (__pyx_t_double_complex, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1erfc", (void (**)(void))&__pyx_fuse_1__pyx_f_5scipy_7special_14cython_special_erfc, "double (double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0exp1", (void (**)(void))&__pyx_fuse_0__pyx_f_5scipy_7special_14cython_special_exp1, "__pyx_t_double_complex (__pyx_t_double_complex, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1exp1", (void (**)(void))&__pyx_fuse_1__pyx_f_5scipy_7special_14cython_special_exp1, "double (double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("treams.special._misc"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0exp", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_5_misc_exp, "__pyx_t_double_complex (__pyx_t_double_complex)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1exp", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_5_misc_exp, "double (double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0pow", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_5_misc_pow, "__pyx_t_double_complex (__pyx_t_double_complex, __pyx_t_double_complex)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1pow", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_5_misc_pow, "double (double, double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0sqrt", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_5_misc_sqrt, "__pyx_t_double_complex (__pyx_t_double_complex)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1sqrt", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_5_misc_sqrt, "double (double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_integrals(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_integrals(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__integrals(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__integrals(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__integrals(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_integrals' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__integrals(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_integrals", __pyx_methods, __pyx_k_Numerical_evaluation_of_integral, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_treams__special___integrals) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "treams.special._integrals")) {
      if (unlikely(PyDict_SetItemString(modules, "treams.special._integrals", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_init_code();
  (void)__Pyx_modinit_type_import_code();
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "treams/special/_integrals.pyx":1
 * """Numerical evaluation of integrals"""             # <<<<<<<<<<<<<<
 * 
 * cimport scipy.special.cython_special as cs
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init treams.special._integrals", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init treams.special._integrals");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr_0_29_36
#define __PYX_HAVE_RT_ImportVoidPtr_0_29_36
static int __Pyx_ImportVoidPtr_0_29_36(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_0_29_36
#define __PYX_HAVE_RT_ImportFunction_0_29_36
static int __Pyx_ImportFunction_0_29_36(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
