/* Generated by Cython 0.29.36 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include"
        ],
        "name": "treams.sw",
        "sources": [
            "src/treams/sw.pyx"
        ]
    },
    "module_name": "treams.sw"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_36"
#define CYTHON_HEX_VERSION 0x001D24F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS ((PY_VERSION_HEX >= 0x030600B1) && (PY_VERSION_HEX < 0x030C00A5))
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__treams__sw
#define __PYX_HAVE_API__treams__sw
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#include <math.h>
#include <stdlib.h>
#include <complex.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/treams/sw.pyx",
  "__init__.pxd",
  "type.pxd",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":690
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":691
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":693
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":697
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":698
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":700
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":704
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":705
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":714
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":715
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":716
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":718
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":719
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":720
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":722
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":723
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":725
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":726
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":727
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);


/*--- Type declarations ---*/
struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic;
struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr;
struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr;
struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr;
struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":729
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":730
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":731
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":733
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* "treams/sw.pyx":577
 * 
 * 
 * def translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, poltype=None, eta=0, func=lattice.lsumsw):             # <<<<<<<<<<<<<<
 *     """translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, helicity=True, eta=0)
 * 
 */
struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic {
  PyObject_HEAD
  PyObject *__pyx_v_in_;
  PyObject *__pyx_v_out;
};


/* "treams/sw.pyx":612
 *     if in_ is None:
 *         in_ = out
 *     out = (*(np.array(o) for o in out),)             # <<<<<<<<<<<<<<
 *     in_ = (*(np.array(i) for i in in_),)
 *     if len(out) < 3 or len(out) > 4:
 */
struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *__pyx_outer_scope;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "treams/sw.pyx":613
 *         in_ = out
 *     out = (*(np.array(o) for o in out),)
 *     in_ = (*(np.array(i) for i in in_),)             # <<<<<<<<<<<<<<
 *     if len(out) < 3 or len(out) > 4:
 *         raise ValueError(f"invalid length of output modes {len(out)}, must be 3 or 4")
 */
struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *__pyx_outer_scope;
  PyObject *__pyx_v_i;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "treams/sw.pyx":648
 *     if poltype == "helicity":
 *         return _translate_periodic_h(
 *             *(o[:, None] for o in out[1:]),             # <<<<<<<<<<<<<<
 *             *in_[1:],
 *             dlms[out[0][:, None], in_[0], 0, :],
 */
struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *__pyx_outer_scope;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "treams/sw.pyx":654
 *         )
 *     return _translate_periodic_p(
 *         *(o[:, None] for o in out[1:]),             # <<<<<<<<<<<<<<
 *         *in_[1:],
 *         dlms[out[0][:, None], in_[0], 0, :],
 */
struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr {
  PyObject_HEAD
  struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *__pyx_outer_scope;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_str(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_str(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_0_29_36
#define __PYX_HAVE_RT_ImportType_proto_0_29_36
#if __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if __STDC_VERSION__ >= 201112L || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_0_29_36(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_0_29_36(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_0_29_36 {
   __Pyx_ImportType_CheckSize_Error_0_29_36 = 0,
   __Pyx_ImportType_CheckSize_Warn_0_29_36 = 1,
   __Pyx_ImportType_CheckSize_Ignore_0_29_36 = 2
};
static PyTypeObject *__Pyx_ImportType_0_29_36(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_0_29_36 check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_0_29_36(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_0_29_36(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void); /*proto*/

/* Module declarations from 'libc.math' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'treams.special.cython_special' */
static double (*__pyx_f_6treams_7special_14cython_special_wigner3j)(long, long, long, long, long, long, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_pi_fun)(double, double, __pyx_t_double_complex, int __pyx_skip_dispatch); /*proto*/
static double (*__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_pi_fun)(double, double, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tau_fun)(double, double, __pyx_t_double_complex, int __pyx_skip_dispatch); /*proto*/
static double (*__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tau_fun)(double, double, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_A)(long, long, long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_A)(long, long, long, long, __pyx_t_double_complex, double, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_B)(long, long, long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_B)(long, long, long, long, __pyx_t_double_complex, double, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_rA)(long, long, long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_rA)(long, long, long, long, double, double, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_rB)(long, long, long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_rB)(long, long, long, long, double, double, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_wignerd)(long, long, long, double, __pyx_t_double_complex, double, int __pyx_skip_dispatch); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_wignerd)(long, long, long, double, double, double, int __pyx_skip_dispatch); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'treams.special._misc' */
static double *__pyx_vp_6treams_7special_5_misc_SQPI = 0;
#define __pyx_v_6treams_7special_5_misc_SQPI (*__pyx_vp_6treams_7special_5_misc_SQPI)
static long (*__pyx_f_6treams_7special_5_misc_minusonepow)(long); /*proto*/
static double (*__pyx_fuse_0__pyx_f_6treams_7special_5_misc_abs)(__pyx_t_double_complex); /*proto*/
static double (*__pyx_fuse_1__pyx_f_6treams_7special_5_misc_abs)(double); /*proto*/
static __pyx_t_double_complex (*__pyx_fuse_0__pyx_f_6treams_7special_5_misc_sqrt)(__pyx_t_double_complex); /*proto*/
static double (*__pyx_fuse_1__pyx_f_6treams_7special_5_misc_sqrt)(double); /*proto*/

/* Module declarations from 'treams.sw' */
static PyTypeObject *__pyx_ptype_6treams_2sw___pyx_scope_struct__translate_periodic = 0;
static PyTypeObject *__pyx_ptype_6treams_2sw___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_6treams_2sw___pyx_scope_struct_2_genexpr = 0;
static PyTypeObject *__pyx_ptype_6treams_2sw___pyx_scope_struct_3_genexpr = 0;
static PyTypeObject *__pyx_ptype_6treams_2sw___pyx_scope_struct_4_genexpr = 0;
static PyUFuncGenericFunction __pyx_v_6treams_2sw_ufunc_translate_loops[2];
static void *__pyx_v_6treams_2sw_ufunc_translate_sh_data[2];
static void *__pyx_v_6treams_2sw_ufunc_translate_rh_data[2];
static void *__pyx_v_6treams_2sw_ufunc_translate_sp_data[2];
static void *__pyx_v_6treams_2sw_ufunc_translate_rp_data[2];
static char __pyx_v_6treams_2sw_ufunc_translate_types[(2 * 10)];
static PyUFuncGenericFunction __pyx_v_6treams_2sw_ufunc_rotate_loops[1];
static void *__pyx_v_6treams_2sw_ufunc_rotate_data[1];
static char __pyx_v_6treams_2sw_ufunc_rotate_types[10];
static PyUFuncGenericFunction __pyx_v_6treams_2sw_gufunc_translate_periodic_h_loops[1];
static void *__pyx_v_6treams_2sw_gufunc_translate_periodic_h_data[1];
static char __pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[9];
static PyUFuncGenericFunction __pyx_v_6treams_2sw_gufunc_translate_periodic_p_loops[1];
static void *__pyx_v_6treams_2sw_gufunc_translate_periodic_p_data[1];
static char __pyx_v_6treams_2sw_gufunc_translate_periodic_p_types[8];
static PyUFuncGenericFunction __pyx_v_6treams_2sw_ufunc_pw_loops[2];
static void *__pyx_v_6treams_2sw_ufunc_pw_h_data[2];
static void *__pyx_v_6treams_2sw_ufunc_pw_p_data[2];
static char __pyx_v_6treams_2sw_ufunc_pw_types[(2 * 9)];
static PyUFuncGenericFunction __pyx_v_6treams_2sw_ufunc_cw_loops[1];
static void *__pyx_v_6treams_2sw_ufunc_cw_h_data[1];
static void *__pyx_v_6treams_2sw_ufunc_cw_p_data[1];
static char __pyx_v_6treams_2sw_ufunc_cw_types[9];
static void __pyx_f_6treams_2sw__loop_D_llllllddd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_2sw__loop_D_llllllDdd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static __pyx_t_double_complex __pyx_f_6treams_2sw__crotate(long, long, long, long, long, long, double, double, double); /*proto*/
static __pyx_t_double_complex __pyx_f_6treams_2sw__transl_A_lattice(long, long, long, long, __pyx_t_double_complex *, long); /*proto*/
static __pyx_t_double_complex __pyx_f_6treams_2sw__transl_B_lattice(long, long, long, long, __pyx_t_double_complex *, long); /*proto*/
static __pyx_t_double_complex __pyx_f_6treams_2sw__ctranslate_periodic_p(long, long, long, long, long, long, __pyx_t_double_complex *, long); /*proto*/
static __pyx_t_double_complex __pyx_f_6treams_2sw__ctranslate_periodic_h(long, long, long, long, long, long, __pyx_t_double_complex *, long); /*proto*/
static void __pyx_f_6treams_2sw__loop_translate_periodic_h(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_2sw__loop_translate_periodic_p(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_2sw__loop_periodic_d(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_2sw__loop_periodic_D(char **, npy_intp *, npy_intp *, void *); /*proto*/
static __pyx_t_double_complex __pyx_f_6treams_2sw__cperiodic_to_cw_h(double, long, long, long, long, long, __pyx_t_double_complex, double); /*proto*/
static __pyx_t_double_complex __pyx_f_6treams_2sw__cperiodic_to_cw_p(double, long, long, long, long, long, __pyx_t_double_complex, double); /*proto*/
static void __pyx_f_6treams_2sw__loop_periodic_cw_D(char **, npy_intp *, npy_intp *, void *); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_sh(long, long, long, long, long, long, __pyx_t_double_complex, double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_sh(long, long, long, long, long, long, double, double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_rh(long, long, long, long, long, long, __pyx_t_double_complex, double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_rh(long, long, long, long, long, long, double, double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_sp(long, long, long, long, long, long, __pyx_t_double_complex, double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_sp(long, long, long, long, long, long, double, double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_rp(long, long, long, long, long, long, __pyx_t_double_complex, double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_rp(long, long, long, long, long, long, double, double, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__cperiodic_to_pw_h(double, double, __pyx_t_double_complex, long, long, long, long, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__cperiodic_to_pw_h(double, double, double, long, long, long, long, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__cperiodic_to_pw_p(double, double, __pyx_t_double_complex, long, long, long, long, double); /*proto*/
static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__cperiodic_to_pw_p(double, double, double, long, long, long, long, double); /*proto*/
#define __Pyx_MODULE_NAME "treams.sw"
extern int __pyx_module_is_main_treams__sw;
int __pyx_module_is_main_treams__sw = 0;

/* Implementation of 'treams.sw' */
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_a[] = "a";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_l[] = "l";
static const char __pyx_k_m[] = "m";
static const char __pyx_k__2[] = "'";
static const char __pyx_k_in[] = "in_";
static const char __pyx_k_kr[] = "kr";
static const char __pyx_k_ks[] = "ks";
static const char __pyx_k_kx[] = "kx";
static const char __pyx_k_ky[] = "ky";
static const char __pyx_k_kz[] = "kz";
static const char __pyx_k_mu[] = "mu";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_rs[] = "rs";
static const char __pyx_k_dim[] = "dim";
static const char __pyx_k_eta[] = "eta";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_phi[] = "phi";
static const char __pyx_k_pol[] = "pol";
static const char __pyx_k_psi[] = "psi";
static const char __pyx_k_qol[] = "qol";
static const char __pyx_k_TODO[] = "TODO";
static const char __pyx_k_area[] = "area";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_dlms[] = "dlms";
static const char __pyx_k_func[] = "func";
static const char __pyx_k_kpar[] = "kpar";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_rsin[] = "rsin";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_array[] = "array";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_modes[] = "modes";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_theta[] = "theta";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_config[] = "config";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_lambda[] = "lambda_";
static const char __pyx_k_lsumsw[] = "lsumsw";
static const char __pyx_k_parity[] = "parity";
static const char __pyx_k_rotate[] = "_rotate";
static const char __pyx_k_rsdiff[] = "rsdiff";
static const char __pyx_k_treams[] = "treams";
static const char __pyx_k_POLTYPE[] = "POLTYPE";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_lattice[] = "lattice";
static const char __pyx_k_poltype[] = "poltype";
static const char __pyx_k_reshape[] = "reshape";
static const char __pyx_k_helicity[] = "helicity";
static const char __pyx_k_rotate_2[] = "rotate";
static const char __pyx_k_singular[] = "singular";
static const char __pyx_k_translate[] = "translate";
static const char __pyx_k_treams_sw[] = "treams.sw";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_zeros_like[] = "zeros_like";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_translate_rh[] = "_translate_rh";
static const char __pyx_k_translate_rp[] = "_translate_rp";
static const char __pyx_k_translate_sh[] = "_translate_sh";
static const char __pyx_k_translate_sp[] = "_translate_sp";
static const char __pyx_k_must_be_3_or_4[] = ", must be 3 or 4";
static const char __pyx_k_periodic_to_cw[] = "periodic_to_cw";
static const char __pyx_k_periodic_to_pw[] = "periodic_to_pw";
static const char __pyx_k_invalid_poltype[] = "invalid poltype '";
static const char __pyx_k_periodic_to_cw_h[] = "_periodic_to_cw_h";
static const char __pyx_k_periodic_to_cw_p[] = "_periodic_to_cw_p";
static const char __pyx_k_periodic_to_pw_h[] = "_periodic_to_pw_h";
static const char __pyx_k_periodic_to_pw_p[] = "_periodic_to_pw_p";
static const char __pyx_k_src_treams_sw_pyx[] = "src/treams/sw.pyx";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_translate_periodic[] = "translate_periodic";
static const char __pyx_k_translate_periodic_h[] = "_translate_periodic_h";
static const char __pyx_k_translate_periodic_p[] = "_translate_periodic_p";
static const char __pyx_k_invalid_length_of_input_modes[] = "invalid length of input modes ";
static const char __pyx_k_invalid_length_of_output_modes[] = "invalid length of output modes ";
static const char __pyx_k_Polarization_must_be_defined_by[] = "Polarization must be defined by 0 or 1";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_Spherical_wave_module_autosummar[] = "Spherical wave module.\n\n.. autosummary::\n   :toctree:\n\n   periodic_to_cw\n   periodic_to_pw\n   rotate\n   translate\n   translate_periodic\n\n";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_translate_periodic_locals_genexp[] = "translate_periodic.<locals>.genexpr";
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_POLTYPE;
static PyObject *__pyx_kp_u_Polarization_must_be_defined_by;
static PyObject *__pyx_n_u_TODO;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_u__2;
static PyObject *__pyx_n_s_a;
static PyObject *__pyx_n_s_area;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_array;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_config;
static PyObject *__pyx_n_s_dim;
static PyObject *__pyx_n_s_dlms;
static PyObject *__pyx_n_s_eta;
static PyObject *__pyx_n_s_func;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_u_helicity;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_in;
static PyObject *__pyx_kp_u_invalid_length_of_input_modes;
static PyObject *__pyx_kp_u_invalid_length_of_output_modes;
static PyObject *__pyx_kp_u_invalid_poltype;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_kpar;
static PyObject *__pyx_n_s_kr;
static PyObject *__pyx_n_s_ks;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_kx;
static PyObject *__pyx_n_s_ky;
static PyObject *__pyx_n_s_kz;
static PyObject *__pyx_n_s_l;
static PyObject *__pyx_n_s_lambda;
static PyObject *__pyx_n_s_lattice;
static PyObject *__pyx_n_s_lsumsw;
static PyObject *__pyx_n_s_m;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_modes;
static PyObject *__pyx_n_s_mu;
static PyObject *__pyx_kp_u_must_be_3_or_4;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_u_parity;
static PyObject *__pyx_n_s_periodic_to_cw;
static PyObject *__pyx_n_s_periodic_to_cw_h;
static PyObject *__pyx_n_s_periodic_to_cw_p;
static PyObject *__pyx_n_s_periodic_to_pw;
static PyObject *__pyx_n_s_periodic_to_pw_h;
static PyObject *__pyx_n_s_periodic_to_pw_p;
static PyObject *__pyx_n_s_phi;
static PyObject *__pyx_n_s_pol;
static PyObject *__pyx_n_s_poltype;
static PyObject *__pyx_n_s_psi;
static PyObject *__pyx_n_s_qol;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reshape;
static PyObject *__pyx_n_s_rotate;
static PyObject *__pyx_n_s_rotate_2;
static PyObject *__pyx_n_s_rs;
static PyObject *__pyx_n_s_rsdiff;
static PyObject *__pyx_n_s_rsin;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_singular;
static PyObject *__pyx_kp_s_src_treams_sw_pyx;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_theta;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_translate;
static PyObject *__pyx_n_s_translate_periodic;
static PyObject *__pyx_n_s_translate_periodic_h;
static PyObject *__pyx_n_s_translate_periodic_locals_genexp;
static PyObject *__pyx_n_s_translate_periodic_p;
static PyObject *__pyx_n_s_translate_rh;
static PyObject *__pyx_n_s_translate_rp;
static PyObject *__pyx_n_s_translate_sh;
static PyObject *__pyx_n_s_translate_sp;
static PyObject *__pyx_n_s_treams;
static PyObject *__pyx_n_s_treams_sw;
static PyObject *__pyx_n_s_zeros_like;
static PyObject *__pyx_pf_6treams_2sw_translate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_lambda_, PyObject *__pyx_v_mu, PyObject *__pyx_v_pol, PyObject *__pyx_v_l, PyObject *__pyx_v_m, PyObject *__pyx_v_qol, PyObject *__pyx_v_kr, PyObject *__pyx_v_theta, PyObject *__pyx_v_phi, PyObject *__pyx_v_poltype, PyObject *__pyx_v_singular, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6treams_2sw_2rotate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_lambda_, PyObject *__pyx_v_mu, PyObject *__pyx_v_pol, PyObject *__pyx_v_l, PyObject *__pyx_v_m, PyObject *__pyx_v_qol, PyObject *__pyx_v_phi, PyObject *__pyx_v_theta, PyObject *__pyx_v_psi, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6treams_2sw_18translate_periodic_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6treams_2sw_18translate_periodic_3genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6treams_2sw_18translate_periodic_6genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6treams_2sw_18translate_periodic_9genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_6treams_2sw_4translate_periodic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ks, PyObject *__pyx_v_kpar, PyObject *__pyx_v_a, PyObject *__pyx_v_rs, PyObject *__pyx_v_out, PyObject *__pyx_v_in_, PyObject *__pyx_v_rsin, PyObject *__pyx_v_poltype, PyObject *__pyx_v_eta, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_6treams_2sw_6periodic_to_pw(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_kx, PyObject *__pyx_v_ky, PyObject *__pyx_v_kz, PyObject *__pyx_v_pol, PyObject *__pyx_v_l, PyObject *__pyx_v_m, PyObject *__pyx_v_qol, PyObject *__pyx_v_area, PyObject *__pyx_v_poltype, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_6treams_2sw_8periodic_to_cw(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_kz, PyObject *__pyx_v_m, PyObject *__pyx_v_pol, PyObject *__pyx_v_l, PyObject *__pyx_v_mu, PyObject *__pyx_v_qol, PyObject *__pyx_v_k, PyObject *__pyx_v_area, PyObject *__pyx_v_poltype, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct__translate_periodic(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct_4_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_k__4;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_slice__5;
static PyObject *__pyx_slice__6;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__11;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
/* Late includes */

/* "treams/sw.pyx":41
 * 
 * 
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_sh(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol, long __pyx_v_l, long __pyx_v_m, long __pyx_v_qol, __pyx_t_double_complex __pyx_v_kr, double __pyx_v_theta, double __pyx_v_phi) {
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_0_ctranslate_sh", 1);

  /* "treams/sw.pyx":42
 * 
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
  /*try:*/ {
    __pyx_t_1 = ((__pyx_fuse_0__pyx_f_6treams_7special_5_misc_abs(__pyx_v_kr) < 1e-16) != 0);
    if (__pyx_t_1) {

      /* "treams/sw.pyx":43
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi) + (
 */
      __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":42
 * 
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
    }

    /* "treams/sw.pyx":44
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
    __pyx_t_2 = ((__pyx_v_pol == __pyx_v_qol) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L8_bool_binop_done;
    }
    switch (__pyx_v_pol) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":45
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi) + (             # <<<<<<<<<<<<<<
 *             2 * pol - 1
 *         ) * cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __Pyx_c_sum_double(__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_A(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_v_theta, __pyx_v_phi, 0), __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((2 * __pyx_v_pol) - 1), 0), __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_B(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_v_theta, __pyx_v_phi, 0)));
      goto __pyx_L3_return;

      /* "treams/sw.pyx":44
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
    }

    /* "treams/sw.pyx":48
 *             2 * pol - 1
 *         ) * cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    __pyx_t_2 = ((__pyx_v_pol == 1) != 0);
    if (!__pyx_t_2) {
      goto __pyx_L12_next_or;
    } else {
    }
    __pyx_t_2 = ((__pyx_v_qol == 0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_L12_next_or:;
    __pyx_t_2 = ((__pyx_v_pol == 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_qol == 1) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":49
 *         ) * cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     raise ValueError("Polarization must be defined by 0 or 1")
 * 
 */
      __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":48
 *             2 * pol - 1
 *         ) * cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    }

    /* "treams/sw.pyx":50
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 50, __pyx_L16_error)
        }
        /*finally:*/ {
          __pyx_L16_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }

  /* "treams/sw.pyx":42
 * 
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/sw.pyx":41
 * 
 * 
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.sw._ctranslate_sh", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_sh(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol, long __pyx_v_l, long __pyx_v_m, long __pyx_v_qol, double __pyx_v_kr, double __pyx_v_theta, double __pyx_v_phi) {
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_1_ctranslate_sh", 1);

  /* "treams/sw.pyx":42
 * 
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
  /*try:*/ {
    __pyx_t_1 = ((__pyx_fuse_1__pyx_f_6treams_7special_5_misc_abs(__pyx_v_kr) < 1e-16) != 0);
    if (__pyx_t_1) {

      /* "treams/sw.pyx":43
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi) + (
 */
      __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":42
 * 
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
    }

    /* "treams/sw.pyx":44
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
    __pyx_t_2 = ((__pyx_v_pol == __pyx_v_qol) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L8_bool_binop_done;
    }
    switch (__pyx_v_pol) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":45
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi) + (             # <<<<<<<<<<<<<<
 *             2 * pol - 1
 *         ) * cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __Pyx_c_sum_double(__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_A(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_t_double_complex_from_parts(__pyx_v_kr, 0), __pyx_v_theta, __pyx_v_phi, 0), __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((2 * __pyx_v_pol) - 1), 0), __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_B(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_t_double_complex_from_parts(__pyx_v_kr, 0), __pyx_v_theta, __pyx_v_phi, 0)));
      goto __pyx_L3_return;

      /* "treams/sw.pyx":44
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
    }

    /* "treams/sw.pyx":48
 *             2 * pol - 1
 *         ) * cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    __pyx_t_2 = ((__pyx_v_pol == 1) != 0);
    if (!__pyx_t_2) {
      goto __pyx_L12_next_or;
    } else {
    }
    __pyx_t_2 = ((__pyx_v_qol == 0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_L12_next_or:;
    __pyx_t_2 = ((__pyx_v_pol == 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_qol == 1) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":49
 *         ) * cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     raise ValueError("Polarization must be defined by 0 or 1")
 * 
 */
      __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":48
 *             2 * pol - 1
 *         ) * cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    }

    /* "treams/sw.pyx":50
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 50, __pyx_L16_error)
        }
        /*finally:*/ {
          __pyx_L16_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }

  /* "treams/sw.pyx":42
 * 
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/sw.pyx":41
 * 
 * 
 * cdef double complex _ctranslate_sh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.sw._ctranslate_sh", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "treams/sw.pyx":53
 * 
 * 
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_rh(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol, long __pyx_v_l, long __pyx_v_m, long __pyx_v_qol, __pyx_t_double_complex __pyx_v_kr, double __pyx_v_theta, double __pyx_v_phi) {
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_0_ctranslate_rh", 1);

  /* "treams/sw.pyx":54
 * 
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
  /*try:*/ {
    __pyx_t_2 = ((__pyx_v_pol == __pyx_v_qol) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L7_bool_binop_done;
    }
    switch (__pyx_v_pol) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":55
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (             # <<<<<<<<<<<<<<
 *             2 * pol - 1
 *         ) * cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __Pyx_c_sum_double(__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_rA(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_t_double_complex_from_parts(__pyx_v_theta, 0), __pyx_v_phi, 0), __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((2 * __pyx_v_pol) - 1), 0), __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_rB(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_t_double_complex_from_parts(__pyx_v_theta, 0), __pyx_v_phi, 0)));
      goto __pyx_L3_return;

      /* "treams/sw.pyx":54
 * 
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
    }

    /* "treams/sw.pyx":58
 *             2 * pol - 1
 *         ) * cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    __pyx_t_2 = ((__pyx_v_pol == 1) != 0);
    if (!__pyx_t_2) {
      goto __pyx_L11_next_or;
    } else {
    }
    __pyx_t_2 = ((__pyx_v_qol == 0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_L11_next_or:;
    __pyx_t_2 = ((__pyx_v_pol == 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_qol == 1) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":59
 *         ) * cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     raise ValueError("Polarization must be defined by 0 or 1")
 * 
 */
      __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":58
 *             2 * pol - 1
 *         ) * cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    }

    /* "treams/sw.pyx":60
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 60, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 60, __pyx_L15_error)
        }
        /*finally:*/ {
          __pyx_L15_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }

  /* "treams/sw.pyx":54
 * 
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/sw.pyx":53
 * 
 * 
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.sw._ctranslate_rh", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_rh(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol, long __pyx_v_l, long __pyx_v_m, long __pyx_v_qol, double __pyx_v_kr, double __pyx_v_theta, double __pyx_v_phi) {
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_1_ctranslate_rh", 1);

  /* "treams/sw.pyx":54
 * 
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
  /*try:*/ {
    __pyx_t_2 = ((__pyx_v_pol == __pyx_v_qol) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L7_bool_binop_done;
    }
    switch (__pyx_v_pol) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":55
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (             # <<<<<<<<<<<<<<
 *             2 * pol - 1
 *         ) * cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __Pyx_c_sum_double(__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_rA(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_v_theta, __pyx_v_phi, 0), __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((2 * __pyx_v_pol) - 1), 0), __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_rB(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_v_theta, __pyx_v_phi, 0)));
      goto __pyx_L3_return;

      /* "treams/sw.pyx":54
 * 
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
    }

    /* "treams/sw.pyx":58
 *             2 * pol - 1
 *         ) * cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    __pyx_t_2 = ((__pyx_v_pol == 1) != 0);
    if (!__pyx_t_2) {
      goto __pyx_L11_next_or;
    } else {
    }
    __pyx_t_2 = ((__pyx_v_qol == 0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_L11_next_or:;
    __pyx_t_2 = ((__pyx_v_pol == 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_qol == 1) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":59
 *         ) * cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     raise ValueError("Polarization must be defined by 0 or 1")
 * 
 */
      __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":58
 *             2 * pol - 1
 *         ) * cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    }

    /* "treams/sw.pyx":60
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 60, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 60, __pyx_L15_error)
        }
        /*finally:*/ {
          __pyx_L15_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }

  /* "treams/sw.pyx":54
 * 
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (
 *             2 * pol - 1
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/sw.pyx":53
 * 
 * 
 * cdef double complex _ctranslate_rh(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi) + (
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.sw._ctranslate_rh", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "treams/sw.pyx":63
 * 
 * 
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_sp(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol, long __pyx_v_l, long __pyx_v_m, long __pyx_v_qol, __pyx_t_double_complex __pyx_v_kr, double __pyx_v_theta, double __pyx_v_phi) {
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_0_ctranslate_sp", 1);

  /* "treams/sw.pyx":64
 * 
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
  /*try:*/ {
    __pyx_t_1 = ((__pyx_fuse_0__pyx_f_6treams_7special_5_misc_abs(__pyx_v_kr) < 1e-16) != 0);
    if (__pyx_t_1) {

      /* "treams/sw.pyx":65
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":64
 * 
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
    }

    /* "treams/sw.pyx":66
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
    __pyx_t_2 = ((__pyx_v_pol == __pyx_v_qol) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L8_bool_binop_done;
    }
    switch (__pyx_v_pol) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":67
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)             # <<<<<<<<<<<<<<
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_A(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_v_theta, __pyx_v_phi, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":66
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
    }

    /* "treams/sw.pyx":68
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    __pyx_t_2 = ((__pyx_v_pol == 1) != 0);
    if (!__pyx_t_2) {
      goto __pyx_L12_next_or;
    } else {
    }
    __pyx_t_2 = ((__pyx_v_qol == 0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_L12_next_or:;
    __pyx_t_2 = ((__pyx_v_pol == 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_qol == 1) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":69
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)             # <<<<<<<<<<<<<<
 *     raise ValueError("Polarization must be defined by 0 or 1")
 * 
 */
      __pyx_r = __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_B(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_v_theta, __pyx_v_phi, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":68
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    }

    /* "treams/sw.pyx":70
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 70, __pyx_L16_error)
        }
        /*finally:*/ {
          __pyx_L16_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }

  /* "treams/sw.pyx":64
 * 
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/sw.pyx":63
 * 
 * 
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.sw._ctranslate_sp", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_sp(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol, long __pyx_v_l, long __pyx_v_m, long __pyx_v_qol, double __pyx_v_kr, double __pyx_v_theta, double __pyx_v_phi) {
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_1_ctranslate_sp", 1);

  /* "treams/sw.pyx":64
 * 
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
  /*try:*/ {
    __pyx_t_1 = ((__pyx_fuse_1__pyx_f_6treams_7special_5_misc_abs(__pyx_v_kr) < 1e-16) != 0);
    if (__pyx_t_1) {

      /* "treams/sw.pyx":65
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":64
 * 
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
    }

    /* "treams/sw.pyx":66
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
    __pyx_t_2 = ((__pyx_v_pol == __pyx_v_qol) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L8_bool_binop_done;
    }
    switch (__pyx_v_pol) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L8_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":67
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)             # <<<<<<<<<<<<<<
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_A(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_t_double_complex_from_parts(__pyx_v_kr, 0), __pyx_v_theta, __pyx_v_phi, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":66
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
    }

    /* "treams/sw.pyx":68
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    __pyx_t_2 = ((__pyx_v_pol == 1) != 0);
    if (!__pyx_t_2) {
      goto __pyx_L12_next_or;
    } else {
    }
    __pyx_t_2 = ((__pyx_v_qol == 0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_L12_next_or:;
    __pyx_t_2 = ((__pyx_v_pol == 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_qol == 1) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":69
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)             # <<<<<<<<<<<<<<
 *     raise ValueError("Polarization must be defined by 0 or 1")
 * 
 */
      __pyx_r = __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_B(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_t_double_complex_from_parts(__pyx_v_kr, 0), __pyx_v_theta, __pyx_v_phi, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":68
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_A(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    }

    /* "treams/sw.pyx":70
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_B(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 70, __pyx_L16_error)
        }
        /*finally:*/ {
          __pyx_L16_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }

  /* "treams/sw.pyx":64
 * 
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if abs(kr) < 1e-16:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     if pol == qol and (pol == 0 or pol == 1):
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/sw.pyx":63
 * 
 * 
 * cdef double complex _ctranslate_sp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if abs(kr) < 1e-16:
 *         return 0.0j
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.sw._ctranslate_sp", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "treams/sw.pyx":73
 * 
 * 
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_rp(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol, long __pyx_v_l, long __pyx_v_m, long __pyx_v_qol, __pyx_t_double_complex __pyx_v_kr, double __pyx_v_theta, double __pyx_v_phi) {
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_0_ctranslate_rp", 1);

  /* "treams/sw.pyx":74
 * 
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
  /*try:*/ {
    __pyx_t_2 = ((__pyx_v_pol == __pyx_v_qol) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L7_bool_binop_done;
    }
    switch (__pyx_v_pol) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":75
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)             # <<<<<<<<<<<<<<
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_rA(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_t_double_complex_from_parts(__pyx_v_theta, 0), __pyx_v_phi, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":74
 * 
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
    }

    /* "treams/sw.pyx":76
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    __pyx_t_2 = ((__pyx_v_pol == 1) != 0);
    if (!__pyx_t_2) {
      goto __pyx_L11_next_or;
    } else {
    }
    __pyx_t_2 = ((__pyx_v_qol == 0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_L11_next_or:;
    __pyx_t_2 = ((__pyx_v_pol == 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_qol == 1) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":77
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)             # <<<<<<<<<<<<<<
 *     raise ValueError("Polarization must be defined by 0 or 1")
 * 
 */
      __pyx_r = __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_rB(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_t_double_complex_from_parts(__pyx_v_theta, 0), __pyx_v_phi, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":76
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    }

    /* "treams/sw.pyx":78
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 78, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 78, __pyx_L15_error)
        }
        /*finally:*/ {
          __pyx_L15_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }

  /* "treams/sw.pyx":74
 * 
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/sw.pyx":73
 * 
 * 
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.sw._ctranslate_rp", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_rp(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol, long __pyx_v_l, long __pyx_v_m, long __pyx_v_qol, double __pyx_v_kr, double __pyx_v_theta, double __pyx_v_phi) {
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("__pyx_fuse_1_ctranslate_rp", 1);

  /* "treams/sw.pyx":74
 * 
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
  /*try:*/ {
    __pyx_t_2 = ((__pyx_v_pol == __pyx_v_qol) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L7_bool_binop_done;
    }
    switch (__pyx_v_pol) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":75
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)             # <<<<<<<<<<<<<<
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 */
      __pyx_r = __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_rA(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_v_theta, __pyx_v_phi, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":74
 * 
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
    }

    /* "treams/sw.pyx":76
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    __pyx_t_2 = ((__pyx_v_pol == 1) != 0);
    if (!__pyx_t_2) {
      goto __pyx_L11_next_or;
    } else {
    }
    __pyx_t_2 = ((__pyx_v_qol == 0) != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_L11_next_or:;
    __pyx_t_2 = ((__pyx_v_pol == 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_qol == 1) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":77
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)             # <<<<<<<<<<<<<<
 *     raise ValueError("Polarization must be defined by 0 or 1")
 * 
 */
      __pyx_r = __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_rB(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_kr, __pyx_v_theta, __pyx_v_phi, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":76
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")
 */
    }

    /* "treams/sw.pyx":78
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return cs.tl_vsw_rB(lambda_, mu, l, m, kr, theta, phi)
 *     raise ValueError("Polarization must be defined by 0 or 1")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 78, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_Raise(__pyx_t_3, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __PYX_ERR(0, 78, __pyx_L15_error)
        }
        /*finally:*/ {
          __pyx_L15_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }

  /* "treams/sw.pyx":74
 * 
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:
 *     if pol == qol and (pol == 0 or pol == 1):             # <<<<<<<<<<<<<<
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/sw.pyx":73
 * 
 * 
 * cdef double complex _ctranslate_rp(long lambda_, long mu, long pol, long l, long m, long qol, number_t kr, double theta, double phi) nogil:             # <<<<<<<<<<<<<<
 *     if pol == qol and (pol == 0 or pol == 1):
 *         return cs.tl_vsw_rA(lambda_, mu, l, m, kr, theta, phi)
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("treams.sw._ctranslate_rp", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "treams/sw.pyx":81
 * 
 * 
 * cdef void _loop_D_llllllddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

static void __pyx_f_6treams_2sw__loop_D_llllllddd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_ip7;
  char *__pyx_v_ip8;
  char *__pyx_v_op0;
  __pyx_t_double_complex __pyx_v_ov0;
  npy_intp __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;

  /* "treams/sw.pyx":82
 * 
 * cdef void _loop_D_llllllddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/sw.pyx":83
 * cdef void _loop_D_llllllddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/sw.pyx":84
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/sw.pyx":85
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/sw.pyx":86
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/sw.pyx":87
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/sw.pyx":88
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/sw.pyx":89
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/sw.pyx":90
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *ip7 = args[7]
 *     cdef char *ip8 = args[8]
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/sw.pyx":91
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]             # <<<<<<<<<<<<<<
 *     cdef char *ip8 = args[8]
 *     cdef char *op0 = args[9]
 */
  __pyx_v_ip7 = (__pyx_v_args[7]);

  /* "treams/sw.pyx":92
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 *     cdef char *ip8 = args[8]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[9]
 *     cdef double complex ov0
 */
  __pyx_v_ip8 = (__pyx_v_args[8]);

  /* "treams/sw.pyx":93
 *     cdef char *ip7 = args[7]
 *     cdef char *ip8 = args[8]
 *     cdef char *op0 = args[9]             # <<<<<<<<<<<<<<
 *     cdef double complex ov0
 *     for i in range(n):
 */
  __pyx_v_op0 = (__pyx_v_args[9]);

  /* "treams/sw.pyx":95
 *     cdef char *op0 = args[9]
 *     cdef double complex ov0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ov0 = (<double complex(*)(long, long, long, long, long, long, double, double, double) nogil>func)(
 *             <long>(<long*>ip0)[0],
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "treams/sw.pyx":96
 *     cdef double complex ov0
 *     for i in range(n):
 *         ov0 = (<double complex(*)(long, long, long, long, long, long, double, double, double) nogil>func)(             # <<<<<<<<<<<<<<
 *             <long>(<long*>ip0)[0],
 *             <long>(<long*>ip1)[0],
 */
    __pyx_v_ov0 = ((__pyx_t_double_complex (*)(long, long, long, long, long, long, double, double, double))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((long)(((long *)__pyx_v_ip2)[0])), ((long)(((long *)__pyx_v_ip3)[0])), ((long)(((long *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), ((double)(((double *)__pyx_v_ip6)[0])), ((double)(((double *)__pyx_v_ip7)[0])), ((double)(((double *)__pyx_v_ip8)[0])));

    /* "treams/sw.pyx":107
 *             <double>(<double*>ip8)[0],
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    (((__pyx_t_double_complex *)__pyx_v_op0)[0]) = __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL(__pyx_v_ov0)), ((double)__Pyx_CIMAG(__pyx_v_ov0)));

    /* "treams/sw.pyx":108
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/sw.pyx":109
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/sw.pyx":110
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/sw.pyx":111
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/sw.pyx":112
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/sw.pyx":113
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/sw.pyx":114
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         ip7 += steps[7]
 *         ip8 += steps[8]
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/sw.pyx":115
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         ip7 += steps[7]             # <<<<<<<<<<<<<<
 *         ip8 += steps[8]
 *         op0 += steps[9]
 */
    __pyx_v_ip7 = (__pyx_v_ip7 + (__pyx_v_steps[7]));

    /* "treams/sw.pyx":116
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 *         ip8 += steps[8]             # <<<<<<<<<<<<<<
 *         op0 += steps[9]
 * 
 */
    __pyx_v_ip8 = (__pyx_v_ip8 + (__pyx_v_steps[8]));

    /* "treams/sw.pyx":117
 *         ip7 += steps[7]
 *         ip8 += steps[8]
 *         op0 += steps[9]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[9]));
  }

  /* "treams/sw.pyx":81
 * 
 * 
 * cdef void _loop_D_llllllddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

  /* function exit code */
}

/* "treams/sw.pyx":120
 * 
 * 
 * cdef void _loop_D_llllllDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

static void __pyx_f_6treams_2sw__loop_D_llllllDdd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_ip7;
  char *__pyx_v_ip8;
  char *__pyx_v_op0;
  __pyx_t_double_complex __pyx_v_ov0;
  npy_intp __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;

  /* "treams/sw.pyx":121
 * 
 * cdef void _loop_D_llllllDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/sw.pyx":122
 * cdef void _loop_D_llllllDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/sw.pyx":123
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/sw.pyx":124
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/sw.pyx":125
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/sw.pyx":126
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/sw.pyx":127
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/sw.pyx":128
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/sw.pyx":129
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *ip7 = args[7]
 *     cdef char *ip8 = args[8]
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/sw.pyx":130
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]             # <<<<<<<<<<<<<<
 *     cdef char *ip8 = args[8]
 *     cdef char *op0 = args[9]
 */
  __pyx_v_ip7 = (__pyx_v_args[7]);

  /* "treams/sw.pyx":131
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 *     cdef char *ip8 = args[8]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[9]
 *     cdef double complex ov0
 */
  __pyx_v_ip8 = (__pyx_v_args[8]);

  /* "treams/sw.pyx":132
 *     cdef char *ip7 = args[7]
 *     cdef char *ip8 = args[8]
 *     cdef char *op0 = args[9]             # <<<<<<<<<<<<<<
 *     cdef double complex ov0
 *     for i in range(n):
 */
  __pyx_v_op0 = (__pyx_v_args[9]);

  /* "treams/sw.pyx":134
 *     cdef char *op0 = args[9]
 *     cdef double complex ov0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ov0 = (<double complex(*)(long, long, long, long, long, long, double complex, double, double) nogil>func)(
 *             <long>(<long*>ip0)[0],
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "treams/sw.pyx":135
 *     cdef double complex ov0
 *     for i in range(n):
 *         ov0 = (<double complex(*)(long, long, long, long, long, long, double complex, double, double) nogil>func)(             # <<<<<<<<<<<<<<
 *             <long>(<long*>ip0)[0],
 *             <long>(<long*>ip1)[0],
 */
    __pyx_v_ov0 = ((__pyx_t_double_complex (*)(long, long, long, long, long, long, __pyx_t_double_complex, double, double))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((long)(((long *)__pyx_v_ip2)[0])), ((long)(((long *)__pyx_v_ip3)[0])), ((long)(((long *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip6)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip6)[0])))), ((double)(((double *)__pyx_v_ip7)[0])), ((double)(((double *)__pyx_v_ip8)[0])));

    /* "treams/sw.pyx":146
 *             <double>(<double*>ip8)[0],
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    (((__pyx_t_double_complex *)__pyx_v_op0)[0]) = __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL(__pyx_v_ov0)), ((double)__Pyx_CIMAG(__pyx_v_ov0)));

    /* "treams/sw.pyx":147
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/sw.pyx":148
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/sw.pyx":149
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/sw.pyx":150
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/sw.pyx":151
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/sw.pyx":152
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/sw.pyx":153
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         ip7 += steps[7]
 *         ip8 += steps[8]
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/sw.pyx":154
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         ip7 += steps[7]             # <<<<<<<<<<<<<<
 *         ip8 += steps[8]
 *         op0 += steps[9]
 */
    __pyx_v_ip7 = (__pyx_v_ip7 + (__pyx_v_steps[7]));

    /* "treams/sw.pyx":155
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 *         ip8 += steps[8]             # <<<<<<<<<<<<<<
 *         op0 += steps[9]
 * 
 */
    __pyx_v_ip8 = (__pyx_v_ip8 + (__pyx_v_steps[8]));

    /* "treams/sw.pyx":156
 *         ip7 += steps[7]
 *         ip8 += steps[8]
 *         op0 += steps[9]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[9]));
  }

  /* "treams/sw.pyx":120
 * 
 * 
 * cdef void _loop_D_llllllDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

  /* function exit code */
}

/* "treams/sw.pyx":251
 * 
 * 
 * def translate(lambda_, mu, pol, l, m, qol, kr, theta, phi, poltype=None, singular=True, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """translate(lambda_, mu, pol, l, m, qol, kr, theta, phi, helicity=True, singular=True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6treams_2sw_1translate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6treams_2sw_translate[] = "translate(lambda_, mu, pol, l, m, qol, kr, theta, phi, helicity=True, singular=True)\n\n    Translation coefficient for spherical modes.\n\n    Returns the correct translation coefficient from :func:`treams.special.tl_vsw_A`,\n    :func:`treams.special.tl_vsw_B`, :func:`treams.special.tl_vsw_rA`, and\n    :func:`treams.special.tl_vsw_rB` or a combination thereof for the specified mode and\n    basis.\n\n    The polarization values `0` and `1` refer to negative and positive helicity\n    waves or, if ``helicity == False``, to TE and TM parity.\n\n    Args:\n        lambda_ (int, array_like): Degree of the destination mode\n        mu (int, array_like): Order of the destination mode\n        pol (int, array_like): Polarization of the destination mode\n        l (int, array_like): Degree of the source mode\n        m (int, array_like): Order of the source mode\n        qol (int, array_like): Polarization of the source mode\n        kr (float or complex, array_like): Translation distance in units of the wave number\n        theta (float, array_like): Polar angle\n        phi (float, array_like): Azimuthal angle\n        helicity (bool, optional): If true, helicity basis is assumed, else parity basis.\n            Defaults to ``True``.\n        singular (bool, optional): If true, singular translation coefficients are used,\n            else regular coefficients. Defaults to ``True``.\n\n    Returns:\n        complex\n    ";
static PyMethodDef __pyx_mdef_6treams_2sw_1translate = {"translate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6treams_2sw_1translate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6treams_2sw_translate};
static PyObject *__pyx_pw_6treams_2sw_1translate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_lambda_ = 0;
  PyObject *__pyx_v_mu = 0;
  PyObject *__pyx_v_pol = 0;
  PyObject *__pyx_v_l = 0;
  PyObject *__pyx_v_m = 0;
  PyObject *__pyx_v_qol = 0;
  PyObject *__pyx_v_kr = 0;
  PyObject *__pyx_v_theta = 0;
  PyObject *__pyx_v_phi = 0;
  PyObject *__pyx_v_poltype = 0;
  PyObject *__pyx_v_singular = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("translate (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 11) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 11, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lambda,&__pyx_n_s_mu,&__pyx_n_s_pol,&__pyx_n_s_l,&__pyx_n_s_m,&__pyx_n_s_qol,&__pyx_n_s_kr,&__pyx_n_s_theta,&__pyx_n_s_phi,&__pyx_n_s_poltype,&__pyx_n_s_singular,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    values[9] = ((PyObject *)Py_None);
    values[10] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lambda)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate", 0, 9, 11, 1); __PYX_ERR(0, 251, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate", 0, 9, 11, 2); __PYX_ERR(0, 251, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_l)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate", 0, 9, 11, 3); __PYX_ERR(0, 251, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate", 0, 9, 11, 4); __PYX_ERR(0, 251, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate", 0, 9, 11, 5); __PYX_ERR(0, 251, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate", 0, 9, 11, 6); __PYX_ERR(0, 251, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theta)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate", 0, 9, 11, 7); __PYX_ERR(0, 251, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_phi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate", 0, 9, 11, 8); __PYX_ERR(0, 251, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_poltype);
          if (value) { values[9] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_singular);
          if (value) { values[10] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 11) ? pos_args : 11;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "translate") < 0)) __PYX_ERR(0, 251, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        default:
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        case  8:
        case  7:
        case  6:
        case  5:
        case  4:
        case  3:
        case  2:
        case  1:
        case  0:
        goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lambda_ = values[0];
    __pyx_v_mu = values[1];
    __pyx_v_pol = values[2];
    __pyx_v_l = values[3];
    __pyx_v_m = values[4];
    __pyx_v_qol = values[5];
    __pyx_v_kr = values[6];
    __pyx_v_theta = values[7];
    __pyx_v_phi = values[8];
    __pyx_v_poltype = values[9];
    __pyx_v_singular = values[10];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("translate", 0, 9, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 251, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("treams.sw.translate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6treams_2sw_translate(__pyx_self, __pyx_v_lambda_, __pyx_v_mu, __pyx_v_pol, __pyx_v_l, __pyx_v_m, __pyx_v_qol, __pyx_v_kr, __pyx_v_theta, __pyx_v_phi, __pyx_v_poltype, __pyx_v_singular, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6treams_2sw_translate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_lambda_, PyObject *__pyx_v_mu, PyObject *__pyx_v_pol, PyObject *__pyx_v_l, PyObject *__pyx_v_m, PyObject *__pyx_v_qol, PyObject *__pyx_v_kr, PyObject *__pyx_v_theta, PyObject *__pyx_v_phi, PyObject *__pyx_v_poltype, PyObject *__pyx_v_singular, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("translate", 0);
  __Pyx_INCREF(__pyx_v_poltype);

  /* "treams/sw.pyx":282
 *         complex
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype             # <<<<<<<<<<<<<<
 *     if poltype == "helicity":
 *         if singular:
 */
  __pyx_t_2 = (__pyx_v_poltype == Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_config); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_POLTYPE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_poltype);
    __pyx_t_1 = __pyx_v_poltype;
  }
  __Pyx_DECREF_SET(__pyx_v_poltype, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "treams/sw.pyx":283
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":             # <<<<<<<<<<<<<<
 *         if singular:
 *             return _translate_sh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_poltype, __pyx_n_u_helicity, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 283, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "treams/sw.pyx":284
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":
 *         if singular:             # <<<<<<<<<<<<<<
 *             return _translate_sh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *         return _translate_rh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_singular); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 284, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "treams/sw.pyx":285
 *     if poltype == "helicity":
 *         if singular:
 *             return _translate_sh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)             # <<<<<<<<<<<<<<
 *         return _translate_rh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *     elif poltype == "parity":
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_translate_sh); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_lambda_);
      __Pyx_GIVEREF(__pyx_v_lambda_);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_lambda_);
      __Pyx_INCREF(__pyx_v_mu);
      __Pyx_GIVEREF(__pyx_v_mu);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_mu);
      __Pyx_INCREF(__pyx_v_pol);
      __Pyx_GIVEREF(__pyx_v_pol);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_pol);
      __Pyx_INCREF(__pyx_v_l);
      __Pyx_GIVEREF(__pyx_v_l);
      PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_l);
      __Pyx_INCREF(__pyx_v_m);
      __Pyx_GIVEREF(__pyx_v_m);
      PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_m);
      __Pyx_INCREF(__pyx_v_qol);
      __Pyx_GIVEREF(__pyx_v_qol);
      PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_qol);
      __Pyx_INCREF(__pyx_v_kr);
      __Pyx_GIVEREF(__pyx_v_kr);
      PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_kr);
      __Pyx_INCREF(__pyx_v_theta);
      __Pyx_GIVEREF(__pyx_v_theta);
      PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_v_theta);
      __Pyx_INCREF(__pyx_v_phi);
      __Pyx_GIVEREF(__pyx_v_phi);
      PyTuple_SET_ITEM(__pyx_t_4, 8, __pyx_v_phi);
      __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "treams/sw.pyx":284
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":
 *         if singular:             # <<<<<<<<<<<<<<
 *             return _translate_sh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *         return _translate_rh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 */
    }

    /* "treams/sw.pyx":286
 *         if singular:
 *             return _translate_sh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *         return _translate_rh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)             # <<<<<<<<<<<<<<
 *     elif poltype == "parity":
 *         if singular:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_translate_rh); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyTuple_New(9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_lambda_);
    __Pyx_GIVEREF(__pyx_v_lambda_);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_lambda_);
    __Pyx_INCREF(__pyx_v_mu);
    __Pyx_GIVEREF(__pyx_v_mu);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_mu);
    __Pyx_INCREF(__pyx_v_pol);
    __Pyx_GIVEREF(__pyx_v_pol);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_pol);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_m);
    __Pyx_GIVEREF(__pyx_v_m);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_m);
    __Pyx_INCREF(__pyx_v_qol);
    __Pyx_GIVEREF(__pyx_v_qol);
    PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_qol);
    __Pyx_INCREF(__pyx_v_kr);
    __Pyx_GIVEREF(__pyx_v_kr);
    PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_kr);
    __Pyx_INCREF(__pyx_v_theta);
    __Pyx_GIVEREF(__pyx_v_theta);
    PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_v_theta);
    __Pyx_INCREF(__pyx_v_phi);
    __Pyx_GIVEREF(__pyx_v_phi);
    PyTuple_SET_ITEM(__pyx_t_4, 8, __pyx_v_phi);
    __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "treams/sw.pyx":283
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":             # <<<<<<<<<<<<<<
 *         if singular:
 *             return _translate_sh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 */
  }

  /* "treams/sw.pyx":287
 *             return _translate_sh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *         return _translate_rh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *     elif poltype == "parity":             # <<<<<<<<<<<<<<
 *         if singular:
 *             return _translate_sp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_poltype, __pyx_n_u_parity, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 287, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "treams/sw.pyx":288
 *         return _translate_rh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *     elif poltype == "parity":
 *         if singular:             # <<<<<<<<<<<<<<
 *             return _translate_sp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *         return _translate_rp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_singular); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 288, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "treams/sw.pyx":289
 *     elif poltype == "parity":
 *         if singular:
 *             return _translate_sp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)             # <<<<<<<<<<<<<<
 *         return _translate_rp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *     raise ValueError(f"invalid poltype '{poltype}'")
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_translate_sp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = PyTuple_New(9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_v_lambda_);
      __Pyx_GIVEREF(__pyx_v_lambda_);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_lambda_);
      __Pyx_INCREF(__pyx_v_mu);
      __Pyx_GIVEREF(__pyx_v_mu);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_mu);
      __Pyx_INCREF(__pyx_v_pol);
      __Pyx_GIVEREF(__pyx_v_pol);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_pol);
      __Pyx_INCREF(__pyx_v_l);
      __Pyx_GIVEREF(__pyx_v_l);
      PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_l);
      __Pyx_INCREF(__pyx_v_m);
      __Pyx_GIVEREF(__pyx_v_m);
      PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_m);
      __Pyx_INCREF(__pyx_v_qol);
      __Pyx_GIVEREF(__pyx_v_qol);
      PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_qol);
      __Pyx_INCREF(__pyx_v_kr);
      __Pyx_GIVEREF(__pyx_v_kr);
      PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_kr);
      __Pyx_INCREF(__pyx_v_theta);
      __Pyx_GIVEREF(__pyx_v_theta);
      PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_v_theta);
      __Pyx_INCREF(__pyx_v_phi);
      __Pyx_GIVEREF(__pyx_v_phi);
      PyTuple_SET_ITEM(__pyx_t_4, 8, __pyx_v_phi);
      __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "treams/sw.pyx":288
 *         return _translate_rh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *     elif poltype == "parity":
 *         if singular:             # <<<<<<<<<<<<<<
 *             return _translate_sp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *         return _translate_rp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 */
    }

    /* "treams/sw.pyx":290
 *         if singular:
 *             return _translate_sp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *         return _translate_rp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)             # <<<<<<<<<<<<<<
 *     raise ValueError(f"invalid poltype '{poltype}'")
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_translate_rp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyTuple_New(9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_lambda_);
    __Pyx_GIVEREF(__pyx_v_lambda_);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_lambda_);
    __Pyx_INCREF(__pyx_v_mu);
    __Pyx_GIVEREF(__pyx_v_mu);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_mu);
    __Pyx_INCREF(__pyx_v_pol);
    __Pyx_GIVEREF(__pyx_v_pol);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_pol);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_m);
    __Pyx_GIVEREF(__pyx_v_m);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_m);
    __Pyx_INCREF(__pyx_v_qol);
    __Pyx_GIVEREF(__pyx_v_qol);
    PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_qol);
    __Pyx_INCREF(__pyx_v_kr);
    __Pyx_GIVEREF(__pyx_v_kr);
    PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_kr);
    __Pyx_INCREF(__pyx_v_theta);
    __Pyx_GIVEREF(__pyx_v_theta);
    PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_v_theta);
    __Pyx_INCREF(__pyx_v_phi);
    __Pyx_GIVEREF(__pyx_v_phi);
    PyTuple_SET_ITEM(__pyx_t_4, 8, __pyx_v_phi);
    __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "treams/sw.pyx":287
 *             return _translate_sh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *         return _translate_rh(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *     elif poltype == "parity":             # <<<<<<<<<<<<<<
 *         if singular:
 *             return _translate_sp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 */
  }

  /* "treams/sw.pyx":291
 *             return _translate_sp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *         return _translate_rp(lambda_, mu, pol, l, m, qol, kr, theta, phi, *args, **kwargs)
 *     raise ValueError(f"invalid poltype '{poltype}'")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = 0;
  __pyx_t_7 = 127;
  __Pyx_INCREF(__pyx_kp_u_invalid_poltype);
  __pyx_t_6 += 17;
  __Pyx_GIVEREF(__pyx_kp_u_invalid_poltype);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_invalid_poltype);
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_poltype, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_7;
  __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_6 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 291, __pyx_L1_error)

  /* "treams/sw.pyx":251
 * 
 * 
 * def translate(lambda_, mu, pol, l, m, qol, kr, theta, phi, poltype=None, singular=True, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """translate(lambda_, mu, pol, l, m, qol, kr, theta, phi, helicity=True, singular=True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("treams.sw.translate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_poltype);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "treams/sw.pyx":294
 * 
 * 
 * cdef double complex _crotate(long lambda_, long mu, long pol1, long l, long m, long pol2, double phi, double theta, double psi) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Rotation coefficient for the rotation around the Euler angles phi, theta and psi.
 */

static __pyx_t_double_complex __pyx_f_6treams_2sw__crotate(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol1, long __pyx_v_l, long __pyx_v_m, long __pyx_v_pol2, double __pyx_v_phi, double __pyx_v_theta, double __pyx_v_psi) {
  __pyx_t_double_complex __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save;
  #endif
  __Pyx_RefNannySetupContext("_crotate", 1);

  /* "treams/sw.pyx":295
 * 
 * cdef double complex _crotate(long lambda_, long mu, long pol1, long l, long m, long pol2, double phi, double theta, double psi) nogil:
 *     """             # <<<<<<<<<<<<<<
 *     Rotation coefficient for the rotation around the Euler angles phi, theta and psi.
 *     It is intended to be used to construct the rotation matrix manually.
 */
  /*try:*/ {

    /* "treams/sw.pyx":299
 *     It is intended to be used to construct the rotation matrix manually.
 *     """
 *     if lambda_ == l and (pol1 == pol2 == 0 or pol1 == pol2 == 1):             # <<<<<<<<<<<<<<
 *         return cs.wignerd(l, mu, m, phi, theta, psi)
 *     if (pol1 == 0 or pol1 == 1) and (pol2 == 0 or pol2 == 1):
 */
    __pyx_t_2 = ((__pyx_v_lambda_ == __pyx_v_l) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_pol1 == __pyx_v_pol2);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_pol2 == 0);
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = (__pyx_v_pol1 == __pyx_v_pol2);
    if (__pyx_t_3) {
      __pyx_t_3 = (__pyx_v_pol2 == 1);
    }
    __pyx_t_2 = (__pyx_t_3 != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":300
 *     """
 *     if lambda_ == l and (pol1 == pol2 == 0 or pol1 == pol2 == 1):
 *         return cs.wignerd(l, mu, m, phi, theta, psi)             # <<<<<<<<<<<<<<
 *     if (pol1 == 0 or pol1 == 1) and (pol2 == 0 or pol2 == 1):
 *         return 0
 */
      __pyx_r = __pyx_fuse_1__pyx_f_6treams_7special_14cython_special_wignerd(__pyx_v_l, __pyx_v_mu, __pyx_v_m, __pyx_v_phi, __pyx_v_theta, __pyx_v_psi, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":299
 *     It is intended to be used to construct the rotation matrix manually.
 *     """
 *     if lambda_ == l and (pol1 == pol2 == 0 or pol1 == pol2 == 1):             # <<<<<<<<<<<<<<
 *         return cs.wignerd(l, mu, m, phi, theta, psi)
 *     if (pol1 == 0 or pol1 == 1) and (pol2 == 0 or pol2 == 1):
 */
    }

    /* "treams/sw.pyx":301
 *     if lambda_ == l and (pol1 == pol2 == 0 or pol1 == pol2 == 1):
 *         return cs.wignerd(l, mu, m, phi, theta, psi)
 *     if (pol1 == 0 or pol1 == 1) and (pol2 == 0 or pol2 == 1):             # <<<<<<<<<<<<<<
 *         return 0
 *     raise ValueError('TODO')
 */
    switch (__pyx_v_pol1) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    switch (__pyx_v_pol2) {
      case 0:
      case 1:
      __pyx_t_2 = 1;
      break;
      default:
      __pyx_t_2 = 0;
      break;
    }
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_1) {

      /* "treams/sw.pyx":302
 *         return cs.wignerd(l, mu, m, phi, theta, psi)
 *     if (pol1 == 0 or pol1 == 1) and (pol2 == 0 or pol2 == 1):
 *         return 0             # <<<<<<<<<<<<<<
 *     raise ValueError('TODO')
 * 
 */
      __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
      goto __pyx_L3_return;

      /* "treams/sw.pyx":301
 *     if lambda_ == l and (pol1 == pol2 == 0 or pol1 == pol2 == 1):
 *         return cs.wignerd(l, mu, m, phi, theta, psi)
 *     if (pol1 == 0 or pol1 == 1) and (pol2 == 0 or pol2 == 1):             # <<<<<<<<<<<<<<
 *         return 0
 *     raise ValueError('TODO')
 */
    }

    /* "treams/sw.pyx":303
 *     if (pol1 == 0 or pol1 == 1) and (pol2 == 0 or pol2 == 1):
 *         return 0
 *     raise ValueError('TODO')             # <<<<<<<<<<<<<<
 * 
 * 
 */
    {
        #ifdef WITH_THREAD
        PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
        #endif
        /*try:*/ {
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 303, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_Raise(__pyx_t_4, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __PYX_ERR(0, 303, __pyx_L14_error)
        }
        /*finally:*/ {
          __pyx_L14_error: {
            #ifdef WITH_THREAD
            __Pyx_PyGILState_Release(__pyx_gilstate_save);
            #endif
            goto __pyx_L4_error;
          }
        }
    }
  }

  /* "treams/sw.pyx":295
 * 
 * cdef double complex _crotate(long lambda_, long mu, long pol1, long l, long m, long pol2, double phi, double theta, double psi) nogil:
 *     """             # <<<<<<<<<<<<<<
 *     Rotation coefficient for the rotation around the Euler angles phi, theta and psi.
 *     It is intended to be used to construct the rotation matrix manually.
 */
  /*finally:*/ {
    __pyx_L3_return: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L0;
    }
    __pyx_L4_error: {
      #ifdef WITH_THREAD
      __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
      #endif
      goto __pyx_L1_error;
    }
  }

  /* "treams/sw.pyx":294
 * 
 * 
 * cdef double complex _crotate(long lambda_, long mu, long pol1, long l, long m, long pol2, double phi, double theta, double psi) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Rotation coefficient for the rotation around the Euler angles phi, theta and psi.
 */

  /* function exit code */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("treams.sw._crotate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "treams/sw.pyx":337
 * 
 * 
 * def rotate(lambda_, mu, pol, l, m, qol, phi, theta=0, psi=0, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """rotate(lambda_, mu, pol, l, m, qol, phi, theta=0, psi=0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6treams_2sw_3rotate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6treams_2sw_2rotate[] = "rotate(lambda_, mu, pol, l, m, qol, phi, theta=0, psi=0)\n\n    Rotation coefficient for spherical modes.\n\n    Returns the correct rotation coefficient from :func:`treams.special.wignerd`. The\n    angles are given as Euler angles in `z-y-z`-convention. In the intrinsic (object\n    fixed coordinate system) convention the rotations are applied in the order phi\n    first, theta second, psi third. In the extrinsic (global or reference frame fixed\n    coordinate system) the rotations are applied psi first, theta second, phi third.\n\n    Args:\n        lambda_ (int, array_like): Degree of the destination mode\n        mu (int, array_like): Order of the destination mode\n        pol (int, array_like): Polarization of the destination mode\n        l (int, array_like): Degree of the source mode\n        m (int, array_like): Order of the source mode\n        qol (int, array_like): Polarization of the source mode\n        phi (float or complex, array_like): First Euler angle\n        theta (float, array_like): Second Euler angle\n        psi (float, array_like): Third Euler angle\n\n    Returns:\n        complex\n    ";
static PyMethodDef __pyx_mdef_6treams_2sw_3rotate = {"rotate", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6treams_2sw_3rotate, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6treams_2sw_2rotate};
static PyObject *__pyx_pw_6treams_2sw_3rotate(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_lambda_ = 0;
  PyObject *__pyx_v_mu = 0;
  PyObject *__pyx_v_pol = 0;
  PyObject *__pyx_v_l = 0;
  PyObject *__pyx_v_m = 0;
  PyObject *__pyx_v_qol = 0;
  PyObject *__pyx_v_phi = 0;
  PyObject *__pyx_v_theta = 0;
  PyObject *__pyx_v_psi = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("rotate (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 9) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 9, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_lambda,&__pyx_n_s_mu,&__pyx_n_s_pol,&__pyx_n_s_l,&__pyx_n_s_m,&__pyx_n_s_qol,&__pyx_n_s_phi,&__pyx_n_s_theta,&__pyx_n_s_psi,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[7] = ((PyObject *)__pyx_int_0);
    values[8] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lambda)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rotate", 0, 7, 9, 1); __PYX_ERR(0, 337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rotate", 0, 7, 9, 2); __PYX_ERR(0, 337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_l)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rotate", 0, 7, 9, 3); __PYX_ERR(0, 337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rotate", 0, 7, 9, 4); __PYX_ERR(0, 337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rotate", 0, 7, 9, 5); __PYX_ERR(0, 337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_phi)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("rotate", 0, 7, 9, 6); __PYX_ERR(0, 337, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_theta);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_psi);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 9) ? pos_args : 9;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "rotate") < 0)) __PYX_ERR(0, 337, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        default:
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        case  6:
        case  5:
        case  4:
        case  3:
        case  2:
        case  1:
        case  0:
        goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_lambda_ = values[0];
    __pyx_v_mu = values[1];
    __pyx_v_pol = values[2];
    __pyx_v_l = values[3];
    __pyx_v_m = values[4];
    __pyx_v_qol = values[5];
    __pyx_v_phi = values[6];
    __pyx_v_theta = values[7];
    __pyx_v_psi = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("rotate", 0, 7, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 337, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("treams.sw.rotate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6treams_2sw_2rotate(__pyx_self, __pyx_v_lambda_, __pyx_v_mu, __pyx_v_pol, __pyx_v_l, __pyx_v_m, __pyx_v_qol, __pyx_v_phi, __pyx_v_theta, __pyx_v_psi, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6treams_2sw_2rotate(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_lambda_, PyObject *__pyx_v_mu, PyObject *__pyx_v_pol, PyObject *__pyx_v_l, PyObject *__pyx_v_m, PyObject *__pyx_v_qol, PyObject *__pyx_v_phi, PyObject *__pyx_v_theta, PyObject *__pyx_v_psi, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("rotate", 0);

  /* "treams/sw.pyx":362
 *         complex
 *     """
 *     return _rotate(lambda_, mu, pol, l, m, qol, phi, theta, psi, *args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_rotate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_lambda_);
  __Pyx_GIVEREF(__pyx_v_lambda_);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_lambda_);
  __Pyx_INCREF(__pyx_v_mu);
  __Pyx_GIVEREF(__pyx_v_mu);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_mu);
  __Pyx_INCREF(__pyx_v_pol);
  __Pyx_GIVEREF(__pyx_v_pol);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_pol);
  __Pyx_INCREF(__pyx_v_l);
  __Pyx_GIVEREF(__pyx_v_l);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_l);
  __Pyx_INCREF(__pyx_v_m);
  __Pyx_GIVEREF(__pyx_v_m);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_v_m);
  __Pyx_INCREF(__pyx_v_qol);
  __Pyx_GIVEREF(__pyx_v_qol);
  PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_v_qol);
  __Pyx_INCREF(__pyx_v_phi);
  __Pyx_GIVEREF(__pyx_v_phi);
  PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_v_phi);
  __Pyx_INCREF(__pyx_v_theta);
  __Pyx_GIVEREF(__pyx_v_theta);
  PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_v_theta);
  __Pyx_INCREF(__pyx_v_psi);
  __Pyx_GIVEREF(__pyx_v_psi);
  PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_v_psi);
  __pyx_t_3 = PyNumber_Add(__pyx_t_2, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 362, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "treams/sw.pyx":337
 * 
 * 
 * def rotate(lambda_, mu, pol, l, m, qol, phi, theta=0, psi=0, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """rotate(lambda_, mu, pol, l, m, qol, phi, theta=0, psi=0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("treams.sw.rotate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "treams/sw.pyx":365
 * 
 * 
 * cdef double complex _transl_A_lattice(long lambda_, long mu, long l, long m, double complex *dlms, long step) nogil:             # <<<<<<<<<<<<<<
 *     cdef double complex pref = (
 *         minusonepow(m)
 */

static __pyx_t_double_complex __pyx_f_6treams_2sw__transl_A_lattice(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_l, long __pyx_v_m, __pyx_t_double_complex *__pyx_v_dlms, long __pyx_v_step) {
  __pyx_t_double_complex __pyx_v_pref;
  __pyx_t_double_complex __pyx_v_dlm;
  __pyx_t_double_complex __pyx_v_res;
  long __pyx_v_p;
  __pyx_t_double_complex __pyx_r;
  double __pyx_t_1;
  double __pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":371
 *             pi
 *             * (2 * l + 1)
 *             * (2 * lambda_ + 1)             # <<<<<<<<<<<<<<
 *             / <double>(l * (l + 1) * lambda_ * (lambda_ + 1))
 *         )
 */
  __pyx_t_1 = ((M_PI * ((2 * __pyx_v_l) + 1)) * ((2 * __pyx_v_lambda_) + 1));

  /* "treams/sw.pyx":372
 *             * (2 * l + 1)
 *             * (2 * lambda_ + 1)
 *             / <double>(l * (l + 1) * lambda_ * (lambda_ + 1))             # <<<<<<<<<<<<<<
 *         )
 *         * cpow(1.0j, lambda_ - l)
 */
  __pyx_t_2 = ((double)(((__pyx_v_l * (__pyx_v_l + 1)) * __pyx_v_lambda_) * (__pyx_v_lambda_ + 1)));
  if (unlikely(__pyx_t_2 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 372, __pyx_L1_error)
  }

  /* "treams/sw.pyx":374
 *             / <double>(l * (l + 1) * lambda_ * (lambda_ + 1))
 *         )
 *         * cpow(1.0j, lambda_ - l)             # <<<<<<<<<<<<<<
 *     )
 *     cdef double complex dlm, res = 0
 */
  __pyx_v_pref = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts((__pyx_f_6treams_7special_5_misc_minusonepow(__pyx_v_m) * sqrt((__pyx_t_1 / __pyx_t_2))), 0), cpow(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts((__pyx_v_lambda_ - __pyx_v_l), 0)));

  /* "treams/sw.pyx":376
 *         * cpow(1.0j, lambda_ - l)
 *     )
 *     cdef double complex dlm, res = 0             # <<<<<<<<<<<<<<
 *     cdef long p
 *     for p in range(l + lambda_, max(labs(lambda_ - l), labs(mu - m)) - 1, -2):
 */
  __pyx_v_res = __pyx_t_double_complex_from_parts(0, 0);

  /* "treams/sw.pyx":378
 *     cdef double complex dlm, res = 0
 *     cdef long p
 *     for p in range(l + lambda_, max(labs(lambda_ - l), labs(mu - m)) - 1, -2):             # <<<<<<<<<<<<<<
 *         dlm = dlms[(p * (p + 1) + m - mu) * step]
 *         res += (
 */
  __pyx_t_3 = labs((__pyx_v_mu - __pyx_v_m));
  __pyx_t_4 = labs((__pyx_v_lambda_ - __pyx_v_l));
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_t_3 = (__pyx_t_5 - 1);
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_4 = (__pyx_v_l + __pyx_v_lambda_); __pyx_t_4 > __pyx_t_5; __pyx_t_4-=2) {
    __pyx_v_p = __pyx_t_4;

    /* "treams/sw.pyx":379
 *     cdef long p
 *     for p in range(l + lambda_, max(labs(lambda_ - l), labs(mu - m)) - 1, -2):
 *         dlm = dlms[(p * (p + 1) + m - mu) * step]             # <<<<<<<<<<<<<<
 *         res += (
 *             dlm
 */
    __pyx_v_dlm = (__pyx_v_dlms[((((__pyx_v_p * (__pyx_v_p + 1)) + __pyx_v_m) - __pyx_v_mu) * __pyx_v_step)]);

    /* "treams/sw.pyx":380
 *     for p in range(l + lambda_, max(labs(lambda_ - l), labs(mu - m)) - 1, -2):
 *         dlm = dlms[(p * (p + 1) + m - mu) * step]
 *         res += (             # <<<<<<<<<<<<<<
 *             dlm
 *             * cpow(1.0j, p)
 */
    __pyx_v_res = __Pyx_c_sum_double(__pyx_v_res, __Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_v_dlm, cpow(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts(__pyx_v_p, 0))), __pyx_t_double_complex_from_parts(sqrt(((2 * __pyx_v_p) + 1)), 0)), __pyx_t_double_complex_from_parts(__pyx_f_6treams_7special_14cython_special_wigner3j(__pyx_v_l, __pyx_v_lambda_, __pyx_v_p, __pyx_v_m, (-__pyx_v_mu), ((-__pyx_v_m) + __pyx_v_mu), 0), 0)), __pyx_t_double_complex_from_parts(__pyx_f_6treams_7special_14cython_special_wigner3j(__pyx_v_l, __pyx_v_lambda_, __pyx_v_p, 0, 0, 0, 0), 0)), __pyx_t_double_complex_from_parts((((__pyx_v_l * (__pyx_v_l + 1)) + (__pyx_v_lambda_ * (__pyx_v_lambda_ + 1))) - (__pyx_v_p * (__pyx_v_p + 1))), 0)));
  }

  /* "treams/sw.pyx":388
 *             * (l * (l + 1) + lambda_ * (lambda_ + 1) - p * (p + 1))
 *         )
 *     return res * pref             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __Pyx_c_prod_double(__pyx_v_res, __pyx_v_pref);
  goto __pyx_L0;

  /* "treams/sw.pyx":365
 * 
 * 
 * cdef double complex _transl_A_lattice(long lambda_, long mu, long l, long m, double complex *dlms, long step) nogil:             # <<<<<<<<<<<<<<
 *     cdef double complex pref = (
 *         minusonepow(m)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._transl_A_lattice", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/sw.pyx":391
 * 
 * 
 * cdef double complex _transl_B_lattice(long lambda_, long mu, long l, long m, double complex *dlms, long step) nogil:             # <<<<<<<<<<<<<<
 *     cdef double complex pref = (
 *         minusonepow(m)
 */

static __pyx_t_double_complex __pyx_f_6treams_2sw__transl_B_lattice(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_l, long __pyx_v_m, __pyx_t_double_complex *__pyx_v_dlms, long __pyx_v_step) {
  __pyx_t_double_complex __pyx_v_pref;
  __pyx_t_double_complex __pyx_v_dlm;
  __pyx_t_double_complex __pyx_v_res;
  long __pyx_v_p;
  __pyx_t_double_complex __pyx_r;
  double __pyx_t_1;
  double __pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":397
 *             pi
 *             * (2 * l + 1)
 *             * (2 * lambda_ + 1)             # <<<<<<<<<<<<<<
 *             / <double>(l * (l + 1) * lambda_ * (lambda_ + 1))
 *         )
 */
  __pyx_t_1 = ((M_PI * ((2 * __pyx_v_l) + 1)) * ((2 * __pyx_v_lambda_) + 1));

  /* "treams/sw.pyx":398
 *             * (2 * l + 1)
 *             * (2 * lambda_ + 1)
 *             / <double>(l * (l + 1) * lambda_ * (lambda_ + 1))             # <<<<<<<<<<<<<<
 *         )
 *         * cpow(1.0j, lambda_ - l)
 */
  __pyx_t_2 = ((double)(((__pyx_v_l * (__pyx_v_l + 1)) * __pyx_v_lambda_) * (__pyx_v_lambda_ + 1)));
  if (unlikely(__pyx_t_2 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 398, __pyx_L1_error)
  }

  /* "treams/sw.pyx":400
 *             / <double>(l * (l + 1) * lambda_ * (lambda_ + 1))
 *         )
 *         * cpow(1.0j, lambda_ - l)             # <<<<<<<<<<<<<<
 *     )
 *     cdef double complex dlm, res = 0
 */
  __pyx_v_pref = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts((__pyx_f_6treams_7special_5_misc_minusonepow(__pyx_v_m) * sqrt((__pyx_t_1 / __pyx_t_2))), 0), cpow(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts((__pyx_v_lambda_ - __pyx_v_l), 0)));

  /* "treams/sw.pyx":402
 *         * cpow(1.0j, lambda_ - l)
 *     )
 *     cdef double complex dlm, res = 0             # <<<<<<<<<<<<<<
 *     cdef long p
 *     for p in range(
 */
  __pyx_v_res = __pyx_t_double_complex_from_parts(0, 0);

  /* "treams/sw.pyx":405
 *     cdef long p
 *     for p in range(
 *         l + lambda_ - 1, max(labs(lambda_ - l) + 1, labs(mu - m)) - 1, -2             # <<<<<<<<<<<<<<
 *     ):
 *         dlm = dlms[(p * (p + 1) + m - mu) * step]
 */
  __pyx_t_3 = labs((__pyx_v_mu - __pyx_v_m));
  __pyx_t_4 = (labs((__pyx_v_lambda_ - __pyx_v_l)) + 1);
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_t_3 = (__pyx_t_5 - 1);

  /* "treams/sw.pyx":404
 *     cdef double complex dlm, res = 0
 *     cdef long p
 *     for p in range(             # <<<<<<<<<<<<<<
 *         l + lambda_ - 1, max(labs(lambda_ - l) + 1, labs(mu - m)) - 1, -2
 *     ):
 */
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_4 = ((__pyx_v_l + __pyx_v_lambda_) - 1); __pyx_t_4 > __pyx_t_5; __pyx_t_4-=2) {
    __pyx_v_p = __pyx_t_4;

    /* "treams/sw.pyx":407
 *         l + lambda_ - 1, max(labs(lambda_ - l) + 1, labs(mu - m)) - 1, -2
 *     ):
 *         dlm = dlms[(p * (p + 1) + m - mu) * step]             # <<<<<<<<<<<<<<
 *         res += (
 *             dlm
 */
    __pyx_v_dlm = (__pyx_v_dlms[((((__pyx_v_p * (__pyx_v_p + 1)) + __pyx_v_m) - __pyx_v_mu) * __pyx_v_step)]);

    /* "treams/sw.pyx":408
 *     ):
 *         dlm = dlms[(p * (p + 1) + m - mu) * step]
 *         res += (             # <<<<<<<<<<<<<<
 *             dlm
 *             * cpow(1.0j, p)
 */
    __pyx_v_res = __Pyx_c_sum_double(__pyx_v_res, __Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_v_dlm, cpow(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts(__pyx_v_p, 0))), __pyx_t_double_complex_from_parts(sqrt(((2 * __pyx_v_p) + 1)), 0)), __pyx_t_double_complex_from_parts(__pyx_f_6treams_7special_14cython_special_wigner3j(__pyx_v_l, __pyx_v_lambda_, __pyx_v_p, __pyx_v_m, (-__pyx_v_mu), ((-__pyx_v_m) + __pyx_v_mu), 0), 0)), __pyx_t_double_complex_from_parts(__pyx_f_6treams_7special_14cython_special_wigner3j(__pyx_v_l, __pyx_v_lambda_, (__pyx_v_p - 1), 0, 0, 0, 0), 0)), __pyx_t_double_complex_from_parts(sqrt(((((((__pyx_v_l + __pyx_v_lambda_) + 1) + __pyx_v_p) * (((__pyx_v_l + __pyx_v_lambda_) + 1) - __pyx_v_p)) * ((__pyx_v_p - __pyx_v_lambda_) + __pyx_v_l)) * ((__pyx_v_p + __pyx_v_lambda_) - __pyx_v_l))), 0)));
  }

  /* "treams/sw.pyx":421
 *             )
 *         )
 *     return res * pref             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __Pyx_c_prod_double(__pyx_v_res, __pyx_v_pref);
  goto __pyx_L0;

  /* "treams/sw.pyx":391
 * 
 * 
 * cdef double complex _transl_B_lattice(long lambda_, long mu, long l, long m, double complex *dlms, long step) nogil:             # <<<<<<<<<<<<<<
 *     cdef double complex pref = (
 *         minusonepow(m)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._transl_B_lattice", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/sw.pyx":424
 * 
 * 
 * cdef double complex _ctranslate_periodic_p(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:             # <<<<<<<<<<<<<<
 *     if pol1 == pol2:
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step)
 */

static __pyx_t_double_complex __pyx_f_6treams_2sw__ctranslate_periodic_p(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol1, long __pyx_v_l, long __pyx_v_m, long __pyx_v_pol2, __pyx_t_double_complex *__pyx_v_dlms, long __pyx_v_step) {
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;

  /* "treams/sw.pyx":425
 * 
 * cdef double complex _ctranslate_periodic_p(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:
 *     if pol1 == pol2:             # <<<<<<<<<<<<<<
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step)
 *     return _transl_B_lattice(lambda_, mu, l, m, dlms, step)
 */
  __pyx_t_1 = ((__pyx_v_pol1 == __pyx_v_pol2) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":426
 * cdef double complex _ctranslate_periodic_p(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:
 *     if pol1 == pol2:
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step)             # <<<<<<<<<<<<<<
 *     return _transl_B_lattice(lambda_, mu, l, m, dlms, step)
 * 
 */
    __pyx_r = __pyx_f_6treams_2sw__transl_A_lattice(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_dlms, __pyx_v_step);
    goto __pyx_L0;

    /* "treams/sw.pyx":425
 * 
 * cdef double complex _ctranslate_periodic_p(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:
 *     if pol1 == pol2:             # <<<<<<<<<<<<<<
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step)
 *     return _transl_B_lattice(lambda_, mu, l, m, dlms, step)
 */
  }

  /* "treams/sw.pyx":427
 *     if pol1 == pol2:
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step)
 *     return _transl_B_lattice(lambda_, mu, l, m, dlms, step)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_f_6treams_2sw__transl_B_lattice(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_dlms, __pyx_v_step);
  goto __pyx_L0;

  /* "treams/sw.pyx":424
 * 
 * 
 * cdef double complex _ctranslate_periodic_p(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:             # <<<<<<<<<<<<<<
 *     if pol1 == pol2:
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/sw.pyx":430
 * 
 * 
 * cdef double complex _ctranslate_periodic_h(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:             # <<<<<<<<<<<<<<
 *     if pol1 == pol2:
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step) + (
 */

static __pyx_t_double_complex __pyx_f_6treams_2sw__ctranslate_periodic_h(long __pyx_v_lambda_, long __pyx_v_mu, long __pyx_v_pol1, long __pyx_v_l, long __pyx_v_m, long __pyx_v_pol2, __pyx_t_double_complex *__pyx_v_dlms, long __pyx_v_step) {
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;

  /* "treams/sw.pyx":431
 * 
 * cdef double complex _ctranslate_periodic_h(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:
 *     if pol1 == pol2:             # <<<<<<<<<<<<<<
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step) + (
 *             2 * pol1 - 1
 */
  __pyx_t_1 = ((__pyx_v_pol1 == __pyx_v_pol2) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":432
 * cdef double complex _ctranslate_periodic_h(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:
 *     if pol1 == pol2:
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step) + (             # <<<<<<<<<<<<<<
 *             2 * pol1 - 1
 *         ) * _transl_B_lattice(lambda_, mu, l, m, dlms, step)
 */
    __pyx_r = __Pyx_c_sum_double(__pyx_f_6treams_2sw__transl_A_lattice(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_dlms, __pyx_v_step), __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((2 * __pyx_v_pol1) - 1), 0), __pyx_f_6treams_2sw__transl_B_lattice(__pyx_v_lambda_, __pyx_v_mu, __pyx_v_l, __pyx_v_m, __pyx_v_dlms, __pyx_v_step)));
    goto __pyx_L0;

    /* "treams/sw.pyx":431
 * 
 * cdef double complex _ctranslate_periodic_h(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:
 *     if pol1 == pol2:             # <<<<<<<<<<<<<<
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step) + (
 *             2 * pol1 - 1
 */
  }

  /* "treams/sw.pyx":435
 *             2 * pol1 - 1
 *         ) * _transl_B_lattice(lambda_, mu, l, m, dlms, step)
 *     return 0.0j             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
  goto __pyx_L0;

  /* "treams/sw.pyx":430
 * 
 * 
 * cdef double complex _ctranslate_periodic_h(long lambda_, long mu, long pol1, long l, long m, long pol2, double complex *dlms, long step) nogil:             # <<<<<<<<<<<<<<
 *     if pol1 == pol2:
 *         return _transl_A_lattice(lambda_, mu, l, m, dlms, step) + (
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/sw.pyx":438
 * 
 * 
 * cdef void _loop_translate_periodic_h(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long dlmstep, step1 = steps[9] // sizeof(double complex), step2 = steps[10] // sizeof(double complex)
 */

static void __pyx_f_6treams_2sw__loop_translate_periodic_h(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_dlmstep;
  long __pyx_v_step1;
  long __pyx_v_step2;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_ip7;
  char *__pyx_v_op0;
  char *__pyx_v_dlmchoice;
  __pyx_t_double_complex __pyx_v_ov0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":439
 * 
 * cdef void _loop_translate_periodic_h(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long dlmstep, step1 = steps[9] // sizeof(double complex), step2 = steps[10] // sizeof(double complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/sw.pyx":440
 * cdef void _loop_translate_periodic_h(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long dlmstep, step1 = steps[9] // sizeof(double complex), step2 = steps[10] // sizeof(double complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 440, __pyx_L1_error)
  }
  __pyx_v_step1 = ((__pyx_v_steps[9]) / __pyx_t_1);
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 440, __pyx_L1_error)
  }
  __pyx_v_step2 = ((__pyx_v_steps[10]) / __pyx_t_1);

  /* "treams/sw.pyx":441
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long dlmstep, step1 = steps[9] // sizeof(double complex), step2 = steps[10] // sizeof(double complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/sw.pyx":442
 *     cdef long dlmstep, step1 = steps[9] // sizeof(double complex), step2 = steps[10] // sizeof(double complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/sw.pyx":443
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/sw.pyx":444
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/sw.pyx":445
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/sw.pyx":446
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/sw.pyx":447
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/sw.pyx":448
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *ip7 = args[7]
 *     cdef char *op0 = args[8]
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/sw.pyx":449
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[8]
 *     cdef char *dlmchoice
 */
  __pyx_v_ip7 = (__pyx_v_args[7]);

  /* "treams/sw.pyx":450
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 *     cdef char *op0 = args[8]             # <<<<<<<<<<<<<<
 *     cdef char *dlmchoice
 *     cdef double complex ov0
 */
  __pyx_v_op0 = (__pyx_v_args[8]);

  /* "treams/sw.pyx":453
 *     cdef char *dlmchoice
 *     cdef double complex ov0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         if <long>(<long*>ip2)[0] == 0:
 *             dlmchoice = ip6
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/sw.pyx":454
 *     cdef double complex ov0
 *     for i in range(n):
 *         if <long>(<long*>ip2)[0] == 0:             # <<<<<<<<<<<<<<
 *             dlmchoice = ip6
 *             dlmstep = step1
 */
    __pyx_t_5 = ((((long)(((long *)__pyx_v_ip2)[0])) == 0) != 0);
    if (__pyx_t_5) {

      /* "treams/sw.pyx":455
 *     for i in range(n):
 *         if <long>(<long*>ip2)[0] == 0:
 *             dlmchoice = ip6             # <<<<<<<<<<<<<<
 *             dlmstep = step1
 *         else:
 */
      __pyx_v_dlmchoice = __pyx_v_ip6;

      /* "treams/sw.pyx":456
 *         if <long>(<long*>ip2)[0] == 0:
 *             dlmchoice = ip6
 *             dlmstep = step1             # <<<<<<<<<<<<<<
 *         else:
 *             dlmchoice = ip7
 */
      __pyx_v_dlmstep = __pyx_v_step1;

      /* "treams/sw.pyx":454
 *     cdef double complex ov0
 *     for i in range(n):
 *         if <long>(<long*>ip2)[0] == 0:             # <<<<<<<<<<<<<<
 *             dlmchoice = ip6
 *             dlmstep = step1
 */
      goto __pyx_L5;
    }

    /* "treams/sw.pyx":458
 *             dlmstep = step1
 *         else:
 *             dlmchoice = ip7             # <<<<<<<<<<<<<<
 *             dlmstep = step2
 *         ov0 = (<double complex (*)(long, long, long, long, long, long, double complex*, long) nogil>func)(
 */
    /*else*/ {
      __pyx_v_dlmchoice = __pyx_v_ip7;

      /* "treams/sw.pyx":459
 *         else:
 *             dlmchoice = ip7
 *             dlmstep = step2             # <<<<<<<<<<<<<<
 *         ov0 = (<double complex (*)(long, long, long, long, long, long, double complex*, long) nogil>func)(
 *             <long>(<long*>ip0)[0],
 */
      __pyx_v_dlmstep = __pyx_v_step2;
    }
    __pyx_L5:;

    /* "treams/sw.pyx":460
 *             dlmchoice = ip7
 *             dlmstep = step2
 *         ov0 = (<double complex (*)(long, long, long, long, long, long, double complex*, long) nogil>func)(             # <<<<<<<<<<<<<<
 *             <long>(<long*>ip0)[0],
 *             <long>(<long*>ip1)[0],
 */
    __pyx_v_ov0 = ((__pyx_t_double_complex (*)(long, long, long, long, long, long, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((long)(((long *)__pyx_v_ip2)[0])), ((long)(((long *)__pyx_v_ip3)[0])), ((long)(((long *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), ((__pyx_t_double_complex *)__pyx_v_dlmchoice), __pyx_v_dlmstep);

    /* "treams/sw.pyx":470
 *             dlmstep,
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    (((__pyx_t_double_complex *)__pyx_v_op0)[0]) = __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL(__pyx_v_ov0)), ((double)__Pyx_CIMAG(__pyx_v_ov0)));

    /* "treams/sw.pyx":471
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/sw.pyx":472
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/sw.pyx":473
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/sw.pyx":474
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/sw.pyx":475
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/sw.pyx":476
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/sw.pyx":477
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         ip7 += steps[7]
 *         op0 += steps[8]
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/sw.pyx":478
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         ip7 += steps[7]             # <<<<<<<<<<<<<<
 *         op0 += steps[8]
 * 
 */
    __pyx_v_ip7 = (__pyx_v_ip7 + (__pyx_v_steps[7]));

    /* "treams/sw.pyx":479
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 *         op0 += steps[8]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[8]));
  }

  /* "treams/sw.pyx":438
 * 
 * 
 * cdef void _loop_translate_periodic_h(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long dlmstep, step1 = steps[9] // sizeof(double complex), step2 = steps[10] // sizeof(double complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._loop_translate_periodic_h", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/sw.pyx":482
 * 
 * 
 * cdef void _loop_translate_periodic_p(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long dlmstep = steps[8] // sizeof(double complex)
 */

static void __pyx_f_6treams_2sw__loop_translate_periodic_p(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_dlmstep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_op0;
  __pyx_t_double_complex __pyx_v_ov0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":483
 * 
 * cdef void _loop_translate_periodic_p(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long dlmstep = steps[8] // sizeof(double complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/sw.pyx":484
 * cdef void _loop_translate_periodic_p(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long dlmstep = steps[8] // sizeof(double complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 484, __pyx_L1_error)
  }
  __pyx_v_dlmstep = ((__pyx_v_steps[8]) / __pyx_t_1);

  /* "treams/sw.pyx":485
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long dlmstep = steps[8] // sizeof(double complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/sw.pyx":486
 *     cdef long dlmstep = steps[8] // sizeof(double complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/sw.pyx":487
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/sw.pyx":488
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/sw.pyx":489
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/sw.pyx":490
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/sw.pyx":491
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/sw.pyx":492
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[7]
 *     cdef double complex ov0
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/sw.pyx":493
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]             # <<<<<<<<<<<<<<
 *     cdef double complex ov0
 *     for i in range(n):
 */
  __pyx_v_op0 = (__pyx_v_args[7]);

  /* "treams/sw.pyx":495
 *     cdef char *op0 = args[7]
 *     cdef double complex ov0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ov0 = (<double complex (*)(long, long, long, long, long, long, double complex*, long) nogil>func)(
 *             <long>(<long*>ip0)[0],
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/sw.pyx":496
 *     cdef double complex ov0
 *     for i in range(n):
 *         ov0 = (<double complex (*)(long, long, long, long, long, long, double complex*, long) nogil>func)(             # <<<<<<<<<<<<<<
 *             <long>(<long*>ip0)[0],
 *             <long>(<long*>ip1)[0],
 */
    __pyx_v_ov0 = ((__pyx_t_double_complex (*)(long, long, long, long, long, long, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((long)(((long *)__pyx_v_ip2)[0])), ((long)(((long *)__pyx_v_ip3)[0])), ((long)(((long *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), ((__pyx_t_double_complex *)__pyx_v_ip6), __pyx_v_dlmstep);

    /* "treams/sw.pyx":506
 *             dlmstep,
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    (((__pyx_t_double_complex *)__pyx_v_op0)[0]) = __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL(__pyx_v_ov0)), ((double)__Pyx_CIMAG(__pyx_v_ov0)));

    /* "treams/sw.pyx":507
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/sw.pyx":508
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/sw.pyx":509
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/sw.pyx":510
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/sw.pyx":511
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/sw.pyx":512
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         op0 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/sw.pyx":513
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         op0 += steps[7]
 * 
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/sw.pyx":514
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         op0 += steps[7]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[7]));
  }

  /* "treams/sw.pyx":482
 * 
 * 
 * cdef void _loop_translate_periodic_p(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long dlmstep = steps[8] // sizeof(double complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._loop_translate_periodic_p", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/sw.pyx":577
 * 
 * 
 * def translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, poltype=None, eta=0, func=lattice.lsumsw):             # <<<<<<<<<<<<<<
 *     """translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, helicity=True, eta=0)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6treams_2sw_5translate_periodic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6treams_2sw_4translate_periodic[] = "translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, helicity=True, eta=0)\n\n    Translation coefficients in a lattice.\n\n    Returns the translation coefficents for the given modes in a lattice. The calculation\n    uses the fast converging sums of :mod:`treams.lattice`.\n\n    The polarization values `0` and `1` refer to negative and positive helicity\n    waves or, if ``helicity == False``, to TE and TM parity.\n\n    Args:\n        ks (float or complex, scalar or (2,)-array): Wave number(s) in the medium, use\n            two values in chiral media, indexed analogous to the polarization values\n        kpar (float, (D,)-array): Parallel component of the wave, defines the dimension with `1 <= D <= 3`\n        a (float, (D,D)-array): Lattice vectors in each row of the array\n        rs (float, (M, 3)-array): Shift vectors with respect to one lattice point\n        out (3- or 4-tuple of integer arrays): Output modes\n        in_ (3- or 4-tuple of integer arrays): Input modes, if none are given equal to\n            the output modes\n        rsin (float): Shift vectors to use with the input modes, if non are given equal\n            to `rs`\n        helicity (bool, optional): If true, helicity basis is assumed, else parity basis.\n            Defaults to ``True``.\n        eta (float or complex, optional): Cut between real and reciprocal space\n            summation, if equal to zero, an estimation for the optimal value is done.\n\n    Returns:\n        complex array\n    ";
static PyMethodDef __pyx_mdef_6treams_2sw_5translate_periodic = {"translate_periodic", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6treams_2sw_5translate_periodic, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6treams_2sw_4translate_periodic};
static PyObject *__pyx_pw_6treams_2sw_5translate_periodic(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ks = 0;
  PyObject *__pyx_v_kpar = 0;
  PyObject *__pyx_v_a = 0;
  PyObject *__pyx_v_rs = 0;
  PyObject *__pyx_v_out = 0;
  PyObject *__pyx_v_in_ = 0;
  PyObject *__pyx_v_rsin = 0;
  PyObject *__pyx_v_poltype = 0;
  PyObject *__pyx_v_eta = 0;
  PyObject *__pyx_v_func = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("translate_periodic (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ks,&__pyx_n_s_kpar,&__pyx_n_s_a,&__pyx_n_s_rs,&__pyx_n_s_out,&__pyx_n_s_in,&__pyx_n_s_rsin,&__pyx_n_s_poltype,&__pyx_n_s_eta,&__pyx_n_s_func,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject *)Py_None);
    values[7] = ((PyObject *)Py_None);
    values[8] = ((PyObject *)__pyx_int_0);
    values[9] = __pyx_k__4;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ks)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kpar)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate_periodic", 0, 5, 10, 1); __PYX_ERR(0, 577, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_a)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate_periodic", 0, 5, 10, 2); __PYX_ERR(0, 577, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate_periodic", 0, 5, 10, 3); __PYX_ERR(0, 577, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("translate_periodic", 0, 5, 10, 4); __PYX_ERR(0, 577, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_in);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rsin);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_poltype);
          if (value) { values[7] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eta);
          if (value) { values[8] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_func);
          if (value) { values[9] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "translate_periodic") < 0)) __PYX_ERR(0, 577, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_ks = values[0];
    __pyx_v_kpar = values[1];
    __pyx_v_a = values[2];
    __pyx_v_rs = values[3];
    __pyx_v_out = values[4];
    __pyx_v_in_ = values[5];
    __pyx_v_rsin = values[6];
    __pyx_v_poltype = values[7];
    __pyx_v_eta = values[8];
    __pyx_v_func = values[9];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("translate_periodic", 0, 5, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 577, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("treams.sw.translate_periodic", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6treams_2sw_4translate_periodic(__pyx_self, __pyx_v_ks, __pyx_v_kpar, __pyx_v_a, __pyx_v_rs, __pyx_v_out, __pyx_v_in_, __pyx_v_rsin, __pyx_v_poltype, __pyx_v_eta, __pyx_v_func);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6treams_2sw_18translate_periodic_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "treams/sw.pyx":612
 *     if in_ is None:
 *         in_ = out
 *     out = (*(np.array(o) for o in out),)             # <<<<<<<<<<<<<<
 *     in_ = (*(np.array(i) for i in in_),)
 *     if len(out) < 3 or len(out) > 4:
 */

static PyObject *__pyx_pf_6treams_2sw_18translate_periodic_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *)__pyx_tp_new_6treams_2sw___pyx_scope_struct_1_genexpr(__pyx_ptype_6treams_2sw___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 612, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6treams_2sw_18translate_periodic_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_translate_periodic_locals_genexp, __pyx_n_s_treams_sw); if (unlikely(!gen)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("treams.sw.translate_periodic.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6treams_2sw_18translate_periodic_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 612, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_out)) { __Pyx_RaiseClosureNameError("out"); __PYX_ERR(0, 612, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_out)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_out)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_out; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_out); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 612, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 612, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 612, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 612, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 612, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 612, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_cur_scope->__pyx_v_o) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 612, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6treams_2sw_18translate_periodic_5generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "treams/sw.pyx":613
 *         in_ = out
 *     out = (*(np.array(o) for o in out),)
 *     in_ = (*(np.array(i) for i in in_),)             # <<<<<<<<<<<<<<
 *     if len(out) < 3 or len(out) > 4:
 *         raise ValueError(f"invalid length of output modes {len(out)}, must be 3 or 4")
 */

static PyObject *__pyx_pf_6treams_2sw_18translate_periodic_3genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *)__pyx_tp_new_6treams_2sw___pyx_scope_struct_2_genexpr(__pyx_ptype_6treams_2sw___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 613, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6treams_2sw_18translate_periodic_5generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_translate_periodic_locals_genexp, __pyx_n_s_treams_sw); if (unlikely(!gen)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("treams.sw.translate_periodic.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6treams_2sw_18translate_periodic_5generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 613, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_in_)) { __Pyx_RaiseClosureNameError("in_"); __PYX_ERR(0, 613, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_in_)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_in_)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_in_; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_in_); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 613, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 613, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 613, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 613, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 613, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 613, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_i, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_np); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_array); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_5, __pyx_cur_scope->__pyx_v_i) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_cur_scope->__pyx_v_i);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 613, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6treams_2sw_18translate_periodic_8generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "treams/sw.pyx":648
 *     if poltype == "helicity":
 *         return _translate_periodic_h(
 *             *(o[:, None] for o in out[1:]),             # <<<<<<<<<<<<<<
 *             *in_[1:],
 *             dlms[out[0][:, None], in_[0], 0, :],
 */

static PyObject *__pyx_pf_6treams_2sw_18translate_periodic_6genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *)__pyx_tp_new_6treams_2sw___pyx_scope_struct_3_genexpr(__pyx_ptype_6treams_2sw___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 648, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6treams_2sw_18translate_periodic_8generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_translate_periodic_locals_genexp, __pyx_n_s_treams_sw); if (unlikely(!gen)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("treams.sw.translate_periodic.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6treams_2sw_18translate_periodic_8generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 648, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_out)) { __Pyx_RaiseClosureNameError("out"); __PYX_ERR(0, 648, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_out, 1, 0, NULL, NULL, &__pyx_slice__5, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 648, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 648, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 648, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 648, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_o, __pyx_tuple__7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 648, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_6treams_2sw_18translate_periodic_11generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "treams/sw.pyx":654
 *         )
 *     return _translate_periodic_p(
 *         *(o[:, None] for o in out[1:]),             # <<<<<<<<<<<<<<
 *         *in_[1:],
 *         dlms[out[0][:, None], in_[0], 0, :],
 */

static PyObject *__pyx_pf_6treams_2sw_18translate_periodic_9genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *)__pyx_tp_new_6treams_2sw___pyx_scope_struct_4_genexpr(__pyx_ptype_6treams_2sw___pyx_scope_struct_4_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 654, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_6treams_2sw_18translate_periodic_11generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_translate_periodic_locals_genexp, __pyx_n_s_treams_sw); if (unlikely(!gen)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("treams.sw.translate_periodic.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_6treams_2sw_18translate_periodic_11generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *__pyx_cur_scope = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 654, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_out)) { __Pyx_RaiseClosureNameError("out"); __PYX_ERR(0, 654, __pyx_L1_error) }
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_out, 1, 0, NULL, NULL, &__pyx_slice__5, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 654, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 654, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 654, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 654, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_cur_scope->__pyx_v_o, __pyx_tuple__7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 654, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_XGIVEREF(__pyx_t_2);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 654, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "treams/sw.pyx":577
 * 
 * 
 * def translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, poltype=None, eta=0, func=lattice.lsumsw):             # <<<<<<<<<<<<<<
 *     """translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, helicity=True, eta=0)
 * 
 */

static PyObject *__pyx_pf_6treams_2sw_4translate_periodic(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_ks, PyObject *__pyx_v_kpar, PyObject *__pyx_v_a, PyObject *__pyx_v_rs, PyObject *__pyx_v_out, PyObject *__pyx_v_in_, PyObject *__pyx_v_rsin, PyObject *__pyx_v_poltype, PyObject *__pyx_v_eta, PyObject *__pyx_v_func) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *__pyx_cur_scope;
  PyObject *__pyx_v_modes = NULL;
  PyObject *__pyx_v_rsdiff = NULL;
  PyObject *__pyx_v_dim = NULL;
  PyObject *__pyx_v_dlms = NULL;
  PyObject *__pyx_gb_6treams_2sw_18translate_periodic_2generator = 0;
  PyObject *__pyx_gb_6treams_2sw_18translate_periodic_5generator1 = 0;
  PyObject *__pyx_8genexpr2__pyx_v_l = NULL;
  PyObject *__pyx_8genexpr2__pyx_v_m = NULL;
  PyObject *__pyx_gb_6treams_2sw_18translate_periodic_8generator2 = 0;
  PyObject *__pyx_gb_6treams_2sw_18translate_periodic_11generator3 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *(*__pyx_t_15)(PyObject *);
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("translate_periodic", 0);
  __pyx_cur_scope = (struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *)__pyx_tp_new_6treams_2sw___pyx_scope_struct__translate_periodic(__pyx_ptype_6treams_2sw___pyx_scope_struct__translate_periodic, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 577, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_out = __pyx_v_out;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_out);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_out);
  __pyx_cur_scope->__pyx_v_in_ = __pyx_v_in_;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_in_);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_in_);
  __Pyx_INCREF(__pyx_v_ks);
  __Pyx_INCREF(__pyx_v_kpar);
  __Pyx_INCREF(__pyx_v_rs);
  __Pyx_INCREF(__pyx_v_rsin);
  __Pyx_INCREF(__pyx_v_poltype);

  /* "treams/sw.pyx":607
 *         complex array
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype             # <<<<<<<<<<<<<<
 *     if poltype not in ("helicity", "parity"):
 *         raise ValueError(f"invalid poltype '{poltype}'")
 */
  __pyx_t_2 = (__pyx_v_poltype == Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_config); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_POLTYPE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 607, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_poltype);
    __pyx_t_1 = __pyx_v_poltype;
  }
  __Pyx_DECREF_SET(__pyx_v_poltype, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "treams/sw.pyx":608
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype not in ("helicity", "parity"):             # <<<<<<<<<<<<<<
 *         raise ValueError(f"invalid poltype '{poltype}'")
 *     if in_ is None:
 */
  __Pyx_INCREF(__pyx_v_poltype);
  __pyx_t_1 = __pyx_v_poltype;
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_helicity, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 608, __pyx_L1_error)
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_parity, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 608, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_5)) {

    /* "treams/sw.pyx":609
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype not in ("helicity", "parity"):
 *         raise ValueError(f"invalid poltype '{poltype}'")             # <<<<<<<<<<<<<<
 *     if in_ is None:
 *         in_ = out
 */
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_7 = 127;
    __Pyx_INCREF(__pyx_kp_u_invalid_poltype);
    __pyx_t_6 += 17;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_poltype);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_invalid_poltype);
    __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_poltype, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_7;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_kp_u__2);
    __pyx_t_6 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__2);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__2);
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 609, __pyx_L1_error)

    /* "treams/sw.pyx":608
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype not in ("helicity", "parity"):             # <<<<<<<<<<<<<<
 *         raise ValueError(f"invalid poltype '{poltype}'")
 *     if in_ is None:
 */
  }

  /* "treams/sw.pyx":610
 *     if poltype not in ("helicity", "parity"):
 *         raise ValueError(f"invalid poltype '{poltype}'")
 *     if in_ is None:             # <<<<<<<<<<<<<<
 *         in_ = out
 *     out = (*(np.array(o) for o in out),)
 */
  __pyx_t_5 = (__pyx_cur_scope->__pyx_v_in_ == Py_None);
  __pyx_t_2 = (__pyx_t_5 != 0);
  if (__pyx_t_2) {

    /* "treams/sw.pyx":611
 *         raise ValueError(f"invalid poltype '{poltype}'")
 *     if in_ is None:
 *         in_ = out             # <<<<<<<<<<<<<<
 *     out = (*(np.array(o) for o in out),)
 *     in_ = (*(np.array(i) for i in in_),)
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_out);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_in_);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_in_, __pyx_cur_scope->__pyx_v_out);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_out);

    /* "treams/sw.pyx":610
 *     if poltype not in ("helicity", "parity"):
 *         raise ValueError(f"invalid poltype '{poltype}'")
 *     if in_ is None:             # <<<<<<<<<<<<<<
 *         in_ = out
 *     out = (*(np.array(o) for o in out),)
 */
  }

  /* "treams/sw.pyx":612
 *     if in_ is None:
 *         in_ = out
 *     out = (*(np.array(o) for o in out),)             # <<<<<<<<<<<<<<
 *     in_ = (*(np.array(i) for i in in_),)
 *     if len(out) < 3 or len(out) > 4:
 */
  __pyx_t_4 = __pyx_pf_6treams_2sw_18translate_periodic_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  {
    PyObject *__pyx_temp = PyList_AsTuple(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_temp; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_out);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_out, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "treams/sw.pyx":613
 *         in_ = out
 *     out = (*(np.array(o) for o in out),)
 *     in_ = (*(np.array(i) for i in in_),)             # <<<<<<<<<<<<<<
 *     if len(out) < 3 or len(out) > 4:
 *         raise ValueError(f"invalid length of output modes {len(out)}, must be 3 or 4")
 */
  __pyx_t_4 = __pyx_pf_6treams_2sw_18translate_periodic_3genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  {
    PyObject *__pyx_temp = PyList_AsTuple(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_temp; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_in_);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_in_, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "treams/sw.pyx":614
 *     out = (*(np.array(o) for o in out),)
 *     in_ = (*(np.array(i) for i in in_),)
 *     if len(out) < 3 or len(out) > 4:             # <<<<<<<<<<<<<<
 *         raise ValueError(f"invalid length of output modes {len(out)}, must be 3 or 4")
 *     elif len(out) == 3:
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_out;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((__pyx_t_6 < 3) != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_out;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 614, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((__pyx_t_6 > 4) != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L8_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "treams/sw.pyx":615
 *     in_ = (*(np.array(i) for i in in_),)
 *     if len(out) < 3 or len(out) > 4:
 *         raise ValueError(f"invalid length of output modes {len(out)}, must be 3 or 4")             # <<<<<<<<<<<<<<
 *     elif len(out) == 3:
 *         out = (np.zeros_like(out[0]),) + out
 */
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_7 = 127;
    __Pyx_INCREF(__pyx_kp_u_invalid_length_of_output_modes);
    __pyx_t_6 += 31;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_length_of_output_modes);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_invalid_length_of_output_modes);
    __pyx_t_4 = __pyx_cur_scope->__pyx_v_out;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_8 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_8, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_kp_u_must_be_3_or_4);
    __pyx_t_6 += 16;
    __Pyx_GIVEREF(__pyx_kp_u_must_be_3_or_4);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_must_be_3_or_4);
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 615, __pyx_L1_error)

    /* "treams/sw.pyx":614
 *     out = (*(np.array(o) for o in out),)
 *     in_ = (*(np.array(i) for i in in_),)
 *     if len(out) < 3 or len(out) > 4:             # <<<<<<<<<<<<<<
 *         raise ValueError(f"invalid length of output modes {len(out)}, must be 3 or 4")
 *     elif len(out) == 3:
 */
  }

  /* "treams/sw.pyx":616
 *     if len(out) < 3 or len(out) > 4:
 *         raise ValueError(f"invalid length of output modes {len(out)}, must be 3 or 4")
 *     elif len(out) == 3:             # <<<<<<<<<<<<<<
 *         out = (np.zeros_like(out[0]),) + out
 *     if len(in_) < 3 or len(in_) > 4:
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_out;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 616, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = ((__pyx_t_6 == 3) != 0);
  if (__pyx_t_2) {

    /* "treams/sw.pyx":617
 *         raise ValueError(f"invalid length of output modes {len(out)}, must be 3 or 4")
 *     elif len(out) == 3:
 *         out = (np.zeros_like(out[0]),) + out             # <<<<<<<<<<<<<<
 *     if len(in_) < 3 or len(in_) > 4:
 *         raise ValueError(f"invalid length of input modes {len(in_)}, must be 3 or 4")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_zeros_like); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_out, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_cur_scope->__pyx_v_out); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_out);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_out, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "treams/sw.pyx":616
 *     if len(out) < 3 or len(out) > 4:
 *         raise ValueError(f"invalid length of output modes {len(out)}, must be 3 or 4")
 *     elif len(out) == 3:             # <<<<<<<<<<<<<<
 *         out = (np.zeros_like(out[0]),) + out
 *     if len(in_) < 3 or len(in_) > 4:
 */
  }

  /* "treams/sw.pyx":618
 *     elif len(out) == 3:
 *         out = (np.zeros_like(out[0]),) + out
 *     if len(in_) < 3 or len(in_) > 4:             # <<<<<<<<<<<<<<
 *         raise ValueError(f"invalid length of input modes {len(in_)}, must be 3 or 4")
 *     elif len(in_) == 3:
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_in_;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((__pyx_t_6 < 3) != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_in_;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 618, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = ((__pyx_t_6 > 4) != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L11_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "treams/sw.pyx":619
 *         out = (np.zeros_like(out[0]),) + out
 *     if len(in_) < 3 or len(in_) > 4:
 *         raise ValueError(f"invalid length of input modes {len(in_)}, must be 3 or 4")             # <<<<<<<<<<<<<<
 *     elif len(in_) == 3:
 *         in_ = (np.zeros_like(in_[0]),) + in_
 */
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_7 = 127;
    __Pyx_INCREF(__pyx_kp_u_invalid_length_of_input_modes);
    __pyx_t_6 += 30;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_length_of_input_modes);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_invalid_length_of_input_modes);
    __pyx_t_3 = __pyx_cur_scope->__pyx_v_in_;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_8 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_8, 0, ' ', 'd'); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_kp_u_must_be_3_or_4);
    __pyx_t_6 += 16;
    __Pyx_GIVEREF(__pyx_kp_u_must_be_3_or_4);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_must_be_3_or_4);
    __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 619, __pyx_L1_error)

    /* "treams/sw.pyx":618
 *     elif len(out) == 3:
 *         out = (np.zeros_like(out[0]),) + out
 *     if len(in_) < 3 or len(in_) > 4:             # <<<<<<<<<<<<<<
 *         raise ValueError(f"invalid length of input modes {len(in_)}, must be 3 or 4")
 *     elif len(in_) == 3:
 */
  }

  /* "treams/sw.pyx":620
 *     if len(in_) < 3 or len(in_) > 4:
 *         raise ValueError(f"invalid length of input modes {len(in_)}, must be 3 or 4")
 *     elif len(in_) == 3:             # <<<<<<<<<<<<<<
 *         in_ = (np.zeros_like(in_[0]),) + in_
 *     if rsin is None:
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_in_;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_6 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = ((__pyx_t_6 == 3) != 0);
  if (__pyx_t_2) {

    /* "treams/sw.pyx":621
 *         raise ValueError(f"invalid length of input modes {len(in_)}, must be 3 or 4")
 *     elif len(in_) == 3:
 *         in_ = (np.zeros_like(in_[0]),) + in_             # <<<<<<<<<<<<<<
 *     if rsin is None:
 *         rsin = rs
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_zeros_like); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_in_, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_9, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Add(__pyx_t_4, __pyx_cur_scope->__pyx_v_in_); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 621, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_in_);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_in_, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "treams/sw.pyx":620
 *     if len(in_) < 3 or len(in_) > 4:
 *         raise ValueError(f"invalid length of input modes {len(in_)}, must be 3 or 4")
 *     elif len(in_) == 3:             # <<<<<<<<<<<<<<
 *         in_ = (np.zeros_like(in_[0]),) + in_
 *     if rsin is None:
 */
  }

  /* "treams/sw.pyx":622
 *     elif len(in_) == 3:
 *         in_ = (np.zeros_like(in_[0]),) + in_
 *     if rsin is None:             # <<<<<<<<<<<<<<
 *         rsin = rs
 *     modes = np.array([
 */
  __pyx_t_2 = (__pyx_v_rsin == Py_None);
  __pyx_t_5 = (__pyx_t_2 != 0);
  if (__pyx_t_5) {

    /* "treams/sw.pyx":623
 *         in_ = (np.zeros_like(in_[0]),) + in_
 *     if rsin is None:
 *         rsin = rs             # <<<<<<<<<<<<<<
 *     modes = np.array([
 *         [l, m]
 */
    __Pyx_INCREF(__pyx_v_rs);
    __Pyx_DECREF_SET(__pyx_v_rsin, __pyx_v_rs);

    /* "treams/sw.pyx":622
 *     elif len(in_) == 3:
 *         in_ = (np.zeros_like(in_[0]),) + in_
 *     if rsin is None:             # <<<<<<<<<<<<<<
 *         rsin = rs
 *     modes = np.array([
 */
  }

  /* "treams/sw.pyx":624
 *     if rsin is None:
 *         rsin = rs
 *     modes = np.array([             # <<<<<<<<<<<<<<
 *         [l, m]
 *         for l in range(np.max(out[1]) + np.max(in_[1]) + 1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  { /* enter inner scope */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 624, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "treams/sw.pyx":626
 *     modes = np.array([
 *         [l, m]
 *         for l in range(np.max(out[1]) + np.max(in_[1]) + 1)             # <<<<<<<<<<<<<<
 *         for m in range(-l, l + 1)
 *     ])
 */
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_max); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_out, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_9 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_np); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_max); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_in_, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_13 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_11 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_13, __pyx_t_10) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_10);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = PyNumber_Add(__pyx_t_9, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_t_12, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 626, __pyx_L16_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (likely(PyList_CheckExact(__pyx_t_12)) || PyTuple_CheckExact(__pyx_t_12)) {
      __pyx_t_11 = __pyx_t_12; __Pyx_INCREF(__pyx_t_11); __pyx_t_6 = 0;
      __pyx_t_14 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 626, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_14 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 626, __pyx_L16_error)
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    for (;;) {
      if (likely(!__pyx_t_14)) {
        if (likely(PyList_CheckExact(__pyx_t_11))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_11)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_12 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_6); __Pyx_INCREF(__pyx_t_12); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 626, __pyx_L16_error)
          #else
          __pyx_t_12 = PySequence_ITEM(__pyx_t_11, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 626, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_12);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_6); __Pyx_INCREF(__pyx_t_12); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 626, __pyx_L16_error)
          #else
          __pyx_t_12 = PySequence_ITEM(__pyx_t_11, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 626, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_12);
          #endif
        }
      } else {
        __pyx_t_12 = __pyx_t_14(__pyx_t_11);
        if (unlikely(!__pyx_t_12)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 626, __pyx_L16_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_12);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_l, __pyx_t_12);
      __pyx_t_12 = 0;

      /* "treams/sw.pyx":627
 *         [l, m]
 *         for l in range(np.max(out[1]) + np.max(in_[1]) + 1)
 *         for m in range(-l, l + 1)             # <<<<<<<<<<<<<<
 *     ])
 *     ks = np.array(ks)
 */
      __pyx_t_12 = PyNumber_Negative(__pyx_8genexpr2__pyx_v_l); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 627, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_9 = __Pyx_PyInt_AddObjC(__pyx_8genexpr2__pyx_v_l, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 627, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 627, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_12);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_12);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_9);
      __pyx_t_12 = 0;
      __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_10, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 627, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (likely(PyList_CheckExact(__pyx_t_9)) || PyTuple_CheckExact(__pyx_t_9)) {
        __pyx_t_10 = __pyx_t_9; __Pyx_INCREF(__pyx_t_10); __pyx_t_8 = 0;
        __pyx_t_15 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 627, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_15 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 627, __pyx_L16_error)
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      for (;;) {
        if (likely(!__pyx_t_15)) {
          if (likely(PyList_CheckExact(__pyx_t_10))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_10)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_8); __Pyx_INCREF(__pyx_t_9); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 627, __pyx_L16_error)
            #else
            __pyx_t_9 = PySequence_ITEM(__pyx_t_10, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 627, __pyx_L16_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_8); __Pyx_INCREF(__pyx_t_9); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(0, 627, __pyx_L16_error)
            #else
            __pyx_t_9 = PySequence_ITEM(__pyx_t_10, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 627, __pyx_L16_error)
            __Pyx_GOTREF(__pyx_t_9);
            #endif
          }
        } else {
          __pyx_t_9 = __pyx_t_15(__pyx_t_10);
          if (unlikely(!__pyx_t_9)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 627, __pyx_L16_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_9);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_m, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "treams/sw.pyx":625
 *         rsin = rs
 *     modes = np.array([
 *         [l, m]             # <<<<<<<<<<<<<<
 *         for l in range(np.max(out[1]) + np.max(in_[1]) + 1)
 *         for m in range(-l, l + 1)
 */
        __pyx_t_9 = PyList_New(2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 625, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_INCREF(__pyx_8genexpr2__pyx_v_l);
        __Pyx_GIVEREF(__pyx_8genexpr2__pyx_v_l);
        PyList_SET_ITEM(__pyx_t_9, 0, __pyx_8genexpr2__pyx_v_l);
        __Pyx_INCREF(__pyx_8genexpr2__pyx_v_m);
        __Pyx_GIVEREF(__pyx_8genexpr2__pyx_v_m);
        PyList_SET_ITEM(__pyx_t_9, 1, __pyx_8genexpr2__pyx_v_m);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 624, __pyx_L16_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "treams/sw.pyx":627
 *         [l, m]
 *         for l in range(np.max(out[1]) + np.max(in_[1]) + 1)
 *         for m in range(-l, l + 1)             # <<<<<<<<<<<<<<
 *     ])
 *     ks = np.array(ks)
 */
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "treams/sw.pyx":626
 *     modes = np.array([
 *         [l, m]
 *         for l in range(np.max(out[1]) + np.max(in_[1]) + 1)             # <<<<<<<<<<<<<<
 *         for m in range(-l, l + 1)
 *     ])
 */
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_l); __pyx_8genexpr2__pyx_v_l = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_m); __pyx_8genexpr2__pyx_v_m = 0;
    goto __pyx_L21_exit_scope;
    __pyx_L16_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_l); __pyx_8genexpr2__pyx_v_l = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_m); __pyx_8genexpr2__pyx_v_m = 0;
    goto __pyx_L1_error;
    __pyx_L21_exit_scope:;
  } /* exit inner scope */
  __pyx_t_11 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_11, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_modes = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "treams/sw.pyx":629
 *         for m in range(-l, l + 1)
 *     ])
 *     ks = np.array(ks)             # <<<<<<<<<<<<<<
 *     ks = ks.reshape((-1, 1))
 *     if ks.shape[0] == 2 and ks[0, 0] == ks[1, 0]:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_ks) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_ks);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_ks, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "treams/sw.pyx":630
 *     ])
 *     ks = np.array(ks)
 *     ks = ks.reshape((-1, 1))             # <<<<<<<<<<<<<<
 *     if ks.shape[0] == 2 and ks[0, 0] == ks[1, 0]:
 *         ks = ks[:1, :]
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_ks, __pyx_n_s_reshape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_tuple__8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_tuple__8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_ks, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "treams/sw.pyx":631
 *     ks = np.array(ks)
 *     ks = ks.reshape((-1, 1))
 *     if ks.shape[0] == 2 and ks[0, 0] == ks[1, 0]:             # <<<<<<<<<<<<<<
 *         ks = ks[:1, :]
 *     kpar = np.array(kpar)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_ks, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {
  } else {
    __pyx_t_5 = __pyx_t_2;
    goto __pyx_L23_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_ks, __pyx_tuple__9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_ks, __pyx_tuple__10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __pyx_t_2;
  __pyx_L23_bool_binop_done:;
  if (__pyx_t_5) {

    /* "treams/sw.pyx":632
 *     ks = ks.reshape((-1, 1))
 *     if ks.shape[0] == 2 and ks[0, 0] == ks[1, 0]:
 *         ks = ks[:1, :]             # <<<<<<<<<<<<<<
 *     kpar = np.array(kpar)
 *     rs = np.array(rs)
 */
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_ks, __pyx_tuple__12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 632, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_ks, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "treams/sw.pyx":631
 *     ks = np.array(ks)
 *     ks = ks.reshape((-1, 1))
 *     if ks.shape[0] == 2 and ks[0, 0] == ks[1, 0]:             # <<<<<<<<<<<<<<
 *         ks = ks[:1, :]
 *     kpar = np.array(kpar)
 */
  }

  /* "treams/sw.pyx":633
 *     if ks.shape[0] == 2 and ks[0, 0] == ks[1, 0]:
 *         ks = ks[:1, :]
 *     kpar = np.array(kpar)             # <<<<<<<<<<<<<<
 *     rs = np.array(rs)
 *     if rs.ndim == 1:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_np); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_v_kpar) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_kpar);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_kpar, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "treams/sw.pyx":634
 *         ks = ks[:1, :]
 *     kpar = np.array(kpar)
 *     rs = np.array(rs)             # <<<<<<<<<<<<<<
 *     if rs.ndim == 1:
 *         rs = rs.reshape((1, -1))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_np); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_array); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_v_rs) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_rs);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_rs, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "treams/sw.pyx":635
 *     kpar = np.array(kpar)
 *     rs = np.array(rs)
 *     if rs.ndim == 1:             # <<<<<<<<<<<<<<
 *         rs = rs.reshape((1, -1))
 *     rsin = np.array(rsin)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rs, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_3, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "treams/sw.pyx":636
 *     rs = np.array(rs)
 *     if rs.ndim == 1:
 *         rs = rs.reshape((1, -1))             # <<<<<<<<<<<<<<
 *     rsin = np.array(rsin)
 *     if rsin.ndim == 1:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_rs, __pyx_n_s_reshape); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_tuple__13) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_tuple__13);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 636, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_rs, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "treams/sw.pyx":635
 *     kpar = np.array(kpar)
 *     rs = np.array(rs)
 *     if rs.ndim == 1:             # <<<<<<<<<<<<<<
 *         rs = rs.reshape((1, -1))
 *     rsin = np.array(rsin)
 */
  }

  /* "treams/sw.pyx":637
 *     if rs.ndim == 1:
 *         rs = rs.reshape((1, -1))
 *     rsin = np.array(rsin)             # <<<<<<<<<<<<<<
 *     if rsin.ndim == 1:
 *         rsin = rsin.reshape((1, -1))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_np); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_array); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_rsin) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_rsin);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_rsin, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "treams/sw.pyx":638
 *         rs = rs.reshape((1, -1))
 *     rsin = np.array(rsin)
 *     if rsin.ndim == 1:             # <<<<<<<<<<<<<<
 *         rsin = rsin.reshape((1, -1))
 *     rsdiff = -rs[:, None, None, None, :] + rsin[:, None, None, :]
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_rsin, __pyx_n_s_ndim); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_t_4, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 638, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "treams/sw.pyx":639
 *     rsin = np.array(rsin)
 *     if rsin.ndim == 1:
 *         rsin = rsin.reshape((1, -1))             # <<<<<<<<<<<<<<
 *     rsdiff = -rs[:, None, None, None, :] + rsin[:, None, None, :]
 * 
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_rsin, __pyx_n_s_reshape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_tuple__13) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_tuple__13);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_rsin, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "treams/sw.pyx":638
 *         rs = rs.reshape((1, -1))
 *     rsin = np.array(rsin)
 *     if rsin.ndim == 1:             # <<<<<<<<<<<<<<
 *         rsin = rsin.reshape((1, -1))
 *     rsdiff = -rs[:, None, None, None, :] + rsin[:, None, None, :]
 */
  }

  /* "treams/sw.pyx":640
 *     if rsin.ndim == 1:
 *         rsin = rsin.reshape((1, -1))
 *     rsdiff = -rs[:, None, None, None, :] + rsin[:, None, None, :]             # <<<<<<<<<<<<<<
 * 
 *     dim = 1 if kpar.ndim == 0 else kpar.shape[-1]
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_rs, __pyx_tuple__14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyNumber_Negative(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_rsin, __pyx_tuple__15); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_rsdiff = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "treams/sw.pyx":642
 *     rsdiff = -rs[:, None, None, None, :] + rsin[:, None, None, :]
 * 
 *     dim = 1 if kpar.ndim == 0 else kpar.shape[-1]             # <<<<<<<<<<<<<<
 *     # The result has the shape (n_rs, n_rs, n_ks, n_modes)
 *     dlms = func(dim, modes[:, 0], modes[:, 1], ks, kpar, a, rsdiff, eta)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_kpar, __pyx_n_s_ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
    __Pyx_INCREF(__pyx_int_1);
    __pyx_t_3 = __pyx_int_1;
  } else {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_kpar, __pyx_n_s_shape); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_4, -1L, long, 1, __Pyx_PyInt_From_long, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_3 = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_v_dim = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "treams/sw.pyx":644
 *     dim = 1 if kpar.ndim == 0 else kpar.shape[-1]
 *     # The result has the shape (n_rs, n_rs, n_ks, n_modes)
 *     dlms = func(dim, modes[:, 0], modes[:, 1], ks, kpar, a, rsdiff, eta)             # <<<<<<<<<<<<<<
 * 
 *     if poltype == "helicity":
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_modes, __pyx_tuple__16); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_modes, __pyx_tuple__17); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_func);
  __pyx_t_11 = __pyx_v_func; __pyx_t_10 = NULL;
  __pyx_t_16 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_11, function);
      __pyx_t_16 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_11)) {
    PyObject *__pyx_temp[9] = {__pyx_t_10, __pyx_v_dim, __pyx_t_1, __pyx_t_4, __pyx_v_ks, __pyx_v_kpar, __pyx_v_a, __pyx_v_rsdiff, __pyx_v_eta};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_16, 8+__pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
    PyObject *__pyx_temp[9] = {__pyx_t_10, __pyx_v_dim, __pyx_t_1, __pyx_t_4, __pyx_v_ks, __pyx_v_kpar, __pyx_v_a, __pyx_v_rsdiff, __pyx_v_eta};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_16, 8+__pyx_t_16); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else
  #endif
  {
    __pyx_t_9 = PyTuple_New(8+__pyx_t_16); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_10) {
      __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_10); __pyx_t_10 = NULL;
    }
    __Pyx_INCREF(__pyx_v_dim);
    __Pyx_GIVEREF(__pyx_v_dim);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_16, __pyx_v_dim);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_16, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_16, __pyx_t_4);
    __Pyx_INCREF(__pyx_v_ks);
    __Pyx_GIVEREF(__pyx_v_ks);
    PyTuple_SET_ITEM(__pyx_t_9, 3+__pyx_t_16, __pyx_v_ks);
    __Pyx_INCREF(__pyx_v_kpar);
    __Pyx_GIVEREF(__pyx_v_kpar);
    PyTuple_SET_ITEM(__pyx_t_9, 4+__pyx_t_16, __pyx_v_kpar);
    __Pyx_INCREF(__pyx_v_a);
    __Pyx_GIVEREF(__pyx_v_a);
    PyTuple_SET_ITEM(__pyx_t_9, 5+__pyx_t_16, __pyx_v_a);
    __Pyx_INCREF(__pyx_v_rsdiff);
    __Pyx_GIVEREF(__pyx_v_rsdiff);
    PyTuple_SET_ITEM(__pyx_t_9, 6+__pyx_t_16, __pyx_v_rsdiff);
    __Pyx_INCREF(__pyx_v_eta);
    __Pyx_GIVEREF(__pyx_v_eta);
    PyTuple_SET_ITEM(__pyx_t_9, 7+__pyx_t_16, __pyx_v_eta);
    __pyx_t_1 = 0;
    __pyx_t_4 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 644, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_v_dlms = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "treams/sw.pyx":646
 *     dlms = func(dim, modes[:, 0], modes[:, 1], ks, kpar, a, rsdiff, eta)
 * 
 *     if poltype == "helicity":             # <<<<<<<<<<<<<<
 *         return _translate_periodic_h(
 *             *(o[:, None] for o in out[1:]),
 */
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_poltype, __pyx_n_u_helicity, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "treams/sw.pyx":647
 * 
 *     if poltype == "helicity":
 *         return _translate_periodic_h(             # <<<<<<<<<<<<<<
 *             *(o[:, None] for o in out[1:]),
 *             *in_[1:],
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_translate_periodic_h); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "treams/sw.pyx":648
 *     if poltype == "helicity":
 *         return _translate_periodic_h(
 *             *(o[:, None] for o in out[1:]),             # <<<<<<<<<<<<<<
 *             *in_[1:],
 *             dlms[out[0][:, None], in_[0], 0, :],
 */
    __pyx_t_11 = __pyx_pf_6treams_2sw_18translate_periodic_6genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 648, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "treams/sw.pyx":647
 * 
 *     if poltype == "helicity":
 *         return _translate_periodic_h(             # <<<<<<<<<<<<<<
 *             *(o[:, None] for o in out[1:]),
 *             *in_[1:],
 */
    __pyx_t_9 = __Pyx_PySequence_Tuple(__pyx_t_11); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

    /* "treams/sw.pyx":649
 *         return _translate_periodic_h(
 *             *(o[:, None] for o in out[1:]),
 *             *in_[1:],             # <<<<<<<<<<<<<<
 *             dlms[out[0][:, None], in_[0], 0, :],
 *             dlms[out[0][:, None], in_[0], ks.shape[0] - 1, :],
 */
    __pyx_t_11 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_v_in_, 1, 0, NULL, NULL, &__pyx_slice__5, 1, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 649, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "treams/sw.pyx":647
 * 
 *     if poltype == "helicity":
 *         return _translate_periodic_h(             # <<<<<<<<<<<<<<
 *             *(o[:, None] for o in out[1:]),
 *             *in_[1:],
 */
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PyNumber_Add(__pyx_t_9, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "treams/sw.pyx":650
 *             *(o[:, None] for o in out[1:]),
 *             *in_[1:],
 *             dlms[out[0][:, None], in_[0], 0, :],             # <<<<<<<<<<<<<<
 *             dlms[out[0][:, None], in_[0], ks.shape[0] - 1, :],
 *         )
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_out, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_t_4, __pyx_tuple__7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_in_, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_int_0);
    __Pyx_INCREF(__pyx_slice__6);
    __Pyx_GIVEREF(__pyx_slice__6);
    PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_slice__6);
    __pyx_t_9 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetItem(__pyx_v_dlms, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "treams/sw.pyx":651
 *             *in_[1:],
 *             dlms[out[0][:, None], in_[0], 0, :],
 *             dlms[out[0][:, None], in_[0], ks.shape[0] - 1, :],             # <<<<<<<<<<<<<<
 *         )
 *     return _translate_periodic_p(
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_out, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_tuple__7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_in_, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_ks, __pyx_n_s_shape); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_10, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyInt_SubtractObjC(__pyx_t_12, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = PyTuple_New(4); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_t_10);
    __Pyx_INCREF(__pyx_slice__6);
    __Pyx_GIVEREF(__pyx_slice__6);
    PyTuple_SET_ITEM(__pyx_t_12, 3, __pyx_slice__6);
    __pyx_t_9 = 0;
    __pyx_t_1 = 0;
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_dlms, __pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 651, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "treams/sw.pyx":647
 * 
 *     if poltype == "helicity":
 *         return _translate_periodic_h(             # <<<<<<<<<<<<<<
 *             *(o[:, None] for o in out[1:]),
 *             *in_[1:],
 */
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_10);
    __pyx_t_4 = 0;
    __pyx_t_10 = 0;
    __pyx_t_10 = PyNumber_Add(__pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 647, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_r = __pyx_t_12;
    __pyx_t_12 = 0;
    goto __pyx_L0;

    /* "treams/sw.pyx":646
 *     dlms = func(dim, modes[:, 0], modes[:, 1], ks, kpar, a, rsdiff, eta)
 * 
 *     if poltype == "helicity":             # <<<<<<<<<<<<<<
 *         return _translate_periodic_h(
 *             *(o[:, None] for o in out[1:]),
 */
  }

  /* "treams/sw.pyx":653
 *             dlms[out[0][:, None], in_[0], ks.shape[0] - 1, :],
 *         )
 *     return _translate_periodic_p(             # <<<<<<<<<<<<<<
 *         *(o[:, None] for o in out[1:]),
 *         *in_[1:],
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_translate_periodic_p); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "treams/sw.pyx":654
 *         )
 *     return _translate_periodic_p(
 *         *(o[:, None] for o in out[1:]),             # <<<<<<<<<<<<<<
 *         *in_[1:],
 *         dlms[out[0][:, None], in_[0], 0, :],
 */
  __pyx_t_10 = __pyx_pf_6treams_2sw_18translate_periodic_9genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "treams/sw.pyx":653
 *             dlms[out[0][:, None], in_[0], ks.shape[0] - 1, :],
 *         )
 *     return _translate_periodic_p(             # <<<<<<<<<<<<<<
 *         *(o[:, None] for o in out[1:]),
 *         *in_[1:],
 */
  __pyx_t_3 = __Pyx_PySequence_Tuple(__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "treams/sw.pyx":655
 *     return _translate_periodic_p(
 *         *(o[:, None] for o in out[1:]),
 *         *in_[1:],             # <<<<<<<<<<<<<<
 *         dlms[out[0][:, None], in_[0], 0, :],
 *     )
 */
  __pyx_t_10 = __Pyx_PyObject_GetSlice(__pyx_cur_scope->__pyx_v_in_, 1, 0, NULL, NULL, &__pyx_slice__5, 1, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "treams/sw.pyx":653
 *             dlms[out[0][:, None], in_[0], ks.shape[0] - 1, :],
 *         )
 *     return _translate_periodic_p(             # <<<<<<<<<<<<<<
 *         *(o[:, None] for o in out[1:]),
 *         *in_[1:],
 */
  __pyx_t_11 = __Pyx_PySequence_Tuple(__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = PyNumber_Add(__pyx_t_3, __pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "treams/sw.pyx":656
 *         *(o[:, None] for o in out[1:]),
 *         *in_[1:],
 *         dlms[out[0][:, None], in_[0], 0, :],             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_11 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_out, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_11, __pyx_tuple__7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_GetItemInt(__pyx_cur_scope->__pyx_v_in_, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_11);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_int_0);
  __Pyx_INCREF(__pyx_slice__6);
  __Pyx_GIVEREF(__pyx_slice__6);
  PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_slice__6);
  __pyx_t_3 = 0;
  __pyx_t_11 = 0;
  __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_v_dlms, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "treams/sw.pyx":653
 *             dlms[out[0][:, None], in_[0], ks.shape[0] - 1, :],
 *         )
 *     return _translate_periodic_p(             # <<<<<<<<<<<<<<
 *         *(o[:, None] for o in out[1:]),
 *         *in_[1:],
 */
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_11);
  __pyx_t_11 = 0;
  __pyx_t_11 = PyNumber_Add(__pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "treams/sw.pyx":577
 * 
 * 
 * def translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, poltype=None, eta=0, func=lattice.lsumsw):             # <<<<<<<<<<<<<<
 *     """translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, helicity=True, eta=0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("treams.sw.translate_periodic", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_modes);
  __Pyx_XDECREF(__pyx_v_rsdiff);
  __Pyx_XDECREF(__pyx_v_dim);
  __Pyx_XDECREF(__pyx_v_dlms);
  __Pyx_XDECREF(__pyx_gb_6treams_2sw_18translate_periodic_2generator);
  __Pyx_XDECREF(__pyx_gb_6treams_2sw_18translate_periodic_5generator1);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_l);
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_m);
  __Pyx_XDECREF(__pyx_gb_6treams_2sw_18translate_periodic_8generator2);
  __Pyx_XDECREF(__pyx_gb_6treams_2sw_18translate_periodic_11generator3);
  __Pyx_XDECREF(__pyx_v_ks);
  __Pyx_XDECREF(__pyx_v_kpar);
  __Pyx_XDECREF(__pyx_v_rs);
  __Pyx_XDECREF(__pyx_v_rsin);
  __Pyx_XDECREF(__pyx_v_poltype);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "treams/sw.pyx":660
 * 
 * 
 * cdef double complex _cperiodic_to_pw_h(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:             # <<<<<<<<<<<<<<
 *     if polvsw != polpw:
 *         return 0.0j
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__cperiodic_to_pw_h(double __pyx_v_kx, double __pyx_v_ky, __pyx_t_double_complex __pyx_v_kz, long __pyx_v_polpw, long __pyx_v_l, long __pyx_v_m, long __pyx_v_polvsw, double __pyx_v_area) {
  __pyx_t_double_complex __pyx_v_k;
  __pyx_t_double_complex __pyx_v_costheta;
  double __pyx_v_phi;
  __pyx_t_double_complex __pyx_v_kz_s;
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  __pyx_t_double_complex __pyx_t_5;
  __pyx_t_double_complex __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":661
 * 
 * cdef double complex _cperiodic_to_pw_h(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     if polvsw != polpw:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 */
  __pyx_t_1 = ((__pyx_v_polvsw != __pyx_v_polpw) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":662
 * cdef double complex _cperiodic_to_pw_h(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     if polvsw != polpw:
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k
 */
    __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
    goto __pyx_L0;

    /* "treams/sw.pyx":661
 * 
 * cdef double complex _cperiodic_to_pw_h(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     if polvsw != polpw:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 */
  }

  /* "treams/sw.pyx":663
 *     if polvsw != polpw:
 *         return 0.0j
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)             # <<<<<<<<<<<<<<
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)
 */
  __pyx_v_k = __pyx_fuse_0__pyx_f_6treams_7special_5_misc_sqrt(__Pyx_c_sum_double(__pyx_t_double_complex_from_parts(((__pyx_v_kx * __pyx_v_kx) + (__pyx_v_ky * __pyx_v_ky)), 0), __Pyx_c_prod_double(__pyx_v_kz, __pyx_v_kz)));

  /* "treams/sw.pyx":664
 *         return 0.0j
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k             # <<<<<<<<<<<<<<
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 */
  if (unlikely(__Pyx_c_is_zero_double(__pyx_v_k))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 664, __pyx_L1_error)
  }
  __pyx_v_costheta = __Pyx_c_quot_double(__pyx_v_kz, __pyx_v_k);

  /* "treams/sw.pyx":665
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)             # <<<<<<<<<<<<<<
 *     cdef double complex kz_s = kz
 *     if kz == 0:
 */
  __pyx_v_phi = atan2(__pyx_v_ky, __pyx_v_kx);

  /* "treams/sw.pyx":666
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz             # <<<<<<<<<<<<<<
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 */
  __pyx_v_kz_s = __pyx_v_kz;

  /* "treams/sw.pyx":667
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 *     if kz == 0:             # <<<<<<<<<<<<<<
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 */
  __pyx_t_1 = ((__Pyx_c_eq_double(__pyx_v_kz, __pyx_t_double_complex_from_parts(0, 0))) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":668
 *     cdef double complex kz_s = kz
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j             # <<<<<<<<<<<<<<
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 *         kz_s = -kz_s
 */
    __pyx_v_kz_s = __Pyx_c_sum_double(__pyx_t_double_complex_from_parts(1e-20, 0), __pyx_t_double_complex_from_parts(0, 1e-20));

    /* "treams/sw.pyx":667
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 *     if kz == 0:             # <<<<<<<<<<<<<<
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 */
    goto __pyx_L4;
  }

  /* "treams/sw.pyx":669
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):             # <<<<<<<<<<<<<<
 *         kz_s = -kz_s
 *     return (
 */
  __pyx_t_2 = ((cimag(__pyx_v_kz_s) < 0.0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((cimag(__pyx_v_kz_s) == 0.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((creal(__pyx_v_kz_s) < 0.0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/sw.pyx":670
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 *         kz_s = -kz_s             # <<<<<<<<<<<<<<
 *     return (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))
 */
    __pyx_v_kz_s = __Pyx_c_neg_double(__pyx_v_kz_s);

    /* "treams/sw.pyx":669
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):             # <<<<<<<<<<<<<<
 *         kz_s = -kz_s
 *     return (
 */
  }
  __pyx_L4:;

  /* "treams/sw.pyx":672
 *         kz_s = -kz_s
 *     return (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))             # <<<<<<<<<<<<<<
 *         * expd((lgamma(l - m + 1) - lgamma(l + m + 1)) * 0.5)
 *         * cexp(1j * m * phi)
 */
  __pyx_t_3 = (M_PI * ((2 * __pyx_v_l) + 1));
  __pyx_t_4 = ((double)(__pyx_v_l * (__pyx_v_l + 1)));
  if (unlikely(__pyx_t_4 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 672, __pyx_L1_error)
  }

  /* "treams/sw.pyx":676
 *         * cexp(1j * m * phi)
 *         * cpow(-1j, l)
 *         * (cs.tau_fun(l, m, costheta) + (2 * polpw - 1) * cs.pi_fun(l, m, costheta))             # <<<<<<<<<<<<<<
 *         / (area * k * kz_s)
 *     )
 */
  __pyx_t_5 = __Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts((sqrt((__pyx_t_3 / __pyx_t_4)) * exp(((lgamma(((__pyx_v_l - __pyx_v_m) + 1)) - lgamma(((__pyx_v_l + __pyx_v_m) + 1))) * 0.5))), 0), cexp(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts(__pyx_v_m, 0)), __pyx_t_double_complex_from_parts(__pyx_v_phi, 0)))), cpow(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_t_double_complex_from_parts(__pyx_v_l, 0))), __Pyx_c_sum_double(__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tau_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0), __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((2 * __pyx_v_polpw) - 1), 0), __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_pi_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0))));

  /* "treams/sw.pyx":677
 *         * cpow(-1j, l)
 *         * (cs.tau_fun(l, m, costheta) + (2 * polpw - 1) * cs.pi_fun(l, m, costheta))
 *         / (area * k * kz_s)             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_6 = __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(__pyx_v_area, 0), __pyx_v_k), __pyx_v_kz_s);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_6))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 677, __pyx_L1_error)
  }
  __pyx_r = __Pyx_c_quot_double(__pyx_t_5, __pyx_t_6);
  goto __pyx_L0;

  /* "treams/sw.pyx":660
 * 
 * 
 * cdef double complex _cperiodic_to_pw_h(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:             # <<<<<<<<<<<<<<
 *     if polvsw != polpw:
 *         return 0.0j
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._cperiodic_to_pw_h", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__cperiodic_to_pw_h(double __pyx_v_kx, double __pyx_v_ky, double __pyx_v_kz, long __pyx_v_polpw, long __pyx_v_l, long __pyx_v_m, long __pyx_v_polvsw, double __pyx_v_area) {
  double __pyx_v_k;
  __pyx_t_double_complex __pyx_v_costheta;
  double __pyx_v_phi;
  __pyx_t_double_complex __pyx_v_kz_s;
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  __pyx_t_double_complex __pyx_t_5;
  __pyx_t_double_complex __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":661
 * 
 * cdef double complex _cperiodic_to_pw_h(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     if polvsw != polpw:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 */
  __pyx_t_1 = ((__pyx_v_polvsw != __pyx_v_polpw) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":662
 * cdef double complex _cperiodic_to_pw_h(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     if polvsw != polpw:
 *         return 0.0j             # <<<<<<<<<<<<<<
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k
 */
    __pyx_r = __pyx_t_double_complex_from_parts(0, 0.0);
    goto __pyx_L0;

    /* "treams/sw.pyx":661
 * 
 * cdef double complex _cperiodic_to_pw_h(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     if polvsw != polpw:             # <<<<<<<<<<<<<<
 *         return 0.0j
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 */
  }

  /* "treams/sw.pyx":663
 *     if polvsw != polpw:
 *         return 0.0j
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)             # <<<<<<<<<<<<<<
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)
 */
  __pyx_v_k = __pyx_fuse_1__pyx_f_6treams_7special_5_misc_sqrt((((__pyx_v_kx * __pyx_v_kx) + (__pyx_v_ky * __pyx_v_ky)) + (__pyx_v_kz * __pyx_v_kz)));

  /* "treams/sw.pyx":664
 *         return 0.0j
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k             # <<<<<<<<<<<<<<
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 */
  if (unlikely(__pyx_v_k == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 664, __pyx_L1_error)
  }
  __pyx_v_costheta = __pyx_t_double_complex_from_parts((__pyx_v_kz / __pyx_v_k), 0);

  /* "treams/sw.pyx":665
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)             # <<<<<<<<<<<<<<
 *     cdef double complex kz_s = kz
 *     if kz == 0:
 */
  __pyx_v_phi = atan2(__pyx_v_ky, __pyx_v_kx);

  /* "treams/sw.pyx":666
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz             # <<<<<<<<<<<<<<
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 */
  __pyx_v_kz_s = __pyx_t_double_complex_from_parts(__pyx_v_kz, 0);

  /* "treams/sw.pyx":667
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 *     if kz == 0:             # <<<<<<<<<<<<<<
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 */
  __pyx_t_1 = ((__pyx_v_kz == 0.0) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":668
 *     cdef double complex kz_s = kz
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j             # <<<<<<<<<<<<<<
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 *         kz_s = -kz_s
 */
    __pyx_v_kz_s = __Pyx_c_sum_double(__pyx_t_double_complex_from_parts(1e-20, 0), __pyx_t_double_complex_from_parts(0, 1e-20));

    /* "treams/sw.pyx":667
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 *     if kz == 0:             # <<<<<<<<<<<<<<
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 */
    goto __pyx_L4;
  }

  /* "treams/sw.pyx":669
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):             # <<<<<<<<<<<<<<
 *         kz_s = -kz_s
 *     return (
 */
  __pyx_t_2 = ((cimag(__pyx_v_kz_s) < 0.0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((cimag(__pyx_v_kz_s) == 0.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_2 = ((creal(__pyx_v_kz_s) < 0.0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/sw.pyx":670
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 *         kz_s = -kz_s             # <<<<<<<<<<<<<<
 *     return (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))
 */
    __pyx_v_kz_s = __Pyx_c_neg_double(__pyx_v_kz_s);

    /* "treams/sw.pyx":669
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):             # <<<<<<<<<<<<<<
 *         kz_s = -kz_s
 *     return (
 */
  }
  __pyx_L4:;

  /* "treams/sw.pyx":672
 *         kz_s = -kz_s
 *     return (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))             # <<<<<<<<<<<<<<
 *         * expd((lgamma(l - m + 1) - lgamma(l + m + 1)) * 0.5)
 *         * cexp(1j * m * phi)
 */
  __pyx_t_3 = (M_PI * ((2 * __pyx_v_l) + 1));
  __pyx_t_4 = ((double)(__pyx_v_l * (__pyx_v_l + 1)));
  if (unlikely(__pyx_t_4 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 672, __pyx_L1_error)
  }

  /* "treams/sw.pyx":676
 *         * cexp(1j * m * phi)
 *         * cpow(-1j, l)
 *         * (cs.tau_fun(l, m, costheta) + (2 * polpw - 1) * cs.pi_fun(l, m, costheta))             # <<<<<<<<<<<<<<
 *         / (area * k * kz_s)
 *     )
 */
  __pyx_t_5 = __Pyx_c_prod_double(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts((sqrt((__pyx_t_3 / __pyx_t_4)) * exp(((lgamma(((__pyx_v_l - __pyx_v_m) + 1)) - lgamma(((__pyx_v_l + __pyx_v_m) + 1))) * 0.5))), 0), cexp(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts(__pyx_v_m, 0)), __pyx_t_double_complex_from_parts(__pyx_v_phi, 0)))), cpow(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_t_double_complex_from_parts(__pyx_v_l, 0))), __Pyx_c_sum_double(__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tau_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0), __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((2 * __pyx_v_polpw) - 1), 0), __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_pi_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0))));

  /* "treams/sw.pyx":677
 *         * cpow(-1j, l)
 *         * (cs.tau_fun(l, m, costheta) + (2 * polpw - 1) * cs.pi_fun(l, m, costheta))
 *         / (area * k * kz_s)             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_6 = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts((__pyx_v_area * __pyx_v_k), 0), __pyx_v_kz_s);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_6))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 677, __pyx_L1_error)
  }
  __pyx_r = __Pyx_c_quot_double(__pyx_t_5, __pyx_t_6);
  goto __pyx_L0;

  /* "treams/sw.pyx":660
 * 
 * 
 * cdef double complex _cperiodic_to_pw_h(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:             # <<<<<<<<<<<<<<
 *     if polvsw != polpw:
 *         return 0.0j
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._cperiodic_to_pw_h", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/sw.pyx":681
 * 
 * 
 * cdef double complex _cperiodic_to_pw_p(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:             # <<<<<<<<<<<<<<
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k
 */

static __pyx_t_double_complex __pyx_fuse_0__pyx_f_6treams_2sw__cperiodic_to_pw_p(double __pyx_v_kx, double __pyx_v_ky, __pyx_t_double_complex __pyx_v_kz, long __pyx_v_polpw, long __pyx_v_l, long __pyx_v_m, long __pyx_v_polvsw, double __pyx_v_area) {
  __pyx_t_double_complex __pyx_v_k;
  __pyx_t_double_complex __pyx_v_costheta;
  double __pyx_v_phi;
  __pyx_t_double_complex __pyx_v_kz_s;
  __pyx_t_double_complex __pyx_v_prefactor;
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  __pyx_t_double_complex __pyx_t_5;
  __pyx_t_double_complex __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":682
 * 
 * cdef double complex _cperiodic_to_pw_p(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)             # <<<<<<<<<<<<<<
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)
 */
  __pyx_v_k = __pyx_fuse_0__pyx_f_6treams_7special_5_misc_sqrt(__Pyx_c_sum_double(__pyx_t_double_complex_from_parts(((__pyx_v_kx * __pyx_v_kx) + (__pyx_v_ky * __pyx_v_ky)), 0), __Pyx_c_prod_double(__pyx_v_kz, __pyx_v_kz)));

  /* "treams/sw.pyx":683
 * cdef double complex _cperiodic_to_pw_p(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k             # <<<<<<<<<<<<<<
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 */
  if (unlikely(__Pyx_c_is_zero_double(__pyx_v_k))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 683, __pyx_L1_error)
  }
  __pyx_v_costheta = __Pyx_c_quot_double(__pyx_v_kz, __pyx_v_k);

  /* "treams/sw.pyx":684
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)             # <<<<<<<<<<<<<<
 *     cdef double complex kz_s = kz
 *     if kz == 0:
 */
  __pyx_v_phi = atan2(__pyx_v_ky, __pyx_v_kx);

  /* "treams/sw.pyx":685
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz             # <<<<<<<<<<<<<<
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 */
  __pyx_v_kz_s = __pyx_v_kz;

  /* "treams/sw.pyx":686
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 *     if kz == 0:             # <<<<<<<<<<<<<<
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 */
  __pyx_t_1 = ((__Pyx_c_eq_double(__pyx_v_kz, __pyx_t_double_complex_from_parts(0, 0))) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":687
 *     cdef double complex kz_s = kz
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j             # <<<<<<<<<<<<<<
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 *         kz_s = -kz_s
 */
    __pyx_v_kz_s = __Pyx_c_sum_double(__pyx_t_double_complex_from_parts(1e-20, 0), __pyx_t_double_complex_from_parts(0, 1e-20));

    /* "treams/sw.pyx":686
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 *     if kz == 0:             # <<<<<<<<<<<<<<
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 */
    goto __pyx_L3;
  }

  /* "treams/sw.pyx":688
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):             # <<<<<<<<<<<<<<
 *         kz_s = -kz_s
 *     cdef double complex prefactor = (
 */
  __pyx_t_2 = ((cimag(__pyx_v_kz_s) < 0.0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((cimag(__pyx_v_kz_s) == 0.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((creal(__pyx_v_kz_s) < 0.0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/sw.pyx":689
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 *         kz_s = -kz_s             # <<<<<<<<<<<<<<
 *     cdef double complex prefactor = (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))
 */
    __pyx_v_kz_s = __Pyx_c_neg_double(__pyx_v_kz_s);

    /* "treams/sw.pyx":688
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):             # <<<<<<<<<<<<<<
 *         kz_s = -kz_s
 *     cdef double complex prefactor = (
 */
  }
  __pyx_L3:;

  /* "treams/sw.pyx":691
 *         kz_s = -kz_s
 *     cdef double complex prefactor = (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))             # <<<<<<<<<<<<<<
 *         * expd((lgamma(l - m + 1) - lgamma(l + m + 1)) * 0.5)
 *         * cexp(1j * m * phi)
 */
  __pyx_t_3 = (M_PI * ((2 * __pyx_v_l) + 1));
  __pyx_t_4 = ((double)(__pyx_v_l * (__pyx_v_l + 1)));
  if (unlikely(__pyx_t_4 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 691, __pyx_L1_error)
  }

  /* "treams/sw.pyx":694
 *         * expd((lgamma(l - m + 1) - lgamma(l + m + 1)) * 0.5)
 *         * cexp(1j * m * phi)
 *         * cpow(-1j, l)             # <<<<<<<<<<<<<<
 *         / (area * k * kz_s)
 *     )
 */
  __pyx_t_5 = __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts((sqrt((__pyx_t_3 / __pyx_t_4)) * exp(((lgamma(((__pyx_v_l - __pyx_v_m) + 1)) - lgamma(((__pyx_v_l + __pyx_v_m) + 1))) * 0.5))), 0), cexp(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts(__pyx_v_m, 0)), __pyx_t_double_complex_from_parts(__pyx_v_phi, 0)))), cpow(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_t_double_complex_from_parts(__pyx_v_l, 0)));

  /* "treams/sw.pyx":695
 *         * cexp(1j * m * phi)
 *         * cpow(-1j, l)
 *         / (area * k * kz_s)             # <<<<<<<<<<<<<<
 *     )
 *     if polvsw == polpw:
 */
  __pyx_t_6 = __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(__pyx_v_area, 0), __pyx_v_k), __pyx_v_kz_s);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_6))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 695, __pyx_L1_error)
  }
  __pyx_v_prefactor = __Pyx_c_quot_double(__pyx_t_5, __pyx_t_6);

  /* "treams/sw.pyx":697
 *         / (area * k * kz_s)
 *     )
 *     if polvsw == polpw:             # <<<<<<<<<<<<<<
 *         return prefactor * cs.tau_fun(l, m, costheta)
 *     return prefactor * cs.pi_fun(l, m, costheta)
 */
  __pyx_t_1 = ((__pyx_v_polvsw == __pyx_v_polpw) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":698
 *     )
 *     if polvsw == polpw:
 *         return prefactor * cs.tau_fun(l, m, costheta)             # <<<<<<<<<<<<<<
 *     return prefactor * cs.pi_fun(l, m, costheta)
 * 
 */
    __pyx_r = __Pyx_c_prod_double(__pyx_v_prefactor, __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tau_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0));
    goto __pyx_L0;

    /* "treams/sw.pyx":697
 *         / (area * k * kz_s)
 *     )
 *     if polvsw == polpw:             # <<<<<<<<<<<<<<
 *         return prefactor * cs.tau_fun(l, m, costheta)
 *     return prefactor * cs.pi_fun(l, m, costheta)
 */
  }

  /* "treams/sw.pyx":699
 *     if polvsw == polpw:
 *         return prefactor * cs.tau_fun(l, m, costheta)
 *     return prefactor * cs.pi_fun(l, m, costheta)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __Pyx_c_prod_double(__pyx_v_prefactor, __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_pi_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0));
  goto __pyx_L0;

  /* "treams/sw.pyx":681
 * 
 * 
 * cdef double complex _cperiodic_to_pw_p(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:             # <<<<<<<<<<<<<<
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._cperiodic_to_pw_p", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

static __pyx_t_double_complex __pyx_fuse_1__pyx_f_6treams_2sw__cperiodic_to_pw_p(double __pyx_v_kx, double __pyx_v_ky, double __pyx_v_kz, long __pyx_v_polpw, long __pyx_v_l, long __pyx_v_m, long __pyx_v_polvsw, double __pyx_v_area) {
  double __pyx_v_k;
  __pyx_t_double_complex __pyx_v_costheta;
  double __pyx_v_phi;
  __pyx_t_double_complex __pyx_v_kz_s;
  __pyx_t_double_complex __pyx_v_prefactor;
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  __pyx_t_double_complex __pyx_t_5;
  __pyx_t_double_complex __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":682
 * 
 * cdef double complex _cperiodic_to_pw_p(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)             # <<<<<<<<<<<<<<
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)
 */
  __pyx_v_k = __pyx_fuse_1__pyx_f_6treams_7special_5_misc_sqrt((((__pyx_v_kx * __pyx_v_kx) + (__pyx_v_ky * __pyx_v_ky)) + (__pyx_v_kz * __pyx_v_kz)));

  /* "treams/sw.pyx":683
 * cdef double complex _cperiodic_to_pw_p(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k             # <<<<<<<<<<<<<<
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 */
  if (unlikely(__pyx_v_k == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 683, __pyx_L1_error)
  }
  __pyx_v_costheta = __pyx_t_double_complex_from_parts((__pyx_v_kz / __pyx_v_k), 0);

  /* "treams/sw.pyx":684
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)             # <<<<<<<<<<<<<<
 *     cdef double complex kz_s = kz
 *     if kz == 0:
 */
  __pyx_v_phi = atan2(__pyx_v_ky, __pyx_v_kx);

  /* "treams/sw.pyx":685
 *     cdef double complex costheta = kz / k
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz             # <<<<<<<<<<<<<<
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 */
  __pyx_v_kz_s = __pyx_t_double_complex_from_parts(__pyx_v_kz, 0);

  /* "treams/sw.pyx":686
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 *     if kz == 0:             # <<<<<<<<<<<<<<
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 */
  __pyx_t_1 = ((__pyx_v_kz == 0.0) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":687
 *     cdef double complex kz_s = kz
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j             # <<<<<<<<<<<<<<
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 *         kz_s = -kz_s
 */
    __pyx_v_kz_s = __Pyx_c_sum_double(__pyx_t_double_complex_from_parts(1e-20, 0), __pyx_t_double_complex_from_parts(0, 1e-20));

    /* "treams/sw.pyx":686
 *     cdef double phi = atan2(ky, kx)
 *     cdef double complex kz_s = kz
 *     if kz == 0:             # <<<<<<<<<<<<<<
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 */
    goto __pyx_L3;
  }

  /* "treams/sw.pyx":688
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):             # <<<<<<<<<<<<<<
 *         kz_s = -kz_s
 *     cdef double complex prefactor = (
 */
  __pyx_t_2 = ((cimag(__pyx_v_kz_s) < 0.0) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((cimag(__pyx_v_kz_s) == 0.0) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((creal(__pyx_v_kz_s) < 0.0) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/sw.pyx":689
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):
 *         kz_s = -kz_s             # <<<<<<<<<<<<<<
 *     cdef double complex prefactor = (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))
 */
    __pyx_v_kz_s = __Pyx_c_neg_double(__pyx_v_kz_s);

    /* "treams/sw.pyx":688
 *     if kz == 0:
 *         kz_s = 1e-20 + 1e-20j
 *     elif (cimag(kz_s) < 0) or (cimag(kz_s) == 0 and creal(kz_s) < 0):             # <<<<<<<<<<<<<<
 *         kz_s = -kz_s
 *     cdef double complex prefactor = (
 */
  }
  __pyx_L3:;

  /* "treams/sw.pyx":691
 *         kz_s = -kz_s
 *     cdef double complex prefactor = (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))             # <<<<<<<<<<<<<<
 *         * expd((lgamma(l - m + 1) - lgamma(l + m + 1)) * 0.5)
 *         * cexp(1j * m * phi)
 */
  __pyx_t_3 = (M_PI * ((2 * __pyx_v_l) + 1));
  __pyx_t_4 = ((double)(__pyx_v_l * (__pyx_v_l + 1)));
  if (unlikely(__pyx_t_4 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 691, __pyx_L1_error)
  }

  /* "treams/sw.pyx":694
 *         * expd((lgamma(l - m + 1) - lgamma(l + m + 1)) * 0.5)
 *         * cexp(1j * m * phi)
 *         * cpow(-1j, l)             # <<<<<<<<<<<<<<
 *         / (area * k * kz_s)
 *     )
 */
  __pyx_t_5 = __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts((sqrt((__pyx_t_3 / __pyx_t_4)) * exp(((lgamma(((__pyx_v_l - __pyx_v_m) + 1)) - lgamma(((__pyx_v_l + __pyx_v_m) + 1))) * 0.5))), 0), cexp(__Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(0, 1.0), __pyx_t_double_complex_from_parts(__pyx_v_m, 0)), __pyx_t_double_complex_from_parts(__pyx_v_phi, 0)))), cpow(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_t_double_complex_from_parts(__pyx_v_l, 0)));

  /* "treams/sw.pyx":695
 *         * cexp(1j * m * phi)
 *         * cpow(-1j, l)
 *         / (area * k * kz_s)             # <<<<<<<<<<<<<<
 *     )
 *     if polvsw == polpw:
 */
  __pyx_t_6 = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts((__pyx_v_area * __pyx_v_k), 0), __pyx_v_kz_s);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_6))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 695, __pyx_L1_error)
  }
  __pyx_v_prefactor = __Pyx_c_quot_double(__pyx_t_5, __pyx_t_6);

  /* "treams/sw.pyx":697
 *         / (area * k * kz_s)
 *     )
 *     if polvsw == polpw:             # <<<<<<<<<<<<<<
 *         return prefactor * cs.tau_fun(l, m, costheta)
 *     return prefactor * cs.pi_fun(l, m, costheta)
 */
  __pyx_t_1 = ((__pyx_v_polvsw == __pyx_v_polpw) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":698
 *     )
 *     if polvsw == polpw:
 *         return prefactor * cs.tau_fun(l, m, costheta)             # <<<<<<<<<<<<<<
 *     return prefactor * cs.pi_fun(l, m, costheta)
 * 
 */
    __pyx_r = __Pyx_c_prod_double(__pyx_v_prefactor, __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tau_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0));
    goto __pyx_L0;

    /* "treams/sw.pyx":697
 *         / (area * k * kz_s)
 *     )
 *     if polvsw == polpw:             # <<<<<<<<<<<<<<
 *         return prefactor * cs.tau_fun(l, m, costheta)
 *     return prefactor * cs.pi_fun(l, m, costheta)
 */
  }

  /* "treams/sw.pyx":699
 *     if polvsw == polpw:
 *         return prefactor * cs.tau_fun(l, m, costheta)
 *     return prefactor * cs.pi_fun(l, m, costheta)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __Pyx_c_prod_double(__pyx_v_prefactor, __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_pi_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0));
  goto __pyx_L0;

  /* "treams/sw.pyx":681
 * 
 * 
 * cdef double complex _cperiodic_to_pw_p(double kx, double ky, number_t kz, long polpw, long l, long m, long polvsw, double area) nogil:             # <<<<<<<<<<<<<<
 *     cdef number_t k = sqrt(kx * kx + ky * ky + kz * kz)
 *     cdef double complex costheta = kz / k
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._cperiodic_to_pw_p", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/sw.pyx":702
 * 
 * 
 * cdef void _loop_periodic_d(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

static void __pyx_f_6treams_2sw__loop_periodic_d(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_ip7;
  char *__pyx_v_op0;
  __pyx_t_double_complex __pyx_v_ov0;
  npy_intp __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;

  /* "treams/sw.pyx":703
 * 
 * cdef void _loop_periodic_d(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/sw.pyx":704
 * cdef void _loop_periodic_d(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/sw.pyx":705
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/sw.pyx":706
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/sw.pyx":707
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/sw.pyx":708
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/sw.pyx":709
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/sw.pyx":710
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/sw.pyx":711
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *ip7 = args[7]
 *     cdef char *op0 = args[8]
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/sw.pyx":712
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[8]
 *     cdef double complex ov0
 */
  __pyx_v_ip7 = (__pyx_v_args[7]);

  /* "treams/sw.pyx":713
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 *     cdef char *op0 = args[8]             # <<<<<<<<<<<<<<
 *     cdef double complex ov0
 *     for i in range(n):
 */
  __pyx_v_op0 = (__pyx_v_args[8]);

  /* "treams/sw.pyx":715
 *     cdef char *op0 = args[8]
 *     cdef double complex ov0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ov0 = (<double complex(*)(double, double, double, long, long, long, long, double) nogil>func)(
 *             <double>(<double*>ip0)[0],
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "treams/sw.pyx":716
 *     cdef double complex ov0
 *     for i in range(n):
 *         ov0 = (<double complex(*)(double, double, double, long, long, long, long, double) nogil>func)(             # <<<<<<<<<<<<<<
 *             <double>(<double*>ip0)[0],
 *             <double>(<double*>ip1)[0],
 */
    __pyx_v_ov0 = ((__pyx_t_double_complex (*)(double, double, double, long, long, long, long, double))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((double)(((double *)__pyx_v_ip1)[0])), ((double)(((double *)__pyx_v_ip2)[0])), ((long)(((long *)__pyx_v_ip3)[0])), ((long)(((long *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), ((long)(((long *)__pyx_v_ip6)[0])), ((double)(((double *)__pyx_v_ip7)[0])));

    /* "treams/sw.pyx":726
 *             <double>(<double*>ip7)[0],
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    (((__pyx_t_double_complex *)__pyx_v_op0)[0]) = __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL(__pyx_v_ov0)), ((double)__Pyx_CIMAG(__pyx_v_ov0)));

    /* "treams/sw.pyx":727
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/sw.pyx":728
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/sw.pyx":729
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/sw.pyx":730
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/sw.pyx":731
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/sw.pyx":732
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/sw.pyx":733
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         ip7 += steps[7]
 *         op0 += steps[8]
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/sw.pyx":734
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         ip7 += steps[7]             # <<<<<<<<<<<<<<
 *         op0 += steps[8]
 * 
 */
    __pyx_v_ip7 = (__pyx_v_ip7 + (__pyx_v_steps[7]));

    /* "treams/sw.pyx":735
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 *         op0 += steps[8]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[8]));
  }

  /* "treams/sw.pyx":702
 * 
 * 
 * cdef void _loop_periodic_d(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

  /* function exit code */
}

/* "treams/sw.pyx":738
 * 
 * 
 * cdef void _loop_periodic_D(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

static void __pyx_f_6treams_2sw__loop_periodic_D(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_ip7;
  char *__pyx_v_op0;
  __pyx_t_double_complex __pyx_v_ov0;
  npy_intp __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;

  /* "treams/sw.pyx":739
 * 
 * cdef void _loop_periodic_D(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/sw.pyx":740
 * cdef void _loop_periodic_D(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/sw.pyx":741
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/sw.pyx":742
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/sw.pyx":743
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/sw.pyx":744
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/sw.pyx":745
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/sw.pyx":746
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/sw.pyx":747
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *ip7 = args[7]
 *     cdef char *op0 = args[8]
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/sw.pyx":748
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[8]
 *     cdef double complex ov0
 */
  __pyx_v_ip7 = (__pyx_v_args[7]);

  /* "treams/sw.pyx":749
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 *     cdef char *op0 = args[8]             # <<<<<<<<<<<<<<
 *     cdef double complex ov0
 *     for i in range(n):
 */
  __pyx_v_op0 = (__pyx_v_args[8]);

  /* "treams/sw.pyx":751
 *     cdef char *op0 = args[8]
 *     cdef double complex ov0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ov0 = (<double complex(*)(double, double, double complex, long, long, long, long, double) nogil>func)(
 *             <double>(<double*>ip0)[0],
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "treams/sw.pyx":752
 *     cdef double complex ov0
 *     for i in range(n):
 *         ov0 = (<double complex(*)(double, double, double complex, long, long, long, long, double) nogil>func)(             # <<<<<<<<<<<<<<
 *             <double>(<double*>ip0)[0],
 *             <double>(<double*>ip1)[0],
 */
    __pyx_v_ov0 = ((__pyx_t_double_complex (*)(double, double, __pyx_t_double_complex, long, long, long, long, double))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((double)(((double *)__pyx_v_ip1)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), ((long)(((long *)__pyx_v_ip3)[0])), ((long)(((long *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), ((long)(((long *)__pyx_v_ip6)[0])), ((double)(((double *)__pyx_v_ip7)[0])));

    /* "treams/sw.pyx":762
 *             <double>(<double*>ip7)[0],
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    (((__pyx_t_double_complex *)__pyx_v_op0)[0]) = __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL(__pyx_v_ov0)), ((double)__Pyx_CIMAG(__pyx_v_ov0)));

    /* "treams/sw.pyx":763
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/sw.pyx":764
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/sw.pyx":765
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/sw.pyx":766
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/sw.pyx":767
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/sw.pyx":768
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/sw.pyx":769
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         ip7 += steps[7]
 *         op0 += steps[8]
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/sw.pyx":770
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         ip7 += steps[7]             # <<<<<<<<<<<<<<
 *         op0 += steps[8]
 * 
 */
    __pyx_v_ip7 = (__pyx_v_ip7 + (__pyx_v_steps[7]));

    /* "treams/sw.pyx":771
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 *         op0 += steps[8]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[8]));
  }

  /* "treams/sw.pyx":738
 * 
 * 
 * cdef void _loop_periodic_D(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

  /* function exit code */
}

/* "treams/sw.pyx":830
 * 
 * 
 * def periodic_to_pw(kx, ky, kz, pol, l, m, qol, area, poltype=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """periodic_to_pw(kx, ky, kz, pol, l, m, qol, area, helicity=True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6treams_2sw_7periodic_to_pw(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6treams_2sw_6periodic_to_pw[] = "periodic_to_pw(kx, ky, kz, pol, l, m, qol, area, helicity=True)\n\n    Convert periodic spherical wave to plane wave.\n\n    Returns the coefficient for the basis change in a periodic arrangement of spherical\n    modes to plane waves. For multiple positions only diagonal values (with respect to\n    the position) are returned. A correct phase factor is still necessary for the full\n    result.\n\n    The polarization values `0` and `1` refer to negative and positive helicity\n    waves or, if ``helicity == False``, to TE and TM parity.\n\n    Args:\n        kx (float, array_like): X component of destination mode wave vector\n        ky (float, array_like): Y component of destination mode wave vector\n        kz (float or complex, array_like): Z component of destination mode wave vector\n        pol (int, array_like): Polarization of the destination mode\n        l (int, array_like): Degree of the source mode\n        m (int, array_like): Order of the source mode\n        qol (int, array_like): Polarization of the source mode\n        area (float, array_like): Unit cell area\n        helicity (bool, optional): If true, helicity basis is assumed, else parity basis.\n            Defaults to ``True``.\n\n    Returns:\n        complex\n    ";
static PyMethodDef __pyx_mdef_6treams_2sw_7periodic_to_pw = {"periodic_to_pw", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6treams_2sw_7periodic_to_pw, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6treams_2sw_6periodic_to_pw};
static PyObject *__pyx_pw_6treams_2sw_7periodic_to_pw(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_kx = 0;
  PyObject *__pyx_v_ky = 0;
  PyObject *__pyx_v_kz = 0;
  PyObject *__pyx_v_pol = 0;
  PyObject *__pyx_v_l = 0;
  PyObject *__pyx_v_m = 0;
  PyObject *__pyx_v_qol = 0;
  PyObject *__pyx_v_area = 0;
  PyObject *__pyx_v_poltype = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("periodic_to_pw (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 9) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 9, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_kx,&__pyx_n_s_ky,&__pyx_n_s_kz,&__pyx_n_s_pol,&__pyx_n_s_l,&__pyx_n_s_m,&__pyx_n_s_qol,&__pyx_n_s_area,&__pyx_n_s_poltype,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[8] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kx)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ky)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_pw", 0, 8, 9, 1); __PYX_ERR(0, 830, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_pw", 0, 8, 9, 2); __PYX_ERR(0, 830, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_pw", 0, 8, 9, 3); __PYX_ERR(0, 830, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_l)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_pw", 0, 8, 9, 4); __PYX_ERR(0, 830, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_pw", 0, 8, 9, 5); __PYX_ERR(0, 830, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_pw", 0, 8, 9, 6); __PYX_ERR(0, 830, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_area)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_pw", 0, 8, 9, 7); __PYX_ERR(0, 830, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_poltype);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 9) ? pos_args : 9;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "periodic_to_pw") < 0)) __PYX_ERR(0, 830, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        default:
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        case  7:
        case  6:
        case  5:
        case  4:
        case  3:
        case  2:
        case  1:
        case  0:
        goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_kx = values[0];
    __pyx_v_ky = values[1];
    __pyx_v_kz = values[2];
    __pyx_v_pol = values[3];
    __pyx_v_l = values[4];
    __pyx_v_m = values[5];
    __pyx_v_qol = values[6];
    __pyx_v_area = values[7];
    __pyx_v_poltype = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("periodic_to_pw", 0, 8, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 830, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("treams.sw.periodic_to_pw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6treams_2sw_6periodic_to_pw(__pyx_self, __pyx_v_kx, __pyx_v_ky, __pyx_v_kz, __pyx_v_pol, __pyx_v_l, __pyx_v_m, __pyx_v_qol, __pyx_v_area, __pyx_v_poltype, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6treams_2sw_6periodic_to_pw(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_kx, PyObject *__pyx_v_ky, PyObject *__pyx_v_kz, PyObject *__pyx_v_pol, PyObject *__pyx_v_l, PyObject *__pyx_v_m, PyObject *__pyx_v_qol, PyObject *__pyx_v_area, PyObject *__pyx_v_poltype, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("periodic_to_pw", 0);
  __Pyx_INCREF(__pyx_v_poltype);

  /* "treams/sw.pyx":858
 *         complex
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype             # <<<<<<<<<<<<<<
 *     if poltype == "helicity":
 *         return _periodic_to_pw_h(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 */
  __pyx_t_2 = (__pyx_v_poltype == Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_config); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_POLTYPE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 858, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_poltype);
    __pyx_t_1 = __pyx_v_poltype;
  }
  __Pyx_DECREF_SET(__pyx_v_poltype, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "treams/sw.pyx":859
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":             # <<<<<<<<<<<<<<
 *         return _periodic_to_pw_h(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 *     elif poltype == "parity":
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_poltype, __pyx_n_u_helicity, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 859, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "treams/sw.pyx":860
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":
 *         return _periodic_to_pw_h(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)             # <<<<<<<<<<<<<<
 *     elif poltype == "parity":
 *         return _periodic_to_pw_p(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_periodic_to_pw_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_kx);
    __Pyx_GIVEREF(__pyx_v_kx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_kx);
    __Pyx_INCREF(__pyx_v_ky);
    __Pyx_GIVEREF(__pyx_v_ky);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_ky);
    __Pyx_INCREF(__pyx_v_kz);
    __Pyx_GIVEREF(__pyx_v_kz);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_kz);
    __Pyx_INCREF(__pyx_v_pol);
    __Pyx_GIVEREF(__pyx_v_pol);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_pol);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_m);
    __Pyx_GIVEREF(__pyx_v_m);
    PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_m);
    __Pyx_INCREF(__pyx_v_qol);
    __Pyx_GIVEREF(__pyx_v_qol);
    PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_qol);
    __Pyx_INCREF(__pyx_v_area);
    __Pyx_GIVEREF(__pyx_v_area);
    PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_v_area);
    __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "treams/sw.pyx":859
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":             # <<<<<<<<<<<<<<
 *         return _periodic_to_pw_h(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 *     elif poltype == "parity":
 */
  }

  /* "treams/sw.pyx":861
 *     if poltype == "helicity":
 *         return _periodic_to_pw_h(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 *     elif poltype == "parity":             # <<<<<<<<<<<<<<
 *         return _periodic_to_pw_p(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 *     raise ValueError(f"invalid poltype '{poltype}'")
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_poltype, __pyx_n_u_parity, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 861, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "treams/sw.pyx":862
 *         return _periodic_to_pw_h(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 *     elif poltype == "parity":
 *         return _periodic_to_pw_p(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)             # <<<<<<<<<<<<<<
 *     raise ValueError(f"invalid poltype '{poltype}'")
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_periodic_to_pw_p); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyTuple_New(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_kx);
    __Pyx_GIVEREF(__pyx_v_kx);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_kx);
    __Pyx_INCREF(__pyx_v_ky);
    __Pyx_GIVEREF(__pyx_v_ky);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_ky);
    __Pyx_INCREF(__pyx_v_kz);
    __Pyx_GIVEREF(__pyx_v_kz);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_kz);
    __Pyx_INCREF(__pyx_v_pol);
    __Pyx_GIVEREF(__pyx_v_pol);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_pol);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_m);
    __Pyx_GIVEREF(__pyx_v_m);
    PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_m);
    __Pyx_INCREF(__pyx_v_qol);
    __Pyx_GIVEREF(__pyx_v_qol);
    PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_qol);
    __Pyx_INCREF(__pyx_v_area);
    __Pyx_GIVEREF(__pyx_v_area);
    PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_v_area);
    __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "treams/sw.pyx":861
 *     if poltype == "helicity":
 *         return _periodic_to_pw_h(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 *     elif poltype == "parity":             # <<<<<<<<<<<<<<
 *         return _periodic_to_pw_p(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 *     raise ValueError(f"invalid poltype '{poltype}'")
 */
  }

  /* "treams/sw.pyx":863
 *     elif poltype == "parity":
 *         return _periodic_to_pw_p(kx, ky, kz, pol, l, m, qol, area, *args, **kwargs)
 *     raise ValueError(f"invalid poltype '{poltype}'")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = 0;
  __pyx_t_7 = 127;
  __Pyx_INCREF(__pyx_kp_u_invalid_poltype);
  __pyx_t_6 += 17;
  __Pyx_GIVEREF(__pyx_kp_u_invalid_poltype);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_invalid_poltype);
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_poltype, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_7;
  __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_6 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 863, __pyx_L1_error)

  /* "treams/sw.pyx":830
 * 
 * 
 * def periodic_to_pw(kx, ky, kz, pol, l, m, qol, area, poltype=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """periodic_to_pw(kx, ky, kz, pol, l, m, qol, area, helicity=True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("treams.sw.periodic_to_pw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_poltype);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "treams/sw.pyx":866
 * 
 * 
 * cdef double complex _cperiodic_to_cw_h(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:             # <<<<<<<<<<<<<<
 *     if mu != m or polsw != polcw:
 *         return 0
 */

static __pyx_t_double_complex __pyx_f_6treams_2sw__cperiodic_to_cw_h(double __pyx_v_kz, long __pyx_v_mu, long __pyx_v_polcw, long __pyx_v_l, long __pyx_v_m, long __pyx_v_polsw, __pyx_t_double_complex __pyx_v_k, double __pyx_v_a) {
  __pyx_t_double_complex __pyx_v_costheta;
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  __pyx_t_double_complex __pyx_t_3;
  double __pyx_t_4;
  double __pyx_t_5;
  __pyx_t_double_complex __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":867
 * 
 * cdef double complex _cperiodic_to_cw_h(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:
 *     if mu != m or polsw != polcw:             # <<<<<<<<<<<<<<
 *         return 0
 *     cdef double complex costheta = kz / k
 */
  __pyx_t_2 = ((__pyx_v_mu != __pyx_v_m) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_polsw != __pyx_v_polcw) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "treams/sw.pyx":868
 * cdef double complex _cperiodic_to_cw_h(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:
 *     if mu != m or polsw != polcw:
 *         return 0             # <<<<<<<<<<<<<<
 *     cdef double complex costheta = kz / k
 *     return (
 */
    __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
    goto __pyx_L0;

    /* "treams/sw.pyx":867
 * 
 * cdef double complex _cperiodic_to_cw_h(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:
 *     if mu != m or polsw != polcw:             # <<<<<<<<<<<<<<
 *         return 0
 *     cdef double complex costheta = kz / k
 */
  }

  /* "treams/sw.pyx":869
 *     if mu != m or polsw != polcw:
 *         return 0
 *     cdef double complex costheta = kz / k             # <<<<<<<<<<<<<<
 *     return (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))
 */
  __pyx_t_3 = __pyx_t_double_complex_from_parts(__pyx_v_kz, 0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_v_k))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 869, __pyx_L1_error)
  }
  __pyx_v_costheta = __Pyx_c_quot_double(__pyx_t_3, __pyx_v_k);

  /* "treams/sw.pyx":871
 *     cdef double complex costheta = kz / k
 *     return (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))             # <<<<<<<<<<<<<<
 *         * expd((lgamma(l - m + 1) - lgamma(l + m + 1)) * 0.5)
 *         * 0.5
 */
  __pyx_t_4 = (M_PI * ((2 * __pyx_v_l) + 1));
  __pyx_t_5 = ((double)(__pyx_v_l * (__pyx_v_l + 1)));
  if (unlikely(__pyx_t_5 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 871, __pyx_L1_error)
  }

  /* "treams/sw.pyx":875
 *         * 0.5
 *         * cpow(-1j, l - m)
 *         * (cs.tau_fun(l, m, costheta) + (2 * polcw - 1) * cs.pi_fun(l, m, costheta))             # <<<<<<<<<<<<<<
 *         / (a * k)
 *     )
 */
  __pyx_t_3 = __Pyx_c_prod_double(__Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((sqrt((__pyx_t_4 / __pyx_t_5)) * exp(((lgamma(((__pyx_v_l - __pyx_v_m) + 1)) - lgamma(((__pyx_v_l + __pyx_v_m) + 1))) * 0.5))) * 0.5), 0), cpow(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_t_double_complex_from_parts((__pyx_v_l - __pyx_v_m), 0))), __Pyx_c_sum_double(__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tau_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0), __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((2 * __pyx_v_polcw) - 1), 0), __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_pi_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0))));

  /* "treams/sw.pyx":876
 *         * cpow(-1j, l - m)
 *         * (cs.tau_fun(l, m, costheta) + (2 * polcw - 1) * cs.pi_fun(l, m, costheta))
 *         / (a * k)             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_6 = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(__pyx_v_a, 0), __pyx_v_k);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_6))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 876, __pyx_L1_error)
  }
  __pyx_r = __Pyx_c_quot_double(__pyx_t_3, __pyx_t_6);
  goto __pyx_L0;

  /* "treams/sw.pyx":866
 * 
 * 
 * cdef double complex _cperiodic_to_cw_h(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:             # <<<<<<<<<<<<<<
 *     if mu != m or polsw != polcw:
 *         return 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._cperiodic_to_cw_h", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/sw.pyx":880
 * 
 * 
 * cdef double complex _cperiodic_to_cw_p(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:             # <<<<<<<<<<<<<<
 *     if mu != m:
 *         return 0
 */

static __pyx_t_double_complex __pyx_f_6treams_2sw__cperiodic_to_cw_p(double __pyx_v_kz, long __pyx_v_mu, long __pyx_v_polcw, long __pyx_v_l, long __pyx_v_m, long __pyx_v_polsw, __pyx_t_double_complex __pyx_v_k, double __pyx_v_a) {
  __pyx_t_double_complex __pyx_v_costheta;
  __pyx_t_double_complex __pyx_v_prefactor;
  __pyx_t_double_complex __pyx_r;
  int __pyx_t_1;
  __pyx_t_double_complex __pyx_t_2;
  double __pyx_t_3;
  double __pyx_t_4;
  __pyx_t_double_complex __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/sw.pyx":881
 * 
 * cdef double complex _cperiodic_to_cw_p(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:
 *     if mu != m:             # <<<<<<<<<<<<<<
 *         return 0
 *     cdef double complex costheta = kz / k
 */
  __pyx_t_1 = ((__pyx_v_mu != __pyx_v_m) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":882
 * cdef double complex _cperiodic_to_cw_p(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:
 *     if mu != m:
 *         return 0             # <<<<<<<<<<<<<<
 *     cdef double complex costheta = kz / k
 *     cdef double complex prefactor = (
 */
    __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
    goto __pyx_L0;

    /* "treams/sw.pyx":881
 * 
 * cdef double complex _cperiodic_to_cw_p(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:
 *     if mu != m:             # <<<<<<<<<<<<<<
 *         return 0
 *     cdef double complex costheta = kz / k
 */
  }

  /* "treams/sw.pyx":883
 *     if mu != m:
 *         return 0
 *     cdef double complex costheta = kz / k             # <<<<<<<<<<<<<<
 *     cdef double complex prefactor = (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))
 */
  __pyx_t_2 = __pyx_t_double_complex_from_parts(__pyx_v_kz, 0);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_v_k))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 883, __pyx_L1_error)
  }
  __pyx_v_costheta = __Pyx_c_quot_double(__pyx_t_2, __pyx_v_k);

  /* "treams/sw.pyx":885
 *     cdef double complex costheta = kz / k
 *     cdef double complex prefactor = (
 *         sqrtd(pi * (2 * l + 1) / <double>(l * (l + 1)))             # <<<<<<<<<<<<<<
 *         * expd((lgamma(l - m + 1) - lgamma(l + m + 1)) * 0.5)
 *         * 0.5
 */
  __pyx_t_3 = (M_PI * ((2 * __pyx_v_l) + 1));
  __pyx_t_4 = ((double)(__pyx_v_l * (__pyx_v_l + 1)));
  if (unlikely(__pyx_t_4 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 885, __pyx_L1_error)
  }

  /* "treams/sw.pyx":888
 *         * expd((lgamma(l - m + 1) - lgamma(l + m + 1)) * 0.5)
 *         * 0.5
 *         * cpow(-1j, l - m)             # <<<<<<<<<<<<<<
 *         / (a * k)
 *     )
 */
  __pyx_t_2 = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(((sqrt((__pyx_t_3 / __pyx_t_4)) * exp(((lgamma(((__pyx_v_l - __pyx_v_m) + 1)) - lgamma(((__pyx_v_l + __pyx_v_m) + 1))) * 0.5))) * 0.5), 0), cpow(__Pyx_c_neg_double(__pyx_t_double_complex_from_parts(0, 1.0)), __pyx_t_double_complex_from_parts((__pyx_v_l - __pyx_v_m), 0)));

  /* "treams/sw.pyx":889
 *         * 0.5
 *         * cpow(-1j, l - m)
 *         / (a * k)             # <<<<<<<<<<<<<<
 *     )
 *     if polcw == polsw:
 */
  __pyx_t_5 = __Pyx_c_prod_double(__pyx_t_double_complex_from_parts(__pyx_v_a, 0), __pyx_v_k);
  if (unlikely(__Pyx_c_is_zero_double(__pyx_t_5))) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 889, __pyx_L1_error)
  }
  __pyx_v_prefactor = __Pyx_c_quot_double(__pyx_t_2, __pyx_t_5);

  /* "treams/sw.pyx":891
 *         / (a * k)
 *     )
 *     if polcw == polsw:             # <<<<<<<<<<<<<<
 *         return prefactor * cs.tau_fun(l, m, costheta)
 *     return prefactor * cs.pi_fun(l, m, costheta)
 */
  __pyx_t_1 = ((__pyx_v_polcw == __pyx_v_polsw) != 0);
  if (__pyx_t_1) {

    /* "treams/sw.pyx":892
 *     )
 *     if polcw == polsw:
 *         return prefactor * cs.tau_fun(l, m, costheta)             # <<<<<<<<<<<<<<
 *     return prefactor * cs.pi_fun(l, m, costheta)
 * 
 */
    __pyx_r = __Pyx_c_prod_double(__pyx_v_prefactor, __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tau_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0));
    goto __pyx_L0;

    /* "treams/sw.pyx":891
 *         / (a * k)
 *     )
 *     if polcw == polsw:             # <<<<<<<<<<<<<<
 *         return prefactor * cs.tau_fun(l, m, costheta)
 *     return prefactor * cs.pi_fun(l, m, costheta)
 */
  }

  /* "treams/sw.pyx":893
 *     if polcw == polsw:
 *         return prefactor * cs.tau_fun(l, m, costheta)
 *     return prefactor * cs.pi_fun(l, m, costheta)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __Pyx_c_prod_double(__pyx_v_prefactor, __pyx_fuse_0__pyx_f_6treams_7special_14cython_special_pi_fun(__pyx_v_l, __pyx_v_m, __pyx_v_costheta, 0));
  goto __pyx_L0;

  /* "treams/sw.pyx":880
 * 
 * 
 * cdef double complex _cperiodic_to_cw_p(double kz, long mu, long polcw, long l, long m, long polsw, double complex k, double a) nogil:             # <<<<<<<<<<<<<<
 *     if mu != m:
 *         return 0
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.sw._cperiodic_to_cw_p", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_r = __pyx_t_double_complex_from_parts(0, 0);
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/sw.pyx":896
 * 
 * 
 * cdef void _loop_periodic_cw_D(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

static void __pyx_f_6treams_2sw__loop_periodic_cw_D(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_ip7;
  char *__pyx_v_op0;
  __pyx_t_double_complex __pyx_v_ov0;
  npy_intp __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;

  /* "treams/sw.pyx":897
 * 
 * cdef void _loop_periodic_cw_D(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/sw.pyx":898
 * cdef void _loop_periodic_cw_D(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/sw.pyx":899
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/sw.pyx":900
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/sw.pyx":901
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/sw.pyx":902
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/sw.pyx":903
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/sw.pyx":904
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/sw.pyx":905
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *ip7 = args[7]
 *     cdef char *op0 = args[8]
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/sw.pyx":906
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[8]
 *     cdef double complex ov0
 */
  __pyx_v_ip7 = (__pyx_v_args[7]);

  /* "treams/sw.pyx":907
 *     cdef char *ip6 = args[6]
 *     cdef char *ip7 = args[7]
 *     cdef char *op0 = args[8]             # <<<<<<<<<<<<<<
 *     cdef double complex ov0
 *     for i in range(n):
 */
  __pyx_v_op0 = (__pyx_v_args[8]);

  /* "treams/sw.pyx":909
 *     cdef char *op0 = args[8]
 *     cdef double complex ov0
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         ov0 = (<double complex(*)(double, long, long, long, long, long, double complex, double) nogil>func)(
 *             <double>(<double*>ip0)[0],
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "treams/sw.pyx":910
 *     cdef double complex ov0
 *     for i in range(n):
 *         ov0 = (<double complex(*)(double, long, long, long, long, long, double complex, double) nogil>func)(             # <<<<<<<<<<<<<<
 *             <double>(<double*>ip0)[0],
 *             <long>(<long*>ip1)[0],
 */
    __pyx_v_ov0 = ((__pyx_t_double_complex (*)(double, long, long, long, long, long, __pyx_t_double_complex, double))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((long)(((long *)__pyx_v_ip2)[0])), ((long)(((long *)__pyx_v_ip3)[0])), ((long)(((long *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip6)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip6)[0])))), ((double)(((double *)__pyx_v_ip7)[0])));

    /* "treams/sw.pyx":920
 *             <double>(<double*>ip7)[0],
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    (((__pyx_t_double_complex *)__pyx_v_op0)[0]) = __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL(__pyx_v_ov0)), ((double)__Pyx_CIMAG(__pyx_v_ov0)));

    /* "treams/sw.pyx":921
 *         )
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/sw.pyx":922
 *         (<double complex*>op0)[0] = <double complex>ov0
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/sw.pyx":923
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/sw.pyx":924
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/sw.pyx":925
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/sw.pyx":926
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/sw.pyx":927
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         ip7 += steps[7]
 *         op0 += steps[8]
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/sw.pyx":928
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         ip7 += steps[7]             # <<<<<<<<<<<<<<
 *         op0 += steps[8]
 * 
 */
    __pyx_v_ip7 = (__pyx_v_ip7 + (__pyx_v_steps[7]));

    /* "treams/sw.pyx":929
 *         ip6 += steps[6]
 *         ip7 += steps[7]
 *         op0 += steps[8]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[8]));
  }

  /* "treams/sw.pyx":896
 * 
 * 
 * cdef void _loop_periodic_cw_D(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef void *func = <void*>data
 */

  /* function exit code */
}

/* "treams/sw.pyx":976
 * 
 * 
 * def periodic_to_cw(kz, m, pol, l, mu, qol, k, area, poltype=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """periodic_to_cw(kz, m, pol, l, mu, qol, k, area, helicity=True)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6treams_2sw_9periodic_to_cw(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_6treams_2sw_8periodic_to_cw[] = "periodic_to_cw(kz, m, pol, l, mu, qol, k, area, helicity=True)\n\n    Convert periodic spherical wave to plane wave.\n\n    Returns the coefficient for the basis change in a periodic arrangement of spherical\n    modes to plane waves. For multiple positions only diagonal values (with respect to\n    the position) are returned. A correct phase factor is still necessary for the full\n    result.\n\n    The polarization values `0` and `1` refer to negative and positive helicity\n    waves or, if ``helicity == False``, to TE and TM parity.\n\n    Args:\n        kz (float, array_like): Z component of destination mode wave vector\n        m (int, array_like): Order or the destination mode\n        pol (int, array_like): Polarization of the destination mode\n        l (int, array_like): Degree of the source mode\n        mu (int, array_like): Order of the source mode\n        qol (int, array_like): Polarization of the source mode\n        k (float or complex, array_like): Wave number\n        area (float, array_like): Unit cell area\n        helicity (bool, optional): If true, helicity basis is assumed, else parity basis.\n            Defaults to ``True``.\n\n    Returns:\n        complex\n    ";
static PyMethodDef __pyx_mdef_6treams_2sw_9periodic_to_cw = {"periodic_to_cw", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_6treams_2sw_9periodic_to_cw, METH_VARARGS|METH_KEYWORDS, __pyx_doc_6treams_2sw_8periodic_to_cw};
static PyObject *__pyx_pw_6treams_2sw_9periodic_to_cw(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_kz = 0;
  PyObject *__pyx_v_m = 0;
  PyObject *__pyx_v_pol = 0;
  PyObject *__pyx_v_l = 0;
  PyObject *__pyx_v_mu = 0;
  PyObject *__pyx_v_qol = 0;
  PyObject *__pyx_v_k = 0;
  PyObject *__pyx_v_area = 0;
  PyObject *__pyx_v_poltype = 0;
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("periodic_to_cw (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  if (PyTuple_GET_SIZE(__pyx_args) > 9) {
    __pyx_v_args = PyTuple_GetSlice(__pyx_args, 9, PyTuple_GET_SIZE(__pyx_args));
    if (unlikely(!__pyx_v_args)) {
      __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
      __Pyx_RefNannyFinishContext();
      return NULL;
    }
    __Pyx_GOTREF(__pyx_v_args);
  } else {
    __pyx_v_args = __pyx_empty_tuple; __Pyx_INCREF(__pyx_empty_tuple);
  }
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_kz,&__pyx_n_s_m,&__pyx_n_s_pol,&__pyx_n_s_l,&__pyx_n_s_mu,&__pyx_n_s_qol,&__pyx_n_s_k,&__pyx_n_s_area,&__pyx_n_s_poltype,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    values[8] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        default:
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kz)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_cw", 0, 8, 9, 1); __PYX_ERR(0, 976, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_cw", 0, 8, 9, 2); __PYX_ERR(0, 976, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_l)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_cw", 0, 8, 9, 3); __PYX_ERR(0, 976, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_cw", 0, 8, 9, 4); __PYX_ERR(0, 976, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_qol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_cw", 0, 8, 9, 5); __PYX_ERR(0, 976, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_cw", 0, 8, 9, 6); __PYX_ERR(0, 976, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_area)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("periodic_to_cw", 0, 8, 9, 7); __PYX_ERR(0, 976, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_poltype);
          if (value) { values[8] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t used_pos_args = (pos_args < 9) ? pos_args : 9;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, used_pos_args, "periodic_to_cw") < 0)) __PYX_ERR(0, 976, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        default:
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        case  7:
        case  6:
        case  5:
        case  4:
        case  3:
        case  2:
        case  1:
        case  0:
        goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_kz = values[0];
    __pyx_v_m = values[1];
    __pyx_v_pol = values[2];
    __pyx_v_l = values[3];
    __pyx_v_mu = values[4];
    __pyx_v_qol = values[5];
    __pyx_v_k = values[6];
    __pyx_v_area = values[7];
    __pyx_v_poltype = values[8];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("periodic_to_cw", 0, 8, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 976, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_args); __pyx_v_args = 0;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("treams.sw.periodic_to_cw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6treams_2sw_8periodic_to_cw(__pyx_self, __pyx_v_kz, __pyx_v_m, __pyx_v_pol, __pyx_v_l, __pyx_v_mu, __pyx_v_qol, __pyx_v_k, __pyx_v_area, __pyx_v_poltype, __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6treams_2sw_8periodic_to_cw(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_kz, PyObject *__pyx_v_m, PyObject *__pyx_v_pol, PyObject *__pyx_v_l, PyObject *__pyx_v_mu, PyObject *__pyx_v_qol, PyObject *__pyx_v_k, PyObject *__pyx_v_area, PyObject *__pyx_v_poltype, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("periodic_to_cw", 0);
  __Pyx_INCREF(__pyx_v_poltype);

  /* "treams/sw.pyx":1004
 *         complex
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype             # <<<<<<<<<<<<<<
 *     if poltype == "helicity":
 *         return _periodic_to_cw_h(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 */
  __pyx_t_2 = (__pyx_v_poltype == Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_config); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1004, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_POLTYPE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1004, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(__pyx_v_poltype);
    __pyx_t_1 = __pyx_v_poltype;
  }
  __Pyx_DECREF_SET(__pyx_v_poltype, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "treams/sw.pyx":1005
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":             # <<<<<<<<<<<<<<
 *         return _periodic_to_cw_h(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 *     elif poltype == "parity":
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_poltype, __pyx_n_u_helicity, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1005, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "treams/sw.pyx":1006
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":
 *         return _periodic_to_cw_h(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)             # <<<<<<<<<<<<<<
 *     elif poltype == "parity":
 *         return _periodic_to_cw_p(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_periodic_to_cw_h); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_kz);
    __Pyx_GIVEREF(__pyx_v_kz);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_kz);
    __Pyx_INCREF(__pyx_v_m);
    __Pyx_GIVEREF(__pyx_v_m);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_m);
    __Pyx_INCREF(__pyx_v_pol);
    __Pyx_GIVEREF(__pyx_v_pol);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_pol);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_mu);
    __Pyx_GIVEREF(__pyx_v_mu);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_mu);
    __Pyx_INCREF(__pyx_v_qol);
    __Pyx_GIVEREF(__pyx_v_qol);
    PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_qol);
    __Pyx_INCREF(__pyx_v_k);
    __Pyx_GIVEREF(__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_k);
    __Pyx_INCREF(__pyx_v_area);
    __Pyx_GIVEREF(__pyx_v_area);
    PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_v_area);
    __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "treams/sw.pyx":1005
 *     """
 *     poltype = config.POLTYPE if poltype is None else poltype
 *     if poltype == "helicity":             # <<<<<<<<<<<<<<
 *         return _periodic_to_cw_h(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 *     elif poltype == "parity":
 */
  }

  /* "treams/sw.pyx":1007
 *     if poltype == "helicity":
 *         return _periodic_to_cw_h(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 *     elif poltype == "parity":             # <<<<<<<<<<<<<<
 *         return _periodic_to_cw_p(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 *     raise ValueError(f"invalid poltype '{poltype}'")
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_poltype, __pyx_n_u_parity, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1007, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "treams/sw.pyx":1008
 *         return _periodic_to_cw_h(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 *     elif poltype == "parity":
 *         return _periodic_to_cw_p(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)             # <<<<<<<<<<<<<<
 *     raise ValueError(f"invalid poltype '{poltype}'")
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_periodic_to_cw_p); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PyTuple_New(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_kz);
    __Pyx_GIVEREF(__pyx_v_kz);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_kz);
    __Pyx_INCREF(__pyx_v_m);
    __Pyx_GIVEREF(__pyx_v_m);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_m);
    __Pyx_INCREF(__pyx_v_pol);
    __Pyx_GIVEREF(__pyx_v_pol);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_pol);
    __Pyx_INCREF(__pyx_v_l);
    __Pyx_GIVEREF(__pyx_v_l);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_v_l);
    __Pyx_INCREF(__pyx_v_mu);
    __Pyx_GIVEREF(__pyx_v_mu);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_v_mu);
    __Pyx_INCREF(__pyx_v_qol);
    __Pyx_GIVEREF(__pyx_v_qol);
    PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_v_qol);
    __Pyx_INCREF(__pyx_v_k);
    __Pyx_GIVEREF(__pyx_v_k);
    PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_v_k);
    __Pyx_INCREF(__pyx_v_area);
    __Pyx_GIVEREF(__pyx_v_area);
    PyTuple_SET_ITEM(__pyx_t_4, 7, __pyx_v_area);
    __pyx_t_3 = PyNumber_Add(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1008, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "treams/sw.pyx":1007
 *     if poltype == "helicity":
 *         return _periodic_to_cw_h(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 *     elif poltype == "parity":             # <<<<<<<<<<<<<<
 *         return _periodic_to_cw_p(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 *     raise ValueError(f"invalid poltype '{poltype}'")
 */
  }

  /* "treams/sw.pyx":1009
 *     elif poltype == "parity":
 *         return _periodic_to_cw_p(kz, m, pol, l, mu, qol, k, area, *args, **kwargs)
 *     raise ValueError(f"invalid poltype '{poltype}'")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = 0;
  __pyx_t_7 = 127;
  __Pyx_INCREF(__pyx_kp_u_invalid_poltype);
  __pyx_t_6 += 17;
  __Pyx_GIVEREF(__pyx_kp_u_invalid_poltype);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_invalid_poltype);
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_v_poltype, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_7;
  __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__2);
  __pyx_t_6 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__2);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__2);
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 1009, __pyx_L1_error)

  /* "treams/sw.pyx":976
 * 
 * 
 * def periodic_to_cw(kz, m, pol, l, mu, qol, k, area, poltype=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """periodic_to_cw(kz, m, pol, l, mu, qol, k, area, helicity=True)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("treams.sw.periodic_to_cw", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_poltype);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":736
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":739
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":742
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":745
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":748
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":752
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":754
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":930
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":931
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":934
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":936
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":937
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":943
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 943, __pyx_L3_error)

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":944
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 944, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 945, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 945, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":949
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 949, __pyx_L3_error)

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":950
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 950, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 951, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 951, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":955
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 955, __pyx_L3_error)

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":956
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 956, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":957
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 957, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 957, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":979
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":994
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1004
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1011
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1018
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_misc.pxd":29
 * cdef long ipow(long base, long exponent) nogil
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a > b else b)
 * 
 */

static CYTHON_INLINE double __pyx_fuse_0__pyx_f_6treams_7special_5_misc_max(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  double __pyx_t_1;

  /* "treams/special/_misc.pxd":30
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:
 *     return (a if a > b else b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 */
  if (((__pyx_v_a > __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":29
 * cdef long ipow(long base, long exponent) nogil
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a > b else b)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static CYTHON_INLINE long __pyx_fuse_1__pyx_f_6treams_7special_5_misc_max(long __pyx_v_a, long __pyx_v_b) {
  long __pyx_r;
  long __pyx_t_1;

  /* "treams/special/_misc.pxd":30
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:
 *     return (a if a > b else b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 */
  if (((__pyx_v_a > __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":29
 * cdef long ipow(long base, long exponent) nogil
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a > b else b)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

static CYTHON_INLINE double __pyx_fuse_0__pyx_f_6treams_7special_5_misc_min(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  double __pyx_t_1;

  /* "treams/special/_misc.pxd":33
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 *     return (a if a < b else b)             # <<<<<<<<<<<<<<
 */
  if (((__pyx_v_a < __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static CYTHON_INLINE long __pyx_fuse_1__pyx_f_6treams_7special_5_misc_min(long __pyx_v_a, long __pyx_v_b) {
  long __pyx_r;
  long __pyx_t_1;

  /* "treams/special/_misc.pxd":33
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 *     return (a if a < b else b)             # <<<<<<<<<<<<<<
 */
  if (((__pyx_v_a < __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *__pyx_freelist_6treams_2sw___pyx_scope_struct__translate_periodic[8];
static int __pyx_freecount_6treams_2sw___pyx_scope_struct__translate_periodic = 0;

static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct__translate_periodic(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6treams_2sw___pyx_scope_struct__translate_periodic > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic)))) {
    o = (PyObject*)__pyx_freelist_6treams_2sw___pyx_scope_struct__translate_periodic[--__pyx_freecount_6treams_2sw___pyx_scope_struct__translate_periodic];
    memset(o, 0, sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct__translate_periodic(PyObject *o) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_in_);
  Py_CLEAR(p->__pyx_v_out);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6treams_2sw___pyx_scope_struct__translate_periodic < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic)))) {
    __pyx_freelist_6treams_2sw___pyx_scope_struct__translate_periodic[__pyx_freecount_6treams_2sw___pyx_scope_struct__translate_periodic++] = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6treams_2sw___pyx_scope_struct__translate_periodic(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *)o;
  if (p->__pyx_v_in_) {
    e = (*v)(p->__pyx_v_in_, a); if (e) return e;
  }
  if (p->__pyx_v_out) {
    e = (*v)(p->__pyx_v_out, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_6treams_2sw___pyx_scope_struct__translate_periodic(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic *)o;
  tmp = ((PyObject*)p->__pyx_v_in_);
  p->__pyx_v_in_ = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_out);
  p->__pyx_v_out = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_6treams_2sw___pyx_scope_struct__translate_periodic = {
  PyVarObject_HEAD_INIT(0, 0)
  "treams.sw.__pyx_scope_struct__translate_periodic", /*tp_name*/
  sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct__translate_periodic), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct__translate_periodic, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6treams_2sw___pyx_scope_struct__translate_periodic, /*tp_traverse*/
  __pyx_tp_clear_6treams_2sw___pyx_scope_struct__translate_periodic, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6treams_2sw___pyx_scope_struct__translate_periodic, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *__pyx_freelist_6treams_2sw___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_6treams_2sw___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6treams_2sw___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6treams_2sw___pyx_scope_struct_1_genexpr[--__pyx_freecount_6treams_2sw___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6treams_2sw___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_6treams_2sw___pyx_scope_struct_1_genexpr[__pyx_freecount_6treams_2sw___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6treams_2sw___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6treams_2sw___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "treams.sw.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6treams_2sw___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6treams_2sw___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *__pyx_freelist_6treams_2sw___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_6treams_2sw___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6treams_2sw___pyx_scope_struct_2_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6treams_2sw___pyx_scope_struct_2_genexpr[--__pyx_freecount_6treams_2sw___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_i);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6treams_2sw___pyx_scope_struct_2_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_6treams_2sw___pyx_scope_struct_2_genexpr[__pyx_freecount_6treams_2sw___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6treams_2sw___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_i) {
    e = (*v)(p->__pyx_v_i, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6treams_2sw___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "treams.sw.__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6treams_2sw___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6treams_2sw___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *__pyx_freelist_6treams_2sw___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_6treams_2sw___pyx_scope_struct_3_genexpr = 0;

static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6treams_2sw___pyx_scope_struct_3_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6treams_2sw___pyx_scope_struct_3_genexpr[--__pyx_freecount_6treams_2sw___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6treams_2sw___pyx_scope_struct_3_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_6treams_2sw___pyx_scope_struct_3_genexpr[__pyx_freecount_6treams_2sw___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6treams_2sw___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6treams_2sw___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "treams.sw.__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6treams_2sw___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6treams_2sw___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *__pyx_freelist_6treams_2sw___pyx_scope_struct_4_genexpr[8];
static int __pyx_freecount_6treams_2sw___pyx_scope_struct_4_genexpr = 0;

static PyObject *__pyx_tp_new_6treams_2sw___pyx_scope_struct_4_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_6treams_2sw___pyx_scope_struct_4_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr)))) {
    o = (PyObject*)__pyx_freelist_6treams_2sw___pyx_scope_struct_4_genexpr[--__pyx_freecount_6treams_2sw___pyx_scope_struct_4_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct_4_genexpr(PyObject *o) {
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_6treams_2sw___pyx_scope_struct_4_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr)))) {
    __pyx_freelist_6treams_2sw___pyx_scope_struct_4_genexpr[__pyx_freecount_6treams_2sw___pyx_scope_struct_4_genexpr++] = ((struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_6treams_2sw___pyx_scope_struct_4_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_6treams_2sw___pyx_scope_struct_4_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "treams.sw.__pyx_scope_struct_4_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_6treams_2sw___pyx_scope_struct_4_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_6treams_2sw___pyx_scope_struct_4_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_6treams_2sw___pyx_scope_struct_4_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_6treams_2sw___pyx_scope_struct_4_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_sw(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_sw},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "sw",
    __pyx_k_Spherical_wave_module_autosummar, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_POLTYPE, __pyx_k_POLTYPE, sizeof(__pyx_k_POLTYPE), 0, 0, 1, 1},
  {&__pyx_kp_u_Polarization_must_be_defined_by, __pyx_k_Polarization_must_be_defined_by, sizeof(__pyx_k_Polarization_must_be_defined_by), 0, 1, 0, 0},
  {&__pyx_n_u_TODO, __pyx_k_TODO, sizeof(__pyx_k_TODO), 0, 1, 0, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
  {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
  {&__pyx_n_s_area, __pyx_k_area, sizeof(__pyx_k_area), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_array, __pyx_k_array, sizeof(__pyx_k_array), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_config, __pyx_k_config, sizeof(__pyx_k_config), 0, 0, 1, 1},
  {&__pyx_n_s_dim, __pyx_k_dim, sizeof(__pyx_k_dim), 0, 0, 1, 1},
  {&__pyx_n_s_dlms, __pyx_k_dlms, sizeof(__pyx_k_dlms), 0, 0, 1, 1},
  {&__pyx_n_s_eta, __pyx_k_eta, sizeof(__pyx_k_eta), 0, 0, 1, 1},
  {&__pyx_n_s_func, __pyx_k_func, sizeof(__pyx_k_func), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_u_helicity, __pyx_k_helicity, sizeof(__pyx_k_helicity), 0, 1, 0, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_in, __pyx_k_in, sizeof(__pyx_k_in), 0, 0, 1, 1},
  {&__pyx_kp_u_invalid_length_of_input_modes, __pyx_k_invalid_length_of_input_modes, sizeof(__pyx_k_invalid_length_of_input_modes), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_length_of_output_modes, __pyx_k_invalid_length_of_output_modes, sizeof(__pyx_k_invalid_length_of_output_modes), 0, 1, 0, 0},
  {&__pyx_kp_u_invalid_poltype, __pyx_k_invalid_poltype, sizeof(__pyx_k_invalid_poltype), 0, 1, 0, 0},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_kpar, __pyx_k_kpar, sizeof(__pyx_k_kpar), 0, 0, 1, 1},
  {&__pyx_n_s_kr, __pyx_k_kr, sizeof(__pyx_k_kr), 0, 0, 1, 1},
  {&__pyx_n_s_ks, __pyx_k_ks, sizeof(__pyx_k_ks), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_kx, __pyx_k_kx, sizeof(__pyx_k_kx), 0, 0, 1, 1},
  {&__pyx_n_s_ky, __pyx_k_ky, sizeof(__pyx_k_ky), 0, 0, 1, 1},
  {&__pyx_n_s_kz, __pyx_k_kz, sizeof(__pyx_k_kz), 0, 0, 1, 1},
  {&__pyx_n_s_l, __pyx_k_l, sizeof(__pyx_k_l), 0, 0, 1, 1},
  {&__pyx_n_s_lambda, __pyx_k_lambda, sizeof(__pyx_k_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_lattice, __pyx_k_lattice, sizeof(__pyx_k_lattice), 0, 0, 1, 1},
  {&__pyx_n_s_lsumsw, __pyx_k_lsumsw, sizeof(__pyx_k_lsumsw), 0, 0, 1, 1},
  {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_modes, __pyx_k_modes, sizeof(__pyx_k_modes), 0, 0, 1, 1},
  {&__pyx_n_s_mu, __pyx_k_mu, sizeof(__pyx_k_mu), 0, 0, 1, 1},
  {&__pyx_kp_u_must_be_3_or_4, __pyx_k_must_be_3_or_4, sizeof(__pyx_k_must_be_3_or_4), 0, 1, 0, 0},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_u_parity, __pyx_k_parity, sizeof(__pyx_k_parity), 0, 1, 0, 1},
  {&__pyx_n_s_periodic_to_cw, __pyx_k_periodic_to_cw, sizeof(__pyx_k_periodic_to_cw), 0, 0, 1, 1},
  {&__pyx_n_s_periodic_to_cw_h, __pyx_k_periodic_to_cw_h, sizeof(__pyx_k_periodic_to_cw_h), 0, 0, 1, 1},
  {&__pyx_n_s_periodic_to_cw_p, __pyx_k_periodic_to_cw_p, sizeof(__pyx_k_periodic_to_cw_p), 0, 0, 1, 1},
  {&__pyx_n_s_periodic_to_pw, __pyx_k_periodic_to_pw, sizeof(__pyx_k_periodic_to_pw), 0, 0, 1, 1},
  {&__pyx_n_s_periodic_to_pw_h, __pyx_k_periodic_to_pw_h, sizeof(__pyx_k_periodic_to_pw_h), 0, 0, 1, 1},
  {&__pyx_n_s_periodic_to_pw_p, __pyx_k_periodic_to_pw_p, sizeof(__pyx_k_periodic_to_pw_p), 0, 0, 1, 1},
  {&__pyx_n_s_phi, __pyx_k_phi, sizeof(__pyx_k_phi), 0, 0, 1, 1},
  {&__pyx_n_s_pol, __pyx_k_pol, sizeof(__pyx_k_pol), 0, 0, 1, 1},
  {&__pyx_n_s_poltype, __pyx_k_poltype, sizeof(__pyx_k_poltype), 0, 0, 1, 1},
  {&__pyx_n_s_psi, __pyx_k_psi, sizeof(__pyx_k_psi), 0, 0, 1, 1},
  {&__pyx_n_s_qol, __pyx_k_qol, sizeof(__pyx_k_qol), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reshape, __pyx_k_reshape, sizeof(__pyx_k_reshape), 0, 0, 1, 1},
  {&__pyx_n_s_rotate, __pyx_k_rotate, sizeof(__pyx_k_rotate), 0, 0, 1, 1},
  {&__pyx_n_s_rotate_2, __pyx_k_rotate_2, sizeof(__pyx_k_rotate_2), 0, 0, 1, 1},
  {&__pyx_n_s_rs, __pyx_k_rs, sizeof(__pyx_k_rs), 0, 0, 1, 1},
  {&__pyx_n_s_rsdiff, __pyx_k_rsdiff, sizeof(__pyx_k_rsdiff), 0, 0, 1, 1},
  {&__pyx_n_s_rsin, __pyx_k_rsin, sizeof(__pyx_k_rsin), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_singular, __pyx_k_singular, sizeof(__pyx_k_singular), 0, 0, 1, 1},
  {&__pyx_kp_s_src_treams_sw_pyx, __pyx_k_src_treams_sw_pyx, sizeof(__pyx_k_src_treams_sw_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_theta, __pyx_k_theta, sizeof(__pyx_k_theta), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_translate, __pyx_k_translate, sizeof(__pyx_k_translate), 0, 0, 1, 1},
  {&__pyx_n_s_translate_periodic, __pyx_k_translate_periodic, sizeof(__pyx_k_translate_periodic), 0, 0, 1, 1},
  {&__pyx_n_s_translate_periodic_h, __pyx_k_translate_periodic_h, sizeof(__pyx_k_translate_periodic_h), 0, 0, 1, 1},
  {&__pyx_n_s_translate_periodic_locals_genexp, __pyx_k_translate_periodic_locals_genexp, sizeof(__pyx_k_translate_periodic_locals_genexp), 0, 0, 1, 1},
  {&__pyx_n_s_translate_periodic_p, __pyx_k_translate_periodic_p, sizeof(__pyx_k_translate_periodic_p), 0, 0, 1, 1},
  {&__pyx_n_s_translate_rh, __pyx_k_translate_rh, sizeof(__pyx_k_translate_rh), 0, 0, 1, 1},
  {&__pyx_n_s_translate_rp, __pyx_k_translate_rp, sizeof(__pyx_k_translate_rp), 0, 0, 1, 1},
  {&__pyx_n_s_translate_sh, __pyx_k_translate_sh, sizeof(__pyx_k_translate_sh), 0, 0, 1, 1},
  {&__pyx_n_s_translate_sp, __pyx_k_translate_sp, sizeof(__pyx_k_translate_sp), 0, 0, 1, 1},
  {&__pyx_n_s_treams, __pyx_k_treams, sizeof(__pyx_k_treams), 0, 0, 1, 1},
  {&__pyx_n_s_treams_sw, __pyx_k_treams_sw, sizeof(__pyx_k_treams_sw), 0, 0, 1, 1},
  {&__pyx_n_s_zeros_like, __pyx_k_zeros_like, sizeof(__pyx_k_zeros_like), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 50, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 95, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 945, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "treams/sw.pyx":50
 *     if (pol == 1 and qol == 0) or (pol == 0 and qol == 1):
 *         return 0.0j
 *     raise ValueError("Polarization must be defined by 0 or 1")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_Polarization_must_be_defined_by); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "treams/sw.pyx":303
 *     if (pol1 == 0 or pol1 == 1) and (pol2 == 0 or pol2 == 1):
 *         return 0
 *     raise ValueError('TODO')             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_n_u_TODO); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "treams/sw.pyx":648
 *     if poltype == "helicity":
 *         return _translate_periodic_h(
 *             *(o[:, None] for o in out[1:]),             # <<<<<<<<<<<<<<
 *             *in_[1:],
 *             dlms[out[0][:, None], in_[0], 0, :],
 */
  __pyx_slice__5 = PySlice_New(__pyx_int_1, Py_None, Py_None); if (unlikely(!__pyx_slice__5)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__5);
  __Pyx_GIVEREF(__pyx_slice__5);
  __pyx_slice__6 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__6)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__6);
  __Pyx_GIVEREF(__pyx_slice__6);
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_slice__6, Py_None); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "treams/sw.pyx":630
 *     ])
 *     ks = np.array(ks)
 *     ks = ks.reshape((-1, 1))             # <<<<<<<<<<<<<<
 *     if ks.shape[0] == 2 and ks[0, 0] == ks[1, 0]:
 *         ks = ks[:1, :]
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_int_neg_1, __pyx_int_1); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "treams/sw.pyx":631
 *     ks = np.array(ks)
 *     ks = ks.reshape((-1, 1))
 *     if ks.shape[0] == 2 and ks[0, 0] == ks[1, 0]:             # <<<<<<<<<<<<<<
 *         ks = ks[:1, :]
 *     kpar = np.array(kpar)
 */
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_0); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_tuple__10 = PyTuple_Pack(2, __pyx_int_1, __pyx_int_0); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "treams/sw.pyx":632
 *     ks = ks.reshape((-1, 1))
 *     if ks.shape[0] == 2 and ks[0, 0] == ks[1, 0]:
 *         ks = ks[:1, :]             # <<<<<<<<<<<<<<
 *     kpar = np.array(kpar)
 *     rs = np.array(rs)
 */
  __pyx_slice__11 = PySlice_New(Py_None, __pyx_int_1, Py_None); if (unlikely(!__pyx_slice__11)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__11);
  __Pyx_GIVEREF(__pyx_slice__11);
  __pyx_tuple__12 = PyTuple_Pack(2, __pyx_slice__11, __pyx_slice__6); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "treams/sw.pyx":636
 *     rs = np.array(rs)
 *     if rs.ndim == 1:
 *         rs = rs.reshape((1, -1))             # <<<<<<<<<<<<<<
 *     rsin = np.array(rsin)
 *     if rsin.ndim == 1:
 */
  __pyx_tuple__13 = PyTuple_Pack(2, __pyx_int_1, __pyx_int_neg_1); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "treams/sw.pyx":640
 *     if rsin.ndim == 1:
 *         rsin = rsin.reshape((1, -1))
 *     rsdiff = -rs[:, None, None, None, :] + rsin[:, None, None, :]             # <<<<<<<<<<<<<<
 * 
 *     dim = 1 if kpar.ndim == 0 else kpar.shape[-1]
 */
  __pyx_tuple__14 = PyTuple_Pack(5, __pyx_slice__6, Py_None, Py_None, Py_None, __pyx_slice__6); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_tuple__15 = PyTuple_Pack(4, __pyx_slice__6, Py_None, Py_None, __pyx_slice__6); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "treams/sw.pyx":644
 *     dim = 1 if kpar.ndim == 0 else kpar.shape[-1]
 *     # The result has the shape (n_rs, n_rs, n_ks, n_modes)
 *     dlms = func(dim, modes[:, 0], modes[:, 1], ks, kpar, a, rsdiff, eta)             # <<<<<<<<<<<<<<
 * 
 *     if poltype == "helicity":
 */
  __pyx_tuple__16 = PyTuple_Pack(2, __pyx_slice__6, __pyx_int_0); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_tuple__17 = PyTuple_Pack(2, __pyx_slice__6, __pyx_int_1); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "treams/sw.pyx":251
 * 
 * 
 * def translate(lambda_, mu, pol, l, m, qol, kr, theta, phi, poltype=None, singular=True, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """translate(lambda_, mu, pol, l, m, qol, kr, theta, phi, helicity=True, singular=True)
 * 
 */
  __pyx_tuple__20 = PyTuple_Pack(13, __pyx_n_s_lambda, __pyx_n_s_mu, __pyx_n_s_pol, __pyx_n_s_l, __pyx_n_s_m, __pyx_n_s_qol, __pyx_n_s_kr, __pyx_n_s_theta, __pyx_n_s_phi, __pyx_n_s_poltype, __pyx_n_s_singular, __pyx_n_s_args, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(11, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_treams_sw_pyx, __pyx_n_s_translate, 251, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 251, __pyx_L1_error)

  /* "treams/sw.pyx":337
 * 
 * 
 * def rotate(lambda_, mu, pol, l, m, qol, phi, theta=0, psi=0, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """rotate(lambda_, mu, pol, l, m, qol, phi, theta=0, psi=0)
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(11, __pyx_n_s_lambda, __pyx_n_s_mu, __pyx_n_s_pol, __pyx_n_s_l, __pyx_n_s_m, __pyx_n_s_qol, __pyx_n_s_phi, __pyx_n_s_theta, __pyx_n_s_psi, __pyx_n_s_args, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(9, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_treams_sw_pyx, __pyx_n_s_rotate_2, 337, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 337, __pyx_L1_error)

  /* "treams/sw.pyx":577
 * 
 * 
 * def translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, poltype=None, eta=0, func=lattice.lsumsw):             # <<<<<<<<<<<<<<
 *     """translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, helicity=True, eta=0)
 * 
 */
  __pyx_tuple__24 = PyTuple_Pack(21, __pyx_n_s_ks, __pyx_n_s_kpar, __pyx_n_s_a, __pyx_n_s_rs, __pyx_n_s_out, __pyx_n_s_in, __pyx_n_s_rsin, __pyx_n_s_poltype, __pyx_n_s_eta, __pyx_n_s_func, __pyx_n_s_modes, __pyx_n_s_rsdiff, __pyx_n_s_dim, __pyx_n_s_dlms, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_l, __pyx_n_s_m, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(10, 0, 21, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_treams_sw_pyx, __pyx_n_s_translate_periodic, 577, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 577, __pyx_L1_error)

  /* "treams/sw.pyx":830
 * 
 * 
 * def periodic_to_pw(kx, ky, kz, pol, l, m, qol, area, poltype=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """periodic_to_pw(kx, ky, kz, pol, l, m, qol, area, helicity=True)
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(11, __pyx_n_s_kx, __pyx_n_s_ky, __pyx_n_s_kz, __pyx_n_s_pol, __pyx_n_s_l, __pyx_n_s_m, __pyx_n_s_qol, __pyx_n_s_area, __pyx_n_s_poltype, __pyx_n_s_args, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(9, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_treams_sw_pyx, __pyx_n_s_periodic_to_pw, 830, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 830, __pyx_L1_error)

  /* "treams/sw.pyx":976
 * 
 * 
 * def periodic_to_cw(kz, m, pol, l, mu, qol, k, area, poltype=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """periodic_to_cw(kz, m, pol, l, mu, qol, k, area, helicity=True)
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(11, __pyx_n_s_kz, __pyx_n_s_m, __pyx_n_s_pol, __pyx_n_s_l, __pyx_n_s_mu, __pyx_n_s_qol, __pyx_n_s_k, __pyx_n_s_area, __pyx_n_s_poltype, __pyx_n_s_args, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(9, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARARGS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_treams_sw_pyx, __pyx_n_s_periodic_to_cw, 976, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_6treams_2sw___pyx_scope_struct__translate_periodic) < 0) __PYX_ERR(0, 577, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6treams_2sw___pyx_scope_struct__translate_periodic.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6treams_2sw___pyx_scope_struct__translate_periodic.tp_dictoffset && __pyx_type_6treams_2sw___pyx_scope_struct__translate_periodic.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6treams_2sw___pyx_scope_struct__translate_periodic.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6treams_2sw___pyx_scope_struct__translate_periodic = &__pyx_type_6treams_2sw___pyx_scope_struct__translate_periodic;
  if (PyType_Ready(&__pyx_type_6treams_2sw___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 612, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6treams_2sw___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6treams_2sw___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_6treams_2sw___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6treams_2sw___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6treams_2sw___pyx_scope_struct_1_genexpr = &__pyx_type_6treams_2sw___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_6treams_2sw___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 613, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6treams_2sw___pyx_scope_struct_2_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6treams_2sw___pyx_scope_struct_2_genexpr.tp_dictoffset && __pyx_type_6treams_2sw___pyx_scope_struct_2_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6treams_2sw___pyx_scope_struct_2_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6treams_2sw___pyx_scope_struct_2_genexpr = &__pyx_type_6treams_2sw___pyx_scope_struct_2_genexpr;
  if (PyType_Ready(&__pyx_type_6treams_2sw___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 648, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6treams_2sw___pyx_scope_struct_3_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6treams_2sw___pyx_scope_struct_3_genexpr.tp_dictoffset && __pyx_type_6treams_2sw___pyx_scope_struct_3_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6treams_2sw___pyx_scope_struct_3_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6treams_2sw___pyx_scope_struct_3_genexpr = &__pyx_type_6treams_2sw___pyx_scope_struct_3_genexpr;
  if (PyType_Ready(&__pyx_type_6treams_2sw___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 654, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_6treams_2sw___pyx_scope_struct_4_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_6treams_2sw___pyx_scope_struct_4_genexpr.tp_dictoffset && __pyx_type_6treams_2sw___pyx_scope_struct_4_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_6treams_2sw___pyx_scope_struct_4_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_6treams_2sw___pyx_scope_struct_4_genexpr = &__pyx_type_6treams_2sw___pyx_scope_struct_4_genexpr;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_0_29_36(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyArray_Descr),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 200, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyArrayIterObject),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 223, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyArrayMultiIterObject),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 227, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyArrayObject),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 239, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_generic) __PYX_ERR(1, 771, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "number", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_number) __PYX_ERR(1, 773, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_integer) __PYX_ERR(1, 775, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 777, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 779, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 781, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_floating) __PYX_ERR(1, 783, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 785, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 787, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "character", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_character) __PYX_ERR(1, 789, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyUFuncObject),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("treams.special._misc"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_0_29_36(__pyx_t_1, "SQPI", (void **)&__pyx_vp_6treams_7special_5_misc_SQPI, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("treams.special.cython_special"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "wigner3j", (void (**)(void))&__pyx_f_6treams_7special_14cython_special_wigner3j, "double (long, long, long, long, long, long, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0pi_fun", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_pi_fun, "__pyx_t_double_complex (double, double, __pyx_t_double_complex, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1pi_fun", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_pi_fun, "double (double, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0tau_fun", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tau_fun, "__pyx_t_double_complex (double, double, __pyx_t_double_complex, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1tau_fun", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tau_fun, "double (double, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0tl_vsw_A", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_A, "__pyx_t_double_complex (long, long, long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1tl_vsw_A", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_A, "__pyx_t_double_complex (long, long, long, long, __pyx_t_double_complex, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0tl_vsw_B", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_B, "__pyx_t_double_complex (long, long, long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1tl_vsw_B", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_B, "__pyx_t_double_complex (long, long, long, long, __pyx_t_double_complex, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0tl_vsw_rA", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_rA, "__pyx_t_double_complex (long, long, long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1tl_vsw_rA", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_rA, "__pyx_t_double_complex (long, long, long, long, double, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0tl_vsw_rB", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_tl_vsw_rB, "__pyx_t_double_complex (long, long, long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1tl_vsw_rB", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_tl_vsw_rB, "__pyx_t_double_complex (long, long, long, long, double, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0wignerd", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_14cython_special_wignerd, "__pyx_t_double_complex (long, long, long, double, __pyx_t_double_complex, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1wignerd", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_14cython_special_wignerd, "__pyx_t_double_complex (long, long, long, double, double, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("treams.special._misc"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "minusonepow", (void (**)(void))&__pyx_f_6treams_7special_5_misc_minusonepow, "long (long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0abs", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_5_misc_abs, "double (__pyx_t_double_complex)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1abs", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_5_misc_abs, "double (double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0sqrt", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_5_misc_sqrt, "__pyx_t_double_complex (__pyx_t_double_complex)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1sqrt", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_5_misc_sqrt, "double (double)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initsw(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initsw(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_sw(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_sw(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_sw(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'sw' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_sw(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("sw", __pyx_methods, __pyx_k_Spherical_wave_module_autosummar, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_treams__sw) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "treams.sw")) {
      if (unlikely(PyDict_SetItemString(modules, "treams.sw", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "treams/sw.pyx":14
 * """
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * from treams import config, lattice
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/sw.pyx":16
 * import numpy as np
 * 
 * from treams import config, lattice             # <<<<<<<<<<<<<<
 * 
 * cimport numpy as np
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_config);
  __Pyx_GIVEREF(__pyx_n_s_config);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_config);
  __Pyx_INCREF(__pyx_n_s_lattice);
  __Pyx_GIVEREF(__pyx_n_s_lattice);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_lattice);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_treams, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_config); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_config, __pyx_t_1) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_lattice); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_lattice, __pyx_t_1) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":159
 * 
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * np.import_ufunc()
 * 
 */
  __pyx_t_3 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 159, __pyx_L1_error)

  /* "treams/sw.pyx":160
 * 
 * np.import_array()
 * np.import_ufunc()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __pyx_f_5numpy_import_ufunc(); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 160, __pyx_L1_error)

  /* "treams/sw.pyx":170
 * cdef char ufunc_translate_types[2 * 10]
 * 
 * ufunc_translate_loops[0] = <np.PyUFuncGenericFunction>_loop_D_llllllddd             # <<<<<<<<<<<<<<
 * ufunc_translate_loops[1] = <np.PyUFuncGenericFunction>_loop_D_llllllDdd
 * ufunc_translate_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_2sw__loop_D_llllllddd);

  /* "treams/sw.pyx":171
 * 
 * ufunc_translate_loops[0] = <np.PyUFuncGenericFunction>_loop_D_llllllddd
 * ufunc_translate_loops[1] = <np.PyUFuncGenericFunction>_loop_D_llllllDdd             # <<<<<<<<<<<<<<
 * ufunc_translate_types[0] = <char>np.NPY_LONG
 * ufunc_translate_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_2sw__loop_D_llllllDdd);

  /* "treams/sw.pyx":172
 * ufunc_translate_loops[0] = <np.PyUFuncGenericFunction>_loop_D_llllllddd
 * ufunc_translate_loops[1] = <np.PyUFuncGenericFunction>_loop_D_llllllDdd
 * ufunc_translate_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[1] = <char>np.NPY_LONG
 * ufunc_translate_types[2] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[0]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":173
 * ufunc_translate_loops[1] = <np.PyUFuncGenericFunction>_loop_D_llllllDdd
 * ufunc_translate_types[0] = <char>np.NPY_LONG
 * ufunc_translate_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[2] = <char>np.NPY_LONG
 * ufunc_translate_types[3] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[1]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":174
 * ufunc_translate_types[0] = <char>np.NPY_LONG
 * ufunc_translate_types[1] = <char>np.NPY_LONG
 * ufunc_translate_types[2] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[3] = <char>np.NPY_LONG
 * ufunc_translate_types[4] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[2]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":175
 * ufunc_translate_types[1] = <char>np.NPY_LONG
 * ufunc_translate_types[2] = <char>np.NPY_LONG
 * ufunc_translate_types[3] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[4] = <char>np.NPY_LONG
 * ufunc_translate_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[3]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":176
 * ufunc_translate_types[2] = <char>np.NPY_LONG
 * ufunc_translate_types[3] = <char>np.NPY_LONG
 * ufunc_translate_types[4] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[5] = <char>np.NPY_LONG
 * ufunc_translate_types[6] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[4]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":177
 * ufunc_translate_types[3] = <char>np.NPY_LONG
 * ufunc_translate_types[4] = <char>np.NPY_LONG
 * ufunc_translate_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[6] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[7] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[5]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":178
 * ufunc_translate_types[4] = <char>np.NPY_LONG
 * ufunc_translate_types[5] = <char>np.NPY_LONG
 * ufunc_translate_types[6] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_translate_types[7] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[8] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[6]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":179
 * ufunc_translate_types[5] = <char>np.NPY_LONG
 * ufunc_translate_types[6] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[7] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_translate_types[8] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[7]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":180
 * ufunc_translate_types[6] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[7] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[8] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_translate_types[9] = <char>np.NPY_CDOUBLE
 * ufunc_translate_types[10] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[8]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":181
 * ufunc_translate_types[7] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[8] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * ufunc_translate_types[10] = <char>np.NPY_LONG
 * ufunc_translate_types[11] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":182
 * ufunc_translate_types[8] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[9] = <char>np.NPY_CDOUBLE
 * ufunc_translate_types[10] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[11] = <char>np.NPY_LONG
 * ufunc_translate_types[12] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[10]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":183
 * ufunc_translate_types[9] = <char>np.NPY_CDOUBLE
 * ufunc_translate_types[10] = <char>np.NPY_LONG
 * ufunc_translate_types[11] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[12] = <char>np.NPY_LONG
 * ufunc_translate_types[13] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[11]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":184
 * ufunc_translate_types[10] = <char>np.NPY_LONG
 * ufunc_translate_types[11] = <char>np.NPY_LONG
 * ufunc_translate_types[12] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[13] = <char>np.NPY_LONG
 * ufunc_translate_types[14] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[12]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":185
 * ufunc_translate_types[11] = <char>np.NPY_LONG
 * ufunc_translate_types[12] = <char>np.NPY_LONG
 * ufunc_translate_types[13] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[14] = <char>np.NPY_LONG
 * ufunc_translate_types[15] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[13]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":186
 * ufunc_translate_types[12] = <char>np.NPY_LONG
 * ufunc_translate_types[13] = <char>np.NPY_LONG
 * ufunc_translate_types[14] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[15] = <char>np.NPY_LONG
 * ufunc_translate_types[16] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[14]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":187
 * ufunc_translate_types[13] = <char>np.NPY_LONG
 * ufunc_translate_types[14] = <char>np.NPY_LONG
 * ufunc_translate_types[15] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_translate_types[16] = <char>np.NPY_CDOUBLE
 * ufunc_translate_types[17] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[15]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":188
 * ufunc_translate_types[14] = <char>np.NPY_LONG
 * ufunc_translate_types[15] = <char>np.NPY_LONG
 * ufunc_translate_types[16] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * ufunc_translate_types[17] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[18] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[16]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":189
 * ufunc_translate_types[15] = <char>np.NPY_LONG
 * ufunc_translate_types[16] = <char>np.NPY_CDOUBLE
 * ufunc_translate_types[17] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_translate_types[18] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[19] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[17]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":190
 * ufunc_translate_types[16] = <char>np.NPY_CDOUBLE
 * ufunc_translate_types[17] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[18] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_translate_types[19] = <char>np.NPY_CDOUBLE
 * ufunc_translate_sh_data[0] = <void*>_ctranslate_sh[double]
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[18]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":191
 * ufunc_translate_types[17] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[18] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[19] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * ufunc_translate_sh_data[0] = <void*>_ctranslate_sh[double]
 * ufunc_translate_sh_data[1] = <void*>_ctranslate_sh[double_complex]
 */
  (__pyx_v_6treams_2sw_ufunc_translate_types[19]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":192
 * ufunc_translate_types[18] = <char>np.NPY_DOUBLE
 * ufunc_translate_types[19] = <char>np.NPY_CDOUBLE
 * ufunc_translate_sh_data[0] = <void*>_ctranslate_sh[double]             # <<<<<<<<<<<<<<
 * ufunc_translate_sh_data[1] = <void*>_ctranslate_sh[double_complex]
 * ufunc_translate_rh_data[0] = <void*>_ctranslate_rh[double]
 */
  (__pyx_v_6treams_2sw_ufunc_translate_sh_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_sh);

  /* "treams/sw.pyx":193
 * ufunc_translate_types[19] = <char>np.NPY_CDOUBLE
 * ufunc_translate_sh_data[0] = <void*>_ctranslate_sh[double]
 * ufunc_translate_sh_data[1] = <void*>_ctranslate_sh[double_complex]             # <<<<<<<<<<<<<<
 * ufunc_translate_rh_data[0] = <void*>_ctranslate_rh[double]
 * ufunc_translate_rh_data[1] = <void*>_ctranslate_rh[double_complex]
 */
  (__pyx_v_6treams_2sw_ufunc_translate_sh_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_sh);

  /* "treams/sw.pyx":194
 * ufunc_translate_sh_data[0] = <void*>_ctranslate_sh[double]
 * ufunc_translate_sh_data[1] = <void*>_ctranslate_sh[double_complex]
 * ufunc_translate_rh_data[0] = <void*>_ctranslate_rh[double]             # <<<<<<<<<<<<<<
 * ufunc_translate_rh_data[1] = <void*>_ctranslate_rh[double_complex]
 * ufunc_translate_sp_data[0] = <void*>_ctranslate_sp[double]
 */
  (__pyx_v_6treams_2sw_ufunc_translate_rh_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_rh);

  /* "treams/sw.pyx":195
 * ufunc_translate_sh_data[1] = <void*>_ctranslate_sh[double_complex]
 * ufunc_translate_rh_data[0] = <void*>_ctranslate_rh[double]
 * ufunc_translate_rh_data[1] = <void*>_ctranslate_rh[double_complex]             # <<<<<<<<<<<<<<
 * ufunc_translate_sp_data[0] = <void*>_ctranslate_sp[double]
 * ufunc_translate_sp_data[1] = <void*>_ctranslate_sp[double_complex]
 */
  (__pyx_v_6treams_2sw_ufunc_translate_rh_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_rh);

  /* "treams/sw.pyx":196
 * ufunc_translate_rh_data[0] = <void*>_ctranslate_rh[double]
 * ufunc_translate_rh_data[1] = <void*>_ctranslate_rh[double_complex]
 * ufunc_translate_sp_data[0] = <void*>_ctranslate_sp[double]             # <<<<<<<<<<<<<<
 * ufunc_translate_sp_data[1] = <void*>_ctranslate_sp[double_complex]
 * ufunc_translate_rp_data[0] = <void*>_ctranslate_rp[double]
 */
  (__pyx_v_6treams_2sw_ufunc_translate_sp_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_sp);

  /* "treams/sw.pyx":197
 * ufunc_translate_rh_data[1] = <void*>_ctranslate_rh[double_complex]
 * ufunc_translate_sp_data[0] = <void*>_ctranslate_sp[double]
 * ufunc_translate_sp_data[1] = <void*>_ctranslate_sp[double_complex]             # <<<<<<<<<<<<<<
 * ufunc_translate_rp_data[0] = <void*>_ctranslate_rp[double]
 * ufunc_translate_rp_data[1] = <void*>_ctranslate_rp[double_complex]
 */
  (__pyx_v_6treams_2sw_ufunc_translate_sp_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_sp);

  /* "treams/sw.pyx":198
 * ufunc_translate_sp_data[0] = <void*>_ctranslate_sp[double]
 * ufunc_translate_sp_data[1] = <void*>_ctranslate_sp[double_complex]
 * ufunc_translate_rp_data[0] = <void*>_ctranslate_rp[double]             # <<<<<<<<<<<<<<
 * ufunc_translate_rp_data[1] = <void*>_ctranslate_rp[double_complex]
 * 
 */
  (__pyx_v_6treams_2sw_ufunc_translate_rp_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_2sw__ctranslate_rp);

  /* "treams/sw.pyx":199
 * ufunc_translate_sp_data[1] = <void*>_ctranslate_sp[double_complex]
 * ufunc_translate_rp_data[0] = <void*>_ctranslate_rp[double]
 * ufunc_translate_rp_data[1] = <void*>_ctranslate_rp[double_complex]             # <<<<<<<<<<<<<<
 * 
 * _translate_sh = np.PyUFunc_FromFuncAndData(
 */
  (__pyx_v_6treams_2sw_ufunc_translate_rp_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_2sw__ctranslate_rp);

  /* "treams/sw.pyx":201
 * ufunc_translate_rp_data[1] = <void*>_ctranslate_rp[double_complex]
 * 
 * _translate_sh = np.PyUFunc_FromFuncAndData(             # <<<<<<<<<<<<<<
 *     ufunc_translate_loops,
 *     ufunc_translate_sh_data,
 */
  __pyx_t_2 = PyUFunc_FromFuncAndData(__pyx_v_6treams_2sw_ufunc_translate_loops, __pyx_v_6treams_2sw_ufunc_translate_sh_data, __pyx_v_6treams_2sw_ufunc_translate_types, 2, 9, 1, 0, ((char *)"_translate_sh"), ((char *)""), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_translate_sh, __pyx_t_2) < 0) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":213
 *     0,
 * )
 * _translate_rh = np.PyUFunc_FromFuncAndData(             # <<<<<<<<<<<<<<
 *     ufunc_translate_loops,
 *     ufunc_translate_rh_data,
 */
  __pyx_t_2 = PyUFunc_FromFuncAndData(__pyx_v_6treams_2sw_ufunc_translate_loops, __pyx_v_6treams_2sw_ufunc_translate_rh_data, __pyx_v_6treams_2sw_ufunc_translate_types, 2, 9, 1, 0, ((char *)"_translate_rh"), ((char *)""), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_translate_rh, __pyx_t_2) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":225
 *     0,
 * )
 * _translate_sp = np.PyUFunc_FromFuncAndData(             # <<<<<<<<<<<<<<
 *     ufunc_translate_loops,
 *     ufunc_translate_sp_data,
 */
  __pyx_t_2 = PyUFunc_FromFuncAndData(__pyx_v_6treams_2sw_ufunc_translate_loops, __pyx_v_6treams_2sw_ufunc_translate_sp_data, __pyx_v_6treams_2sw_ufunc_translate_types, 2, 9, 1, 0, ((char *)"_translate_sp"), ((char *)""), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_translate_sp, __pyx_t_2) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":237
 *     0,
 * )
 * _translate_rp = np.PyUFunc_FromFuncAndData(             # <<<<<<<<<<<<<<
 *     ufunc_translate_loops,
 *     ufunc_translate_rp_data,
 */
  __pyx_t_2 = PyUFunc_FromFuncAndData(__pyx_v_6treams_2sw_ufunc_translate_loops, __pyx_v_6treams_2sw_ufunc_translate_rp_data, __pyx_v_6treams_2sw_ufunc_translate_types, 2, 9, 1, 0, ((char *)"_translate_rp"), ((char *)""), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_translate_rp, __pyx_t_2) < 0) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":251
 * 
 * 
 * def translate(lambda_, mu, pol, l, m, qol, kr, theta, phi, poltype=None, singular=True, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """translate(lambda_, mu, pol, l, m, qol, kr, theta, phi, helicity=True, singular=True)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6treams_2sw_1translate, NULL, __pyx_n_s_treams_sw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_translate, __pyx_t_2) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":310
 * cdef char ufunc_rotate_types[10]
 * 
 * ufunc_rotate_loops[0] = <np.PyUFuncGenericFunction>_loop_D_llllllddd             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[0] = <char>np.NPY_LONG
 * ufunc_rotate_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_2sw__loop_D_llllllddd);

  /* "treams/sw.pyx":311
 * 
 * ufunc_rotate_loops[0] = <np.PyUFuncGenericFunction>_loop_D_llllllddd
 * ufunc_rotate_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[1] = <char>np.NPY_LONG
 * ufunc_rotate_types[2] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[0]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":312
 * ufunc_rotate_loops[0] = <np.PyUFuncGenericFunction>_loop_D_llllllddd
 * ufunc_rotate_types[0] = <char>np.NPY_LONG
 * ufunc_rotate_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[2] = <char>np.NPY_LONG
 * ufunc_rotate_types[3] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[1]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":313
 * ufunc_rotate_types[0] = <char>np.NPY_LONG
 * ufunc_rotate_types[1] = <char>np.NPY_LONG
 * ufunc_rotate_types[2] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[3] = <char>np.NPY_LONG
 * ufunc_rotate_types[4] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[2]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":314
 * ufunc_rotate_types[1] = <char>np.NPY_LONG
 * ufunc_rotate_types[2] = <char>np.NPY_LONG
 * ufunc_rotate_types[3] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[4] = <char>np.NPY_LONG
 * ufunc_rotate_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[3]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":315
 * ufunc_rotate_types[2] = <char>np.NPY_LONG
 * ufunc_rotate_types[3] = <char>np.NPY_LONG
 * ufunc_rotate_types[4] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[5] = <char>np.NPY_LONG
 * ufunc_rotate_types[6] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[4]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":316
 * ufunc_rotate_types[3] = <char>np.NPY_LONG
 * ufunc_rotate_types[4] = <char>np.NPY_LONG
 * ufunc_rotate_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[6] = <char>np.NPY_DOUBLE
 * ufunc_rotate_types[7] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[5]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":317
 * ufunc_rotate_types[4] = <char>np.NPY_LONG
 * ufunc_rotate_types[5] = <char>np.NPY_LONG
 * ufunc_rotate_types[6] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[7] = <char>np.NPY_DOUBLE
 * ufunc_rotate_types[8] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[6]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":318
 * ufunc_rotate_types[5] = <char>np.NPY_LONG
 * ufunc_rotate_types[6] = <char>np.NPY_DOUBLE
 * ufunc_rotate_types[7] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[8] = <char>np.NPY_DOUBLE
 * ufunc_rotate_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[7]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":319
 * ufunc_rotate_types[6] = <char>np.NPY_DOUBLE
 * ufunc_rotate_types[7] = <char>np.NPY_DOUBLE
 * ufunc_rotate_types[8] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_rotate_types[9] = <char>np.NPY_CDOUBLE
 * ufunc_rotate_data[0] = <void*>_crotate
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[8]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":320
 * ufunc_rotate_types[7] = <char>np.NPY_DOUBLE
 * ufunc_rotate_types[8] = <char>np.NPY_DOUBLE
 * ufunc_rotate_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * ufunc_rotate_data[0] = <void*>_crotate
 * 
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":321
 * ufunc_rotate_types[8] = <char>np.NPY_DOUBLE
 * ufunc_rotate_types[9] = <char>np.NPY_CDOUBLE
 * ufunc_rotate_data[0] = <void*>_crotate             # <<<<<<<<<<<<<<
 * 
 * _rotate = np.PyUFunc_FromFuncAndData(
 */
  (__pyx_v_6treams_2sw_ufunc_rotate_data[0]) = ((void *)__pyx_f_6treams_2sw__crotate);

  /* "treams/sw.pyx":323
 * ufunc_rotate_data[0] = <void*>_crotate
 * 
 * _rotate = np.PyUFunc_FromFuncAndData(             # <<<<<<<<<<<<<<
 *     ufunc_rotate_loops,
 *     ufunc_rotate_data,
 */
  __pyx_t_2 = PyUFunc_FromFuncAndData(__pyx_v_6treams_2sw_ufunc_rotate_loops, __pyx_v_6treams_2sw_ufunc_rotate_data, __pyx_v_6treams_2sw_ufunc_rotate_types, 1, 9, 1, 0, ((char *)"_rotate"), ((char *)""), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_rotate, __pyx_t_2) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":337
 * 
 * 
 * def rotate(lambda_, mu, pol, l, m, qol, phi, theta=0, psi=0, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """rotate(lambda_, mu, pol, l, m, qol, phi, theta=0, psi=0)
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_6treams_2sw_3rotate, NULL, __pyx_n_s_treams_sw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_rotate_2, __pyx_t_2) < 0) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":521
 * cdef char gufunc_translate_periodic_h_types[9]
 * 
 * gufunc_translate_periodic_h_loops[0] = <np.PyUFuncGenericFunction>_loop_translate_periodic_h             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_types[0] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_2sw__loop_translate_periodic_h);

  /* "treams/sw.pyx":522
 * 
 * gufunc_translate_periodic_h_loops[0] = <np.PyUFuncGenericFunction>_loop_translate_periodic_h
 * gufunc_translate_periodic_h_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_types[1] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[2] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[0]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":523
 * gufunc_translate_periodic_h_loops[0] = <np.PyUFuncGenericFunction>_loop_translate_periodic_h
 * gufunc_translate_periodic_h_types[0] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_types[2] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[3] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[1]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":524
 * gufunc_translate_periodic_h_types[0] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[1] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[2] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_types[3] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[4] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[2]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":525
 * gufunc_translate_periodic_h_types[1] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[2] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[3] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_types[4] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[3]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":526
 * gufunc_translate_periodic_h_types[2] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[3] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[4] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_types[5] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[6] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[4]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":527
 * gufunc_translate_periodic_h_types[3] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[4] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_h_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[5]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":528
 * gufunc_translate_periodic_h_types[4] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[5] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[6] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_h_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[6]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":529
 * gufunc_translate_periodic_h_types[5] = <char>np.NPY_LONG
 * gufunc_translate_periodic_h_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_h_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_h_data[0] = <void*>_ctranslate_periodic_h
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":530
 * gufunc_translate_periodic_h_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_h_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_h_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_h_data[0] = <void*>_ctranslate_periodic_h
 * 
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":531
 * gufunc_translate_periodic_h_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_h_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_h_data[0] = <void*>_ctranslate_periodic_h             # <<<<<<<<<<<<<<
 * 
 * _translate_periodic_h = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_h_data[0]) = ((void *)__pyx_f_6treams_2sw__ctranslate_periodic_h);

  /* "treams/sw.pyx":533
 * gufunc_translate_periodic_h_data[0] = <void*>_ctranslate_periodic_h
 * 
 * _translate_periodic_h = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_translate_periodic_h_loops,
 *     gufunc_translate_periodic_h_data,
 */
  __pyx_t_2 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_2sw_gufunc_translate_periodic_h_loops, __pyx_v_6treams_2sw_gufunc_translate_periodic_h_data, __pyx_v_6treams_2sw_gufunc_translate_periodic_h_types, 1, 8, 1, 0, ((char *)"_translate_periodic_h"), ((char *)""), 0, ((char *)"(),(),(),(),(),(),(a),(a)->()")); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_translate_periodic_h, __pyx_t_2) < 0) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":551
 * cdef char gufunc_translate_periodic_p_types[8]
 * 
 * gufunc_translate_periodic_p_loops[0] = <np.PyUFuncGenericFunction>_loop_translate_periodic_p             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_p_types[0] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_2sw__loop_translate_periodic_p);

  /* "treams/sw.pyx":552
 * 
 * gufunc_translate_periodic_p_loops[0] = <np.PyUFuncGenericFunction>_loop_translate_periodic_p
 * gufunc_translate_periodic_p_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_p_types[1] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[2] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_types[0]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":553
 * gufunc_translate_periodic_p_loops[0] = <np.PyUFuncGenericFunction>_loop_translate_periodic_p
 * gufunc_translate_periodic_p_types[0] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_p_types[2] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[3] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_types[1]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":554
 * gufunc_translate_periodic_p_types[0] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[1] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[2] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_p_types[3] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[4] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_types[2]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":555
 * gufunc_translate_periodic_p_types[1] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[2] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[3] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_p_types[4] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_types[3]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":556
 * gufunc_translate_periodic_p_types[2] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[3] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[4] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_p_types[5] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[6] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_types[4]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":557
 * gufunc_translate_periodic_p_types[3] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[4] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_p_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_p_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_types[5]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":558
 * gufunc_translate_periodic_p_types[4] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[5] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[6] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_p_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_p_data[0] = <void*>_ctranslate_periodic_p
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_types[6]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":559
 * gufunc_translate_periodic_p_types[5] = <char>np.NPY_LONG
 * gufunc_translate_periodic_p_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_p_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_translate_periodic_p_data[0] = <void*>_ctranslate_periodic_p
 * 
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":560
 * gufunc_translate_periodic_p_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_p_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_translate_periodic_p_data[0] = <void*>_ctranslate_periodic_p             # <<<<<<<<<<<<<<
 * 
 * _translate_periodic_p = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_2sw_gufunc_translate_periodic_p_data[0]) = ((void *)__pyx_f_6treams_2sw__ctranslate_periodic_p);

  /* "treams/sw.pyx":562
 * gufunc_translate_periodic_p_data[0] = <void*>_ctranslate_periodic_p
 * 
 * _translate_periodic_p = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_translate_periodic_p_loops,
 *     gufunc_translate_periodic_p_data,
 */
  __pyx_t_2 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_2sw_gufunc_translate_periodic_p_loops, __pyx_v_6treams_2sw_gufunc_translate_periodic_p_data, __pyx_v_6treams_2sw_gufunc_translate_periodic_p_types, 1, 7, 1, 0, ((char *)"_translate_periodic_p"), ((char *)""), 0, ((char *)"(),(),(),(),(),(),(a)->()")); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_translate_periodic_p, __pyx_t_2) < 0) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "treams/sw.pyx":577
 * 
 * 
 * def translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, poltype=None, eta=0, func=lattice.lsumsw):             # <<<<<<<<<<<<<<
 *     """translate_periodic(ks, kpar, a, rs, out, in_=None, rsin=None, helicity=True, eta=0)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_lattice); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_lsumsw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_k__4 = __pyx_t_1;
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6treams_2sw_5translate_periodic, NULL, __pyx_n_s_treams_sw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_translate_periodic, __pyx_t_1) < 0) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/sw.pyx":779
 * cdef char ufunc_pw_types[2 * 9]
 * 
 * ufunc_pw_loops[0] = <np.PyUFuncGenericFunction>_loop_periodic_d             # <<<<<<<<<<<<<<
 * ufunc_pw_loops[1] = <np.PyUFuncGenericFunction>_loop_periodic_D
 * ufunc_pw_types[0] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_2sw__loop_periodic_d);

  /* "treams/sw.pyx":780
 * 
 * ufunc_pw_loops[0] = <np.PyUFuncGenericFunction>_loop_periodic_d
 * ufunc_pw_loops[1] = <np.PyUFuncGenericFunction>_loop_periodic_D             # <<<<<<<<<<<<<<
 * ufunc_pw_types[0] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_2sw__loop_periodic_D);

  /* "treams/sw.pyx":781
 * ufunc_pw_loops[0] = <np.PyUFuncGenericFunction>_loop_periodic_d
 * ufunc_pw_loops[1] = <np.PyUFuncGenericFunction>_loop_periodic_D
 * ufunc_pw_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_types[1] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":782
 * ufunc_pw_loops[1] = <np.PyUFuncGenericFunction>_loop_periodic_D
 * ufunc_pw_types[0] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_types[2] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[3] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":783
 * ufunc_pw_types[0] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[1] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_types[3] = <char>np.NPY_LONG
 * ufunc_pw_types[4] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":784
 * ufunc_pw_types[1] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[2] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[3] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_pw_types[4] = <char>np.NPY_LONG
 * ufunc_pw_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[3]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":785
 * ufunc_pw_types[2] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[3] = <char>np.NPY_LONG
 * ufunc_pw_types[4] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_pw_types[5] = <char>np.NPY_LONG
 * ufunc_pw_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[4]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":786
 * ufunc_pw_types[3] = <char>np.NPY_LONG
 * ufunc_pw_types[4] = <char>np.NPY_LONG
 * ufunc_pw_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_pw_types[6] = <char>np.NPY_LONG
 * ufunc_pw_types[7] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[5]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":787
 * ufunc_pw_types[4] = <char>np.NPY_LONG
 * ufunc_pw_types[5] = <char>np.NPY_LONG
 * ufunc_pw_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_pw_types[7] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[6]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":788
 * ufunc_pw_types[5] = <char>np.NPY_LONG
 * ufunc_pw_types[6] = <char>np.NPY_LONG
 * ufunc_pw_types[7] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_types[8] = <char>np.NPY_CDOUBLE
 * ufunc_pw_types[9] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[7]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":789
 * ufunc_pw_types[6] = <char>np.NPY_LONG
 * ufunc_pw_types[7] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_types[9] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[10] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":790
 * ufunc_pw_types[7] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[8] = <char>np.NPY_CDOUBLE
 * ufunc_pw_types[9] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_types[10] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[11] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[9]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":791
 * ufunc_pw_types[8] = <char>np.NPY_CDOUBLE
 * ufunc_pw_types[9] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[10] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_types[11] = <char>np.NPY_CDOUBLE
 * ufunc_pw_types[12] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[10]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":792
 * ufunc_pw_types[9] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[10] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[11] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_types[12] = <char>np.NPY_LONG
 * ufunc_pw_types[13] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[11]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":793
 * ufunc_pw_types[10] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[11] = <char>np.NPY_CDOUBLE
 * ufunc_pw_types[12] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_pw_types[13] = <char>np.NPY_LONG
 * ufunc_pw_types[14] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[12]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":794
 * ufunc_pw_types[11] = <char>np.NPY_CDOUBLE
 * ufunc_pw_types[12] = <char>np.NPY_LONG
 * ufunc_pw_types[13] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_pw_types[14] = <char>np.NPY_LONG
 * ufunc_pw_types[15] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[13]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":795
 * ufunc_pw_types[12] = <char>np.NPY_LONG
 * ufunc_pw_types[13] = <char>np.NPY_LONG
 * ufunc_pw_types[14] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_pw_types[15] = <char>np.NPY_LONG
 * ufunc_pw_types[16] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[14]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":796
 * ufunc_pw_types[13] = <char>np.NPY_LONG
 * ufunc_pw_types[14] = <char>np.NPY_LONG
 * ufunc_pw_types[15] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_pw_types[16] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[17] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[15]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":797
 * ufunc_pw_types[14] = <char>np.NPY_LONG
 * ufunc_pw_types[15] = <char>np.NPY_LONG
 * ufunc_pw_types[16] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_types[17] = <char>np.NPY_CDOUBLE
 * ufunc_pw_h_data[0] = <void*>_cperiodic_to_pw_h[double]
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[16]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":798
 * ufunc_pw_types[15] = <char>np.NPY_LONG
 * ufunc_pw_types[16] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[17] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * ufunc_pw_h_data[0] = <void*>_cperiodic_to_pw_h[double]
 * ufunc_pw_h_data[1] = <void*>_cperiodic_to_pw_h[double_complex]
 */
  (__pyx_v_6treams_2sw_ufunc_pw_types[17]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":799
 * ufunc_pw_types[16] = <char>np.NPY_DOUBLE
 * ufunc_pw_types[17] = <char>np.NPY_CDOUBLE
 * ufunc_pw_h_data[0] = <void*>_cperiodic_to_pw_h[double]             # <<<<<<<<<<<<<<
 * ufunc_pw_h_data[1] = <void*>_cperiodic_to_pw_h[double_complex]
 * ufunc_pw_p_data[0] = <void*>_cperiodic_to_pw_p[double]
 */
  (__pyx_v_6treams_2sw_ufunc_pw_h_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_2sw__cperiodic_to_pw_h);

  /* "treams/sw.pyx":800
 * ufunc_pw_types[17] = <char>np.NPY_CDOUBLE
 * ufunc_pw_h_data[0] = <void*>_cperiodic_to_pw_h[double]
 * ufunc_pw_h_data[1] = <void*>_cperiodic_to_pw_h[double_complex]             # <<<<<<<<<<<<<<
 * ufunc_pw_p_data[0] = <void*>_cperiodic_to_pw_p[double]
 * ufunc_pw_p_data[1] = <void*>_cperiodic_to_pw_p[double_complex]
 */
  (__pyx_v_6treams_2sw_ufunc_pw_h_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_2sw__cperiodic_to_pw_h);

  /* "treams/sw.pyx":801
 * ufunc_pw_h_data[0] = <void*>_cperiodic_to_pw_h[double]
 * ufunc_pw_h_data[1] = <void*>_cperiodic_to_pw_h[double_complex]
 * ufunc_pw_p_data[0] = <void*>_cperiodic_to_pw_p[double]             # <<<<<<<<<<<<<<
 * ufunc_pw_p_data[1] = <void*>_cperiodic_to_pw_p[double_complex]
 * 
 */
  (__pyx_v_6treams_2sw_ufunc_pw_p_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_2sw__cperiodic_to_pw_p);

  /* "treams/sw.pyx":802
 * ufunc_pw_h_data[1] = <void*>_cperiodic_to_pw_h[double_complex]
 * ufunc_pw_p_data[0] = <void*>_cperiodic_to_pw_p[double]
 * ufunc_pw_p_data[1] = <void*>_cperiodic_to_pw_p[double_complex]             # <<<<<<<<<<<<<<
 * 
 * _periodic_to_pw_h = np.PyUFunc_FromFuncAndData(
 */
  (__pyx_v_6treams_2sw_ufunc_pw_p_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_2sw__cperiodic_to_pw_p);

  /* "treams/sw.pyx":804
 * ufunc_pw_p_data[1] = <void*>_cperiodic_to_pw_p[double_complex]
 * 
 * _periodic_to_pw_h = np.PyUFunc_FromFuncAndData(             # <<<<<<<<<<<<<<
 *     ufunc_pw_loops,
 *     ufunc_pw_h_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndData(__pyx_v_6treams_2sw_ufunc_pw_loops, __pyx_v_6treams_2sw_ufunc_pw_h_data, __pyx_v_6treams_2sw_ufunc_pw_types, 2, 8, 1, 0, ((char *)"_periodic_to_pw_h"), ((char *)""), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_periodic_to_pw_h, __pyx_t_1) < 0) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/sw.pyx":816
 *     0,
 * )
 * _periodic_to_pw_p = np.PyUFunc_FromFuncAndData(             # <<<<<<<<<<<<<<
 *     ufunc_pw_loops,
 *     ufunc_pw_p_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndData(__pyx_v_6treams_2sw_ufunc_pw_loops, __pyx_v_6treams_2sw_ufunc_pw_p_data, __pyx_v_6treams_2sw_ufunc_pw_types, 2, 8, 1, 0, ((char *)"_periodic_to_pw_p"), ((char *)""), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_periodic_to_pw_p, __pyx_t_1) < 0) __PYX_ERR(0, 816, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/sw.pyx":830
 * 
 * 
 * def periodic_to_pw(kx, ky, kz, pol, l, m, qol, area, poltype=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """periodic_to_pw(kx, ky, kz, pol, l, m, qol, area, helicity=True)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6treams_2sw_7periodic_to_pw, NULL, __pyx_n_s_treams_sw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_periodic_to_pw, __pyx_t_1) < 0) __PYX_ERR(0, 830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/sw.pyx":937
 * cdef char ufunc_cw_types[9]
 * 
 * ufunc_cw_loops[0] = <np.PyUFuncGenericFunction>_loop_periodic_cw_D             # <<<<<<<<<<<<<<
 * ufunc_cw_types[0] = <char>np.NPY_DOUBLE
 * ufunc_cw_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_cw_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_2sw__loop_periodic_cw_D);

  /* "treams/sw.pyx":938
 * 
 * ufunc_cw_loops[0] = <np.PyUFuncGenericFunction>_loop_periodic_cw_D
 * ufunc_cw_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_cw_types[1] = <char>np.NPY_LONG
 * ufunc_cw_types[2] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_cw_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":939
 * ufunc_cw_loops[0] = <np.PyUFuncGenericFunction>_loop_periodic_cw_D
 * ufunc_cw_types[0] = <char>np.NPY_DOUBLE
 * ufunc_cw_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_cw_types[2] = <char>np.NPY_LONG
 * ufunc_cw_types[3] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_cw_types[1]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":940
 * ufunc_cw_types[0] = <char>np.NPY_DOUBLE
 * ufunc_cw_types[1] = <char>np.NPY_LONG
 * ufunc_cw_types[2] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_cw_types[3] = <char>np.NPY_LONG
 * ufunc_cw_types[4] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_cw_types[2]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":941
 * ufunc_cw_types[1] = <char>np.NPY_LONG
 * ufunc_cw_types[2] = <char>np.NPY_LONG
 * ufunc_cw_types[3] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_cw_types[4] = <char>np.NPY_LONG
 * ufunc_cw_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_2sw_ufunc_cw_types[3]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":942
 * ufunc_cw_types[2] = <char>np.NPY_LONG
 * ufunc_cw_types[3] = <char>np.NPY_LONG
 * ufunc_cw_types[4] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_cw_types[5] = <char>np.NPY_LONG
 * ufunc_cw_types[6] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_cw_types[4]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":943
 * ufunc_cw_types[3] = <char>np.NPY_LONG
 * ufunc_cw_types[4] = <char>np.NPY_LONG
 * ufunc_cw_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * ufunc_cw_types[6] = <char>np.NPY_CDOUBLE
 * ufunc_cw_types[7] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_cw_types[5]) = ((char)NPY_LONG);

  /* "treams/sw.pyx":944
 * ufunc_cw_types[4] = <char>np.NPY_LONG
 * ufunc_cw_types[5] = <char>np.NPY_LONG
 * ufunc_cw_types[6] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * ufunc_cw_types[7] = <char>np.NPY_DOUBLE
 * ufunc_cw_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_2sw_ufunc_cw_types[6]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":945
 * ufunc_cw_types[5] = <char>np.NPY_LONG
 * ufunc_cw_types[6] = <char>np.NPY_CDOUBLE
 * ufunc_cw_types[7] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * ufunc_cw_types[8] = <char>np.NPY_CDOUBLE
 * ufunc_cw_h_data[0] = <void*>_cperiodic_to_cw_h
 */
  (__pyx_v_6treams_2sw_ufunc_cw_types[7]) = ((char)NPY_DOUBLE);

  /* "treams/sw.pyx":946
 * ufunc_cw_types[6] = <char>np.NPY_CDOUBLE
 * ufunc_cw_types[7] = <char>np.NPY_DOUBLE
 * ufunc_cw_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * ufunc_cw_h_data[0] = <void*>_cperiodic_to_cw_h
 * ufunc_cw_p_data[0] = <void*>_cperiodic_to_cw_p
 */
  (__pyx_v_6treams_2sw_ufunc_cw_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/sw.pyx":947
 * ufunc_cw_types[7] = <char>np.NPY_DOUBLE
 * ufunc_cw_types[8] = <char>np.NPY_CDOUBLE
 * ufunc_cw_h_data[0] = <void*>_cperiodic_to_cw_h             # <<<<<<<<<<<<<<
 * ufunc_cw_p_data[0] = <void*>_cperiodic_to_cw_p
 * 
 */
  (__pyx_v_6treams_2sw_ufunc_cw_h_data[0]) = ((void *)__pyx_f_6treams_2sw__cperiodic_to_cw_h);

  /* "treams/sw.pyx":948
 * ufunc_cw_types[8] = <char>np.NPY_CDOUBLE
 * ufunc_cw_h_data[0] = <void*>_cperiodic_to_cw_h
 * ufunc_cw_p_data[0] = <void*>_cperiodic_to_cw_p             # <<<<<<<<<<<<<<
 * 
 * _periodic_to_cw_h = np.PyUFunc_FromFuncAndData(
 */
  (__pyx_v_6treams_2sw_ufunc_cw_p_data[0]) = ((void *)__pyx_f_6treams_2sw__cperiodic_to_cw_p);

  /* "treams/sw.pyx":950
 * ufunc_cw_p_data[0] = <void*>_cperiodic_to_cw_p
 * 
 * _periodic_to_cw_h = np.PyUFunc_FromFuncAndData(             # <<<<<<<<<<<<<<
 *     ufunc_cw_loops,
 *     ufunc_cw_h_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndData(__pyx_v_6treams_2sw_ufunc_cw_loops, __pyx_v_6treams_2sw_ufunc_cw_h_data, __pyx_v_6treams_2sw_ufunc_cw_types, 1, 8, 1, 0, ((char *)"_periodic_to_cw_h"), ((char *)""), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_periodic_to_cw_h, __pyx_t_1) < 0) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/sw.pyx":962
 *     0,
 * )
 * _periodic_to_cw_p = np.PyUFunc_FromFuncAndData(             # <<<<<<<<<<<<<<
 *     ufunc_cw_loops,
 *     ufunc_cw_p_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndData(__pyx_v_6treams_2sw_ufunc_cw_loops, __pyx_v_6treams_2sw_ufunc_cw_p_data, __pyx_v_6treams_2sw_ufunc_cw_types, 1, 8, 1, 0, ((char *)"_periodic_to_cw_p"), ((char *)""), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_periodic_to_cw_p, __pyx_t_1) < 0) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/sw.pyx":976
 * 
 * 
 * def periodic_to_cw(kz, m, pol, l, mu, qol, k, area, poltype=None, *args, **kwargs):             # <<<<<<<<<<<<<<
 *     """periodic_to_cw(kz, m, pol, l, mu, qol, k, area, helicity=True)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_6treams_2sw_9periodic_to_cw, NULL, __pyx_n_s_treams_sw); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_periodic_to_cw, __pyx_t_1) < 0) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/sw.pyx":1
 * """Spherical wave module.             # <<<<<<<<<<<<<<
 * 
 * .. autosummary::
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init treams.sw", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init treams.sw");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* SliceObject */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    if (last_one_off) {
        assert(*dpos == '0');
        dpos++;
    }
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_0_29_36
#define __PYX_HAVE_RT_ImportType_0_29_36
static PyTypeObject *__Pyx_ImportType_0_29_36(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_0_29_36 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_0_29_36 && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_0_29_36 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_intptr_t neg_one = (Py_intptr_t) -1, const_zero = (Py_intptr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(Py_intptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_intptr_t neg_one = (Py_intptr_t) -1, const_zero = (Py_intptr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (Py_intptr_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (Py_intptr_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(Py_intptr_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            Py_intptr_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (Py_intptr_t) -1;
        }
    } else {
        Py_intptr_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (Py_intptr_t) -1;
        val = __Pyx_PyInt_As_Py_intptr_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to Py_intptr_t");
    return (Py_intptr_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to Py_intptr_t");
    return (Py_intptr_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr_0_29_36
#define __PYX_HAVE_RT_ImportVoidPtr_0_29_36
static int __Pyx_ImportVoidPtr_0_29_36(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_0_29_36
#define __PYX_HAVE_RT_ImportFunction_0_29_36
static int __Pyx_ImportFunction_0_29_36(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
