/* Generated by Cython 0.29.36 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "include_dirs": [
            "/tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/core/include"
        ],
        "name": "treams.special._gufuncs",
        "sources": [
            "src/treams/special/_gufuncs.pyx"
        ]
    },
    "module_name": "treams.special._gufuncs"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_36"
#define CYTHON_HEX_VERSION 0x001D24F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS ((PY_VERSION_HEX >= 0x030600B1) && (PY_VERSION_HEX < 0x030C00A5))
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__treams__special___gufuncs
#define __PYX_HAVE_API__treams__special___gufuncs
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/treams/special/_gufuncs.pyx",
  "__init__.pxd",
  "type.pxd",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()


/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":690
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":691
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":693
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":697
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":698
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":700
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":704
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":705
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":714
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":715
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":716
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":718
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":719
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":720
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":722
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":723
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":725
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":726
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":727
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);


/*--- Type declarations ---*/

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":729
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":730
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":731
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":733
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_0_29_36
#define __PYX_HAVE_RT_ImportType_proto_0_29_36
#if __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if __STDC_VERSION__ >= 201112L || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_0_29_36(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_0_29_36(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_0_29_36 {
   __Pyx_ImportType_CheckSize_Error_0_29_36 = 0,
   __Pyx_ImportType_CheckSize_Warn_0_29_36 = 1,
   __Pyx_ImportType_CheckSize_Ignore_0_29_36 = 2
};
static PyTypeObject *__Pyx_ImportType_0_29_36(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_0_29_36 check_size);
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_0_29_36(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_0_29_36(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/
static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void); /*proto*/

/* Module declarations from 'treams.special' */

/* Module declarations from 'treams.special._coord' */
static void (*__pyx_f_6treams_7special_6_coord_car2cyl)(double *, double *, long, long); /*proto*/
static void (*__pyx_f_6treams_7special_6_coord_car2pol)(double *, double *, long, long); /*proto*/
static void (*__pyx_f_6treams_7special_6_coord_car2sph)(double *, double *, long, long); /*proto*/
static void (*__pyx_f_6treams_7special_6_coord_cyl2car)(double *, double *, long, long); /*proto*/
static void (*__pyx_f_6treams_7special_6_coord_cyl2sph)(double *, double *, long, long); /*proto*/
static void (*__pyx_f_6treams_7special_6_coord_pol2car)(double *, double *, long, long); /*proto*/
static void (*__pyx_f_6treams_7special_6_coord_sph2cyl)(double *, double *, long, long); /*proto*/
static void (*__pyx_f_6treams_7special_6_coord_sph2car)(double *, double *, long, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcar2cyl)(__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcar2cyl)(double *, double *, double *, long, long, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcar2pol)(__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcar2pol)(double *, double *, double *, long, long, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcar2sph)(__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcar2sph)(double *, double *, double *, long, long, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcyl2car)(__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcyl2car)(double *, double *, double *, long, long, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcyl2sph)(__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcyl2sph)(double *, double *, double *, long, long, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vpol2car)(__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vpol2car)(double *, double *, double *, long, long, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vsph2car)(__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vsph2car)(double *, double *, double *, long, long, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vsph2cyl)(__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vsph2cyl)(double *, double *, double *, long, long, long); /*proto*/

/* Module declarations from 'treams.special._waves' */
static void (*__pyx_f_6treams_7special_6_waves_vcw_A)(double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, long, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_f_6treams_7special_6_waves_vcw_M)(double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_f_6treams_7special_6_waves_vcw_N)(double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vcw_rA)(double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, long, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vcw_rA)(double, long, double, double, double, __pyx_t_double_complex, long, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vcw_rM)(double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vcw_rM)(double, long, double, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vcw_rN)(double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vcw_rN)(double, long, double, double, double, __pyx_t_double_complex, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vpw_A)(__pyx_t_double_complex, __pyx_t_double_complex, __pyx_t_double_complex, double, double, double, long, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vpw_A)(double, double, double, double, double, double, long, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vpw_M)(__pyx_t_double_complex, __pyx_t_double_complex, __pyx_t_double_complex, double, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vpw_M)(double, double, double, double, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vpw_N)(__pyx_t_double_complex, __pyx_t_double_complex, __pyx_t_double_complex, double, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vpw_N)(double, double, double, double, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsh_X)(long, long, __pyx_t_double_complex, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsh_X)(long, long, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsh_Y)(long, long, __pyx_t_double_complex, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsh_Y)(long, long, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsh_Z)(long, long, __pyx_t_double_complex, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsh_Z)(long, long, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_A)(long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, long, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_A)(long, long, __pyx_t_double_complex, double, double, long, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_M)(long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_M)(long, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_N)(long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_N)(long, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_rA)(long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, long, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_rA)(long, long, double, double, double, long, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_rM)(long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_rM)(long, long, double, double, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_rN)(long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, __pyx_t_double_complex *, long); /*proto*/
static void (*__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_rN)(long, long, double, double, double, __pyx_t_double_complex *, long); /*proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'treams.special._misc' */
static double *__pyx_vp_6treams_7special_5_misc_SQPI = 0;
#define __pyx_v_6treams_7special_5_misc_SQPI (*__pyx_vp_6treams_7special_5_misc_SQPI)

/* Module declarations from 'treams.special._gufuncs' */
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_types[2];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_types[2];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_types[2];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_types[2];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_types[2];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_types[2];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_types[2];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_types[2];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2cyl_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types[(2 * 3)];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2sph_data[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcyl2car_data[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcyl2sph_data[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsph2car_data[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsph2cyl_data[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2pol_data[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vpol2car_data[2];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[(2 * 5)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[(2 * 5)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[(2 * 5)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[(2 * 6)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[(2 * 6)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[(2 * 6)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[(2 * 6)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[(2 * 7)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[(2 * 7)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_types[6];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[(2 * 6)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_types[7];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[(2 * 7)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_loops[1];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_data[1];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types[8];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[(2 * 8)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[(2 * 7)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[(2 * 7)];
static PyUFuncGenericFunction __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_loops[2];
static void *__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_data[2];
static char __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[(2 * 8)];
static void __pyx_f_6treams_7special_8_gufuncs_loop_d_d(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_d_dd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_Dd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_lldd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llddd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDdd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDDd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_lldddl(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDddl(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDDdl(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dlddd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dlDdd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dldddD(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dlDddD(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dldddDl(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dlDddDl(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dddddd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_DDDddd(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_ddddddl(char **, npy_intp *, npy_intp *, void *); /*proto*/
static void __pyx_f_6treams_7special_8_gufuncs_loop_D_DDDdddl(char **, npy_intp *, npy_intp *, void *); /*proto*/
#define __Pyx_MODULE_NAME "treams.special._gufuncs"
extern int __pyx_module_is_main_treams__special___gufuncs;
int __pyx_module_is_main_treams__special___gufuncs = 0;

/* Implementation of 'treams.special._gufuncs' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_vcw_A[] = "vcw_A";
static const char __pyx_k_vcw_M[] = "vcw_M";
static const char __pyx_k_vcw_N[] = "vcw_N";
static const char __pyx_k_vpw_A[] = "vpw_A";
static const char __pyx_k_vpw_M[] = "vpw_M";
static const char __pyx_k_vpw_N[] = "vpw_N";
static const char __pyx_k_vsh_X[] = "vsh_X";
static const char __pyx_k_vsh_Y[] = "vsh_Y";
static const char __pyx_k_vsh_Z[] = "vsh_Z";
static const char __pyx_k_vsw_A[] = "vsw_A";
static const char __pyx_k_vsw_M[] = "vsw_M";
static const char __pyx_k_vsw_N[] = "vsw_N";
static const char __pyx_k_vcw_rA[] = "vcw_rA";
static const char __pyx_k_vcw_rM[] = "vcw_rM";
static const char __pyx_k_vcw_rN[] = "vcw_rN";
static const char __pyx_k_vsw_rA[] = "vsw_rA";
static const char __pyx_k_vsw_rM[] = "vsw_rM";
static const char __pyx_k_vsw_rN[] = "vsw_rN";
static const char __pyx_k_car2cyl[] = "car2cyl";
static const char __pyx_k_car2pol[] = "car2pol";
static const char __pyx_k_car2sph[] = "car2sph";
static const char __pyx_k_cyl2car[] = "cyl2car";
static const char __pyx_k_cyl2sph[] = "cyl2sph";
static const char __pyx_k_pol2car[] = "pol2car";
static const char __pyx_k_sph2car[] = "sph2car";
static const char __pyx_k_sph2cyl[] = "sph2cyl";
static const char __pyx_k_vcar2cyl[] = "vcar2cyl";
static const char __pyx_k_vcar2pol[] = "vcar2pol";
static const char __pyx_k_vcar2sph[] = "vcar2sph";
static const char __pyx_k_vcyl2car[] = "vcyl2car";
static const char __pyx_k_vcyl2sph[] = "vcyl2sph";
static const char __pyx_k_vpol2car[] = "vpol2car";
static const char __pyx_k_vsph2car[] = "vsph2car";
static const char __pyx_k_vsph2cyl[] = "vsph2cyl";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_Generalized_universal_functions[] = "Generalized universal functions for the special subpackage";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_car2cyl;
static PyObject *__pyx_n_u_car2cyl;
static PyObject *__pyx_n_s_car2pol;
static PyObject *__pyx_n_u_car2pol;
static PyObject *__pyx_n_s_car2sph;
static PyObject *__pyx_n_u_car2sph;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_cyl2car;
static PyObject *__pyx_n_u_cyl2car;
static PyObject *__pyx_n_s_cyl2sph;
static PyObject *__pyx_n_u_cyl2sph;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_pol2car;
static PyObject *__pyx_n_u_pol2car;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_sph2car;
static PyObject *__pyx_n_u_sph2car;
static PyObject *__pyx_n_s_sph2cyl;
static PyObject *__pyx_n_u_sph2cyl;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_vcar2cyl;
static PyObject *__pyx_n_u_vcar2cyl;
static PyObject *__pyx_n_s_vcar2pol;
static PyObject *__pyx_n_u_vcar2pol;
static PyObject *__pyx_n_s_vcar2sph;
static PyObject *__pyx_n_u_vcar2sph;
static PyObject *__pyx_n_s_vcw_A;
static PyObject *__pyx_n_u_vcw_A;
static PyObject *__pyx_n_s_vcw_M;
static PyObject *__pyx_n_u_vcw_M;
static PyObject *__pyx_n_s_vcw_N;
static PyObject *__pyx_n_u_vcw_N;
static PyObject *__pyx_n_s_vcw_rA;
static PyObject *__pyx_n_u_vcw_rA;
static PyObject *__pyx_n_s_vcw_rM;
static PyObject *__pyx_n_u_vcw_rM;
static PyObject *__pyx_n_s_vcw_rN;
static PyObject *__pyx_n_u_vcw_rN;
static PyObject *__pyx_n_s_vcyl2car;
static PyObject *__pyx_n_u_vcyl2car;
static PyObject *__pyx_n_s_vcyl2sph;
static PyObject *__pyx_n_u_vcyl2sph;
static PyObject *__pyx_n_s_vpol2car;
static PyObject *__pyx_n_u_vpol2car;
static PyObject *__pyx_n_s_vpw_A;
static PyObject *__pyx_n_u_vpw_A;
static PyObject *__pyx_n_s_vpw_M;
static PyObject *__pyx_n_u_vpw_M;
static PyObject *__pyx_n_s_vpw_N;
static PyObject *__pyx_n_u_vpw_N;
static PyObject *__pyx_n_s_vsh_X;
static PyObject *__pyx_n_u_vsh_X;
static PyObject *__pyx_n_s_vsh_Y;
static PyObject *__pyx_n_u_vsh_Y;
static PyObject *__pyx_n_s_vsh_Z;
static PyObject *__pyx_n_u_vsh_Z;
static PyObject *__pyx_n_s_vsph2car;
static PyObject *__pyx_n_u_vsph2car;
static PyObject *__pyx_n_s_vsph2cyl;
static PyObject *__pyx_n_u_vsph2cyl;
static PyObject *__pyx_n_s_vsw_A;
static PyObject *__pyx_n_u_vsw_A;
static PyObject *__pyx_n_s_vsw_M;
static PyObject *__pyx_n_u_vsw_M;
static PyObject *__pyx_n_s_vsw_N;
static PyObject *__pyx_n_u_vsw_N;
static PyObject *__pyx_n_s_vsw_rA;
static PyObject *__pyx_n_u_vsw_rA;
static PyObject *__pyx_n_s_vsw_rM;
static PyObject *__pyx_n_u_vsw_rM;
static PyObject *__pyx_n_s_vsw_rN;
static PyObject *__pyx_n_u_vsw_rN;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
/* Late includes */

/* "treams/special/_gufuncs.pyx":46
 * 
 * 
 * cdef void loop_d_d(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep = steps[2] // sizeof(double), ostep = steps[3] // sizeof(double)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_d_d(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_istep;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":47
 * 
 * cdef void loop_d_d(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long istep = steps[2] // sizeof(double), ostep = steps[3] // sizeof(double)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":48
 * cdef void loop_d_d(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep = steps[2] // sizeof(double), ostep = steps[3] // sizeof(double)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(double));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 48, __pyx_L1_error)
  }
  __pyx_v_istep = ((__pyx_v_steps[2]) / __pyx_t_1);
  __pyx_t_1 = (sizeof(double));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 48, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[3]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":49
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep = steps[2] // sizeof(double), ostep = steps[3] // sizeof(double)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *op0 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":50
 *     cdef long istep = steps[2] // sizeof(double), ostep = steps[3] // sizeof(double)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[1]
 *     for i in range(n):
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":51
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *op0 = args[1]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double*, double*, long, long) nogil>func)(<double*>ip0, <double*>op0, istep, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":52
 *     cdef char *ip0 = args[0]
 *     cdef char *op0 = args[1]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double*, double*, long, long) nogil>func)(<double*>ip0, <double*>op0, istep, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":53
 *     cdef char *op0 = args[1]
 *     for i in range(n):
 *         (<void(*)(double*, double*, long, long) nogil>func)(<double*>ip0, <double*>op0, istep, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         op0 += steps[1]
 */
    ((void (*)(double *, double *, long, long))__pyx_v_func)(((double *)__pyx_v_ip0), ((double *)__pyx_v_op0), __pyx_v_istep, __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":54
 *     for i in range(n):
 *         (<void(*)(double*, double*, long, long) nogil>func)(<double*>ip0, <double*>op0, istep, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         op0 += steps[1]
 * 
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":55
 *         (<void(*)(double*, double*, long, long) nogil>func)(<double*>ip0, <double*>op0, istep, ostep)
 *         ip0 += steps[0]
 *         op0 += steps[1]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[1]));
  }

  /* "treams/special/_gufuncs.pyx":46
 * 
 * 
 * cdef void loop_d_d(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep = steps[2] // sizeof(double), ostep = steps[3] // sizeof(double)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_d_d", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":58
 * 
 * 
 * cdef void loop_d_dd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep0 = steps[3] // sizeof(double), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_d_dd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_istep0;
  long __pyx_v_istep1;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":59
 * 
 * cdef void loop_d_dd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long istep0 = steps[3] // sizeof(double), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":60
 * cdef void loop_d_dd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep0 = steps[3] // sizeof(double), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(double));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 60, __pyx_L1_error)
  }
  __pyx_v_istep0 = ((__pyx_v_steps[3]) / __pyx_t_1);
  __pyx_t_1 = (sizeof(double));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 60, __pyx_L1_error)
  }
  __pyx_v_istep1 = ((__pyx_v_steps[4]) / __pyx_t_1);
  __pyx_t_1 = (sizeof(double));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 60, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[5]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":61
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep0 = steps[3] // sizeof(double), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":62
 *     cdef long istep0 = steps[3] // sizeof(double), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *op0 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":63
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[2]
 *     for i in range(n):
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":64
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *op0 = args[2]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double*, double*, double*, long, long, long) nogil>func)(<double*>ip0, <double*>ip1, <double*>op0, istep0, istep1, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":65
 *     cdef char *ip1 = args[1]
 *     cdef char *op0 = args[2]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double*, double*, double*, long, long, long) nogil>func)(<double*>ip0, <double*>ip1, <double*>op0, istep0, istep1, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":66
 *     cdef char *op0 = args[2]
 *     for i in range(n):
 *         (<void(*)(double*, double*, double*, long, long, long) nogil>func)(<double*>ip0, <double*>ip1, <double*>op0, istep0, istep1, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(double *, double *, double *, long, long, long))__pyx_v_func)(((double *)__pyx_v_ip0), ((double *)__pyx_v_ip1), ((double *)__pyx_v_op0), __pyx_v_istep0, __pyx_v_istep1, __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":67
 *     for i in range(n):
 *         (<void(*)(double*, double*, double*, long, long, long) nogil>func)(<double*>ip0, <double*>ip1, <double*>op0, istep0, istep1, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         op0 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":68
 *         (<void(*)(double*, double*, double*, long, long, long) nogil>func)(<double*>ip0, <double*>ip1, <double*>op0, istep0, istep1, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         op0 += steps[2]
 * 
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":69
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         op0 += steps[2]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[2]));
  }

  /* "treams/special/_gufuncs.pyx":58
 * 
 * 
 * cdef void loop_d_dd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep0 = steps[3] // sizeof(double), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_d_dd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":72
 * 
 * 
 * cdef void loop_D_Dd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep0 = steps[3] // sizeof(double complex), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_Dd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_istep0;
  long __pyx_v_istep1;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":73
 * 
 * cdef void loop_D_Dd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long istep0 = steps[3] // sizeof(double complex), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":74
 * cdef void loop_D_Dd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep0 = steps[3] // sizeof(double complex), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 74, __pyx_L1_error)
  }
  __pyx_v_istep0 = ((__pyx_v_steps[3]) / __pyx_t_1);
  __pyx_t_1 = (sizeof(double));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 74, __pyx_L1_error)
  }
  __pyx_v_istep1 = ((__pyx_v_steps[4]) / __pyx_t_1);
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 74, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[5]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":75
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep0 = steps[3] // sizeof(double complex), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":76
 *     cdef long istep0 = steps[3] // sizeof(double complex), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *op0 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":77
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[2]
 *     for i in range(n):
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":78
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *op0 = args[2]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double complex*, double*, double complex*, long, long, long) nogil>func)(<double complex*>ip0, <double*>ip1, <double complex*>op0, istep0, istep1, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":79
 *     cdef char *ip1 = args[1]
 *     cdef char *op0 = args[2]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double complex*, double*, double complex*, long, long, long) nogil>func)(<double complex*>ip0, <double*>ip1, <double complex*>op0, istep0, istep1, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":80
 *     cdef char *op0 = args[2]
 *     for i in range(n):
 *         (<void(*)(double complex*, double*, double complex*, long, long, long) nogil>func)(<double complex*>ip0, <double*>ip1, <double complex*>op0, istep0, istep1, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long))__pyx_v_func)(((__pyx_t_double_complex *)__pyx_v_ip0), ((double *)__pyx_v_ip1), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_istep0, __pyx_v_istep1, __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":81
 *     for i in range(n):
 *         (<void(*)(double complex*, double*, double complex*, long, long, long) nogil>func)(<double complex*>ip0, <double*>ip1, <double complex*>op0, istep0, istep1, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         op0 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":82
 *         (<void(*)(double complex*, double*, double complex*, long, long, long) nogil>func)(<double complex*>ip0, <double*>ip1, <double complex*>op0, istep0, istep1, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         op0 += steps[2]
 * 
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":83
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         op0 += steps[2]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[2]));
  }

  /* "treams/special/_gufuncs.pyx":72
 * 
 * 
 * cdef void loop_D_Dd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long istep0 = steps[3] // sizeof(double complex), istep1 = steps[4] // sizeof(double), ostep = steps[5] // sizeof(double complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_Dd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":86
 * 
 * 
 * cdef void loop_D_lldd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_lldd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":87
 * 
 * cdef void loop_D_lldd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":88
 * cdef void loop_D_lldd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[5] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 88, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[5]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":89
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":90
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":91
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":92
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *op0 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":93
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[4]
 *     for i in range(n):
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":94
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *op0 = args[4]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(long, long, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":95
 *     cdef char *ip3 = args[3]
 *     cdef char *op0 = args[4]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(long, long, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":96
 *     cdef char *op0 = args[4]
 *     for i in range(n):
 *         (<void(*)(long, long, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(long, long, double, double, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((double)(((double *)__pyx_v_ip2)[0])), ((double)(((double *)__pyx_v_ip3)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":97
 *     for i in range(n):
 *         (<void(*)(long, long, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":98
 *         (<void(*)(long, long, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":99
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         op0 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":100
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         op0 += steps[4]
 * 
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":101
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         op0 += steps[4]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[4]));
  }

  /* "treams/special/_gufuncs.pyx":86
 * 
 * 
 * cdef void loop_D_lldd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_lldd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":104
 * 
 * 
 * cdef void loop_D_llDd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":105
 * 
 * cdef void loop_D_llDd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":106
 * cdef void loop_D_llDd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[5] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 106, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[5]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":107
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":108
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":109
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":110
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *op0 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":111
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[4]
 *     for i in range(n):
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":112
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *op0 = args[4]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":113
 *     cdef char *ip3 = args[3]
 *     cdef char *op0 = args[4]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(long, long, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":114
 *     cdef char *op0 = args[4]
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(long, long, __pyx_t_double_complex, double, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), ((double)(((double *)__pyx_v_ip3)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":115
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":116
 *         (<void(*)(long, long, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":117
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         op0 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":118
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         op0 += steps[4]
 * 
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":119
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         op0 += steps[4]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[4]));
  }

  /* "treams/special/_gufuncs.pyx":104
 * 
 * 
 * cdef void loop_D_llDd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[5] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_llDd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":122
 * 
 * 
 * cdef void loop_D_llddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llddd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":123
 * 
 * cdef void loop_D_llddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":124
 * cdef void loop_D_llddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 124, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[6]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":125
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":126
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":127
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":128
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":129
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":130
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":131
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(long, long, double, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":132
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(long, long, double, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":133
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 *         (<void(*)(long, long, double, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(long, long, double, double, double, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((double)(((double *)__pyx_v_ip2)[0])), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":134
 *     for i in range(n):
 *         (<void(*)(long, long, double, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":135
 *         (<void(*)(long, long, double, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":136
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":137
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         op0 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":138
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         op0 += steps[5]
 * 
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":139
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         op0 += steps[5]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[5]));
  }

  /* "treams/special/_gufuncs.pyx":122
 * 
 * 
 * cdef void loop_D_llddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_llddd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":142
 * 
 * 
 * cdef void loop_D_llDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDdd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":143
 * 
 * cdef void loop_D_llDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":144
 * cdef void loop_D_llDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 144, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[6]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":145
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":146
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":147
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":148
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":149
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":150
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":151
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":152
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(long, long, double complex, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":153
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(long, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":154
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":155
 *         (<void(*)(long, long, double complex, double, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":156
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":157
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         op0 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":158
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         op0 += steps[5]
 * 
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":159
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         op0 += steps[5]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[5]));
  }

  /* "treams/special/_gufuncs.pyx":142
 * 
 * 
 * cdef void loop_D_llDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_llDdd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":162
 * 
 * 
 * cdef void loop_D_llDDd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDDd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":163
 * 
 * cdef void loop_D_llDDd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":164
 * cdef void loop_D_llDDd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 164, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[6]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":165
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":166
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":167
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":168
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":169
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":170
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":171
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":172
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(long, long, double complex, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":173
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip3)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip3)[0])))), ((double)(((double *)__pyx_v_ip4)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":174
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":175
 *         (<void(*)(long, long, double complex, double complex, double, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":176
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":177
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         op0 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":178
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         op0 += steps[5]
 * 
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":179
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         op0 += steps[5]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[5]));
  }

  /* "treams/special/_gufuncs.pyx":162
 * 
 * 
 * cdef void loop_D_llDDd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_llDDd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":182
 * 
 * 
 * cdef void loop_D_lldddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_lldddl(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":183
 * 
 * cdef void loop_D_lldddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":184
 * cdef void loop_D_lldddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 184, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[7]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":185
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":186
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":187
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":188
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":189
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":190
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":191
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":192
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(long, long, double, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":193
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(long, long, double, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":194
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 *         (<void(*)(long, long, double, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(long, long, double, double, double, long, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((double)(((double *)__pyx_v_ip2)[0])), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":195
 *     for i in range(n):
 *         (<void(*)(long, long, double, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":196
 *         (<void(*)(long, long, double, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":197
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":198
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":199
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         op0 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":200
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         op0 += steps[6]
 * 
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":201
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         op0 += steps[6]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[6]));
  }

  /* "treams/special/_gufuncs.pyx":182
 * 
 * 
 * cdef void loop_D_lldddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_lldddl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":204
 * 
 * 
 * cdef void loop_D_llDddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDddl(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":205
 * 
 * cdef void loop_D_llDddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":206
 * cdef void loop_D_llDddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 206, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[7]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":207
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":208
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":209
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":210
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":211
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":212
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":213
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":214
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":215
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(long, long, double complex, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":216
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(long, long, __pyx_t_double_complex, double, double, long, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":217
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":218
 *         (<void(*)(long, long, double complex, double, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":219
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":220
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":221
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         op0 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":222
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         op0 += steps[6]
 * 
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":223
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         op0 += steps[6]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[6]));
  }

  /* "treams/special/_gufuncs.pyx":204
 * 
 * 
 * cdef void loop_D_llDddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_llDddl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":226
 * 
 * 
 * cdef void loop_D_llDDdl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_llDDdl(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":227
 * 
 * cdef void loop_D_llDDdl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":228
 * cdef void loop_D_llDDdl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 228, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[7]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":229
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":230
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":231
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":232
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":233
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":234
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":235
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":236
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double complex, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":237
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(long, long, double complex, double complex, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":238
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double complex, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, long, __pyx_t_double_complex *, long))__pyx_v_func)(((long)(((long *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip3)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip3)[0])))), ((double)(((double *)__pyx_v_ip4)[0])), ((long)(((long *)__pyx_v_ip5)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":239
 *     for i in range(n):
 *         (<void(*)(long, long, double complex, double complex, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":240
 *         (<void(*)(long, long, double complex, double complex, double, long, double complex*, long) nogil>func)(<long>(<long*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double complex>(<double complex*>ip3)[0], <double>(<double*>ip4)[0], <long>(<long*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":241
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":242
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":243
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         op0 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":244
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         op0 += steps[6]
 * 
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":245
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         op0 += steps[6]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[6]));
  }

  /* "treams/special/_gufuncs.pyx":226
 * 
 * 
 * cdef void loop_D_llDDdl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_llDDdl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":248
 * 
 * 
 * cdef void loop_D_dlddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dlddd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":249
 * 
 * cdef void loop_D_dlddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":250
 * cdef void loop_D_dlddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 250, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[6]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":251
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":252
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":253
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":254
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":255
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":256
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":257
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double, long, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":258
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double, long, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":259
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 *         (<void(*)(double, long, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(double, long, double, double, double, __pyx_t_double_complex *, long))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((double)(((double *)__pyx_v_ip2)[0])), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":260
 *     for i in range(n):
 *         (<void(*)(double, long, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":261
 *         (<void(*)(double, long, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":262
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":263
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         op0 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":264
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         op0 += steps[5]
 * 
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":265
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         op0 += steps[5]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[5]));
  }

  /* "treams/special/_gufuncs.pyx":248
 * 
 * 
 * cdef void loop_D_dlddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_dlddd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":268
 * 
 * 
 * cdef void loop_D_dlDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dlDdd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":269
 * 
 * cdef void loop_D_dlDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":270
 * cdef void loop_D_dlDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 270, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[6]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":271
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":272
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":273
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":274
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":275
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":276
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":277
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double, long, double complex, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":278
 *     cdef char *ip4 = args[4]
 *     cdef char *op0 = args[5]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double, long, double complex, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":279
 *     cdef char *op0 = args[5]
 *     for i in range(n):
 *         (<void(*)(double, long, double complex, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":280
 *     for i in range(n):
 *         (<void(*)(double, long, double complex, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":281
 *         (<void(*)(double, long, double complex, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":282
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":283
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         op0 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":284
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         op0 += steps[5]
 * 
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":285
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         op0 += steps[5]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[5]));
  }

  /* "treams/special/_gufuncs.pyx":268
 * 
 * 
 * cdef void loop_D_dlDdd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[6] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_dlDdd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":288
 * 
 * 
 * cdef void loop_D_dldddD(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dldddD(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":289
 * 
 * cdef void loop_D_dldddD(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":290
 * cdef void loop_D_dldddD(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 290, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[7]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":291
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":292
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":293
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":294
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":295
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":296
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":297
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":298
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double, long, double, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":299
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double, long, double, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":300
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 *         (<void(*)(double, long, double, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(double, long, double, double, double, __pyx_t_double_complex, __pyx_t_double_complex *, long))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((double)(((double *)__pyx_v_ip2)[0])), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip5)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip5)[0])))), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":301
 *     for i in range(n):
 *         (<void(*)(double, long, double, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":302
 *         (<void(*)(double, long, double, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":303
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":304
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":305
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         op0 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":306
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         op0 += steps[6]
 * 
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":307
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         op0 += steps[6]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[6]));
  }

  /* "treams/special/_gufuncs.pyx":288
 * 
 * 
 * cdef void loop_D_dldddD(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_dldddD", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":310
 * 
 * 
 * cdef void loop_D_dlDddD(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dlDddD(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":311
 * 
 * cdef void loop_D_dlDddD(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":312
 * cdef void loop_D_dlDddD(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 312, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[7]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":313
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":314
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":315
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":316
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":317
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":318
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":319
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":320
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double, long, double complex, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":321
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double, long, double complex, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":322
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 *         (<void(*)(double, long, double complex, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, __pyx_t_double_complex *, long))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip5)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip5)[0])))), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":323
 *     for i in range(n):
 *         (<void(*)(double, long, double complex, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":324
 *         (<void(*)(double, long, double complex, double, double, double complex, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":325
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":326
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":327
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         op0 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":328
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         op0 += steps[6]
 * 
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":329
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         op0 += steps[6]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[6]));
  }

  /* "treams/special/_gufuncs.pyx":310
 * 
 * 
 * cdef void loop_D_dlDddD(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_dlDddD", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":332
 * 
 * 
 * cdef void loop_D_dldddDl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dldddDl(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":333
 * 
 * cdef void loop_D_dldddDl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":334
 * cdef void loop_D_dldddDl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 334, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[8]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":335
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":336
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":337
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":338
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":339
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":340
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":341
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":342
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[7]
 *     for i in range(n):
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":343
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double, long, double, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[7]);

  /* "treams/special/_gufuncs.pyx":344
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double, long, double, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":345
 *     cdef char *op0 = args[7]
 *     for i in range(n):
 *         (<void(*)(double, long, double, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(double, long, double, double, double, __pyx_t_double_complex, long, __pyx_t_double_complex *, long))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), ((double)(((double *)__pyx_v_ip2)[0])), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip5)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip5)[0])))), ((long)(((long *)__pyx_v_ip6)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":346
 *     for i in range(n):
 *         (<void(*)(double, long, double, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":347
 *         (<void(*)(double, long, double, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":348
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":349
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":350
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":351
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         op0 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":352
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         op0 += steps[7]
 * 
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/special/_gufuncs.pyx":353
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         op0 += steps[7]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[7]));
  }

  /* "treams/special/_gufuncs.pyx":332
 * 
 * 
 * cdef void loop_D_dldddDl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_dldddDl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":356
 * 
 * 
 * cdef void loop_D_dlDddDl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dlDddDl(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":357
 * 
 * cdef void loop_D_dlDddDl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":358
 * cdef void loop_D_dlDddDl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 358, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[8]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":359
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":360
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":361
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":362
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":363
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":364
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":365
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":366
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[7]
 *     for i in range(n):
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":367
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double, long, double complex, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[7]);

  /* "treams/special/_gufuncs.pyx":368
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double, long, double complex, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":369
 *     cdef char *op0 = args[7]
 *     for i in range(n):
 *         (<void(*)(double, long, double complex, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, long, __pyx_t_double_complex *, long))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((long)(((long *)__pyx_v_ip1)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip5)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip5)[0])))), ((long)(((long *)__pyx_v_ip6)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":370
 *     for i in range(n):
 *         (<void(*)(double, long, double complex, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":371
 *         (<void(*)(double, long, double complex, double, double, double complex, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <long>(<long*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double complex>(<double complex*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":372
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":373
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":374
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":375
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         op0 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":376
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         op0 += steps[7]
 * 
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/special/_gufuncs.pyx":377
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         op0 += steps[7]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[7]));
  }

  /* "treams/special/_gufuncs.pyx":356
 * 
 * 
 * cdef void loop_D_dlDddDl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_dlDddDl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":380
 * 
 * 
 * cdef void loop_D_dddddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_dddddd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":381
 * 
 * cdef void loop_D_dddddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":382
 * cdef void loop_D_dddddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 382, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[7]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":383
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":384
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":385
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":386
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":387
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":388
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":389
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":390
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double, double, double, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":391
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double, double, double, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":392
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 *         (<void(*)(double, double, double, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(double, double, double, double, double, double, __pyx_t_double_complex *, long))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((double)(((double *)__pyx_v_ip1)[0])), ((double)(((double *)__pyx_v_ip2)[0])), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((double)(((double *)__pyx_v_ip5)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":393
 *     for i in range(n):
 *         (<void(*)(double, double, double, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":394
 *         (<void(*)(double, double, double, double, double, double, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":395
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":396
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":397
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         op0 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":398
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         op0 += steps[6]
 * 
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":399
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         op0 += steps[6]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[6]));
  }

  /* "treams/special/_gufuncs.pyx":380
 * 
 * 
 * cdef void loop_D_dddddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_dddddd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":402
 * 
 * 
 * cdef void loop_D_DDDddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_DDDddd(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":403
 * 
 * cdef void loop_D_DDDddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":404
 * cdef void loop_D_DDDddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 404, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[7]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":405
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":406
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":407
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":408
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":409
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":410
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":411
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":412
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double complex, double complex, double complex, double, double, double, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":413
 *     cdef char *ip5 = args[5]
 *     cdef char *op0 = args[6]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double complex, double complex, double complex, double, double, double, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":414
 *     cdef char *op0 = args[6]
 *     for i in range(n):
 *         (<void(*)(double complex, double complex, double complex, double, double, double, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(__pyx_t_double_complex, __pyx_t_double_complex, __pyx_t_double_complex, double, double, double, __pyx_t_double_complex *, long))__pyx_v_func)(__pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip0)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip0)[0])))), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip1)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip1)[0])))), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((double)(((double *)__pyx_v_ip5)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":415
 *     for i in range(n):
 *         (<void(*)(double complex, double complex, double complex, double, double, double, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":416
 *         (<void(*)(double complex, double complex, double complex, double, double, double, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":417
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":418
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":419
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         op0 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":420
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         op0 += steps[6]
 * 
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":421
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         op0 += steps[6]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[6]));
  }

  /* "treams/special/_gufuncs.pyx":402
 * 
 * 
 * cdef void loop_D_DDDddd(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[7] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_DDDddd", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":424
 * 
 * 
 * cdef void loop_D_ddddddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_ddddddl(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":425
 * 
 * cdef void loop_D_ddddddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":426
 * cdef void loop_D_ddddddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 426, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[8]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":427
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":428
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":429
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":430
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":431
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":432
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":433
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":434
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[7]
 *     for i in range(n):
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":435
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double, double, double, double, double, double, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[7]);

  /* "treams/special/_gufuncs.pyx":436
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double, double, double, double, double, double, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":437
 *     cdef char *op0 = args[7]
 *     for i in range(n):
 *         (<void(*)(double, double, double, double, double, double, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(double, double, double, double, double, double, long, __pyx_t_double_complex *, long))__pyx_v_func)(((double)(((double *)__pyx_v_ip0)[0])), ((double)(((double *)__pyx_v_ip1)[0])), ((double)(((double *)__pyx_v_ip2)[0])), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((double)(((double *)__pyx_v_ip5)[0])), ((long)(((long *)__pyx_v_ip6)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":438
 *     for i in range(n):
 *         (<void(*)(double, double, double, double, double, double, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":439
 *         (<void(*)(double, double, double, double, double, double, long, double complex*, long) nogil>func)(<double>(<double*>ip0)[0], <double>(<double*>ip1)[0], <double>(<double*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":440
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":441
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":442
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":443
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         op0 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":444
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         op0 += steps[7]
 * 
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/special/_gufuncs.pyx":445
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         op0 += steps[7]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[7]));
  }

  /* "treams/special/_gufuncs.pyx":424
 * 
 * 
 * cdef void loop_D_ddddddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_ddddddl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "treams/special/_gufuncs.pyx":448
 * 
 * 
 * cdef void loop_D_DDDdddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 */

static void __pyx_f_6treams_7special_8_gufuncs_loop_D_DDDdddl(char **__pyx_v_args, npy_intp *__pyx_v_dims, npy_intp *__pyx_v_steps, void *__pyx_v_data) {
  CYTHON_UNUSED npy_intp __pyx_v_i;
  npy_intp __pyx_v_n;
  long __pyx_v_ostep;
  void *__pyx_v_func;
  char *__pyx_v_ip0;
  char *__pyx_v_ip1;
  char *__pyx_v_ip2;
  char *__pyx_v_ip3;
  char *__pyx_v_ip4;
  char *__pyx_v_ip5;
  char *__pyx_v_ip6;
  char *__pyx_v_op0;
  size_t __pyx_t_1;
  npy_intp __pyx_t_2;
  npy_intp __pyx_t_3;
  npy_intp __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "treams/special/_gufuncs.pyx":449
 * 
 * cdef void loop_D_DDDdddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]             # <<<<<<<<<<<<<<
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 */
  __pyx_v_n = (__pyx_v_dims[0]);

  /* "treams/special/_gufuncs.pyx":450
 * cdef void loop_D_DDDdddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)             # <<<<<<<<<<<<<<
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 */
  __pyx_t_1 = (sizeof(__pyx_t_double_complex));
  if (unlikely(__pyx_t_1 == 0)) {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
    __PYX_ERR(0, 450, __pyx_L1_error)
  }
  __pyx_v_ostep = ((__pyx_v_steps[8]) / __pyx_t_1);

  /* "treams/special/_gufuncs.pyx":451
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data             # <<<<<<<<<<<<<<
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 */
  __pyx_v_func = ((void *)__pyx_v_data);

  /* "treams/special/_gufuncs.pyx":452
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]             # <<<<<<<<<<<<<<
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 */
  __pyx_v_ip0 = (__pyx_v_args[0]);

  /* "treams/special/_gufuncs.pyx":453
 *     cdef void *func = <void*>data
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]             # <<<<<<<<<<<<<<
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 */
  __pyx_v_ip1 = (__pyx_v_args[1]);

  /* "treams/special/_gufuncs.pyx":454
 *     cdef char *ip0 = args[0]
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]             # <<<<<<<<<<<<<<
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 */
  __pyx_v_ip2 = (__pyx_v_args[2]);

  /* "treams/special/_gufuncs.pyx":455
 *     cdef char *ip1 = args[1]
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]             # <<<<<<<<<<<<<<
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 */
  __pyx_v_ip3 = (__pyx_v_args[3]);

  /* "treams/special/_gufuncs.pyx":456
 *     cdef char *ip2 = args[2]
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]             # <<<<<<<<<<<<<<
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 */
  __pyx_v_ip4 = (__pyx_v_args[4]);

  /* "treams/special/_gufuncs.pyx":457
 *     cdef char *ip3 = args[3]
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]             # <<<<<<<<<<<<<<
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]
 */
  __pyx_v_ip5 = (__pyx_v_args[5]);

  /* "treams/special/_gufuncs.pyx":458
 *     cdef char *ip4 = args[4]
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]             # <<<<<<<<<<<<<<
 *     cdef char *op0 = args[7]
 *     for i in range(n):
 */
  __pyx_v_ip6 = (__pyx_v_args[6]);

  /* "treams/special/_gufuncs.pyx":459
 *     cdef char *ip5 = args[5]
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]             # <<<<<<<<<<<<<<
 *     for i in range(n):
 *         (<void(*)(double complex, double complex, double complex, double, double, double, long, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 */
  __pyx_v_op0 = (__pyx_v_args[7]);

  /* "treams/special/_gufuncs.pyx":460
 *     cdef char *ip6 = args[6]
 *     cdef char *op0 = args[7]
 *     for i in range(n):             # <<<<<<<<<<<<<<
 *         (<void(*)(double complex, double complex, double complex, double, double, double, long, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "treams/special/_gufuncs.pyx":461
 *     cdef char *op0 = args[7]
 *     for i in range(n):
 *         (<void(*)(double complex, double complex, double complex, double, double, double, long, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)             # <<<<<<<<<<<<<<
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 */
    ((void (*)(__pyx_t_double_complex, __pyx_t_double_complex, __pyx_t_double_complex, double, double, double, long, __pyx_t_double_complex *, long))__pyx_v_func)(__pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip0)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip0)[0])))), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip1)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip1)[0])))), __pyx_t_double_complex_from_parts(((double)__Pyx_CREAL((((__pyx_t_double_complex *)__pyx_v_ip2)[0]))), ((double)__Pyx_CIMAG((((__pyx_t_double_complex *)__pyx_v_ip2)[0])))), ((double)(((double *)__pyx_v_ip3)[0])), ((double)(((double *)__pyx_v_ip4)[0])), ((double)(((double *)__pyx_v_ip5)[0])), ((long)(((long *)__pyx_v_ip6)[0])), ((__pyx_t_double_complex *)__pyx_v_op0), __pyx_v_ostep);

    /* "treams/special/_gufuncs.pyx":462
 *     for i in range(n):
 *         (<void(*)(double complex, double complex, double complex, double, double, double, long, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]             # <<<<<<<<<<<<<<
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 */
    __pyx_v_ip0 = (__pyx_v_ip0 + (__pyx_v_steps[0]));

    /* "treams/special/_gufuncs.pyx":463
 *         (<void(*)(double complex, double complex, double complex, double, double, double, long, double complex*, long) nogil>func)(<double complex>(<double complex*>ip0)[0], <double complex>(<double complex*>ip1)[0], <double complex>(<double complex*>ip2)[0], <double>(<double*>ip3)[0], <double>(<double*>ip4)[0], <double>(<double*>ip5)[0], <long>(<long*>ip6)[0], <double complex*>op0, ostep)
 *         ip0 += steps[0]
 *         ip1 += steps[1]             # <<<<<<<<<<<<<<
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 */
    __pyx_v_ip1 = (__pyx_v_ip1 + (__pyx_v_steps[1]));

    /* "treams/special/_gufuncs.pyx":464
 *         ip0 += steps[0]
 *         ip1 += steps[1]
 *         ip2 += steps[2]             # <<<<<<<<<<<<<<
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 */
    __pyx_v_ip2 = (__pyx_v_ip2 + (__pyx_v_steps[2]));

    /* "treams/special/_gufuncs.pyx":465
 *         ip1 += steps[1]
 *         ip2 += steps[2]
 *         ip3 += steps[3]             # <<<<<<<<<<<<<<
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 */
    __pyx_v_ip3 = (__pyx_v_ip3 + (__pyx_v_steps[3]));

    /* "treams/special/_gufuncs.pyx":466
 *         ip2 += steps[2]
 *         ip3 += steps[3]
 *         ip4 += steps[4]             # <<<<<<<<<<<<<<
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 */
    __pyx_v_ip4 = (__pyx_v_ip4 + (__pyx_v_steps[4]));

    /* "treams/special/_gufuncs.pyx":467
 *         ip3 += steps[3]
 *         ip4 += steps[4]
 *         ip5 += steps[5]             # <<<<<<<<<<<<<<
 *         ip6 += steps[6]
 *         op0 += steps[7]
 */
    __pyx_v_ip5 = (__pyx_v_ip5 + (__pyx_v_steps[5]));

    /* "treams/special/_gufuncs.pyx":468
 *         ip4 += steps[4]
 *         ip5 += steps[5]
 *         ip6 += steps[6]             # <<<<<<<<<<<<<<
 *         op0 += steps[7]
 * 
 */
    __pyx_v_ip6 = (__pyx_v_ip6 + (__pyx_v_steps[6]));

    /* "treams/special/_gufuncs.pyx":469
 *         ip5 += steps[5]
 *         ip6 += steps[6]
 *         op0 += steps[7]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_op0 = (__pyx_v_op0 + (__pyx_v_steps[7]));
  }

  /* "treams/special/_gufuncs.pyx":448
 * 
 * 
 * cdef void loop_D_DDDdddl(char **args, np.npy_intp *dims, np.npy_intp *steps, void *data) nogil:             # <<<<<<<<<<<<<<
 *     cdef np.npy_intp i, n = dims[0]
 *     cdef long ostep = steps[8] // sizeof(double_complex)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("treams.special._gufuncs.loop_D_DDDdddl", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 1);
  __pyx_L0:;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":736
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":739
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":742
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":745
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":748
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":752
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":754
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":930
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":931
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":934
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":936
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":937
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":943
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 943, __pyx_L3_error)

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":944
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 944, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 945, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 945, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":949
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 949, __pyx_L3_error)

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":950
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 950, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 951, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 951, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":955
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 955, __pyx_L3_error)

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":956
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 956, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":957
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 957, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 957, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":979
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":994
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1004
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1011
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1018
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_misc.pxd":29
 * cdef long ipow(long base, long exponent) nogil
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a > b else b)
 * 
 */

static CYTHON_INLINE double __pyx_fuse_0__pyx_f_6treams_7special_5_misc_max(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  double __pyx_t_1;

  /* "treams/special/_misc.pxd":30
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:
 *     return (a if a > b else b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 */
  if (((__pyx_v_a > __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":29
 * cdef long ipow(long base, long exponent) nogil
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a > b else b)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static CYTHON_INLINE long __pyx_fuse_1__pyx_f_6treams_7special_5_misc_max(long __pyx_v_a, long __pyx_v_b) {
  long __pyx_r;
  long __pyx_t_1;

  /* "treams/special/_misc.pxd":30
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:
 *     return (a if a > b else b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 */
  if (((__pyx_v_a > __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":29
 * cdef long ipow(long base, long exponent) nogil
 * 
 * cdef inline real_t max(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a > b else b)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

static CYTHON_INLINE double __pyx_fuse_0__pyx_f_6treams_7special_5_misc_min(double __pyx_v_a, double __pyx_v_b) {
  double __pyx_r;
  double __pyx_t_1;

  /* "treams/special/_misc.pxd":33
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 *     return (a if a < b else b)             # <<<<<<<<<<<<<<
 */
  if (((__pyx_v_a < __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static CYTHON_INLINE long __pyx_fuse_1__pyx_f_6treams_7special_5_misc_min(long __pyx_v_a, long __pyx_v_b) {
  long __pyx_r;
  long __pyx_t_1;

  /* "treams/special/_misc.pxd":33
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:
 *     return (a if a < b else b)             # <<<<<<<<<<<<<<
 */
  if (((__pyx_v_a < __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__gufuncs(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__gufuncs},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "_gufuncs",
    __pyx_k_Generalized_universal_functions, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_car2cyl, __pyx_k_car2cyl, sizeof(__pyx_k_car2cyl), 0, 0, 1, 1},
  {&__pyx_n_u_car2cyl, __pyx_k_car2cyl, sizeof(__pyx_k_car2cyl), 0, 1, 0, 1},
  {&__pyx_n_s_car2pol, __pyx_k_car2pol, sizeof(__pyx_k_car2pol), 0, 0, 1, 1},
  {&__pyx_n_u_car2pol, __pyx_k_car2pol, sizeof(__pyx_k_car2pol), 0, 1, 0, 1},
  {&__pyx_n_s_car2sph, __pyx_k_car2sph, sizeof(__pyx_k_car2sph), 0, 0, 1, 1},
  {&__pyx_n_u_car2sph, __pyx_k_car2sph, sizeof(__pyx_k_car2sph), 0, 1, 0, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_cyl2car, __pyx_k_cyl2car, sizeof(__pyx_k_cyl2car), 0, 0, 1, 1},
  {&__pyx_n_u_cyl2car, __pyx_k_cyl2car, sizeof(__pyx_k_cyl2car), 0, 1, 0, 1},
  {&__pyx_n_s_cyl2sph, __pyx_k_cyl2sph, sizeof(__pyx_k_cyl2sph), 0, 0, 1, 1},
  {&__pyx_n_u_cyl2sph, __pyx_k_cyl2sph, sizeof(__pyx_k_cyl2sph), 0, 1, 0, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_pol2car, __pyx_k_pol2car, sizeof(__pyx_k_pol2car), 0, 0, 1, 1},
  {&__pyx_n_u_pol2car, __pyx_k_pol2car, sizeof(__pyx_k_pol2car), 0, 1, 0, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_sph2car, __pyx_k_sph2car, sizeof(__pyx_k_sph2car), 0, 0, 1, 1},
  {&__pyx_n_u_sph2car, __pyx_k_sph2car, sizeof(__pyx_k_sph2car), 0, 1, 0, 1},
  {&__pyx_n_s_sph2cyl, __pyx_k_sph2cyl, sizeof(__pyx_k_sph2cyl), 0, 0, 1, 1},
  {&__pyx_n_u_sph2cyl, __pyx_k_sph2cyl, sizeof(__pyx_k_sph2cyl), 0, 1, 0, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_vcar2cyl, __pyx_k_vcar2cyl, sizeof(__pyx_k_vcar2cyl), 0, 0, 1, 1},
  {&__pyx_n_u_vcar2cyl, __pyx_k_vcar2cyl, sizeof(__pyx_k_vcar2cyl), 0, 1, 0, 1},
  {&__pyx_n_s_vcar2pol, __pyx_k_vcar2pol, sizeof(__pyx_k_vcar2pol), 0, 0, 1, 1},
  {&__pyx_n_u_vcar2pol, __pyx_k_vcar2pol, sizeof(__pyx_k_vcar2pol), 0, 1, 0, 1},
  {&__pyx_n_s_vcar2sph, __pyx_k_vcar2sph, sizeof(__pyx_k_vcar2sph), 0, 0, 1, 1},
  {&__pyx_n_u_vcar2sph, __pyx_k_vcar2sph, sizeof(__pyx_k_vcar2sph), 0, 1, 0, 1},
  {&__pyx_n_s_vcw_A, __pyx_k_vcw_A, sizeof(__pyx_k_vcw_A), 0, 0, 1, 1},
  {&__pyx_n_u_vcw_A, __pyx_k_vcw_A, sizeof(__pyx_k_vcw_A), 0, 1, 0, 1},
  {&__pyx_n_s_vcw_M, __pyx_k_vcw_M, sizeof(__pyx_k_vcw_M), 0, 0, 1, 1},
  {&__pyx_n_u_vcw_M, __pyx_k_vcw_M, sizeof(__pyx_k_vcw_M), 0, 1, 0, 1},
  {&__pyx_n_s_vcw_N, __pyx_k_vcw_N, sizeof(__pyx_k_vcw_N), 0, 0, 1, 1},
  {&__pyx_n_u_vcw_N, __pyx_k_vcw_N, sizeof(__pyx_k_vcw_N), 0, 1, 0, 1},
  {&__pyx_n_s_vcw_rA, __pyx_k_vcw_rA, sizeof(__pyx_k_vcw_rA), 0, 0, 1, 1},
  {&__pyx_n_u_vcw_rA, __pyx_k_vcw_rA, sizeof(__pyx_k_vcw_rA), 0, 1, 0, 1},
  {&__pyx_n_s_vcw_rM, __pyx_k_vcw_rM, sizeof(__pyx_k_vcw_rM), 0, 0, 1, 1},
  {&__pyx_n_u_vcw_rM, __pyx_k_vcw_rM, sizeof(__pyx_k_vcw_rM), 0, 1, 0, 1},
  {&__pyx_n_s_vcw_rN, __pyx_k_vcw_rN, sizeof(__pyx_k_vcw_rN), 0, 0, 1, 1},
  {&__pyx_n_u_vcw_rN, __pyx_k_vcw_rN, sizeof(__pyx_k_vcw_rN), 0, 1, 0, 1},
  {&__pyx_n_s_vcyl2car, __pyx_k_vcyl2car, sizeof(__pyx_k_vcyl2car), 0, 0, 1, 1},
  {&__pyx_n_u_vcyl2car, __pyx_k_vcyl2car, sizeof(__pyx_k_vcyl2car), 0, 1, 0, 1},
  {&__pyx_n_s_vcyl2sph, __pyx_k_vcyl2sph, sizeof(__pyx_k_vcyl2sph), 0, 0, 1, 1},
  {&__pyx_n_u_vcyl2sph, __pyx_k_vcyl2sph, sizeof(__pyx_k_vcyl2sph), 0, 1, 0, 1},
  {&__pyx_n_s_vpol2car, __pyx_k_vpol2car, sizeof(__pyx_k_vpol2car), 0, 0, 1, 1},
  {&__pyx_n_u_vpol2car, __pyx_k_vpol2car, sizeof(__pyx_k_vpol2car), 0, 1, 0, 1},
  {&__pyx_n_s_vpw_A, __pyx_k_vpw_A, sizeof(__pyx_k_vpw_A), 0, 0, 1, 1},
  {&__pyx_n_u_vpw_A, __pyx_k_vpw_A, sizeof(__pyx_k_vpw_A), 0, 1, 0, 1},
  {&__pyx_n_s_vpw_M, __pyx_k_vpw_M, sizeof(__pyx_k_vpw_M), 0, 0, 1, 1},
  {&__pyx_n_u_vpw_M, __pyx_k_vpw_M, sizeof(__pyx_k_vpw_M), 0, 1, 0, 1},
  {&__pyx_n_s_vpw_N, __pyx_k_vpw_N, sizeof(__pyx_k_vpw_N), 0, 0, 1, 1},
  {&__pyx_n_u_vpw_N, __pyx_k_vpw_N, sizeof(__pyx_k_vpw_N), 0, 1, 0, 1},
  {&__pyx_n_s_vsh_X, __pyx_k_vsh_X, sizeof(__pyx_k_vsh_X), 0, 0, 1, 1},
  {&__pyx_n_u_vsh_X, __pyx_k_vsh_X, sizeof(__pyx_k_vsh_X), 0, 1, 0, 1},
  {&__pyx_n_s_vsh_Y, __pyx_k_vsh_Y, sizeof(__pyx_k_vsh_Y), 0, 0, 1, 1},
  {&__pyx_n_u_vsh_Y, __pyx_k_vsh_Y, sizeof(__pyx_k_vsh_Y), 0, 1, 0, 1},
  {&__pyx_n_s_vsh_Z, __pyx_k_vsh_Z, sizeof(__pyx_k_vsh_Z), 0, 0, 1, 1},
  {&__pyx_n_u_vsh_Z, __pyx_k_vsh_Z, sizeof(__pyx_k_vsh_Z), 0, 1, 0, 1},
  {&__pyx_n_s_vsph2car, __pyx_k_vsph2car, sizeof(__pyx_k_vsph2car), 0, 0, 1, 1},
  {&__pyx_n_u_vsph2car, __pyx_k_vsph2car, sizeof(__pyx_k_vsph2car), 0, 1, 0, 1},
  {&__pyx_n_s_vsph2cyl, __pyx_k_vsph2cyl, sizeof(__pyx_k_vsph2cyl), 0, 0, 1, 1},
  {&__pyx_n_u_vsph2cyl, __pyx_k_vsph2cyl, sizeof(__pyx_k_vsph2cyl), 0, 1, 0, 1},
  {&__pyx_n_s_vsw_A, __pyx_k_vsw_A, sizeof(__pyx_k_vsw_A), 0, 0, 1, 1},
  {&__pyx_n_u_vsw_A, __pyx_k_vsw_A, sizeof(__pyx_k_vsw_A), 0, 1, 0, 1},
  {&__pyx_n_s_vsw_M, __pyx_k_vsw_M, sizeof(__pyx_k_vsw_M), 0, 0, 1, 1},
  {&__pyx_n_u_vsw_M, __pyx_k_vsw_M, sizeof(__pyx_k_vsw_M), 0, 1, 0, 1},
  {&__pyx_n_s_vsw_N, __pyx_k_vsw_N, sizeof(__pyx_k_vsw_N), 0, 0, 1, 1},
  {&__pyx_n_u_vsw_N, __pyx_k_vsw_N, sizeof(__pyx_k_vsw_N), 0, 1, 0, 1},
  {&__pyx_n_s_vsw_rA, __pyx_k_vsw_rA, sizeof(__pyx_k_vsw_rA), 0, 0, 1, 1},
  {&__pyx_n_u_vsw_rA, __pyx_k_vsw_rA, sizeof(__pyx_k_vsw_rA), 0, 1, 0, 1},
  {&__pyx_n_s_vsw_rM, __pyx_k_vsw_rM, sizeof(__pyx_k_vsw_rM), 0, 0, 1, 1},
  {&__pyx_n_u_vsw_rM, __pyx_k_vsw_rM, sizeof(__pyx_k_vsw_rM), 0, 1, 0, 1},
  {&__pyx_n_s_vsw_rN, __pyx_k_vsw_rN, sizeof(__pyx_k_vsw_rN), 0, 0, 1, 1},
  {&__pyx_n_u_vsw_rN, __pyx_k_vsw_rN, sizeof(__pyx_k_vsw_rN), 0, 1, 0, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 52, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 945, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple_)) __PYX_ERR(1, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "../../../../../tmp/build-env-crh18hlu/lib/python3.10/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(1, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_0_29_36(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyArray_Descr),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 200, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyArrayIterObject),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 223, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyArrayMultiIterObject),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 227, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyArrayObject),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 239, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_generic) __PYX_ERR(1, 771, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "number", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_number) __PYX_ERR(1, 773, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_integer) __PYX_ERR(1, 775, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 777, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 779, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 781, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_floating) __PYX_ERR(1, 783, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 785, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 787, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "character", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyObject),__Pyx_ImportType_CheckSize_Warn_0_29_36); if (!__pyx_ptype_5numpy_character) __PYX_ERR(1, 789, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType_0_29_36(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_0_29_36(PyUFuncObject),__Pyx_ImportType_CheckSize_Ignore_0_29_36); if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("treams.special._misc"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_0_29_36(__pyx_t_1, "SQPI", (void **)&__pyx_vp_6treams_7special_5_misc_SQPI, "double") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("treams.special._coord"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "car2cyl", (void (**)(void))&__pyx_f_6treams_7special_6_coord_car2cyl, "void (double *, double *, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "car2pol", (void (**)(void))&__pyx_f_6treams_7special_6_coord_car2pol, "void (double *, double *, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "car2sph", (void (**)(void))&__pyx_f_6treams_7special_6_coord_car2sph, "void (double *, double *, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "cyl2car", (void (**)(void))&__pyx_f_6treams_7special_6_coord_cyl2car, "void (double *, double *, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "cyl2sph", (void (**)(void))&__pyx_f_6treams_7special_6_coord_cyl2sph, "void (double *, double *, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "pol2car", (void (**)(void))&__pyx_f_6treams_7special_6_coord_pol2car, "void (double *, double *, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "sph2cyl", (void (**)(void))&__pyx_f_6treams_7special_6_coord_sph2cyl, "void (double *, double *, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "sph2car", (void (**)(void))&__pyx_f_6treams_7special_6_coord_sph2car, "void (double *, double *, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vcar2cyl", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcar2cyl, "void (__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vcar2cyl", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcar2cyl, "void (double *, double *, double *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vcar2pol", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcar2pol, "void (__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vcar2pol", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcar2pol, "void (double *, double *, double *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vcar2sph", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcar2sph, "void (__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vcar2sph", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcar2sph, "void (double *, double *, double *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vcyl2car", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcyl2car, "void (__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vcyl2car", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcyl2car, "void (double *, double *, double *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vcyl2sph", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcyl2sph, "void (__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vcyl2sph", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcyl2sph, "void (double *, double *, double *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vpol2car", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vpol2car, "void (__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vpol2car", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vpol2car, "void (double *, double *, double *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsph2car", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vsph2car, "void (__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsph2car", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vsph2car, "void (double *, double *, double *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsph2cyl", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vsph2cyl, "void (__pyx_t_double_complex *, double *, __pyx_t_double_complex *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsph2cyl", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vsph2cyl, "void (double *, double *, double *, long, long, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("treams.special._waves"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "vcw_A", (void (**)(void))&__pyx_f_6treams_7special_6_waves_vcw_A, "void (double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, long, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "vcw_M", (void (**)(void))&__pyx_f_6treams_7special_6_waves_vcw_M, "void (double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "vcw_N", (void (**)(void))&__pyx_f_6treams_7special_6_waves_vcw_N, "void (double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vcw_rA", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vcw_rA, "void (double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, long, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vcw_rA", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vcw_rA, "void (double, long, double, double, double, __pyx_t_double_complex, long, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vcw_rM", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vcw_rM, "void (double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vcw_rM", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vcw_rM, "void (double, long, double, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vcw_rN", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vcw_rN, "void (double, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vcw_rN", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vcw_rN, "void (double, long, double, double, double, __pyx_t_double_complex, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vpw_A", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vpw_A, "void (__pyx_t_double_complex, __pyx_t_double_complex, __pyx_t_double_complex, double, double, double, long, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vpw_A", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vpw_A, "void (double, double, double, double, double, double, long, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vpw_M", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vpw_M, "void (__pyx_t_double_complex, __pyx_t_double_complex, __pyx_t_double_complex, double, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vpw_M", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vpw_M, "void (double, double, double, double, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vpw_N", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vpw_N, "void (__pyx_t_double_complex, __pyx_t_double_complex, __pyx_t_double_complex, double, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vpw_N", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vpw_N, "void (double, double, double, double, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsh_X", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsh_X, "void (long, long, __pyx_t_double_complex, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsh_X", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsh_X, "void (long, long, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsh_Y", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsh_Y, "void (long, long, __pyx_t_double_complex, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsh_Y", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsh_Y, "void (long, long, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsh_Z", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsh_Z, "void (long, long, __pyx_t_double_complex, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsh_Z", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsh_Z, "void (long, long, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsw_A", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_A, "void (long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, long, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsw_A", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_A, "void (long, long, __pyx_t_double_complex, double, double, long, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsw_M", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_M, "void (long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsw_M", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_M, "void (long, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsw_N", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_N, "void (long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsw_N", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_N, "void (long, long, __pyx_t_double_complex, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsw_rA", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_rA, "void (long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, long, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsw_rA", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_rA, "void (long, long, double, double, double, long, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsw_rM", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_rM, "void (long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsw_rM", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_rM, "void (long, long, double, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_0vsw_rN", (void (**)(void))&__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_rN, "void (long, long, __pyx_t_double_complex, __pyx_t_double_complex, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction_0_29_36(__pyx_t_1, "__pyx_fuse_1vsw_rN", (void (**)(void))&__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_rN, "void (long, long, double, double, double, __pyx_t_double_complex *, long)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_gufuncs(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_gufuncs(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__gufuncs(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__gufuncs(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__gufuncs(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_gufuncs' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__gufuncs(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_gufuncs", __pyx_methods, __pyx_k_Generalized_universal_functions, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_treams__special___gufuncs) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "treams.special._gufuncs")) {
      if (unlikely(PyDict_SetItemString(modules, "treams.special._gufuncs", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  (void)__Pyx_modinit_type_init_code();
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "treams/special/_gufuncs.pyx":8
 * from treams.special._misc cimport double_complex
 * 
 * __all__ = [             # <<<<<<<<<<<<<<
 *     "car2cyl",
 *     "car2sph",
 */
  __pyx_t_1 = PyList_New(34); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_car2cyl);
  __Pyx_GIVEREF(__pyx_n_u_car2cyl);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_car2cyl);
  __Pyx_INCREF(__pyx_n_u_car2sph);
  __Pyx_GIVEREF(__pyx_n_u_car2sph);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_car2sph);
  __Pyx_INCREF(__pyx_n_u_cyl2car);
  __Pyx_GIVEREF(__pyx_n_u_cyl2car);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_u_cyl2car);
  __Pyx_INCREF(__pyx_n_u_cyl2sph);
  __Pyx_GIVEREF(__pyx_n_u_cyl2sph);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_u_cyl2sph);
  __Pyx_INCREF(__pyx_n_u_sph2car);
  __Pyx_GIVEREF(__pyx_n_u_sph2car);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_sph2car);
  __Pyx_INCREF(__pyx_n_u_sph2cyl);
  __Pyx_GIVEREF(__pyx_n_u_sph2cyl);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_u_sph2cyl);
  __Pyx_INCREF(__pyx_n_u_vcar2cyl);
  __Pyx_GIVEREF(__pyx_n_u_vcar2cyl);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_u_vcar2cyl);
  __Pyx_INCREF(__pyx_n_u_vcar2sph);
  __Pyx_GIVEREF(__pyx_n_u_vcar2sph);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_u_vcar2sph);
  __Pyx_INCREF(__pyx_n_u_vcyl2car);
  __Pyx_GIVEREF(__pyx_n_u_vcyl2car);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_u_vcyl2car);
  __Pyx_INCREF(__pyx_n_u_vcyl2sph);
  __Pyx_GIVEREF(__pyx_n_u_vcyl2sph);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_u_vcyl2sph);
  __Pyx_INCREF(__pyx_n_u_vsph2car);
  __Pyx_GIVEREF(__pyx_n_u_vsph2car);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_u_vsph2car);
  __Pyx_INCREF(__pyx_n_u_vsph2cyl);
  __Pyx_GIVEREF(__pyx_n_u_vsph2cyl);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_u_vsph2cyl);
  __Pyx_INCREF(__pyx_n_u_car2pol);
  __Pyx_GIVEREF(__pyx_n_u_car2pol);
  PyList_SET_ITEM(__pyx_t_1, 12, __pyx_n_u_car2pol);
  __Pyx_INCREF(__pyx_n_u_pol2car);
  __Pyx_GIVEREF(__pyx_n_u_pol2car);
  PyList_SET_ITEM(__pyx_t_1, 13, __pyx_n_u_pol2car);
  __Pyx_INCREF(__pyx_n_u_vcar2pol);
  __Pyx_GIVEREF(__pyx_n_u_vcar2pol);
  PyList_SET_ITEM(__pyx_t_1, 14, __pyx_n_u_vcar2pol);
  __Pyx_INCREF(__pyx_n_u_vpol2car);
  __Pyx_GIVEREF(__pyx_n_u_vpol2car);
  PyList_SET_ITEM(__pyx_t_1, 15, __pyx_n_u_vpol2car);
  __Pyx_INCREF(__pyx_n_u_vsh_X);
  __Pyx_GIVEREF(__pyx_n_u_vsh_X);
  PyList_SET_ITEM(__pyx_t_1, 16, __pyx_n_u_vsh_X);
  __Pyx_INCREF(__pyx_n_u_vsh_Y);
  __Pyx_GIVEREF(__pyx_n_u_vsh_Y);
  PyList_SET_ITEM(__pyx_t_1, 17, __pyx_n_u_vsh_Y);
  __Pyx_INCREF(__pyx_n_u_vsh_Z);
  __Pyx_GIVEREF(__pyx_n_u_vsh_Z);
  PyList_SET_ITEM(__pyx_t_1, 18, __pyx_n_u_vsh_Z);
  __Pyx_INCREF(__pyx_n_u_vsw_M);
  __Pyx_GIVEREF(__pyx_n_u_vsw_M);
  PyList_SET_ITEM(__pyx_t_1, 19, __pyx_n_u_vsw_M);
  __Pyx_INCREF(__pyx_n_u_vsw_N);
  __Pyx_GIVEREF(__pyx_n_u_vsw_N);
  PyList_SET_ITEM(__pyx_t_1, 20, __pyx_n_u_vsw_N);
  __Pyx_INCREF(__pyx_n_u_vsw_A);
  __Pyx_GIVEREF(__pyx_n_u_vsw_A);
  PyList_SET_ITEM(__pyx_t_1, 21, __pyx_n_u_vsw_A);
  __Pyx_INCREF(__pyx_n_u_vsw_rM);
  __Pyx_GIVEREF(__pyx_n_u_vsw_rM);
  PyList_SET_ITEM(__pyx_t_1, 22, __pyx_n_u_vsw_rM);
  __Pyx_INCREF(__pyx_n_u_vsw_rN);
  __Pyx_GIVEREF(__pyx_n_u_vsw_rN);
  PyList_SET_ITEM(__pyx_t_1, 23, __pyx_n_u_vsw_rN);
  __Pyx_INCREF(__pyx_n_u_vsw_rA);
  __Pyx_GIVEREF(__pyx_n_u_vsw_rA);
  PyList_SET_ITEM(__pyx_t_1, 24, __pyx_n_u_vsw_rA);
  __Pyx_INCREF(__pyx_n_u_vcw_M);
  __Pyx_GIVEREF(__pyx_n_u_vcw_M);
  PyList_SET_ITEM(__pyx_t_1, 25, __pyx_n_u_vcw_M);
  __Pyx_INCREF(__pyx_n_u_vcw_N);
  __Pyx_GIVEREF(__pyx_n_u_vcw_N);
  PyList_SET_ITEM(__pyx_t_1, 26, __pyx_n_u_vcw_N);
  __Pyx_INCREF(__pyx_n_u_vcw_A);
  __Pyx_GIVEREF(__pyx_n_u_vcw_A);
  PyList_SET_ITEM(__pyx_t_1, 27, __pyx_n_u_vcw_A);
  __Pyx_INCREF(__pyx_n_u_vcw_rM);
  __Pyx_GIVEREF(__pyx_n_u_vcw_rM);
  PyList_SET_ITEM(__pyx_t_1, 28, __pyx_n_u_vcw_rM);
  __Pyx_INCREF(__pyx_n_u_vcw_rN);
  __Pyx_GIVEREF(__pyx_n_u_vcw_rN);
  PyList_SET_ITEM(__pyx_t_1, 29, __pyx_n_u_vcw_rN);
  __Pyx_INCREF(__pyx_n_u_vcw_rA);
  __Pyx_GIVEREF(__pyx_n_u_vcw_rA);
  PyList_SET_ITEM(__pyx_t_1, 30, __pyx_n_u_vcw_rA);
  __Pyx_INCREF(__pyx_n_u_vpw_M);
  __Pyx_GIVEREF(__pyx_n_u_vpw_M);
  PyList_SET_ITEM(__pyx_t_1, 31, __pyx_n_u_vpw_M);
  __Pyx_INCREF(__pyx_n_u_vpw_N);
  __Pyx_GIVEREF(__pyx_n_u_vpw_N);
  PyList_SET_ITEM(__pyx_t_1, 32, __pyx_n_u_vpw_N);
  __Pyx_INCREF(__pyx_n_u_vpw_A);
  __Pyx_GIVEREF(__pyx_n_u_vpw_A);
  PyList_SET_ITEM(__pyx_t_1, 33, __pyx_n_u_vpw_A);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":472
 * 
 * 
 * np.import_array()             # <<<<<<<<<<<<<<
 * np.import_ufunc()
 * 
 */
  __pyx_t_2 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 472, __pyx_L1_error)

  /* "treams/special/_gufuncs.pyx":473
 * 
 * np.import_array()
 * np.import_ufunc()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_f_5numpy_import_ufunc(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 473, __pyx_L1_error)

  /* "treams/special/_gufuncs.pyx":480
 * cdef char gufunc_car2cyl_types[2]
 * 
 * gufunc_car2cyl_loops[0] = <np.PyUFuncGenericFunction>loop_d_d             # <<<<<<<<<<<<<<
 * gufunc_car2cyl_types[0] = <char>np.NPY_DOUBLE
 * gufunc_car2cyl_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_d_d);

  /* "treams/special/_gufuncs.pyx":481
 * 
 * gufunc_car2cyl_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_car2cyl_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_car2cyl_types[1] = <char>np.NPY_DOUBLE
 * gufunc_car2cyl_data[0] = <void*>_coord.car2cyl
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":482
 * gufunc_car2cyl_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_car2cyl_types[0] = <char>np.NPY_DOUBLE
 * gufunc_car2cyl_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_car2cyl_data[0] = <void*>_coord.car2cyl
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":483
 * gufunc_car2cyl_types[0] = <char>np.NPY_DOUBLE
 * gufunc_car2cyl_types[1] = <char>np.NPY_DOUBLE
 * gufunc_car2cyl_data[0] = <void*>_coord.car2cyl             # <<<<<<<<<<<<<<
 * 
 * car2cyl = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_data[0]) = ((void *)__pyx_f_6treams_7special_6_coord_car2cyl);

  /* "treams/special/_gufuncs.pyx":485
 * gufunc_car2cyl_data[0] = <void*>_coord.car2cyl
 * 
 * car2cyl = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_car2cyl_loops,
 *     gufunc_car2cyl_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_car2cyl_types, 1, 1, 1, 0, ((char *)"car2cyl"), ((char *)"car2cyl(r)\n\nConvert cartesian to cylindrical coordinates\n\nThe cartesian coordinates :math:`(x, y, z)` are converted to\n:math:`(\\rho, \\varphi, z)` with :math:`x = \\rho \\cos \\varphi` and\n:math:`y = \\rho \\sin \\varphi`.\n\nArgs:\n    r (float, 3-array): Cartesian coordinates in 3-dimensional space\n\nReturns:\n    float, 3-array\n"), 0, ((char *)"(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_car2cyl, __pyx_t_1) < 0) __PYX_ERR(0, 485, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":517
 * cdef char gufunc_car2sph_types[2]
 * 
 * gufunc_car2sph_loops[0] = <np.PyUFuncGenericFunction>loop_d_d             # <<<<<<<<<<<<<<
 * gufunc_car2sph_types[0] = <char>np.NPY_DOUBLE
 * gufunc_car2sph_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_d_d);

  /* "treams/special/_gufuncs.pyx":518
 * 
 * gufunc_car2sph_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_car2sph_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_car2sph_types[1] = <char>np.NPY_DOUBLE
 * gufunc_car2sph_data[0] = <void*>_coord.car2sph
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":519
 * gufunc_car2sph_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_car2sph_types[0] = <char>np.NPY_DOUBLE
 * gufunc_car2sph_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_car2sph_data[0] = <void*>_coord.car2sph
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":520
 * gufunc_car2sph_types[0] = <char>np.NPY_DOUBLE
 * gufunc_car2sph_types[1] = <char>np.NPY_DOUBLE
 * gufunc_car2sph_data[0] = <void*>_coord.car2sph             # <<<<<<<<<<<<<<
 * 
 * car2sph = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_data[0]) = ((void *)__pyx_f_6treams_7special_6_coord_car2sph);

  /* "treams/special/_gufuncs.pyx":522
 * gufunc_car2sph_data[0] = <void*>_coord.car2sph
 * 
 * car2sph = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_car2sph_loops,
 *     gufunc_car2sph_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_car2sph_types, 1, 1, 1, 0, ((char *)"car2sph"), ((char *)"car2sph(r)\n\nConvert cartesian to spherical coordinates\n\nThe cartesian coordinates :math:`(x, y, z)` are converted to\n:math:`(r, \\theta, \\varphi)` with :math:`x = r \\sin \\theta \\cos \\varphi`,\n:math:`y = r \\sin \\theta \\sin \\varphi`, and :math:`z = r \\cos \\theta`.\n\nArgs:\n    r (float, 3-array): Cartesian coordinates in 3-dimensional space\n\nReturns:\n    float, 3-array\n"), 0, ((char *)"(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_car2sph, __pyx_t_1) < 0) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":554
 * cdef char gufunc_cyl2car_types[2]
 * 
 * gufunc_cyl2car_loops[0] = <np.PyUFuncGenericFunction>loop_d_d             # <<<<<<<<<<<<<<
 * gufunc_cyl2car_types[0] = <char>np.NPY_DOUBLE
 * gufunc_cyl2car_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_d_d);

  /* "treams/special/_gufuncs.pyx":555
 * 
 * gufunc_cyl2car_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_cyl2car_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_cyl2car_types[1] = <char>np.NPY_DOUBLE
 * gufunc_cyl2car_data[0] = <void*>_coord.cyl2car
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":556
 * gufunc_cyl2car_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_cyl2car_types[0] = <char>np.NPY_DOUBLE
 * gufunc_cyl2car_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_cyl2car_data[0] = <void*>_coord.cyl2car
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":557
 * gufunc_cyl2car_types[0] = <char>np.NPY_DOUBLE
 * gufunc_cyl2car_types[1] = <char>np.NPY_DOUBLE
 * gufunc_cyl2car_data[0] = <void*>_coord.cyl2car             # <<<<<<<<<<<<<<
 * 
 * cyl2car = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_data[0]) = ((void *)__pyx_f_6treams_7special_6_coord_cyl2car);

  /* "treams/special/_gufuncs.pyx":559
 * gufunc_cyl2car_data[0] = <void*>_coord.cyl2car
 * 
 * cyl2car = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_cyl2car_loops,
 *     gufunc_cyl2car_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2car_types, 1, 1, 1, 0, ((char *)"cyl2car"), ((char *)"cyl2car(r)\n\nConvert cylindrical to cartesian coordinates\n\nThe cylindrical coordinates :math:`(\\rho, \\varphi, z)` are converted to\n:math:`(x, y, z)` with :math:`x = \\rho \\cos \\varphi` and\n:math:`y = \\rho \\sin \\varphi`.\n\nArgs:\n    r (float, 3-array): Cylindrical coordinates in 3-dimensional space\n\nReturns:\n    float, 3-array\n"), 0, ((char *)"(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cyl2car, __pyx_t_1) < 0) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":591
 * cdef char gufunc_cyl2sph_types[2]
 * 
 * gufunc_cyl2sph_loops[0] = <np.PyUFuncGenericFunction>loop_d_d             # <<<<<<<<<<<<<<
 * gufunc_cyl2sph_types[0] = <char>np.NPY_DOUBLE
 * gufunc_cyl2sph_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_d_d);

  /* "treams/special/_gufuncs.pyx":592
 * 
 * gufunc_cyl2sph_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_cyl2sph_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_cyl2sph_types[1] = <char>np.NPY_DOUBLE
 * gufunc_cyl2sph_data[0] = <void*>_coord.cyl2sph
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":593
 * gufunc_cyl2sph_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_cyl2sph_types[0] = <char>np.NPY_DOUBLE
 * gufunc_cyl2sph_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_cyl2sph_data[0] = <void*>_coord.cyl2sph
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":594
 * gufunc_cyl2sph_types[0] = <char>np.NPY_DOUBLE
 * gufunc_cyl2sph_types[1] = <char>np.NPY_DOUBLE
 * gufunc_cyl2sph_data[0] = <void*>_coord.cyl2sph             # <<<<<<<<<<<<<<
 * 
 * cyl2sph = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_data[0]) = ((void *)__pyx_f_6treams_7special_6_coord_cyl2sph);

  /* "treams/special/_gufuncs.pyx":596
 * gufunc_cyl2sph_data[0] = <void*>_coord.cyl2sph
 * 
 * cyl2sph = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_cyl2sph_loops,
 *     gufunc_cyl2sph_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_cyl2sph_types, 1, 1, 1, 0, ((char *)"cyl2sph"), ((char *)"cyl2sph(r)\n\nConvert cylindrical to spherical coordinates\n\nThe cylindrical coordinates :math:`(\\rho, \\varphi, z)` are converted to\n:math:`(r, \\theta, \\varphi)` with :math:`\\rho = r \\sin \\theta` and\n:math:`z = \\rho \\cos \\theta`.\n\nArgs:\n    r (float, 3-array): Cylindrical coordinates in 3-dimensional space\n\nReturns:\n    float, 3-array\n"), 0, ((char *)"(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cyl2sph, __pyx_t_1) < 0) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":628
 * cdef char gufunc_sph2car_types[2]
 * 
 * gufunc_sph2car_loops[0] = <np.PyUFuncGenericFunction>loop_d_d             # <<<<<<<<<<<<<<
 * gufunc_sph2car_types[0] = <char>np.NPY_DOUBLE
 * gufunc_sph2car_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_d_d);

  /* "treams/special/_gufuncs.pyx":629
 * 
 * gufunc_sph2car_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_sph2car_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_sph2car_types[1] = <char>np.NPY_DOUBLE
 * gufunc_sph2car_data[0] = <void*>_coord.sph2car
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":630
 * gufunc_sph2car_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_sph2car_types[0] = <char>np.NPY_DOUBLE
 * gufunc_sph2car_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_sph2car_data[0] = <void*>_coord.sph2car
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":631
 * gufunc_sph2car_types[0] = <char>np.NPY_DOUBLE
 * gufunc_sph2car_types[1] = <char>np.NPY_DOUBLE
 * gufunc_sph2car_data[0] = <void*>_coord.sph2car             # <<<<<<<<<<<<<<
 * 
 * sph2car = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_data[0]) = ((void *)__pyx_f_6treams_7special_6_coord_sph2car);

  /* "treams/special/_gufuncs.pyx":633
 * gufunc_sph2car_data[0] = <void*>_coord.sph2car
 * 
 * sph2car = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_sph2car_loops,
 *     gufunc_sph2car_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_sph2car_types, 1, 1, 1, 0, ((char *)"sph2car"), ((char *)"sph2car(r)\n\nConvert spherical to cartesian coordinates\n\nThe spherical coordinates :math:`(r, \\theta, \\varphi)` are converted to\n:math:`(x, y, z)` with :math:`x = r \\sin \\theta \\cos \\varphi`,\n:math:`y = r \\sin \\theta \\sin \\varphi`, and :math:`z = r \\cos \\theta`.\n\nArgs:\n    r (float, 3-array): Spherical coordinates in 3-dimensional space\n\nReturns:\n    float, 3-array\n"), 0, ((char *)"(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sph2car, __pyx_t_1) < 0) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":665
 * cdef char gufunc_sph2cyl_types[2]
 * 
 * gufunc_sph2cyl_loops[0] = <np.PyUFuncGenericFunction>loop_d_d             # <<<<<<<<<<<<<<
 * gufunc_sph2cyl_types[0] = <char>np.NPY_DOUBLE
 * gufunc_sph2cyl_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_d_d);

  /* "treams/special/_gufuncs.pyx":666
 * 
 * gufunc_sph2cyl_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_sph2cyl_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_sph2cyl_types[1] = <char>np.NPY_DOUBLE
 * gufunc_sph2cyl_data[0] = <void*>_coord.sph2cyl
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":667
 * gufunc_sph2cyl_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_sph2cyl_types[0] = <char>np.NPY_DOUBLE
 * gufunc_sph2cyl_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_sph2cyl_data[0] = <void*>_coord.sph2cyl
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":668
 * gufunc_sph2cyl_types[0] = <char>np.NPY_DOUBLE
 * gufunc_sph2cyl_types[1] = <char>np.NPY_DOUBLE
 * gufunc_sph2cyl_data[0] = <void*>_coord.sph2cyl             # <<<<<<<<<<<<<<
 * 
 * sph2cyl = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_data[0]) = ((void *)__pyx_f_6treams_7special_6_coord_sph2cyl);

  /* "treams/special/_gufuncs.pyx":670
 * gufunc_sph2cyl_data[0] = <void*>_coord.sph2cyl
 * 
 * sph2cyl = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_sph2cyl_loops,
 *     gufunc_sph2cyl_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_sph2cyl_types, 1, 1, 1, 0, ((char *)"sph2cyl"), ((char *)"sph2cylr(r)\n\nConvert spherical to cylindrical coordinates\n\nThe spherical coordinates :math:`(r, \\theta, \\varphi)` are converted to\n:math:`(\\rho, \\varphi, z)` with :math:`\\rho = r \\sin \\theta` and\n:math:`z = \\rho \\cos \\theta`.\n\nArgs:\n    r (float, 3-array): Spherical coordinates in 3-dimensional space\n\nReturns:\n    float, 3-array\n"), 0, ((char *)"(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sph2cyl, __pyx_t_1) < 0) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":702
 * cdef char gufunc_car2pol_types[2]
 * 
 * gufunc_car2pol_loops[0] = <np.PyUFuncGenericFunction>loop_d_d             # <<<<<<<<<<<<<<
 * gufunc_car2pol_types[0] = <char>np.NPY_DOUBLE
 * gufunc_car2pol_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_d_d);

  /* "treams/special/_gufuncs.pyx":703
 * 
 * gufunc_car2pol_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_car2pol_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_car2pol_types[1] = <char>np.NPY_DOUBLE
 * gufunc_car2pol_data[0] = <void*>_coord.car2pol
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":704
 * gufunc_car2pol_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_car2pol_types[0] = <char>np.NPY_DOUBLE
 * gufunc_car2pol_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_car2pol_data[0] = <void*>_coord.car2pol
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":705
 * gufunc_car2pol_types[0] = <char>np.NPY_DOUBLE
 * gufunc_car2pol_types[1] = <char>np.NPY_DOUBLE
 * gufunc_car2pol_data[0] = <void*>_coord.car2pol             # <<<<<<<<<<<<<<
 * 
 * car2pol = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_data[0]) = ((void *)__pyx_f_6treams_7special_6_coord_car2pol);

  /* "treams/special/_gufuncs.pyx":707
 * gufunc_car2pol_data[0] = <void*>_coord.car2pol
 * 
 * car2pol = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_car2pol_loops,
 *     gufunc_car2pol_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_car2pol_types, 1, 1, 1, 0, ((char *)"car2pol"), ((char *)"car2pol(r)\n\nConvert cartesian to polar coordinates\n\nThe cartesian coordinates :math:`(x, y)` are converted to\n:math:`(\\rho, \\varphi)` with :math:`x = \\rho \\cos \\varphi` and\n:math:`y = \\rho \\sin \\varphi`.\n\nArgs:\n    r (float, 2-array): Cartesian coordinates in 2-dimensional space\n\nReturns:\n    float, 2-array\n"), 0, ((char *)"(2)->(2)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_car2pol, __pyx_t_1) < 0) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":739
 * cdef char gufunc_pol2car_types[2]
 * 
 * gufunc_pol2car_loops[0] = <np.PyUFuncGenericFunction>loop_d_d             # <<<<<<<<<<<<<<
 * gufunc_pol2car_types[0] = <char>np.NPY_DOUBLE
 * gufunc_pol2car_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_d_d);

  /* "treams/special/_gufuncs.pyx":740
 * 
 * gufunc_pol2car_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_pol2car_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_pol2car_types[1] = <char>np.NPY_DOUBLE
 * gufunc_pol2car_data[0] = <void*>_coord.pol2car
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":741
 * gufunc_pol2car_loops[0] = <np.PyUFuncGenericFunction>loop_d_d
 * gufunc_pol2car_types[0] = <char>np.NPY_DOUBLE
 * gufunc_pol2car_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_pol2car_data[0] = <void*>_coord.pol2car
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":742
 * gufunc_pol2car_types[0] = <char>np.NPY_DOUBLE
 * gufunc_pol2car_types[1] = <char>np.NPY_DOUBLE
 * gufunc_pol2car_data[0] = <void*>_coord.pol2car             # <<<<<<<<<<<<<<
 * 
 * pol2car = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_data[0]) = ((void *)__pyx_f_6treams_7special_6_coord_pol2car);

  /* "treams/special/_gufuncs.pyx":744
 * gufunc_pol2car_data[0] = <void*>_coord.pol2car
 * 
 * pol2car = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_pol2car_loops,
 *     gufunc_pol2car_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_pol2car_types, 1, 1, 1, 0, ((char *)"pol2car"), ((char *)"pol2car(r)\n\nConvert polar to cartesian coordinates\n\nThe polar coordinates :math:`(x, y)` are converted to\n:math:`(\\rho, \\varphi)` with :math:`x = \\rho \\cos \\varphi` and\n:math:`y = \\rho \\sin \\varphi`.\n\nArgs:\n    r (float, 2-array): Polar coordinates in 2-dimensional space\n\nReturns:\n    float, 2-array\n"), 0, ((char *)"(2)->(2)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pol2car, __pyx_t_1) < 0) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":776
 * cdef char gufunc_vcoord_types[2 * 3]
 * 
 * gufunc_vcoord_loops[0] = <np.PyUFuncGenericFunction>loop_d_dd             # <<<<<<<<<<<<<<
 * gufunc_vcoord_loops[1] = <np.PyUFuncGenericFunction>loop_D_Dd
 * gufunc_vcoord_types[0] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_d_dd);

  /* "treams/special/_gufuncs.pyx":777
 * 
 * gufunc_vcoord_loops[0] = <np.PyUFuncGenericFunction>loop_d_dd
 * gufunc_vcoord_loops[1] = <np.PyUFuncGenericFunction>loop_D_Dd             # <<<<<<<<<<<<<<
 * gufunc_vcoord_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_Dd);

  /* "treams/special/_gufuncs.pyx":778
 * gufunc_vcoord_loops[0] = <np.PyUFuncGenericFunction>loop_d_dd
 * gufunc_vcoord_loops[1] = <np.PyUFuncGenericFunction>loop_D_Dd
 * gufunc_vcoord_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcoord_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":779
 * gufunc_vcoord_loops[1] = <np.PyUFuncGenericFunction>loop_D_Dd
 * gufunc_vcoord_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcoord_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[3] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":780
 * gufunc_vcoord_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcoord_types[3] = <char>np.NPY_CDOUBLE
 * gufunc_vcoord_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":781
 * gufunc_vcoord_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[3] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcoord_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types[3]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":782
 * gufunc_vcoord_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[3] = <char>np.NPY_CDOUBLE
 * gufunc_vcoord_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcoord_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcar2cyl_data[0] = <void*>_coord.vcar2cyl[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":783
 * gufunc_vcoord_types[3] = <char>np.NPY_CDOUBLE
 * gufunc_vcoord_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcar2cyl_data[0] = <void*>_coord.vcar2cyl[double]
 * gufunc_vcar2cyl_data[1] = <void*>_coord.vcar2cyl[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":784
 * gufunc_vcoord_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcoord_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcar2cyl_data[0] = <void*>_coord.vcar2cyl[double]             # <<<<<<<<<<<<<<
 * gufunc_vcar2cyl_data[1] = <void*>_coord.vcar2cyl[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2cyl_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcar2cyl);

  /* "treams/special/_gufuncs.pyx":785
 * gufunc_vcoord_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcar2cyl_data[0] = <void*>_coord.vcar2cyl[double]
 * gufunc_vcar2cyl_data[1] = <void*>_coord.vcar2cyl[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vcar2cyl = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2cyl_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcar2cyl);

  /* "treams/special/_gufuncs.pyx":787
 * gufunc_vcar2cyl_data[1] = <void*>_coord.vcar2cyl[double_complex]
 * 
 * vcar2cyl = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcoord_loops,
 *     gufunc_vcar2cyl_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2cyl_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types, 2, 2, 1, 0, ((char *)"vcar2cyl"), ((char *)"vcar2cyl(v, r)\n\nConvert vector in cartesian coordinates to vector in cylindrical coordinates\n\nThe cartesian vector :math:`(v_x, v_y, v_z)`  at coordinates :math:`(x, y, z)` is\nconverted to :math:`(v_\\rho, v_\\varphi, v_z)` with\n\n.. math::\n\n    \\begin{pmatrix} v_\\rho \\\\ v_\\varphi \\\\ v_z \\end{pmatrix}\n    =\n    \\begin{pmatrix}\n        \\cos \\varphi & \\sin \\varphi & 0 \\\\\n        -\\sin \\varphi & \\cos \\varphi & 0 \\\\\n        0 & 0 & 1\n    \\end{pmatrix}\n    \\begin{pmatrix} v_x \\\\ v_y \\\\ v_z \\end{pmatrix}\n\nArgs:\n    v (float or complex, 3-array): Vector in 3-dimensional cartesian space\n    r (float, 3-array): Cartesian coordinates in 3-dimensional space\n\nReturns:\n    float or complex, 3-array\n"), 0, ((char *)"(3),(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcar2cyl, __pyx_t_1) < 0) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":828
 * cdef void *gufunc_vcar2sph_data[2]
 * 
 * gufunc_vcar2sph_data[0] = <void*>_coord.vcar2sph[double]             # <<<<<<<<<<<<<<
 * gufunc_vcar2sph_data[1] = <void*>_coord.vcar2sph[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2sph_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcar2sph);

  /* "treams/special/_gufuncs.pyx":829
 * 
 * gufunc_vcar2sph_data[0] = <void*>_coord.vcar2sph[double]
 * gufunc_vcar2sph_data[1] = <void*>_coord.vcar2sph[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vcar2sph = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2sph_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcar2sph);

  /* "treams/special/_gufuncs.pyx":831
 * gufunc_vcar2sph_data[1] = <void*>_coord.vcar2sph[double_complex]
 * 
 * vcar2sph = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcoord_loops,
 *     gufunc_vcar2sph_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2sph_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types, 2, 2, 1, 0, ((char *)"vcar2sph"), ((char *)"vcar2cyl(v, r)\n\nConvert vector in cartesian coordinates to vector in spherical coordinates\n\nThe cartesian vector :math:`(v_x, v_y, v_z)`  at coordinates :math:`(x, y, z)` is\nconverted to :math:`(v_r, v_\\theta, v_\\varphi)` with\n\n.. math::\n\n    \\begin{pmatrix} v_r \\\\ v_\\theta \\\\ v_\\varphi \\end{pmatrix}\n    =\n    \\begin{pmatrix}\n        \\sin \\theta \\cos \\varphi & \\sin \\theta \\sin \\varphi & \\cos \\theta \\\\\n        \\cos \\theta \\cos \\varphi & \\cos \\theta \\sin \\varphi & \\cos \\theta \\\\\n        -\\sin \\varphi & \\cos \\varphi & 0\n    \\end{pmatrix}\n    \\begin{pmatrix} v_x \\\\ v_y \\\\ v_z \\end{pmatrix}\n\nArgs:\n    v (float or complex, 3-array): Vector in 3-dimensional cartesian space\n    r (float, 3-array): Cartesian coordinates in 3-dimensional space\n\nReturns:\n    float or complex, 3-array\n"), 0, ((char *)"(3),(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcar2sph, __pyx_t_1) < 0) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":871
 * 
 * cdef void *gufunc_vcyl2car_data[2]
 * gufunc_vcyl2car_data[0] = <void*>_coord.vcyl2car[double]             # <<<<<<<<<<<<<<
 * gufunc_vcyl2car_data[1] = <void*>_coord.vcyl2car[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcyl2car_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcyl2car);

  /* "treams/special/_gufuncs.pyx":872
 * cdef void *gufunc_vcyl2car_data[2]
 * gufunc_vcyl2car_data[0] = <void*>_coord.vcyl2car[double]
 * gufunc_vcyl2car_data[1] = <void*>_coord.vcyl2car[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vcyl2car = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcyl2car_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcyl2car);

  /* "treams/special/_gufuncs.pyx":874
 * gufunc_vcyl2car_data[1] = <void*>_coord.vcyl2car[double_complex]
 * 
 * vcyl2car = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcoord_loops,
 *     gufunc_vcyl2car_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcyl2car_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types, 2, 2, 1, 0, ((char *)"vcyl2car"), ((char *)"vcyl2car(v, r)\n\nConvert vector in cylindrical coordinates to vector in cartesian coordinates\n\nThe cylindrical vector :math:`(v_\\rho, v_\\varphi, v_z)`  at coordinates\n:math:`(\\rho, \\varphi, z)` is converted to :math:`(v_x, v_y, v_z)` with\n\n.. math::\n\n    \\begin{pmatrix} v_x \\\\ v_y \\\\ v_z \\end{pmatrix}\n    =\n    \\begin{pmatrix}\n        \\cos \\varphi & -\\sin \\varphi & 0 \\\\\n        \\sin \\varphi & \\cos \\varphi & 0 \\\\\n        0 & 0 & 1\n    \\end{pmatrix}\n    \\begin{pmatrix} v_\\rho \\\\ v_\\varphi \\\\ v_z \\end{pmatrix}\n\nArgs:\n    v (float or complex, 3-array): Vector in 3-dimensional cylindrical space\n    r (float, 3-array): Cylindrical coordinates in 3-dimensional space\n\nReturns:\n    float or complex, 3-array\n"), 0, ((char *)"(3),(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 874, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcyl2car, __pyx_t_1) < 0) __PYX_ERR(0, 874, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":914
 * 
 * cdef void *gufunc_vcyl2sph_data[2]
 * gufunc_vcyl2sph_data[0] = <void*>_coord.vcyl2sph[double]             # <<<<<<<<<<<<<<
 * gufunc_vcyl2sph_data[1] = <void*>_coord.vcyl2sph[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcyl2sph_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcyl2sph);

  /* "treams/special/_gufuncs.pyx":915
 * cdef void *gufunc_vcyl2sph_data[2]
 * gufunc_vcyl2sph_data[0] = <void*>_coord.vcyl2sph[double]
 * gufunc_vcyl2sph_data[1] = <void*>_coord.vcyl2sph[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vcyl2sph = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcyl2sph_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcyl2sph);

  /* "treams/special/_gufuncs.pyx":917
 * gufunc_vcyl2sph_data[1] = <void*>_coord.vcyl2sph[double_complex]
 * 
 * vcyl2sph = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcoord_loops,
 *     gufunc_vcyl2sph_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcyl2sph_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types, 2, 2, 1, 0, ((char *)"vcyl2sph"), ((char *)"vcyl2sph(v, r)\n\nConvert vector in cylindrical coordinates to vector in spherical coordinates\n\nThe cylindrical vector :math:`(v_\\rho, v_\\varphi, v_z)`  at coordinates\n:math:`(\\rho, \\varphi, z)` is converted to :math:`(v_r, v_\\theta, v_\\varphi)` with\n\n.. math::\n\n    \\begin{pmatrix} v_r \\\\ v_\\theta \\\\ v_\\varphi \\end{pmatrix}\n    =\n    \\begin{pmatrix}\n        \\sin \\theta & 0 & \\cos \\theta \\\\\n        \\cos \\theta & 0 & -\\sin \\theta \\\\\n        0 & 1 & 0\n    \\end{pmatrix}\n    \\begin{pmatrix} v_\\rho \\\\ v_\\varphi \\\\ v_z \\end{pmatrix}\n\nArgs:\n    v (float or complex, 3-array): Vector in 3-dimensional cylindrical space\n    r (float, 3-array): Cylindrical coordinates in 3-dimensional space\n\nReturns:\n    float or complex, 3-array\n"), 0, ((char *)"(3),(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcyl2sph, __pyx_t_1) < 0) __PYX_ERR(0, 917, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":957
 * 
 * cdef void *gufunc_vsph2car_data[2]
 * gufunc_vsph2car_data[0] = <void*>_coord.vsph2car[double]             # <<<<<<<<<<<<<<
 * gufunc_vsph2car_data[1] = <void*>_coord.vsph2car[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsph2car_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vsph2car);

  /* "treams/special/_gufuncs.pyx":958
 * cdef void *gufunc_vsph2car_data[2]
 * gufunc_vsph2car_data[0] = <void*>_coord.vsph2car[double]
 * gufunc_vsph2car_data[1] = <void*>_coord.vsph2car[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsph2car = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsph2car_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vsph2car);

  /* "treams/special/_gufuncs.pyx":960
 * gufunc_vsph2car_data[1] = <void*>_coord.vsph2car[double_complex]
 * 
 * vsph2car = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcoord_loops,
 *     gufunc_vsph2car_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsph2car_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types, 2, 2, 1, 0, ((char *)"vsph2car"), ((char *)"vsph2car(v, r)\n\nConvert vector in spherical coordinates to vector in cartesian coordinates\n\nThe spherical vector :math:`(v_r, v_\\theta, v_\\varphi)`  at coordinates\n:math:`(r, \\theta,  \\varphi)` is converted to :math:`(v_x, v_y, v_z)` with\n\n.. math::\n\n    \\begin{pmatrix} v_x \\\\ v_y \\\\ v_z \\end{pmatrix}\n    =\n    \\begin{pmatrix}\n        \\sin \\theta \\cos \\varphi & \\cos \\theta \\cos \\varphi & -\\sin \\varphi \\\\\n        \\sin \\theta \\sin \\varphi & \\cos \\theta \\sin \\varphi & \\cos \\varphi \\\\\n        \\cos \\theta & -\\sin \\theta & 0\n    \\end{pmatrix}\n    \\begin{pmatrix} v_r \\\\ v_\\theta \\\\ v_\\varphi \\end{pmatrix}\n\nArgs:\n    v (float or complex, 3-array): Vector in 3-dimensional spherical space\n    r (float, 3-array): Spherical coordinates in 3-dimensional space\n\nReturns:\n    float or complex, 3-array\n"), 0, ((char *)"(3),(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsph2car, __pyx_t_1) < 0) __PYX_ERR(0, 960, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1000
 * 
 * cdef void *gufunc_vsph2cyl_data[2]
 * gufunc_vsph2cyl_data[0] = <void*>_coord.vsph2cyl[double]             # <<<<<<<<<<<<<<
 * gufunc_vsph2cyl_data[1] = <void*>_coord.vsph2cyl[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsph2cyl_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vsph2cyl);

  /* "treams/special/_gufuncs.pyx":1001
 * cdef void *gufunc_vsph2cyl_data[2]
 * gufunc_vsph2cyl_data[0] = <void*>_coord.vsph2cyl[double]
 * gufunc_vsph2cyl_data[1] = <void*>_coord.vsph2cyl[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsph2cyl = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsph2cyl_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vsph2cyl);

  /* "treams/special/_gufuncs.pyx":1003
 * gufunc_vsph2cyl_data[1] = <void*>_coord.vsph2cyl[double_complex]
 * 
 * vsph2cyl = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcoord_loops,
 *     gufunc_vsph2cyl_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsph2cyl_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types, 2, 2, 1, 0, ((char *)"vsph2cyl"), ((char *)"vsph2cyl(v, r)\n\nConvert vector in spherical coordinates to vector in cylindrical coordinates\n\nThe spherical vector :math:`(v_r, v_\\theta, v_\\varphi)`  at coordinates\n:math:`(r, \\theta,  \\varphi)` is converted to :math:`(v_\\rho, v_\\varphi, v_z)` with\n\n.. math::\n\n    \\begin{pmatrix} v_\\rho \\\\ v_\\varphi \\\\ v_z \\end{pmatrix}\n    =\n    \\begin{pmatrix}\n        \\sin \\theta & \\cos \\theta & 0 \\\\\n        0 & 0 & 1 \\\\\n        \\cos \\theta & -\\sin \\theta & 0\n    \\end{pmatrix}\n    \\begin{pmatrix} v_r \\\\ v_\\theta \\\\ v_\\varphi \\end{pmatrix}\n\nArgs:\n    v (float or complex, 3-array): Vector in 3-dimensional spherical space\n    r (float, 3-array): Spherical coordinates in 3-dimensional space\n\nReturns:\n    float or complex, 3-array\n"), 0, ((char *)"(3),(3)->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsph2cyl, __pyx_t_1) < 0) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1043
 * 
 * cdef void *gufunc_vcar2pol_data[2]
 * gufunc_vcar2pol_data[0] = <void*>_coord.vcar2pol[double]             # <<<<<<<<<<<<<<
 * gufunc_vcar2pol_data[1] = <void*>_coord.vcar2pol[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2pol_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vcar2pol);

  /* "treams/special/_gufuncs.pyx":1044
 * cdef void *gufunc_vcar2pol_data[2]
 * gufunc_vcar2pol_data[0] = <void*>_coord.vcar2pol[double]
 * gufunc_vcar2pol_data[1] = <void*>_coord.vcar2pol[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vcar2pol = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2pol_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vcar2pol);

  /* "treams/special/_gufuncs.pyx":1046
 * gufunc_vcar2pol_data[1] = <void*>_coord.vcar2pol[double_complex]
 * 
 * vcar2pol = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcoord_loops,
 *     gufunc_vcar2pol_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcar2pol_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types, 2, 2, 1, 0, ((char *)"vcar2pol"), ((char *)"vcar2pol(v, r)\n\nConvert vector in cartesian coordinates to vector in polar coordinates\n\nThe cartesian vector :math:`(v_x, v_y)`  at coordinates :math:`(x, y)` is\nconverted to :math:`(v_\\rho, v_\\varphi)` with\n\n.. math::\n\n    \\begin{pmatrix} v_\\rho \\\\ v_\\varphi \\end{pmatrix}\n    =\n    \\begin{pmatrix}\n        \\cos \\varphi & \\sin \\varphi \\\\\n        -\\sin \\varphi & \\cos \\varphi\n    \\end{pmatrix}\n    \\begin{pmatrix} v_x \\\\ v_y \\end{pmatrix}\n\nArgs:\n    v (float or complex, 3-array): Vector in 2-dimensional cartesian space\n    r (float, 2-array): Cartesian coordinates in 2-dimensional space\n\nReturns:\n    float or complex, 2-array\n"), 0, ((char *)"(2),(2)->(2)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcar2pol, __pyx_t_1) < 0) __PYX_ERR(0, 1046, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1085
 * 
 * cdef void *gufunc_vpol2car_data[2]
 * gufunc_vpol2car_data[0] = <void*>_coord.vpol2car[double]             # <<<<<<<<<<<<<<
 * gufunc_vpol2car_data[1] = <void*>_coord.vpol2car[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpol2car_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_coord_vpol2car);

  /* "treams/special/_gufuncs.pyx":1086
 * cdef void *gufunc_vpol2car_data[2]
 * gufunc_vpol2car_data[0] = <void*>_coord.vpol2car[double]
 * gufunc_vpol2car_data[1] = <void*>_coord.vpol2car[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vpol2car = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpol2car_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_coord_vpol2car);

  /* "treams/special/_gufuncs.pyx":1088
 * gufunc_vpol2car_data[1] = <void*>_coord.vpol2car[double_complex]
 * 
 * vpol2car = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcoord_loops,
 *     gufunc_vpol2car_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vpol2car_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcoord_types, 2, 2, 1, 0, ((char *)"vpol2car"), ((char *)"vpol2car(v, r)\n\nConvert vector in polar coordinates to vector in cartesian coordinates\n\nThe polar vector :math:`(v_\\rho, v_\\varphi)`  at coordinates\n:math:`(\\rho, \\varphi)` is converted to :math:`(v_x, v_y)` with\n\n.. math::\n\n    \\begin{pmatrix} v_x \\\\ v_y \\end{pmatrix}\n    =\n    \\begin{pmatrix}\n        \\cos \\varphi & -\\sin \\varphi \\\\\n        \\sin \\varphi & \\cos \\varphi\n    \\end{pmatrix}\n    \\begin{pmatrix} v_\\rho \\\\ v_\\varphi \\end{pmatrix}\n\nArgs:\n    v (float or complex, 3-array): Vector in 3-dimensional cylindrical space\n    r (float, 3-array): Cylindrical coordinates in 3-dimensional space\n\nReturns:\n    float or complex, 3-array\n"), 0, ((char *)"(2),(2)->(2)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vpol2car, __pyx_t_1) < 0) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1130
 * cdef char gufunc_vsh_X_types[2 * 5]
 * 
 * gufunc_vsh_X_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldd             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd
 * gufunc_vsh_X_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_lldd);

  /* "treams/special/_gufuncs.pyx":1131
 * 
 * gufunc_vsh_X_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldd
 * gufunc_vsh_X_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[0] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDd);

  /* "treams/special/_gufuncs.pyx":1132
 * gufunc_vsh_X_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldd
 * gufunc_vsh_X_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd
 * gufunc_vsh_X_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[1] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[0]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1133
 * gufunc_vsh_X_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd
 * gufunc_vsh_X_types[0] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsh_X_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1134
 * gufunc_vsh_X_types[0] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[1] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsh_X_types[4] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1135
 * gufunc_vsh_X_types[1] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsh_X_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[4] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_X_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1136
 * gufunc_vsh_X_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsh_X_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsh_X_types[4] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[5] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[4]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1137
 * gufunc_vsh_X_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsh_X_types[4] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_X_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[6] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[5]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1138
 * gufunc_vsh_X_types[4] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_X_types[5] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_X_types[8] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1139
 * gufunc_vsh_X_types[5] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[6] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vsh_X_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1140
 * gufunc_vsh_X_types[6] = <char>np.NPY_LONG
 * gufunc_vsh_X_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_X_types[8] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_X_data[0] = <void*>_waves.vsh_X[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[8]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1141
 * gufunc_vsh_X_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_X_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vsh_X_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_data[0] = <void*>_waves.vsh_X[double]
 * gufunc_vsh_X_data[1] = <void*>_waves.vsh_X[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1142
 * gufunc_vsh_X_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vsh_X_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_X_data[0] = <void*>_waves.vsh_X[double]             # <<<<<<<<<<<<<<
 * gufunc_vsh_X_data[1] = <void*>_waves.vsh_X[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsh_X);

  /* "treams/special/_gufuncs.pyx":1143
 * gufunc_vsh_X_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_X_data[0] = <void*>_waves.vsh_X[double]
 * gufunc_vsh_X_data[1] = <void*>_waves.vsh_X[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsh_X = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsh_X);

  /* "treams/special/_gufuncs.pyx":1145
 * gufunc_vsh_X_data[1] = <void*>_waves.vsh_X[double_complex]
 * 
 * vsh_X = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vsh_X_loops,
 *     gufunc_vsh_X_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_X_types, 2, 4, 1, 0, ((char *)"vsh_X"), ((char *)"vsh_X(l, m, theta, phi)\n\nVector spherical harmonic X in spherical coordinates\n\nThe vector spherical harmonics can be defined via the (scalar) spherical harmonics\n(:func:`treams.special.sph_harm`)\n\n.. math::\n\n    \\boldsymbol X_{lm}(\\theta, \\varphi)\n    = \\frac{1}{\\sqrt{l (l + 1)}} \\nabla Y_{lm}(\\theta, \\varphi)\n\nwhich can be expressed as\n\n.. math::\n\n    \\boldsymbol X_{lm}(\\theta, \\varphi)\n    = \\mathrm i \\sqrt{\\frac{2 l + 1}{4 \\pi l (l + 1)} \\frac{(l - m)!}{(l + m)!}}\n    \\left(\\mathrm i \\pi_{lm}(\\theta, \\varphi) \\boldsymbol{\\hat\\theta}\n    - \\tau_{lm}(\\theta, \\varphi) \\boldsymbol{\\hat\\varphi}\\right)\n\nwith the angular functions :func:`treams.special.pi_fun` and\n:func:`treams.special.tau_fun`.\n\nArgs:\n    l (int, array_like): Degree :math:`l \\geq 0`\n    m (int, array_like): Order :math:`|m| \\leq l`\n    theta (float or complex, array_like): Polar angle\n    phi (float, array_like): Azimuthal angle\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsh_X, __pyx_t_1) < 0) __PYX_ERR(0, 1145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1196
 * cdef char gufunc_vsh_Y_types[2 * 5]
 * 
 * gufunc_vsh_Y_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldd             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd
 * gufunc_vsh_Y_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_lldd);

  /* "treams/special/_gufuncs.pyx":1197
 * 
 * gufunc_vsh_Y_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldd
 * gufunc_vsh_Y_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[0] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDd);

  /* "treams/special/_gufuncs.pyx":1198
 * gufunc_vsh_Y_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldd
 * gufunc_vsh_Y_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd
 * gufunc_vsh_Y_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[1] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[0]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1199
 * gufunc_vsh_Y_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd
 * gufunc_vsh_Y_types[0] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Y_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1200
 * gufunc_vsh_Y_types[0] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[1] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Y_types[4] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1201
 * gufunc_vsh_Y_types[1] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Y_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[4] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Y_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1202
 * gufunc_vsh_Y_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Y_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Y_types[4] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[5] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[4]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1203
 * gufunc_vsh_Y_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Y_types[4] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Y_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[6] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[5]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1204
 * gufunc_vsh_Y_types[4] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Y_types[5] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Y_types[8] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1205
 * gufunc_vsh_Y_types[5] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[6] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Y_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1206
 * gufunc_vsh_Y_types[6] = <char>np.NPY_LONG
 * gufunc_vsh_Y_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Y_types[8] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Y_data[0] = <void*>_waves.vsh_Y[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[8]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1207
 * gufunc_vsh_Y_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Y_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Y_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_data[0] = <void*>_waves.vsh_Y[double]
 * gufunc_vsh_Y_data[1] = <void*>_waves.vsh_Y[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1208
 * gufunc_vsh_Y_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Y_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Y_data[0] = <void*>_waves.vsh_Y[double]             # <<<<<<<<<<<<<<
 * gufunc_vsh_Y_data[1] = <void*>_waves.vsh_Y[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsh_Y);

  /* "treams/special/_gufuncs.pyx":1209
 * gufunc_vsh_Y_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Y_data[0] = <void*>_waves.vsh_Y[double]
 * gufunc_vsh_Y_data[1] = <void*>_waves.vsh_Y[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsh_Y = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsh_Y);

  /* "treams/special/_gufuncs.pyx":1211
 * gufunc_vsh_Y_data[1] = <void*>_waves.vsh_Y[double_complex]
 * 
 * vsh_Y = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vsh_Y_loops,
 *     gufunc_vsh_Y_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Y_types, 2, 4, 1, 0, ((char *)"vsh_Y"), ((char *)"vsh_Y(l, m, theta, phi)\n\nVector spherical harmonic Y in spherical coordinates\n\nThe vector spherical harmonics can be defined by\n:math:`\\boldsymbol Y_{lm}(\\theta, \\varphi) = \\boldsymbol{\\hat r} \\times\n\\boldsymbol X_{lm}(\\theta, \\varphi)` using :func:`treams.special.vsh_X`. Alternatively,\nit can can be expressed as\n\n.. math::\n\n    \\boldsymbol Y_{lm}(\\theta, \\varphi)\n    = \\mathrm i \\sqrt{\\frac{2 l + 1}{4 \\pi l (l + 1)} \\frac{(l - m)!}{(l + m)!}}\n    \\left(\\tau_{lm}(\\theta, \\varphi) \\boldsymbol{\\hat\\theta}\n    + \\mathrm i \\pi_{lm}(\\theta, \\varphi) \\boldsymbol{\\hat\\varphi}\\right)\n\nwith the angular functions :func:`treams.special.pi_fun` and\n:func:`treams.special.tau_fun`.\n\nArgs:\n    l (int, array_like): Degree :math:`l \\geq 0`\n    m (int, array_like): Order :math:`|m| \\leq l`\n    theta (float or complex, array_like): Polar angle\n    phi (float, array_like): Azimuthal angle\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsh_Y, __pyx_t_1) < 0) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1257
 * cdef char gufunc_vsh_Z_types[2 * 5]
 * 
 * gufunc_vsh_Z_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldd             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd
 * gufunc_vsh_Z_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_lldd);

  /* "treams/special/_gufuncs.pyx":1258
 * 
 * gufunc_vsh_Z_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldd
 * gufunc_vsh_Z_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[0] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDd);

  /* "treams/special/_gufuncs.pyx":1259
 * gufunc_vsh_Z_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldd
 * gufunc_vsh_Z_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd
 * gufunc_vsh_Z_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[1] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[0]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1260
 * gufunc_vsh_Z_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDd
 * gufunc_vsh_Z_types[0] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Z_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1261
 * gufunc_vsh_Z_types[0] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[1] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Z_types[4] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1262
 * gufunc_vsh_Z_types[1] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Z_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[4] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Z_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1263
 * gufunc_vsh_Z_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Z_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Z_types[4] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[5] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[4]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1264
 * gufunc_vsh_Z_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Z_types[4] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Z_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[6] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[5]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1265
 * gufunc_vsh_Z_types[4] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Z_types[5] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Z_types[8] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1266
 * gufunc_vsh_Z_types[5] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[6] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Z_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1267
 * gufunc_vsh_Z_types[6] = <char>np.NPY_LONG
 * gufunc_vsh_Z_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Z_types[8] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Z_data[0] = <void*>_waves.vsh_Z[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[8]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1268
 * gufunc_vsh_Z_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Z_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Z_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_data[0] = <void*>_waves.vsh_Z[double]
 * gufunc_vsh_Z_data[1] = <void*>_waves.vsh_Z[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1269
 * gufunc_vsh_Z_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vsh_Z_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Z_data[0] = <void*>_waves.vsh_Z[double]             # <<<<<<<<<<<<<<
 * gufunc_vsh_Z_data[1] = <void*>_waves.vsh_Z[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsh_Z);

  /* "treams/special/_gufuncs.pyx":1270
 * gufunc_vsh_Z_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsh_Z_data[0] = <void*>_waves.vsh_Z[double]
 * gufunc_vsh_Z_data[1] = <void*>_waves.vsh_Z[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsh_Z = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsh_Z);

  /* "treams/special/_gufuncs.pyx":1272
 * gufunc_vsh_Z_data[1] = <void*>_waves.vsh_Z[double_complex]
 * 
 * vsh_Z = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vsh_Z_loops,
 *     gufunc_vsh_Z_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsh_Z_types, 2, 4, 1, 0, ((char *)"vsh_Z"), ((char *)"vsh_Z(l, m, theta, phi)\n\nVector spherical harmonic Z in spherical coordinates\n\nThe vector spherical harmonics can be defined via the (scalar) spherical harmonics\n(:func:`treams.special.sph_harm`) by :math:`\\boldsymbol Z_{lm}(\\theta, \\varphi) =\n\\mathrm i Y_{lm}(\\theta, \\varphi) \\boldsymbol{\\hat r}`.\n\nArgs:\n    l (int, array_like): Degree :math:`l \\geq 0`\n    m (int, array_like): Order :math:`|m| \\leq l`\n    theta (float or complex, array_like): Polar angle\n    phi (float, array_like): Azimuthal angle\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsh_Z, __pyx_t_1) < 0) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1307
 * cdef char gufunc_vsw_M_types[2 * 6]
 * 
 * gufunc_vsw_M_loops[0] = <np.PyUFuncGenericFunction>loop_D_llDdd             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_M_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDdd);

  /* "treams/special/_gufuncs.pyx":1308
 * 
 * gufunc_vsw_M_loops[0] = <np.PyUFuncGenericFunction>loop_D_llDdd
 * gufunc_vsw_M_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDDd);

  /* "treams/special/_gufuncs.pyx":1309
 * gufunc_vsw_M_loops[0] = <np.PyUFuncGenericFunction>loop_D_llDdd
 * gufunc_vsw_M_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_M_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[2] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[0]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1310
 * gufunc_vsw_M_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_M_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1311
 * gufunc_vsw_M_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[2] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_M_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[2]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1312
 * gufunc_vsw_M_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_M_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1313
 * gufunc_vsw_M_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_M_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1314
 * gufunc_vsw_M_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_M_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_M_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[7] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1315
 * gufunc_vsw_M_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_M_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1316
 * gufunc_vsw_M_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[7] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[7]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1317
 * gufunc_vsw_M_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[10] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1318
 * gufunc_vsw_M_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_M_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_M_types[11] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1319
 * gufunc_vsw_M_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[10] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_data[0] = <void*>_waves.vsw_M[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[10]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1320
 * gufunc_vsw_M_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_M_types[11] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_data[0] = <void*>_waves.vsw_M[double]
 * gufunc_vsw_M_data[1] = <void*>_waves.vsw_M[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types[11]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1321
 * gufunc_vsw_M_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_M_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_data[0] = <void*>_waves.vsw_M[double]             # <<<<<<<<<<<<<<
 * gufunc_vsw_M_data[1] = <void*>_waves.vsw_M[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_M);

  /* "treams/special/_gufuncs.pyx":1322
 * gufunc_vsw_M_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_M_data[0] = <void*>_waves.vsw_M[double]
 * gufunc_vsw_M_data[1] = <void*>_waves.vsw_M[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsw_M = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_M);

  /* "treams/special/_gufuncs.pyx":1324
 * gufunc_vsw_M_data[1] = <void*>_waves.vsw_M[double_complex]
 * 
 * vsw_M = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vsw_M_loops,
 *     gufunc_vsw_M_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_M_types, 2, 5, 1, 0, ((char *)"vsw_M"), ((char *)"vsw_M(l, m, x, theta, phi)\n\nSingular vector spherical wave M\n\nThe vector spherical wave is defined by :math:`\\boldsymbol M_{lm}^{(3)}\n(x, \\theta, \\varphi) = h_l^{(1)} (x) \\boldsymbol X_{lm}(\\theta, \\varphi)` using\n:func:`treams.special.spherical_hankel1` and :func:`treams.special.vsh_X`.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation, that is also tangential on a spherical surface. This is often used to\ndescribe a transverse electric (in spherical coordinates) (TE) wave. Additionally,\nthe term magnetic (refferring to the multipole) is used for this solution.\n\nArgs:\n    l (int, array_like): Degree :math:`l \\geq 0`\n    m (int, array_like): Order :math:`|m| \\leq l`\n    x (float or complex, array_like): Distance in units of the wave number :math:`kr`\n    theta (float or complex, array_like): Polar angle\n    phi (float, array_like): Azimuthal angle\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsw_M, __pyx_t_1) < 0) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1365
 * cdef char gufunc_vsw_N_types[2 * 6]
 * 
 * gufunc_vsw_N_loops[0] = <np.PyUFuncGenericFunction>loop_D_llDdd             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_N_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDdd);

  /* "treams/special/_gufuncs.pyx":1366
 * 
 * gufunc_vsw_N_loops[0] = <np.PyUFuncGenericFunction>loop_D_llDdd
 * gufunc_vsw_N_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDDd);

  /* "treams/special/_gufuncs.pyx":1367
 * gufunc_vsw_N_loops[0] = <np.PyUFuncGenericFunction>loop_D_llDdd
 * gufunc_vsw_N_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_N_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[2] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[0]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1368
 * gufunc_vsw_N_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_N_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1369
 * gufunc_vsw_N_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[2] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_N_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[2]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1370
 * gufunc_vsw_N_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_N_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1371
 * gufunc_vsw_N_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_N_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1372
 * gufunc_vsw_N_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_N_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_N_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[7] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1373
 * gufunc_vsw_N_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_N_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1374
 * gufunc_vsw_N_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[7] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[7]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1375
 * gufunc_vsw_N_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[10] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1376
 * gufunc_vsw_N_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_N_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_N_types[11] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1377
 * gufunc_vsw_N_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[10] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_data[0] = <void*>_waves.vsw_N[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[10]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1378
 * gufunc_vsw_N_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_N_types[11] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_data[0] = <void*>_waves.vsw_N[double]
 * gufunc_vsw_N_data[1] = <void*>_waves.vsw_N[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types[11]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1379
 * gufunc_vsw_N_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_N_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_data[0] = <void*>_waves.vsw_N[double]             # <<<<<<<<<<<<<<
 * gufunc_vsw_N_data[1] = <void*>_waves.vsw_N[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_N);

  /* "treams/special/_gufuncs.pyx":1380
 * gufunc_vsw_N_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_N_data[0] = <void*>_waves.vsw_N[double]
 * gufunc_vsw_N_data[1] = <void*>_waves.vsw_N[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsw_N = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_N);

  /* "treams/special/_gufuncs.pyx":1382
 * gufunc_vsw_N_data[1] = <void*>_waves.vsw_N[double_complex]
 * 
 * vsw_N = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vsw_N_loops,
 *     gufunc_vsw_N_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_N_types, 2, 5, 1, 0, ((char *)"vsw_N"), ((char *)"vsw_N(l, m, x, theta, phi)\n\nSingular vector spherical wave N\n\nThe vector spherical wave is defined by\n\n.. math::\n\n    \\boldsymbol N_{lm}^{(3)}(x, \\theta, \\varphi)\n    = \\nabla \\times \\boldsymbol M_{lm}^{(3)}(x, \\theta, \\varphi) \\\\\n    = \\left({h_l^{(1)}}'(x)\n    + \\frac{h_l^{(1)}(x)}{x}\\right) \\boldsymbol Y_{lm}(\\theta, \\varphi)\n    + \\sqrt{l (l + 1)} \\frac{h_l^{(1)}(x)}{x} \\boldsymbol Z_{lm}(\\theta, \\varphi)\n\nwith :func:`treams.special.vsw_M`, :func:`treams.special.vsh_Y`,\n:func:`treams.special.vsh_Z`, and :func:`treams.special.spherical_hankel1`.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. This is often used to describe a transverse magnetic (in spherical\ncoordinates) (TM) wave. Additionally, the term electric (refferring to the\nmultipole) is used for this solution.\n\nArgs:\n    l (int, array_like): Degree :math:`l \\geq 0`\n    m (int, array_like): Order :math:`|m| \\leq l`\n    x (float or complex, array_like): Distance in units of the wave number :math:`kr`\n    theta (float or complex, array_like): Polar angle\n    phi (float, array_like): Azimuthal angle\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1382, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsw_N, __pyx_t_1) < 0) __PYX_ERR(0, 1382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1432
 * cdef char gufunc_vsw_rM_types[2 * 6]
 * 
 * gufunc_vsw_rM_loops[0] = <np.PyUFuncGenericFunction>loop_D_llddd             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_rM_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llddd);

  /* "treams/special/_gufuncs.pyx":1433
 * 
 * gufunc_vsw_rM_loops[0] = <np.PyUFuncGenericFunction>loop_D_llddd
 * gufunc_vsw_rM_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDDd);

  /* "treams/special/_gufuncs.pyx":1434
 * gufunc_vsw_rM_loops[0] = <np.PyUFuncGenericFunction>loop_D_llddd
 * gufunc_vsw_rM_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_rM_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[0]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1435
 * gufunc_vsw_rM_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_rM_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1436
 * gufunc_vsw_rM_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1437
 * gufunc_vsw_rM_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1438
 * gufunc_vsw_rM_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1439
 * gufunc_vsw_rM_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[7] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1440
 * gufunc_vsw_rM_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1441
 * gufunc_vsw_rM_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[7] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[7]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1442
 * gufunc_vsw_rM_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_types[10] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1443
 * gufunc_vsw_rM_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_rM_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[11] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1444
 * gufunc_vsw_rM_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_types[10] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_data[0] = <void*>_waves.vsw_rM[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[10]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1445
 * gufunc_vsw_rM_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[11] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_data[0] = <void*>_waves.vsw_rM[double]
 * gufunc_vsw_rM_data[1] = <void*>_waves.vsw_rM[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types[11]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1446
 * gufunc_vsw_rM_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rM_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_data[0] = <void*>_waves.vsw_rM[double]             # <<<<<<<<<<<<<<
 * gufunc_vsw_rM_data[1] = <void*>_waves.vsw_rM[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_rM);

  /* "treams/special/_gufuncs.pyx":1447
 * gufunc_vsw_rM_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rM_data[0] = <void*>_waves.vsw_rM[double]
 * gufunc_vsw_rM_data[1] = <void*>_waves.vsw_rM[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsw_rM = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_rM);

  /* "treams/special/_gufuncs.pyx":1449
 * gufunc_vsw_rM_data[1] = <void*>_waves.vsw_rM[double_complex]
 * 
 * vsw_rM = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vsw_rM_loops,
 *     gufunc_vsw_rM_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rM_types, 2, 5, 1, 0, ((char *)"vsw_rM"), ((char *)"vsw_rM(l, m, x, theta, phi)\n\nRegular vector spherical wave M\n\nThe vector spherical wave is defined by\n:math:`\\boldsymbol M_{lm}^{(1)}(x, \\theta, \\varphi) = j_l(x)\n\\boldsymbol X_{lm}(\\theta, \\varphi)` using :func:`treams.special.spherical_jn` and\n:func:`treams.special.vsh_X`.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation, that is also tangential on a spherical surface. This is often used to\ndescribe a transverse electric (in spherical coordinates) (TE) wave. Additionally,\nthe term magnetic (refferring to the multipole) is used for this solution.\n\nArgs:\n    l (int, array_like): Degree :math:`l \\geq 0`\n    m (int, array_like): Order :math:`|m| \\leq l`\n    x (float or complex, array_like): Distance in units of the wave number :math:`kr`\n    theta (float or complex, array_like): Polar angle\n    phi (float, array_like): Azimuthal angle\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsw_rM, __pyx_t_1) < 0) __PYX_ERR(0, 1449, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1491
 * cdef char gufunc_vsw_rN_types[2 * 6]
 * 
 * gufunc_vsw_rN_loops[0] = <np.PyUFuncGenericFunction>loop_D_llddd             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_rN_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llddd);

  /* "treams/special/_gufuncs.pyx":1492
 * 
 * gufunc_vsw_rN_loops[0] = <np.PyUFuncGenericFunction>loop_D_llddd
 * gufunc_vsw_rN_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDDd);

  /* "treams/special/_gufuncs.pyx":1493
 * gufunc_vsw_rN_loops[0] = <np.PyUFuncGenericFunction>loop_D_llddd
 * gufunc_vsw_rN_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_rN_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[0]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1494
 * gufunc_vsw_rN_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDd
 * gufunc_vsw_rN_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1495
 * gufunc_vsw_rN_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1496
 * gufunc_vsw_rN_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1497
 * gufunc_vsw_rN_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1498
 * gufunc_vsw_rN_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[7] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1499
 * gufunc_vsw_rN_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1500
 * gufunc_vsw_rN_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[7] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[7]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1501
 * gufunc_vsw_rN_types[6] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_types[10] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1502
 * gufunc_vsw_rN_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_rN_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[11] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1503
 * gufunc_vsw_rN_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_types[10] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_data[0] = <void*>_waves.vsw_rN[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[10]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1504
 * gufunc_vsw_rN_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[11] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_data[0] = <void*>_waves.vsw_rN[double]
 * gufunc_vsw_rN_data[1] = <void*>_waves.vsw_rN[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types[11]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1505
 * gufunc_vsw_rN_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rN_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_data[0] = <void*>_waves.vsw_rN[double]             # <<<<<<<<<<<<<<
 * gufunc_vsw_rN_data[1] = <void*>_waves.vsw_rN[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_rN);

  /* "treams/special/_gufuncs.pyx":1506
 * gufunc_vsw_rN_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rN_data[0] = <void*>_waves.vsw_rN[double]
 * gufunc_vsw_rN_data[1] = <void*>_waves.vsw_rN[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsw_rN = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_rN);

  /* "treams/special/_gufuncs.pyx":1508
 * gufunc_vsw_rN_data[1] = <void*>_waves.vsw_rN[double_complex]
 * 
 * vsw_rN = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vsw_rN_loops,
 *     gufunc_vsw_rN_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rN_types, 2, 5, 1, 0, ((char *)"vsw_rN"), ((char *)"vsw_rN(l, m, x, theta, phi)\n\nRegular vector spherical wave N\n\nThe vector spherical wave is defined by\n\n.. math::\n\n    \\boldsymbol N_{lm}^{(1)}(x, \\theta, \\varphi)\n    = \\nabla \\times \\boldsymbol M_{lm}^{(1)}(x, \\theta, \\varphi) \\\\\n    = \\left(j_l'(x) + \\frac{j_l(x)}{x}\\right) \\boldsymbol Y_{lm}(\\theta, \\varphi)\n    + \\sqrt{l (l + 1)} \\frac{j_l(x)}{x} \\boldsymbol Z_{lm}(\\theta, \\varphi)\n\nwith :func:`treams.special.vsw_M`, :func:`treams.special.vsh_Y`,\n:func:`treams.special.vsh_Z`, and :func:`treams.special.spherical_jn`.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. This is often used to describe a transverse magnetic (in spherical\ncoordinates) (TM) wave. Additionally, the term electric (refferring to the\nmultipole) is used for this solution.\n\nArgs:\n    l (int, array_like): Degree :math:`l \\geq 0`\n    m (int, array_like): Order :math:`|m| \\leq l`\n    x (float or complex, array_like): Distance in units of the wave number :math:`kr`\n    theta (float or complex, array_like): Polar angle\n    phi (float, array_like): Azimuthal angle\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsw_rN, __pyx_t_1) < 0) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1557
 * cdef char gufunc_vsw_rA_types[2 * 7]
 * 
 * gufunc_vsw_rA_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldddl             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDdl
 * gufunc_vsw_rA_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_lldddl);

  /* "treams/special/_gufuncs.pyx":1558
 * 
 * gufunc_vsw_rA_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldddl
 * gufunc_vsw_rA_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDdl             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDDdl);

  /* "treams/special/_gufuncs.pyx":1559
 * gufunc_vsw_rA_loops[0] = <np.PyUFuncGenericFunction>loop_D_lldddl
 * gufunc_vsw_rA_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDdl
 * gufunc_vsw_rA_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[0]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1560
 * gufunc_vsw_rA_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDdl
 * gufunc_vsw_rA_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1561
 * gufunc_vsw_rA_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1562
 * gufunc_vsw_rA_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1563
 * gufunc_vsw_rA_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[5] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[6] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1564
 * gufunc_vsw_rA_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_types[7] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[5]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1565
 * gufunc_vsw_rA_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[5] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[6] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[8] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[6]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1566
 * gufunc_vsw_rA_types[5] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_types[7] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[8] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[7]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1567
 * gufunc_vsw_rA_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[8] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_types[10] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[8]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1568
 * gufunc_vsw_rA_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[8] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_types[11] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1569
 * gufunc_vsw_rA_types[8] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_types[10] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[12] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[10]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1570
 * gufunc_vsw_rA_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_types[11] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[12] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[13] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[11]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1571
 * gufunc_vsw_rA_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[12] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_data[0] = <void*>_waves.vsw_rA[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[12]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1572
 * gufunc_vsw_rA_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vsw_rA_types[12] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[13] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_data[0] = <void*>_waves.vsw_rA[double]
 * gufunc_vsw_rA_data[1] = <void*>_waves.vsw_rA[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types[13]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1573
 * gufunc_vsw_rA_types[12] = <char>np.NPY_LONG
 * gufunc_vsw_rA_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_data[0] = <void*>_waves.vsw_rA[double]             # <<<<<<<<<<<<<<
 * gufunc_vsw_rA_data[1] = <void*>_waves.vsw_rA[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_rA);

  /* "treams/special/_gufuncs.pyx":1574
 * gufunc_vsw_rA_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_rA_data[0] = <void*>_waves.vsw_rA[double]
 * gufunc_vsw_rA_data[1] = <void*>_waves.vsw_rA[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsw_rA = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_rA);

  /* "treams/special/_gufuncs.pyx":1576
 * gufunc_vsw_rA_data[1] = <void*>_waves.vsw_rA[double_complex]
 * 
 * vsw_rA = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vsw_rA_loops,
 *     gufunc_vsw_rA_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_rA_types, 2, 6, 1, 0, ((char *)"vsw_rA"), ((char *)"vsw_rA(l, m, x, theta, phi, p)\n\nRegular helical vector spherical wave\n\nThe vector spherical wave is defined by\n\n.. math::\n\n    \\boldsymbol A_{lm\\pm}^{(1)}(x, \\theta, \\varphi)\n    = \\frac{\\boldsymbol N_{lm}^{(1)}(x, \\theta, \\varphi)\n    \\pm \\boldsymbol M_{lm}^{(1)}(x, \\theta, \\varphi)}{\\sqrt{2}}\n\nwith :func:`treams.special.vsw_rM` and :func:`treams.special.vsw_rN`. The sign is\ndetermined by `p`, where `0` corresponds to negative and `1` to positive helicity.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. Additionally, it has a well-defined helicity.\n\nArgs:\n    l (int, array_like): Degree :math:`l \\geq 0`\n    m (int, array_like): Order :math:`|m| \\leq l`\n    x (float or complex, array_like): Distance in units of the wave number :math:`kr`\n    theta (float or complex, array_like): Polar angle\n    phi (float, array_like): Azimuthal angle\n    p (bool, array_like): Helicity\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsw_rA, __pyx_t_1) < 0) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1623
 * cdef char gufunc_vsw_A_types[2 * 7]
 * 
 * gufunc_vsw_A_loops[0] = <np.PyUFuncGenericFunction>loop_D_llDddl             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDdl
 * gufunc_vsw_A_types[0] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDddl);

  /* "treams/special/_gufuncs.pyx":1624
 * 
 * gufunc_vsw_A_loops[0] = <np.PyUFuncGenericFunction>loop_D_llDddl
 * gufunc_vsw_A_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDdl             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_llDDdl);

  /* "treams/special/_gufuncs.pyx":1625
 * gufunc_vsw_A_loops[0] = <np.PyUFuncGenericFunction>loop_D_llDddl
 * gufunc_vsw_A_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDdl
 * gufunc_vsw_A_types[0] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[2] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[0]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1626
 * gufunc_vsw_A_loops[1] = <np.PyUFuncGenericFunction>loop_D_llDDdl
 * gufunc_vsw_A_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1627
 * gufunc_vsw_A_types[0] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[2] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_A_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[2]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1628
 * gufunc_vsw_A_types[1] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_A_types[5] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1629
 * gufunc_vsw_A_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_A_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[5] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[6] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1630
 * gufunc_vsw_A_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vsw_A_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_A_types[5] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[7] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[5]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1631
 * gufunc_vsw_A_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vsw_A_types[5] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[6] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[8] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[6]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1632
 * gufunc_vsw_A_types[5] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[7] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[8] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[7]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1633
 * gufunc_vsw_A_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[8] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[10] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[8]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1634
 * gufunc_vsw_A_types[7] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[8] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[11] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1635
 * gufunc_vsw_A_types[8] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[10] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vsw_A_types[12] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[10]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1636
 * gufunc_vsw_A_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[11] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[12] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[13] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[11]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1637
 * gufunc_vsw_A_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vsw_A_types[12] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_data[0] = <void*>_waves.vsw_A[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[12]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1638
 * gufunc_vsw_A_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vsw_A_types[12] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[13] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_data[0] = <void*>_waves.vsw_A[double]
 * gufunc_vsw_A_data[1] = <void*>_waves.vsw_A[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types[13]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1639
 * gufunc_vsw_A_types[12] = <char>np.NPY_LONG
 * gufunc_vsw_A_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_data[0] = <void*>_waves.vsw_A[double]             # <<<<<<<<<<<<<<
 * gufunc_vsw_A_data[1] = <void*>_waves.vsw_A[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vsw_A);

  /* "treams/special/_gufuncs.pyx":1640
 * gufunc_vsw_A_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vsw_A_data[0] = <void*>_waves.vsw_A[double]
 * gufunc_vsw_A_data[1] = <void*>_waves.vsw_A[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vsw_A = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vsw_A);

  /* "treams/special/_gufuncs.pyx":1642
 * gufunc_vsw_A_data[1] = <void*>_waves.vsw_A[double_complex]
 * 
 * vsw_A = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vsw_A_loops,
 *     gufunc_vsw_A_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vsw_A_types, 2, 6, 1, 0, ((char *)"vsw_A"), ((char *)"vsw_A(l, m, x, theta, phi, p)\n\nSingular helical vector spherical wave\n\nThe vector spherical wave is defined by\n\n.. math::\n\n    \\boldsymbol A_{lm\\pm}^{(3)}(x, \\theta, \\varphi)\n    = \\frac{\\boldsymbol N_{lm}^{(3)}(x, \\theta, \\varphi)\n    \\pm \\boldsymbol M_{lm}^{(3)}(x, \\theta, \\varphi)}{\\sqrt{2}}\n\nwith :func:`treams.special.vsw_M` and :func:`treams.special.vsw_N`. The sign is\ndetermined by `p`, where `0` corresponds to negative and `1` to positive helicity.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. Additionally, it has a well-defined helicity.\n\nArgs:\n    l (int, array_like): Degree :math:`l \\geq 0`\n    m (int, array_like): Order :math:`|m| \\leq l`\n    x (float or complex, array_like): Distance in units of the wave number :math:`kr`\n    theta (float or complex, array_like): Polar angle\n    phi (float, array_like): Azimuthal angle\n    p (bool, array_like): Helicity\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsw_A, __pyx_t_1) < 0) __PYX_ERR(0, 1642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1689
 * cdef char gufunc_vcw_M_types[6]
 * 
 * gufunc_vcw_M_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlDdd             # <<<<<<<<<<<<<<
 * gufunc_vcw_M_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_M_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dlDdd);

  /* "treams/special/_gufuncs.pyx":1690
 * 
 * gufunc_vcw_M_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlDdd
 * gufunc_vcw_M_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_M_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_M_types[2] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1691
 * gufunc_vcw_M_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlDdd
 * gufunc_vcw_M_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_M_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_M_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_M_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1692
 * gufunc_vcw_M_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_M_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_M_types[2] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_M_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_M_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_types[2]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1693
 * gufunc_vcw_M_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_M_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_M_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_M_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_M_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1694
 * gufunc_vcw_M_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_M_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_M_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_M_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_M_data[0] = <void*>_waves.vcw_M
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1695
 * gufunc_vcw_M_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_M_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_M_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_M_data[0] = <void*>_waves.vcw_M
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1696
 * gufunc_vcw_M_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_M_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_M_data[0] = <void*>_waves.vcw_M             # <<<<<<<<<<<<<<
 * 
 * vcw_M = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_data[0]) = ((void *)__pyx_f_6treams_7special_6_waves_vcw_M);

  /* "treams/special/_gufuncs.pyx":1698
 * gufunc_vcw_M_data[0] = <void*>_waves.vcw_M
 * 
 * vcw_M = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcw_M_loops,
 *     gufunc_vcw_M_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_M_types, 1, 5, 1, 0, ((char *)"vcw_M"), ((char *)"vcw_M(kz, m, xrho, phi, z)\n\nSingular vector cylindrical wave M\n\nThe vector cylindrical wave is defined by\n\n.. math::\n\n    \\boldsymbol M_{k_z m}^{(3)}(x_\\rho, \\varphi, z)\n    = \\left(\\frac{\\mathrm i m}{x_\\rho} H_m^{(1)}(x_\\rho)\\boldsymbol{\\hat{\\rho}}\n    - {H_m^{(1)}}'(x_\\rho) \\boldsymbol{\\hat{\\varphi}}\\right)\n    \\mathrm e^{\\mathrm i (k_z z + m \\varphi)}\n\nusing :func:`treams.special.hankel1`.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation, that is also tangential on a cylindrical surface. This is often used to\ndescribe a transverse electric (in cylindrical coordinates) (TE) wave.\n\nArgs:\n    kz (float, array_like): Z component of the wave\n    m (int, array_like): Order :math:`|m| \\leq l`\n    xrho (float or complex, array_like): Radial in units of the wave number :math:`k_\\rho \\rho`\n    phi (float, array_like): Azimuthal angle\n    z (float, array_like): Z coordinate\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcw_M, __pyx_t_1) < 0) __PYX_ERR(0, 1698, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1745
 * cdef char gufunc_vcw_rM_types[2 * 6]
 * 
 * gufunc_vcw_rM_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlddd             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDdd
 * gufunc_vcw_rM_types[0] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dlddd);

  /* "treams/special/_gufuncs.pyx":1746
 * 
 * gufunc_vcw_rM_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlddd
 * gufunc_vcw_rM_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDdd             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dlDdd);

  /* "treams/special/_gufuncs.pyx":1747
 * gufunc_vcw_rM_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlddd
 * gufunc_vcw_rM_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDdd
 * gufunc_vcw_rM_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_rM_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1748
 * gufunc_vcw_rM_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDdd
 * gufunc_vcw_rM_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1749
 * gufunc_vcw_rM_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_rM_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1750
 * gufunc_vcw_rM_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_rM_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1751
 * gufunc_vcw_rM_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rM_types[6] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1752
 * gufunc_vcw_rM_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[6] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[7] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1753
 * gufunc_vcw_rM_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rM_types[6] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[7] = <char>np.NPY_LONG
 * gufunc_vcw_rM_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[6]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1754
 * gufunc_vcw_rM_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rM_types[6] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[7] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rM_types[9] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[7]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1755
 * gufunc_vcw_rM_types[6] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[7] = <char>np.NPY_LONG
 * gufunc_vcw_rM_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[9] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[10] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1756
 * gufunc_vcw_rM_types[7] = <char>np.NPY_LONG
 * gufunc_vcw_rM_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rM_types[9] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[11] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[9]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1757
 * gufunc_vcw_rM_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rM_types[9] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[10] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rM_data[0] = <void*>_waves.vcw_rM[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[10]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1758
 * gufunc_vcw_rM_types[9] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[11] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_data[0] = <void*>_waves.vcw_rM[double]
 * gufunc_vcw_rM_data[1] = <void*>_waves.vcw_rM[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types[11]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1759
 * gufunc_vcw_rM_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rM_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rM_data[0] = <void*>_waves.vcw_rM[double]             # <<<<<<<<<<<<<<
 * gufunc_vcw_rM_data[1] = <void*>_waves.vcw_rM[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vcw_rM);

  /* "treams/special/_gufuncs.pyx":1760
 * gufunc_vcw_rM_types[11] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rM_data[0] = <void*>_waves.vcw_rM[double]
 * gufunc_vcw_rM_data[1] = <void*>_waves.vcw_rM[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vcw_rM = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vcw_rM);

  /* "treams/special/_gufuncs.pyx":1762
 * gufunc_vcw_rM_data[1] = <void*>_waves.vcw_rM[double_complex]
 * 
 * vcw_rM = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcw_rM_loops,
 *     gufunc_vcw_rM_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rM_types, 2, 5, 1, 0, ((char *)"vcw_rM"), ((char *)"vcw_rM(kz, m, xrho, phi, z)\n\nRegular vector cylindrical wave M\n\nThe vector cylindrical wave is defined by\n\n.. math::\n\n    \\boldsymbol M_{k_z m}^{(1)}(x_\\rho, \\varphi, z)\n    = \\left(\\frac{\\mathrm i m}{x_\\rho} J_m(x_\\rho)\\boldsymbol{\\hat{\\rho}}\n    - J_m'(x_\\rho) \\boldsymbol{\\hat{\\varphi}}\\right)\n    \\mathrm e^{\\mathrm i (k_z z + m \\varphi)}\n\nusing :func:`treams.special.jn`.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation, that is also tangential on a cylindrical surface. This is often used to\ndescribe a transverse electric (in cylindrical coordinates) (TE) wave.\n\nArgs:\n    kz (float, array_like): Z component of the wave\n    m (int, array_like): Order :math:`|m| \\leq l`\n    xrho (float or complex, array_like): Radial in units of the wave number :math:`k_\\rho \\rho`\n    phi (float, array_like): Azimuthal angle\n    z (float, array_like): Z coordinate\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcw_rM, __pyx_t_1) < 0) __PYX_ERR(0, 1762, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1809
 * cdef char gufunc_vcw_N_types[7]
 * 
 * gufunc_vcw_N_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlDddD             # <<<<<<<<<<<<<<
 * gufunc_vcw_N_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_N_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dlDddD);

  /* "treams/special/_gufuncs.pyx":1810
 * 
 * gufunc_vcw_N_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlDddD
 * gufunc_vcw_N_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_N_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_N_types[2] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1811
 * gufunc_vcw_N_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlDddD
 * gufunc_vcw_N_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_N_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_N_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_N_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1812
 * gufunc_vcw_N_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_N_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_N_types[2] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_N_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_N_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_types[2]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1813
 * gufunc_vcw_N_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_N_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_N_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_N_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_N_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1814
 * gufunc_vcw_N_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_N_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_N_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_N_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_N_types[6] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1815
 * gufunc_vcw_N_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_N_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_N_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_N_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_N_data[0] = <void*>_waves.vcw_N
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1816
 * gufunc_vcw_N_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_N_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_N_types[6] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_N_data[0] = <void*>_waves.vcw_N
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_types[6]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1817
 * gufunc_vcw_N_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_N_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_N_data[0] = <void*>_waves.vcw_N             # <<<<<<<<<<<<<<
 * 
 * vcw_N = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_data[0]) = ((void *)__pyx_f_6treams_7special_6_waves_vcw_N);

  /* "treams/special/_gufuncs.pyx":1819
 * gufunc_vcw_N_data[0] = <void*>_waves.vcw_N
 * 
 * vcw_N = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcw_N_loops,
 *     gufunc_vcw_N_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_N_types, 1, 6, 1, 0, ((char *)"vcw_N"), ((char *)"vcw_N(kz, m, xrho, phi, z, k)\n\nSingular vector cylindrical wave N\n\nThe vector cylindrical wave is defined by\n\n.. math::\n\n    \\boldsymbol N_{k_z m}^{(3)}(x_\\rho, \\varphi, z)\n    = \\left(\\frac{\\mathrm i k_z}{k} {H_m^{(1)}}'(x_\\rho)\\boldsymbol{\\hat{\\rho}}\n    - \\frac{m k_z}{k x_\\rho} H_m^{(1)}(x_\\rho) \\boldsymbol{\\hat{\\varphi}}\n    + \\frac{k_\\rho}{k} H_m^{(1)}(x_\\rho) \\boldsymbol{\\hat{z}}\\right)\n    \\mathrm e^{\\mathrm i (k_z z + m \\varphi)}\n\nusing :func:`treams.special.hankel1`.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. This is often used to describe a transverse magnetic (in cylindrical\ncoordinates) (TM) wave.\n\nArgs:\n    kz (float, array_like): Z component of the wave\n    m (int, array_like): Order :math:`|m| \\leq l`\n    xrho (float or complex, array_like): Radial in units of the wave number\n        :math:`k_\\rho \\rho`\n    phi (float, array_like): Azimuthal angle\n    z (float, array_like): Z coordinate\n    k (float or complex): Wave number\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcw_N, __pyx_t_1) < 0) __PYX_ERR(0, 1819, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1869
 * cdef char gufunc_vcw_rN_types[2 * 7]
 * 
 * gufunc_vcw_rN_loops[0] = <np.PyUFuncGenericFunction>loop_D_dldddD             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDddD
 * gufunc_vcw_rN_types[0] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dldddD);

  /* "treams/special/_gufuncs.pyx":1870
 * 
 * gufunc_vcw_rN_loops[0] = <np.PyUFuncGenericFunction>loop_D_dldddD
 * gufunc_vcw_rN_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDddD             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dlDddD);

  /* "treams/special/_gufuncs.pyx":1871
 * gufunc_vcw_rN_loops[0] = <np.PyUFuncGenericFunction>loop_D_dldddD
 * gufunc_vcw_rN_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDddD
 * gufunc_vcw_rN_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_rN_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1872
 * gufunc_vcw_rN_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDddD
 * gufunc_vcw_rN_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1873
 * gufunc_vcw_rN_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_rN_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1874
 * gufunc_vcw_rN_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_rN_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1875
 * gufunc_vcw_rN_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[6] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1876
 * gufunc_vcw_rN_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[7] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1877
 * gufunc_vcw_rN_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[6] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[7] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[8] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[6]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1878
 * gufunc_vcw_rN_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[7] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[8] = <char>np.NPY_LONG
 * gufunc_vcw_rN_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[7]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1879
 * gufunc_vcw_rN_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[7] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[8] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[10] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[8]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1880
 * gufunc_vcw_rN_types[7] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[8] = <char>np.NPY_LONG
 * gufunc_vcw_rN_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[11] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1881
 * gufunc_vcw_rN_types[8] = <char>np.NPY_LONG
 * gufunc_vcw_rN_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[10] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[12] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[10]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1882
 * gufunc_vcw_rN_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[11] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[12] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[13] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[11]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1883
 * gufunc_vcw_rN_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[12] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_data[0] = <void*>_waves.vcw_rN[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[12]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1884
 * gufunc_vcw_rN_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rN_types[12] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[13] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_data[0] = <void*>_waves.vcw_rN[double]
 * gufunc_vcw_rN_data[1] = <void*>_waves.vcw_rN[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types[13]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1885
 * gufunc_vcw_rN_types[12] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_data[0] = <void*>_waves.vcw_rN[double]             # <<<<<<<<<<<<<<
 * gufunc_vcw_rN_data[1] = <void*>_waves.vcw_rN[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vcw_rN);

  /* "treams/special/_gufuncs.pyx":1886
 * gufunc_vcw_rN_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rN_data[0] = <void*>_waves.vcw_rN[double]
 * gufunc_vcw_rN_data[1] = <void*>_waves.vcw_rN[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vcw_rN = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vcw_rN);

  /* "treams/special/_gufuncs.pyx":1888
 * gufunc_vcw_rN_data[1] = <void*>_waves.vcw_rN[double_complex]
 * 
 * vcw_rN = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcw_rN_loops,
 *     gufunc_vcw_rN_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rN_types, 2, 6, 1, 0, ((char *)"vcw_rN"), ((char *)"vcw_N(kz, m, xrho, phi, z, k)\n\nRegular vector cylindrical wave N\n\nThe vector cylindrical wave is defined by\n\n.. math::\n\n    \\boldsymbol N_{k_z m}^{(3)}(x_\\rho, \\varphi, z)\n    = \\left(\\frac{\\mathrm i k_z}{k} J_m'(x_\\rho)\\boldsymbol{\\hat{\\rho}}\n    - \\frac{m k_z}{k x_\\rho} J_m(x_\\rho) \\boldsymbol{\\hat{\\varphi}}\n    + \\frac{k_\\rho}{k} J_m(x_\\rho) \\boldsymbol{\\hat{z}}\\right)\n    \\mathrm e^{\\mathrm i (k_z z + m \\varphi)}\n\nusing :func:`treams.special.jv`.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. This is often used to describe a transverse magnetic (in cylindrical\ncoordinates) (TM) wave.\n\nArgs:\n    kz (float, array_like): Z component of the wave\n    m (int, array_like): Order :math:`|m| \\leq l`\n    xrho (float or complex, array_like): Radial in units of the wave number\n        :math:`k_\\rho \\rho`\n    phi (float, array_like): Azimuthal angle\n    z (float, array_like): Z coordinate\n    k (float or complex): Wave number\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1888, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcw_rN, __pyx_t_1) < 0) __PYX_ERR(0, 1888, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1938
 * cdef char gufunc_vcw_A_types[8]
 * 
 * gufunc_vcw_A_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlDddDl             # <<<<<<<<<<<<<<
 * gufunc_vcw_A_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_A_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dlDddDl);

  /* "treams/special/_gufuncs.pyx":1939
 * 
 * gufunc_vcw_A_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlDddDl
 * gufunc_vcw_A_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_A_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_A_types[2] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1940
 * gufunc_vcw_A_loops[0] = <np.PyUFuncGenericFunction>loop_D_dlDddDl
 * gufunc_vcw_A_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_A_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_A_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_A_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1941
 * gufunc_vcw_A_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_A_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_A_types[2] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_A_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_A_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types[2]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1942
 * gufunc_vcw_A_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_A_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_A_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_A_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_A_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1943
 * gufunc_vcw_A_types[2] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_A_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_A_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_A_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_A_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":1944
 * gufunc_vcw_A_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_A_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_A_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_A_types[6] = <char>np.NPY_LONG
 * gufunc_vcw_A_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1945
 * gufunc_vcw_A_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_A_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_A_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_A_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_A_data[0] = <void*>_waves.vcw_A
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":1946
 * gufunc_vcw_A_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_A_types[6] = <char>np.NPY_LONG
 * gufunc_vcw_A_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_A_data[0] = <void*>_waves.vcw_A
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":1947
 * gufunc_vcw_A_types[6] = <char>np.NPY_LONG
 * gufunc_vcw_A_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_A_data[0] = <void*>_waves.vcw_A             # <<<<<<<<<<<<<<
 * 
 * vcw_A = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_data[0]) = ((void *)__pyx_f_6treams_7special_6_waves_vcw_A);

  /* "treams/special/_gufuncs.pyx":1949
 * gufunc_vcw_A_data[0] = <void*>_waves.vcw_A
 * 
 * vcw_A = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcw_A_loops,
 *     gufunc_vcw_A_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_A_types, 1, 7, 1, 0, ((char *)"vcw_A"), ((char *)"vcw_A(kz, m, xrho, phi, z, k, pol)\n\nSingular helical vector cylindrical wave\n\nThe vector spherical wave is defined by\n\n.. math::\n\n    \\boldsymbol A_{k_z m \\pm}^{(3)}(x_\\rho, \\varphi, z)\n    = \\frac{\\boldsymbol N_{k_z m}^{(3)}(x_\\rho, \\varphi, z)\n    \\pm \\boldsymbol M_{k_z m}^{(3)}(x_\\rho, \\varphi, z)}{\\sqrt{2}}\n\nwith :func:`treams.special.vcw_M` and :func:`treams.special.vcw_N`. The sign is\ndetermined by `p`, where `0` corresponds to negative and `1` to positive helicity.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. Additionally, it has a well-defined helicity.\n\nArgs:\n    kz (float, array_like): Z component of the wave\n    m (int, array_like): Order :math:`|m| \\leq l`\n    xrho (float or complex, array_like): Radial in units of the wave number\n        :math:`k_\\rho \\rho`\n    phi (float, array_like): Azimuthal angle\n    z (float, array_like): Z coordinate\n    k (float or complex): Wave number\n    pol (bool, array_like): Polarization\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcw_A, __pyx_t_1) < 0) __PYX_ERR(0, 1949, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1998
 * cdef char gufunc_vcw_rA_types[2 * 8]
 * 
 * gufunc_vcw_rA_loops[0] = <np.PyUFuncGenericFunction>loop_D_dldddDl             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDddDl
 * gufunc_vcw_rA_types[0] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dldddDl);

  /* "treams/special/_gufuncs.pyx":1999
 * 
 * gufunc_vcw_rA_loops[0] = <np.PyUFuncGenericFunction>loop_D_dldddDl
 * gufunc_vcw_rA_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDddDl             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[1] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dlDddDl);

  /* "treams/special/_gufuncs.pyx":2000
 * gufunc_vcw_rA_loops[0] = <np.PyUFuncGenericFunction>loop_D_dldddDl
 * gufunc_vcw_rA_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDddDl
 * gufunc_vcw_rA_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2001
 * gufunc_vcw_rA_loops[1] = <np.PyUFuncGenericFunction>loop_D_dlDddDl
 * gufunc_vcw_rA_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[1] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[1]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":2002
 * gufunc_vcw_rA_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2003
 * gufunc_vcw_rA_types[1] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[5] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2004
 * gufunc_vcw_rA_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2005
 * gufunc_vcw_rA_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[5] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[6] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[5]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2006
 * gufunc_vcw_rA_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[8] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":2007
 * gufunc_vcw_rA_types[5] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[6] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[9] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2008
 * gufunc_vcw_rA_types[6] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[8] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[9] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[10] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[8]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2009
 * gufunc_vcw_rA_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[9] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[11] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[9]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":2010
 * gufunc_vcw_rA_types[8] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[9] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[10] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[12] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[10]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2011
 * gufunc_vcw_rA_types[9] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[11] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[13] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[11]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2012
 * gufunc_vcw_rA_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[12] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[14] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[12]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2013
 * gufunc_vcw_rA_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[13] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[14] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[15] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[13]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2014
 * gufunc_vcw_rA_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vcw_rA_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[14] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_types[15] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_data[0] = <void*>_waves.vcw_rA[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[14]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":2015
 * gufunc_vcw_rA_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_types[14] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[15] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_data[0] = <void*>_waves.vcw_rA[double]
 * gufunc_vcw_rA_data[1] = <void*>_waves.vcw_rA[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types[15]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2016
 * gufunc_vcw_rA_types[14] = <char>np.NPY_LONG
 * gufunc_vcw_rA_types[15] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_data[0] = <void*>_waves.vcw_rA[double]             # <<<<<<<<<<<<<<
 * gufunc_vcw_rA_data[1] = <void*>_waves.vcw_rA[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vcw_rA);

  /* "treams/special/_gufuncs.pyx":2017
 * gufunc_vcw_rA_types[15] = <char>np.NPY_CDOUBLE
 * gufunc_vcw_rA_data[0] = <void*>_waves.vcw_rA[double]
 * gufunc_vcw_rA_data[1] = <void*>_waves.vcw_rA[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vcw_rA = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vcw_rA);

  /* "treams/special/_gufuncs.pyx":2019
 * gufunc_vcw_rA_data[1] = <void*>_waves.vcw_rA[double_complex]
 * 
 * vcw_rA = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vcw_rA_loops,
 *     gufunc_vcw_rA_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vcw_rA_types, 2, 7, 1, 0, ((char *)"vcw_rA"), ((char *)"vcw_rA(kz, m, xrho, phi, z, k)\n\nRegular helical vector cylindrical wave\n\nThe vector spherical wave is defined by\n\n.. math::\n\n    \\boldsymbol A_{k_z m \\pm}^{(1)}(x_\\rho, \\varphi, z)\n    = \\frac{\\boldsymbol N_{k_z m}^{(1)}(x_\\rho, \\varphi, z) \\pm\n    \\boldsymbol M_{k_z m}^{(1)}(x_\\rho, \\varphi, z)}{\\sqrt{2}}\n\nwith :func:`treams.special.vcw_rM` and :func:`treams.special.vcw_rN`. The sign is\ndetermined by `p`, where `0` corresponds to negative and `1` to positive helicity.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. Additionally, it has a well-defined helicity.\n\nArgs:\n    kz (float, array_like): Z component of the wave\n    m (int, array_like): Order :math:`|m| \\leq l`\n    xrho (float or complex, array_like): Radial in units of the wave number :math:`k_\\rho \\rho`\n    phi (float, array_like): Azimuthal angle\n    z (float, array_like): Z coordinate\n    k (float or complex): Wave number\n    pol (bool, array_like): Polarization\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vcw_rA, __pyx_t_1) < 0) __PYX_ERR(0, 2019, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":2067
 * cdef char gufunc_vpw_M_types[2 * 7]
 * 
 * gufunc_vpw_M_loops[0] = <np.PyUFuncGenericFunction>loop_D_dddddd             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDddd
 * gufunc_vpw_M_types[0] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dddddd);

  /* "treams/special/_gufuncs.pyx":2068
 * 
 * gufunc_vpw_M_loops[0] = <np.PyUFuncGenericFunction>loop_D_dddddd
 * gufunc_vpw_M_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDddd             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_DDDddd);

  /* "treams/special/_gufuncs.pyx":2069
 * gufunc_vpw_M_loops[0] = <np.PyUFuncGenericFunction>loop_D_dddddd
 * gufunc_vpw_M_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDddd
 * gufunc_vpw_M_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2070
 * gufunc_vpw_M_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDddd
 * gufunc_vpw_M_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2071
 * gufunc_vpw_M_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2072
 * gufunc_vpw_M_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[5] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2073
 * gufunc_vpw_M_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[5] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[6] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2074
 * gufunc_vpw_M_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[5] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[5]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2075
 * gufunc_vpw_M_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[5] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[6] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[6]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2076
 * gufunc_vpw_M_types[5] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2077
 * gufunc_vpw_M_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[10] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2078
 * gufunc_vpw_M_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[11] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2079
 * gufunc_vpw_M_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[10] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[12] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[10]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2080
 * gufunc_vpw_M_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[11] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[13] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[11]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2081
 * gufunc_vpw_M_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[12] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_data[0] = <void*>_waves.vpw_M[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[12]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2082
 * gufunc_vpw_M_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[13] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_data[0] = <void*>_waves.vpw_M[double]
 * gufunc_vpw_M_data[1] = <void*>_waves.vpw_M[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types[13]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2083
 * gufunc_vpw_M_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vpw_M_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_data[0] = <void*>_waves.vpw_M[double]             # <<<<<<<<<<<<<<
 * gufunc_vpw_M_data[1] = <void*>_waves.vpw_M[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vpw_M);

  /* "treams/special/_gufuncs.pyx":2084
 * gufunc_vpw_M_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_M_data[0] = <void*>_waves.vpw_M[double]
 * gufunc_vpw_M_data[1] = <void*>_waves.vpw_M[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vpw_M = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vpw_M);

  /* "treams/special/_gufuncs.pyx":2086
 * gufunc_vpw_M_data[1] = <void*>_waves.vpw_M[double_complex]
 * 
 * vpw_M = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vpw_M_loops,
 *     gufunc_vpw_M_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_M_types, 2, 6, 1, 0, ((char *)"vpw_M"), ((char *)"vpw_M(kx, ky, kz, x, y, z)\n\nVector plane wave M\n\nThe vector plane wave is defined by\n\n.. math::\n\n    \\boldsymbol M_{\\boldsymbol k}(\\boldsymbol r)\n    = - \\mathrm i \\boldsymbol{\\hat \\varphi}_{\\boldsymbol k}\n    \\mathrm e^{\\mathrm i \\boldsymbol k \\boldsymbol r}.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation, that is also tangential on a planar surface. This is often used to\ndescribe a transverse electric (TE) wave.\n\nArgs:\n    kx (float or complex, array_like): X component of the wave vector\n    ky (float or complex, array_like): Y component of the wave vector\n    kz (float or complex, array_like): Z component of the wave vector\n    x (float, array_like): Y coordinate\n    y (float, array_like): X coordinate\n    z (float, array_like): Z coordinate\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vpw_M, __pyx_t_1) < 0) __PYX_ERR(0, 2086, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":2131
 * cdef char gufunc_vpw_N_types[2 * 7]
 * 
 * gufunc_vpw_N_loops[0] = <np.PyUFuncGenericFunction>loop_D_dddddd             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDddd
 * gufunc_vpw_N_types[0] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_dddddd);

  /* "treams/special/_gufuncs.pyx":2132
 * 
 * gufunc_vpw_N_loops[0] = <np.PyUFuncGenericFunction>loop_D_dddddd
 * gufunc_vpw_N_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDddd             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_DDDddd);

  /* "treams/special/_gufuncs.pyx":2133
 * gufunc_vpw_N_loops[0] = <np.PyUFuncGenericFunction>loop_D_dddddd
 * gufunc_vpw_N_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDddd
 * gufunc_vpw_N_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2134
 * gufunc_vpw_N_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDddd
 * gufunc_vpw_N_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2135
 * gufunc_vpw_N_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2136
 * gufunc_vpw_N_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[5] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2137
 * gufunc_vpw_N_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[5] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[6] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2138
 * gufunc_vpw_N_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[5] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[5]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2139
 * gufunc_vpw_N_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[5] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[6] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[6]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2140
 * gufunc_vpw_N_types[5] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2141
 * gufunc_vpw_N_types[6] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[10] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2142
 * gufunc_vpw_N_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[11] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2143
 * gufunc_vpw_N_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[10] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[12] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[10]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2144
 * gufunc_vpw_N_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[11] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[13] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[11]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2145
 * gufunc_vpw_N_types[10] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[12] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_data[0] = <void*>_waves.vpw_N[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[12]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2146
 * gufunc_vpw_N_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[13] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_data[0] = <void*>_waves.vpw_N[double]
 * gufunc_vpw_N_data[1] = <void*>_waves.vpw_N[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types[13]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2147
 * gufunc_vpw_N_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vpw_N_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_data[0] = <void*>_waves.vpw_N[double]             # <<<<<<<<<<<<<<
 * gufunc_vpw_N_data[1] = <void*>_waves.vpw_N[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vpw_N);

  /* "treams/special/_gufuncs.pyx":2148
 * gufunc_vpw_N_types[13] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_N_data[0] = <void*>_waves.vpw_N[double]
 * gufunc_vpw_N_data[1] = <void*>_waves.vpw_N[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vpw_N = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vpw_N);

  /* "treams/special/_gufuncs.pyx":2150
 * gufunc_vpw_N_data[1] = <void*>_waves.vpw_N[double_complex]
 * 
 * vpw_N = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vpw_N_loops,
 *     gufunc_vpw_N_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_N_types, 2, 6, 1, 0, ((char *)"vpw_N"), ((char *)"vpw_N(kx, ky, kz, x, y, z)\n\nVector plane wave N\n\nThe vector plane wave is defined by\n\n.. math::\n\n    \\boldsymbol N_{\\boldsymbol k}(\\boldsymbol r)\n    = - \\boldsymbol{\\hat \\theta}_{\\boldsymbol k}\n    \\mathrm e^{\\mathrm i \\boldsymbol k \\boldsymbol r}.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. This is often used to describe a transverse magnetic (TM) wave.\n\nArgs:\n    kx (float or complex, array_like): X component of the wave vector\n    ky (float or complex, array_like): Y component of the wave vector\n    kz (float or complex, array_like): Z component of the wave vector\n    x (float, array_like): Y coordinate\n    y (float, array_like): X coordinate\n    z (float, array_like): Z coordinate\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vpw_N, __pyx_t_1) < 0) __PYX_ERR(0, 2150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":2194
 * cdef char gufunc_vpw_A_types[2 * 8]
 * 
 * gufunc_vpw_A_loops[0] = <np.PyUFuncGenericFunction>loop_D_ddddddl             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDdddl
 * gufunc_vpw_A_types[0] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_loops[0]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_ddddddl);

  /* "treams/special/_gufuncs.pyx":2195
 * 
 * gufunc_vpw_A_loops[0] = <np.PyUFuncGenericFunction>loop_D_ddddddl
 * gufunc_vpw_A_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDdddl             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[1] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_loops[1]) = ((PyUFuncGenericFunction)__pyx_f_6treams_7special_8_gufuncs_loop_D_DDDdddl);

  /* "treams/special/_gufuncs.pyx":2196
 * gufunc_vpw_A_loops[0] = <np.PyUFuncGenericFunction>loop_D_ddddddl
 * gufunc_vpw_A_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDdddl
 * gufunc_vpw_A_types[0] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[2] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[0]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2197
 * gufunc_vpw_A_loops[1] = <np.PyUFuncGenericFunction>loop_D_DDDdddl
 * gufunc_vpw_A_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[1] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[3] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[1]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2198
 * gufunc_vpw_A_types[0] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[2] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[4] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[2]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2199
 * gufunc_vpw_A_types[1] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[3] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[5] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[3]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2200
 * gufunc_vpw_A_types[2] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[4] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[5] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[6] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[4]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2201
 * gufunc_vpw_A_types[3] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[5] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[6] = <char>np.NPY_LONG
 * gufunc_vpw_A_types[7] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[5]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2202
 * gufunc_vpw_A_types[4] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[5] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[6] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[8] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[6]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":2203
 * gufunc_vpw_A_types[5] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[6] = <char>np.NPY_LONG
 * gufunc_vpw_A_types[7] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[9] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[7]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2204
 * gufunc_vpw_A_types[6] = <char>np.NPY_LONG
 * gufunc_vpw_A_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[8] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[10] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[8]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2205
 * gufunc_vpw_A_types[7] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[9] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[11] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[9]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2206
 * gufunc_vpw_A_types[8] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[10] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[12] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[10]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2207
 * gufunc_vpw_A_types[9] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[11] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[13] = <char>np.NPY_DOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[11]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2208
 * gufunc_vpw_A_types[10] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[12] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[13] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[14] = <char>np.NPY_LONG
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[12]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2209
 * gufunc_vpw_A_types[11] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[13] = <char>np.NPY_DOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[14] = <char>np.NPY_LONG
 * gufunc_vpw_A_types[15] = <char>np.NPY_CDOUBLE
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[13]) = ((char)NPY_DOUBLE);

  /* "treams/special/_gufuncs.pyx":2210
 * gufunc_vpw_A_types[12] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[13] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[14] = <char>np.NPY_LONG             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_types[15] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_data[0] = <void*>_waves.vpw_A[double]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[14]) = ((char)NPY_LONG);

  /* "treams/special/_gufuncs.pyx":2211
 * gufunc_vpw_A_types[13] = <char>np.NPY_DOUBLE
 * gufunc_vpw_A_types[14] = <char>np.NPY_LONG
 * gufunc_vpw_A_types[15] = <char>np.NPY_CDOUBLE             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_data[0] = <void*>_waves.vpw_A[double]
 * gufunc_vpw_A_data[1] = <void*>_waves.vpw_A[double_complex]
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types[15]) = ((char)NPY_CDOUBLE);

  /* "treams/special/_gufuncs.pyx":2212
 * gufunc_vpw_A_types[14] = <char>np.NPY_LONG
 * gufunc_vpw_A_types[15] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_data[0] = <void*>_waves.vpw_A[double]             # <<<<<<<<<<<<<<
 * gufunc_vpw_A_data[1] = <void*>_waves.vpw_A[double_complex]
 * 
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_data[0]) = ((void *)__pyx_fuse_1__pyx_f_6treams_7special_6_waves_vpw_A);

  /* "treams/special/_gufuncs.pyx":2213
 * gufunc_vpw_A_types[15] = <char>np.NPY_CDOUBLE
 * gufunc_vpw_A_data[0] = <void*>_waves.vpw_A[double]
 * gufunc_vpw_A_data[1] = <void*>_waves.vpw_A[double_complex]             # <<<<<<<<<<<<<<
 * 
 * vpw_A = np.PyUFunc_FromFuncAndDataAndSignature(
 */
  (__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_data[1]) = ((void *)__pyx_fuse_0__pyx_f_6treams_7special_6_waves_vpw_A);

  /* "treams/special/_gufuncs.pyx":2215
 * gufunc_vpw_A_data[1] = <void*>_waves.vpw_A[double_complex]
 * 
 * vpw_A = np.PyUFunc_FromFuncAndDataAndSignature(             # <<<<<<<<<<<<<<
 *     gufunc_vpw_A_loops,
 *     gufunc_vpw_A_data,
 */
  __pyx_t_1 = PyUFunc_FromFuncAndDataAndSignature(__pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_loops, __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_data, __pyx_v_6treams_7special_8_gufuncs_gufunc_vpw_A_types, 2, 7, 1, 0, ((char *)"vpw_A"), ((char *)"vpw_A(kx, ky, kz, x, y, z, p)\n\nVector plane wave of well-defined helicity\n\nThe vector plane wave is defined by\n\n.. math::\n\n    \\boldsymbol A_{\\boldsymbol k \\pm}(\\boldsymbol r)\n    = \\frac{\\boldsymbol N_{\\boldsymbol k \\pm}(\\boldsymbol r)\n    \\pm \\boldsymbol M_{\\boldsymbol k}(\\boldsymbol r)}{\\sqrt{2}}\n\nwith :func:`treams.special.vpw_M` and :func:`treams.special.vpw_N`. The sign is\ndetermined by `p`, where `0` corresponds to negative and `1` to positive helicity.\n\nThis function is describing a transverse solution to the vectorial Helmholtz wave\nequation. Additionally, it has a well-defined helicity.\n\nArgs:\n    kx (float or complex, array_like): X component of the wave vector\n    ky (float or complex, array_like): Y component of the wave vector\n    kz (float or complex, array_like): Z component of the wave vector\n    x (float, array_like): Y coordinate\n    y (float, array_like): X coordinate\n    z (float, array_like): Z coordinate\n    p (bool, array_like): Helicity\n\nReturns:\n    complex, 3-array\n"), 0, ((char *)"(),(),(),(),(),(),()->(3)")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vpw_A, __pyx_t_1) < 0) __PYX_ERR(0, 2215, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_gufuncs.pyx":1
 * """Generalized universal functions for the special subpackage"""             # <<<<<<<<<<<<<<
 * 
 * cimport numpy as np
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "treams/special/_misc.pxd":32
 *     return (a if a > b else b)
 * 
 * cdef inline real_t min(real_t a, real_t b) nogil:             # <<<<<<<<<<<<<<
 *     return (a if a < b else b)
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init treams.special._gufuncs", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init treams.special._gufuncs");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_0_29_36
#define __PYX_HAVE_RT_ImportType_0_29_36
static PyTypeObject *__Pyx_ImportType_0_29_36(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_0_29_36 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_0_29_36 && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_0_29_36 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_Py_intptr_t(Py_intptr_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_intptr_t neg_one = (Py_intptr_t) -1, const_zero = (Py_intptr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(Py_intptr_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(Py_intptr_t),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE Py_intptr_t __Pyx_PyInt_As_Py_intptr_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_intptr_t neg_one = (Py_intptr_t) -1, const_zero = (Py_intptr_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(Py_intptr_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (Py_intptr_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) >= 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (Py_intptr_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (Py_intptr_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(Py_intptr_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(Py_intptr_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(Py_intptr_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(Py_intptr_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(Py_intptr_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(Py_intptr_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(Py_intptr_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) (((Py_intptr_t)-1)*(((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(Py_intptr_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(Py_intptr_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(Py_intptr_t) - 1 > 4 * PyLong_SHIFT) {
                            return (Py_intptr_t) ((((((((((Py_intptr_t)digits[3]) << PyLong_SHIFT) | (Py_intptr_t)digits[2]) << PyLong_SHIFT) | (Py_intptr_t)digits[1]) << PyLong_SHIFT) | (Py_intptr_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(Py_intptr_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(Py_intptr_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(Py_intptr_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            Py_intptr_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (Py_intptr_t) -1;
        }
    } else {
        Py_intptr_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (Py_intptr_t) -1;
        val = __Pyx_PyInt_As_Py_intptr_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to Py_intptr_t");
    return (Py_intptr_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to Py_intptr_t");
    return (Py_intptr_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr_0_29_36
#define __PYX_HAVE_RT_ImportVoidPtr_0_29_36
static int __Pyx_ImportVoidPtr_0_29_36(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction_0_29_36
#define __PYX_HAVE_RT_ImportFunction_0_29_36
static int __Pyx_ImportFunction_0_29_36(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
